import os
import sys
import re
import h5py
import numpy as np

class CronosFileHandler(object) :
    def set_dataDir(self, DataDir) :
        self.datadir = DataDir
        print " Data dir set to: '" + self.datadir +"'"

    def scan_Projs(self) :
        
        filenames = []
        fullnames = []

        #        for file in os.listdir(self.datadir):
        #            if(self.string_has_patt(file, ".*_float\Z")) :
        #                filenames.append(file)
        #                fullnames.append(self.datadir + "/" + file)
        ##                print file

        for file in os.listdir(self.datadir):
            if(self.string_has_patt(file, ".*_float\Z")) :
                filenames.append(file)
                #                print file

        filenames.sort()
        
        # Now make a list of full filenames
        for iFile in range(len(filenames)) :
            fullnames.append(self.datadir + "/" + filenames[iFile])


        numDirs = len(filenames)

        self.numProjs = 0

        self.projs = []
        self.nFrames = []

        for iDir in range(numDirs) :
            dirloc = filenames[iDir]
            nameLength = len(dirloc)

            pname = dirloc[0:nameLength-6]
            dirname = self.datadir + "/" + pname +"_float";

            numLoc = self.count_files_serial(dirname, ".h5", True)

            if(numLoc > 0) :
                self.numProjs += 1
                self.projs.append(pname)
                self.nFrames.append(numLoc)

        print " Reduced num of projects: " + str(self.numProjs)

        for iProj in range(self.numProjs) :
            message = " Project: " + self.projs[iProj] + " with " + str(self.nFrames[iProj])
            if(self.nFrames[iProj] > 1) :
                message += " files"
            else :
                message += " file"
            print message

        return self.numProjs


    def scan_tsteps(self, i_proj, progress=None) :
        # determine available timesteps
        float_dir = self.datadir + "/";
        float_dir += self.projs[i_proj];
        float_dir += "_float/";
        print " Reading files in dir: " + float_dir

        # Now we get all available files in the float-data dir -- files
        # are saved in vector "files"
        self.numFiles = self.count_files_serial(float_dir, ".h5", True);

        # clear lists
        self.files_steps = [];
        self.files_times = [];
        self.filenames = [];
        times_loc = [];
        pos = np.array(range(self.numFiles), dtype='i')

        self.files.sort()


        # Containing meta-data about all hdf5 files 
        h5_metas = []

        # Array 'files' holds all filenames:
        for ifile in range(self.numFiles) :
            if not progress is None : progress(ifile, self.numFiles)
            filename = float_dir + self.files[ifile];
            
            h5in = None
            try:
                # Open the file
                h5in = h5py.File(filename, 'r');

                # open data group
                h5group = h5in['Data']
    
                time = 0.;
                #                node = "time"
                #                if node in h5group.keys():
                time = h5group.attrs['time']
                timestep = 0;
                #                node = "timestep" 
                #                if node in h5group.keys():
                timestep = h5group.attrs["timestep"];
                times_loc.append(time)
                
                h5_meta = (timestep, time, self.files[ifile])
                h5_metas.append(h5_meta)
                
            except IOError as e:
                print "IOError opening file '" + str(filename) + "': " + str(e)
                raise e
            finally:
                if not h5in is None : h5in.close()

        # Sort meta by their timestep (which is on position 0)
        h5_metas.sort(key = lambda x : x[0])
        for ifile in range(len(h5_metas)):
            (timestep, time, filename) = h5_metas[ifile]
            self.filenames.append(filename)
            self.files_times.append(time);
            self.files_steps.append(timestep);
        
        return len(h5_metas)
        
        ##Deprated version
        '''
        # Now sort the stuff...
        pos[0] = 0;

        for ifile in range(self.numFiles) :
            pos[ifile] = 0

        for ifile in range(1, self.numFiles) :

            jfile=0;

            while(times_loc[ifile] > times_loc[pos[jfile]] and jfile<ifile) :
                print " other file " + str(jfile) + " " + str(ifile)
                jfile += 1

            print " Found jfile: " + str(jfile);
            for ipos in range(ifile, jfile, -1) :
                pos[ipos] = pos[ipos-1];
                print " Copy " + str(ipos-1) + " " + str(ipos)

            pos[jfile] = ifile;
        


        # Now store filenames, timesteps and times in correct order
        for ifile in range(self.numFiles) :

            filenum = pos[ifile];
            print " pos " + str(pos[ifile])
                
            # Open file:
            filename = float_dir + self.files[filenum];

            self.filenames.append(self.files[filenum]);

            # Open the file
            h5in = None
            try:
                h5in = h5py.File(filename, 'r');

                # open data group
                h5group = h5in['Data']
                time = 0.;
                time = h5group.attrs['time']
                
                self.files_times.append(time);
    #            print " time: " + str(ifile) + " " + str(time) + " " + filename
    
                tstep = 0
                tstep = h5group.attrs["timestep"];
                self.files_steps.append(tstep);
            finally:
                if not h5in is None : h5in.close()

        # Return number of timesteps
        return self.numFiles;
        '''


    def scan_DataGroups(self, i_proj, i_time) :
        #! Get number of groups in file

        # Get the directory
        float_dir = self.datadir + "/";
        float_dir += self.projs[i_proj] + "_float/";

        # Get filename:
        filename = float_dir + self.filenames[i_time];
        print " My filename " + filename

        # Now open the hdf5-file
        h5in = None
        try:
            h5in = h5py.File(filename,'r')
    
            # get number of groups
            allGroups = h5in.keys();
    
            self.fluidGroups = []
            self.numGroups = 0
            for iGroup in range(len(allGroups)) :
                self.fluidGroups.append(allGroups[iGroup])
                self.numGroups += 1;
        finally:
            if not h5in is None : h5in.close()

        return self.numGroups

    #        if(self.numGroups == 1) :
    #            h5fluidGroup = allGroups[0]
    #        else :
    #            h5fluidGroup = allGroups[1]

    def scan_TestCase(self, i_proj, i_time) :
        
       # Get the directory
        float_dir = self.datadir + "/";
        float_dir += self.projs[i_proj] + "_float/";

        # Get filename:
        filename = float_dir + self.filenames[i_time];
        print " My filename " + filename

        test_case = "none"
        
        print " Doing weird stuff "

        # Now open the hdf5-file
        h5in = None
        try:
            h5in = h5py.File(filename,'r')
            # open global group
            h5group = h5in['Data']
#            print h5group.attrs.keys()
            # check wheter is a testcase
            node = "TestCase"
            if node in h5group.attrs.keys():
                print " found a testcase "
                test_case = h5group.attrs['TestCase']

        except IOError as e:
            print "IOError opening file '" + str(filename) + "': " + str(e)
            raise e
        finally:
            if not h5in is None : h5in.close()

        return test_case

    def scan_DataFields(self, i_proj, i_time, i_group) :
        #! Generate the list of fields available in hdf5 file
        #!
        #  We begin by obtaining all names of the data fields
        #
        

        # Get the directory
        float_dir = self.datadir + "/";
        float_dir += self.projs[i_proj] + "_float/";

        # Get filename:
        filename = float_dir + self.filenames[i_time];
        print " My filename " + filename

           # Now open the hdf5-file
        h5in = None
        try:
            h5in = h5py.File(filename,'r')
            # Open the overall group
            h5group = h5in['Data']
    
            h5fluidGroup = h5in[self.fluidGroups[i_group]]
    
            print " the Group "
            print self.fluidGroups[i_group]
            print " "
                
    
            # Get number of datasets
            self.numDatasets = 0
            self.numDatasets = h5fluidGroup.attrs['Entries']
    
            # Get number of other datasets
            numOtherDatasets = 0
            if(i_group > 1) :
                print " In thingy "
                print " "
                for i_other in range(1, i_group) :
                    print " Other " + str(i_other)
                    print " "
                    h5otherGroup = h5in[self.fluidGroups[i_other]]
                    numOtherDatasets += h5otherGroup.attrs['Entries']
            print " Number of other fields " + str(numOtherDatasets) + " " + str(i_group)
            
    
            # Get all field names:
            self.fieldNames = [];
    
            print " num "
            print self.numDatasets
            print " "
    
            
    #        for i_field in range(0, self.numDatasets) :
            for i_field in range(numOtherDatasets,
                                 self.numDatasets+numOtherDatasets) :
                
                print str(i_field) + " " + str(h5fluidGroup)
                fieldName = self.GetDatasetName(h5fluidGroup, i_field);
                self.fieldNames.append(fieldName);
    
                
            # Write field names:
            for i_field in range(0, self.numDatasets) :
                print " Field: " + str(i_field) + " --> " + self.fieldNames[i_field];

            h5DataGroup = h5in['Data']

            try:
                # Get Particles stuff
                self.numParticles = 0
                self.particleNames = []
                self.nParticleSpectrumBins = []
                self.momentumBinCenters = []

                self.numParticles = h5DataGroup.attrs['nParticles']

                h5ParticleGroup = h5DataGroup['Particles']

                print("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  adding particles")
                print(self.numParticles)

                for i_particle in range( self.numParticles):
                    particleID = 'particle%02d' % i_particle

                    particleName = h5ParticleGroup.attrs[particleID + '_type']
                    self.particleNames.append(particleName)
                    print('Particle', i_particle, '-->', particleName)

                    self.nParticleSpectrumBins.append(h5ParticleGroup.attrs[particleID + '_bins'])

                    dataset = h5ParticleGroup['momentumCenters_%02d' % i_particle]
                    self.momentumBinCenters.append(dataset[:])
            except:
                pass

            # Get Fluid Type
            try:
                self.fluid_type = h5DataGroup.attrs['fluid_type']
            except:
                self.fluid_type = "hydro"



        finally:
            if not h5in is None : h5in.close()

        return self.numDatasets


    def get_Geometry(self, i_proj, i_time):
        filename = self.make_filename(i_proj, i_time);

        # Now open the hdf5-file
        h5in = None
        try:
            h5in = h5py.File(filename, 'r');
    
            # open data group
            h5group = h5in['Data']
            
            geom = h5group.attrs["geometry"]
            return int(geom)
            
        except KeyError:
            return -1
        except ValueError:
            return -1
            
        finally:
            if not h5in is None: h5in.close()


    def get_Attribute(self, i_proj, i_time, gname, aname) :
        filename = self.make_filename(i_proj, i_time);

        # Now open the hdf5-file
        h5in = None
        try:
            h5in = h5py.File(filename, 'r');
    
            # open data group
            h5group = h5in['Data']
            mygroup = h5group[gname]

            attr = mygroup.attrs[aname]
            return attr
            
        except KeyError:
            return -1
        except ValueError:
            return -1
            
        finally:
            if not h5in is None: h5in.close()

    def get_ArraySizes(self, i_proj, i_time, i_field, i_group) :
        #! Determine extent of arrays
        #! Here we load all properties of the arrays - like, e.g., the
        #  gridpoint-size and the spatial extent.

        # Make filename
        filename = self.make_filename(i_proj, i_time);

        # Now open the hdf5-file
        h5in = None
        try:
            h5in = h5py.File(filename, 'r');
    
            # open data group
            h5group = h5in['Data']
            # open fluid group
            h5fluidGroup = h5in[self.fluidGroups[i_group]]
    
            # Pick array in the file:
            dset = h5fluidGroup[self.fieldNames[i_field]]
    
            # Get attributes of dataset
            # Get extent of array
            Nx_inv = dset.shape
            xb = dset.attrs["origin"]
            dx = dset.attrs["delta"]
#            print " Origin: "
#            print xb
    
            # Save corrected array extent
            self.Nx = [1]*3
            self.Nx[0] = Nx_inv[2]
            self.Nx[1] = Nx_inv[1]
            self.Nx[2] = Nx_inv[0]

            # check whether a nonlinear grid has been used:
            nonLin = "NonlinGrid" in h5group
            if(nonLin) :
                h5GridGroup = h5group["NonlinGrid"];

    
            # Use extent info to build position arrays:
            self.xPosC = np.array(range(self.Nx[0]), dtype='f')
            self.xPosL = np.array(range(self.Nx[0]+1), dtype='f')

            if(nonLin) :
                dataset = h5GridGroup["xCentres"]
                self.xPosC = dataset[:]
                dataset = h5GridGroup["xEdgesL"]
                self.xPosL = dataset[:]

            else :
                for ix in range(self.Nx[0]) :
                    self.xPosC[ix] = xb[0] + dx[0]*ix
                    self.xPosL[ix] = xb[0] + dx[0]*ix -0.5*dx[0]
                self.xPosL[self.Nx[0]] = self.xPosL[self.Nx[0]-1]+dx[0]
    
            self.yPosC = np.array(range(self.Nx[1]), dtype='f')
            self.yPosL = np.array(range(self.Nx[1]+1), dtype='f')

            if(nonLin) :
                dataset = h5GridGroup["yCentres"]
                self.yPosC = dataset[:]
                dataset = h5GridGroup["yEdgesL"]
                self.yPosL = dataset[:]

            else :
                for iy in range(self.Nx[1]) :
                    self.yPosC[iy] = xb[1] + dx[1]*iy
                    self.yPosL[iy] = xb[1] + dx[1]*iy - 0.5*dx[1]
                self.yPosL[self.Nx[1]] = self.yPosL[self.Nx[1]-1]+dx[1]
    

            self.zPosC = np.array(range(self.Nx[2]), dtype='f')
            self.zPosL = np.array(range(self.Nx[2]+1), dtype='f')

            if(nonLin) :
                dataset = h5GridGroup["zCentres"]
                self.zPosC = dataset[:]
                dataset = h5GridGroup["zEdgesL"]
                self.zPosL = dataset[:]

            else :
                for iz in range(self.Nx[2]) :
                    self.zPosC[iz] = xb[2] + dx[2]*iz
                    self.zPosL[iz] = xb[2] + dx[2]*iz - 0.5*dx[2]
                self.zPosL[self.Nx[2]] = self.zPosL[self.Nx[2]-1]+dx[2]

        finally:
            if not h5in is None : h5in.close()
        
        # Now we return a vector just holding the number of gridpoints:
        return self.Nx;

    def get_groupName(self, i_group) :
        return self.fluidGroups[i_group]
        

    def get_pname(self, iproj) :
        if(iproj >= 0 and iproj < self.numProjs) :
            return self.projs[iproj];
        else :
            print " no such iproj %s / %s" % (iproj, self.numProjs)
            sys.exit();

            
    def get_nFrames(self, iproj) :
        if(iproj >= 0 and iproj < self.numProjs) :
            print " my num " + str(self.nFrames[iproj])
            return self.nFrames[iproj];
        else :
            print " no such iproj "
            sys.exit();

    def get_numFields(self) :
        return self.numDatasets
        
    def get_FieldName(self, i_field) :
        #! Return name of chosen dataset.
        return self.fieldNames[i_field];

    def get_FluidType(self):
        return self.fluid_type

    def get_numParticles(self):
        return self.numParticles

    def get_ParticleName(self, i_particle):
        # ! Return name of chosen dataset.
        return self.particleNames[i_particle]

    def get_ParticleNames(self):
        # ! Return name of chosen dataset.
        return self.particleNames

    def get_time(self, i_time) :
        #! Return time saved in hdf5-file
        return self.files_times[i_time]

    def get_step(self, i_time) :
        #! Return timestep saved in hdf5-file
        return self.files_steps[i_time];


    def get_pos(self, i_plane, i_pos) :
        if(i_plane == 0) : # x-y plane:
            return self.zPosC[i_pos];
        elif (i_plane == 1) : # x-z plane
            return self.yPosC[i_pos];
        else : # y-z plane
            return self.xPosC[i_pos];

    def get_PosArrayLeft(self, dir) :
        if(dir==0) :
            PosArr = self.xPosL;
        elif (dir==1) :
            PosArr = self.yPosL
        elif (dir==2) :
            PosArr = self.zPosL
        else :
            print " No such direction " + str(dir)
            sys.exit();
        return PosArr

    def get_PosArrayCen(self, dir) :
        if(dir==0) :
            PosArr = self.xPosC;
        elif (dir==1) :
            PosArr = self.yPosC
        elif (dir==2) :
            PosArr = self.zPosC
        else :
            print " No such direction " + str(dir)
            sys.exit();
        print " "
        print " got pos for " + str(dir)
        print self.Nx[1]
        return PosArr

    def get_nParticleSpectrumBins(self, i_particle):
        return self.nParticleSpectrumBins[i_particle]

    def get_MomentumBinCenter(self, i_particle, i_momentum):
        return self.momentumBinCenters[i_particle][i_momentum]

    def get_MomentumBinCenters(self, i_particle):
        return self.momentumBinCenters[i_particle]

    def get_NormValue(self) :
        return self.normFactor
        

    def get_NormUnit(self) :
        return self.normUnit

    def scan_Norm(self, i_proj, i_time, i_dataset, i_group) :
        # Open the file
        # Set up filename:
        pname = self.projs[i_proj]; # Get name of project

        # directory of project
        proj_dir = self.datadir + "/" + pname + "_float/";
        
        # filename
        filename = proj_dir + self.filenames[i_time]

        self.normUnit = "nil"
        self.normFactor = 1.;

        # Now open the hdf5-file
        h5in = None
        try:
            h5in = h5py.File(filename,'r')
            # open data group
            h5group = h5in[self.fluidGroups[i_group]]
            # Get name of dataset:
            dset_name = self.fieldNames[i_dataset]
            dset = h5group[dset_name]
            # check wheter has normalisation
            node = "NormFactor"
            if node in dset.attrs.keys():
                print " found normalisation "
                self.normFactor = dset.attrs['NormFactor']
                self.normUnit = dset.attrs['NormUnit']
                print " norm factor: " + str(self.normFactor)
            else :
                print " no normalisation found "
                self.normFactor = 1.
                self.normUnit = 'nil'
        except IOError as e:
            print "IOError opening file '" + str(filename) + "': " + str(e)
            raise e
        finally:
            if not h5in is None : h5in.close()

        return self.normFactor


    def Load_2DData(self, i_proj, i_time, i_dataset, i_plotPlane, i_slice, i_group, progress = None) :

        # Open the file
        # Set up filename:
        pname = self.projs[i_proj]; # Get name of project

        # directory of project
        proj_dir = self.datadir + "/" + pname + "_float/";
        
        # filename
        filename = proj_dir + self.filenames[i_time]

        # Open the file
        h5in = None
        try:
            h5in = h5py.File(filename, 'r');
    
            # open data group
            h5group = h5in[self.fluidGroups[i_group]]
            print " Reading from group " + str(i_group) + " " + self.fluidGroups[i_group]
            print " "
            print " "
    #        h5group = h5in['Data']
    
            # Get name of dataset:
            dset_name = self.fieldNames[i_dataset]
    
            # Get the cut-data
            data = self.Read2DFrom3DMatrix(h5group, dset_name,
                                           i_plotPlane, i_slice);
        finally:
            if not h5in is None : h5in.close()

#        print " More of data " + " " + str(data.shape) +  " " + str(data[2,3])
           
        return data   


    def Load_1DData(self, i_proj, i_time, i_dataset, i_dir, i_posPerp, i_group) :

        # Open the file
        # Set up filename:
        pname = self.projs[i_proj]; # Get name of project

        # directory of project
        proj_dir = self.datadir + "/" + pname + "_float/";
        
        # filename
        filename = proj_dir + self.filenames[i_time]

        # Open the file
        h5in = None
        try:
            h5in = h5py.File(filename, 'r');
    
            # open data group
    #        h5group = h5in['Data']
            h5group = h5in[self.fluidGroups[i_group]]
            print " My group is " + self.fluidGroups[i_group]
            print " "
            print str(i_dataset) + " " + str(len(self.fieldNames))
            print " "
    
            # Get name of dataset:
            dset_name = self.fieldNames[i_dataset]
    
            # Get the cut-data
            data = self.Read1DFrom3DMatrix(h5group, dset_name,
                                           i_dir, i_posPerp);
    
    #        print " More of data " + " " + str(data.shape) +  " " + str(data[2,3])

        finally:
            if not h5in is None : h5in.close()

        return data


    def Load_Spectrum_Data(self, i_proj, i_time, i_particle, i_pos, i_group):
        # Open the file
        # Set up filename:
        pname = self.projs[i_proj]  # Get name of project

        # directory of project
        proj_dir = self.datadir + "/" + pname + "_float/"

        # filename
        filename = proj_dir + self.filenames[i_time]

        # Open the file
        h5in = None
        data = []

        try:
            h5in = h5py.File(filename, 'r')

            # open data group
            h5group = h5in[self.fluidGroups[i_group]]
            print(" Loading particle", i_particle, "spectrum at pos", i_pos)
            print(" My group is " + self.fluidGroups[i_group])
            print(self.nParticleSpectrumBins)

            # Get name of dataset:
            for i_bin in range(self.nParticleSpectrumBins[i_particle]):

                dset_name = "particle%02d_%02d" % (i_particle, i_bin)
                print(" Getting Data from", dset_name)

                # Get the cut-data
                data.append(self.Read0DFrom3DMatrix(h5group, dset_name, i_pos))
        finally:
            if not h5in is None: h5in.close()

        return np.asarray(data)


    def Read2DFrom3DMatrix(self, h5group, ArrayName, dir, posPerp) :

        # Open the dataset:
        dataset = h5group[ArrayName]
#        data = dset[0,2:10,1:9:3]

        # Get dimensions of dataset:
        DIM = 3;

        dims_out = dataset.shape
        dimhdf = len(dims_out)
        
        if not (dimhdf == DIM) :
            print " Wrong dimensionality of input data: "
            sys.exit();
         
        # Beware - dimensions are swapped
        # 2D array holding size
        mx = [1]*2

        if(dir==0) :
            mx[0] = dims_out[1];
            mx[1] = dims_out[2];
        elif (dir==1) : # x,z
            mx[0] = dims_out[0];
            mx[1] = dims_out[2];
        elif (dir==2) : # y,z
            mx[0] = dims_out[0];
            mx[1] = dims_out[1];
        else :
            print " Error no such direction "
            sys.exit();
        
        data = np.zeros((mx[0],mx[1]), dtype=np.float32)
        if(dir==0) :#/ x,y plane

            data = dataset[posPerp,:,:]

        elif (dir==1) : # x,z plane

#            data = np.zeros((mx[0],mx[1]), dtype=np.float32)
            data = dataset[:,posPerp,:]

        else :

 #           data = np.zeros((mx[0],mx[1]), dtype=np.float32)
            data = dataset[:,:,posPerp]


#        print " Size of data " + str(dir) + " " + str(data.shape) + " " + str(mx[0]) + " " + str(mx[1]) + " " + str(data[2,3])

        return data




    def Read1DFrom3DMatrix(self, h5group, ArrayName, dir, posPerp) :

        # Open the dataset:
        dataset = h5group[ArrayName]
#        data = dset[0,2:10,1:9:3]

        # Get dimensions of dataset:
        DIM = 3;

        dims_out = dataset.shape
        dimhdf = len(dims_out)
        
        if not (dimhdf == DIM) :
            print " Wrong dimensionality of input data: "
            sys.exit();
         
        # Beware - dimensions are swapped
        # 2D array holding size

        if(dir==0) :
            mx = dims_out[2];
        elif (dir==1) : # x,z
            mx = dims_out[1];
        elif (dir==2) : # y,z
            mx = dims_out[0];
        else :
            print " Error no such direction "
            sys.exit();
        
        data = np.zeros((mx), dtype=np.float32)
        if(dir==0) :#/ x,y plane

            data = dataset[posPerp[1],posPerp[0],:]

        elif (dir==1) : # x,z plane

#            data = np.zeros((mx[0],mx[1]), dtype=np.float32)
            data = dataset[posPerp[1],:,posPerp[0]]

        else :

 #           data = np.zeros((mx[0],mx[1]), dtype=np.float32)
            data = dataset[:,posPerp[1],posPerp[0]]


#        print " Size of data " + str(dir) + " " + str(data.shape) + " " + str(mx[0]) + " " + str(mx[1]) + " " + str(data[2,3])

        return data

    def Read0DFrom3DMatrix(self, h5group, ArrayName, pos):
        # Open the dataset:
        dataset = h5group[ArrayName]

        # Check dimensions of dataset:
        DIM = 3
        dims_out = dataset.shape
        dimhdf = len(dims_out)
        if not (dimhdf == DIM):
            print(" Wrong dimensionality of input data: ")
            sys.exit()

        return dataset[pos[2], pos[1], pos[0]]


    def count_files_serial(self, DirName, patt, from_end):
        
        if from_end == True:
            my_string = ".*" + patt + "\Z"
        else:
#            my_string = "\A" + patt + ".*"
            my_string = ".*" + patt + ".*"

        my_regexp = re.compile(my_string)

        coord_regexp = re.compile(".*_coord.*")

        numFiles = 0
        myNum = 0
        self.files = []
        
        # get all files first:
        for file in os.listdir(DirName):
            myNum += 1

            if self.string_has_regexp(file, my_regexp):
                if not self.string_has_regexp(file, coord_regexp):
                    numFiles += 1
                    self.files.append(file)

        return numFiles

                        
    def string_has_regexp(self, string, regexp):

        number = len(regexp.findall(string))
        if number == 1:
            return True
        else:
            return False
    def string_has_patt(self, string, patt):

        regexp = re.compile(patt)
        number = len(regexp.findall(string))
        if number == 1:
            return True
        else:
            return False


    def GetDatasetName(self, h5group, num) :

        AttrName = "Name_om" + "%2.2d" % num
        print " Attr: " + AttrName
        
        DatasetName = h5group.attrs[AttrName]

        return DatasetName

    def make_filename(self, i_proj, i_time) :
        #! Make the filename

        # Make the directory
        float_dir = self.datadir;
        float_dir += "/";
        float_dir += self.projs[i_proj];
        float_dir += "_float/";

        # Make filename:
        filename = float_dir + self.filenames[i_time];
        return filename;
