from PyQt4 import QtCore, QtGui, uic
import CronosFileHandler
import Parker 
import BetaLaw
import SodToro as Sod
import SRHD_Riemann
import re
import math
import sys
import os
import PlotDataStorage
import numpy as np
import matplotlib
import threading
matplotlib.use('Qt4Agg')
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib import colors, ticker
from itertools import cycle

import mpl_toolkits.axisartist.floating_axes as floating_axes
import  mpl_toolkits.axisartist.angle_helper as angle_helper
from matplotlib.projections import PolarAxes
from mpl_toolkits.axisartist.grid_finder import FixedLocator, MaxNLocator, \
     DictFormatter
from matplotlib.transforms import Affine2D

matplotlib.rcParams['ps.useafm'] = 'True'


class CronosWidgeteer(QtGui.QWidget):
    def __init__(self, parent=None,dataDir=None):
        QtGui.QWidget.__init__(self,parent)

        uic.loadUi(os.path.join(os.path.dirname(__file__), 'CronosWidget.ui'), self)
        
        if dataDir is None:
            self._datadir = "/home/dave/work/Cronos/data"
        else:
            self._datadir = dataDir

        self.i_time = 0
        self.time_t = 0.0
        self.timestep = 0
        self.Proj_Geometry = -1

    def setupWidget(self, Widget):

        self.forFun = True

        # Make file handler object
        self.ProjHandler = CronosFileHandler.CronosFileHandler()
        self.ProjHandler.set_dataDir(self._datadir)

        self.connect(self.butt_Exit, QtCore.SIGNAL('clicked()'),
                     QtGui.qApp, QtCore.SLOT('quit()'))
        self.connect(self.butt_ScanFolder, QtCore.SIGNAL('clicked()'),
                     self.scanProjs)

        self.connect(self.butt_SetDataDir, QtCore.SIGNAL('clicked()'),
                     self.setDataDir)

        # Scan for timesteps and fields
        self.connect(self.butt_ScanFrames, QtCore.SIGNAL('clicked()'),
                     self.scanTSteps);


        # Get chosen project
        self.connect(self.list_projs, QtCore.SIGNAL('itemSelectionChanged()'),
                     self.set_iproj)

        # Selection of fluid (if necessary)
        self.connect(self.select_fluid, QtCore.SIGNAL('activated(int)'),
                     self.set_iFluid);
        self.numFluids = 1
#        self.set_iFluid(0);

        # Selection of field
        self.connect(self.select_fields, QtCore.SIGNAL('activated(int)'),
                     self.set_iQuant);
        self.i_quant = 0
        # Selection of mode
        self.connect(self.select_mod, QtCore.SIGNAL('activated(int)'),
                     self.set_iMod);
        # Selection of vec field
        self.connect(self.select_vecField, QtCore.SIGNAL('activated(int)'),
                     self.set_iVecField)
        self.i_vecField = 0

        # get chosen timestep
        self.spinBox_timestep.setRange(0, 0);
        self.slider_timestep.setRange(0, 0);
        self.connect(self.spinBox_timestep, QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_timestep, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_timestep, QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_timestep, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_timestep, QtCore.SIGNAL('valueChanged(int)'),
                     self.set_iTime);

        # Choice of plane
        self.connect(self.butt_plane_xy, QtCore.SIGNAL('toggled(bool)'),
                     self.set_xyPlane);
        self.connect(self.butt_plane_xz, QtCore.SIGNAL('toggled(bool)'),
                     self.set_xzPlane);
        self.connect(self.butt_plane_yz, QtCore.SIGNAL('toggled(bool)'),
                     self.set_yzPlane);
        self.butt_plane_xy.click()


        # Choice of slice
        self.spinBox_slice.setRange(0, 0);
        self.slider_slice.setRange(0, 0);
        self.connect(self.spinBox_slice, QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_slice, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_slice, QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_slice, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_slice, QtCore.SIGNAL('valueChanged(int)'),
                     self.set_iSlice);
        self.spinBox_slice.setEnabled(False);
        self.slider_slice.setEnabled(False);


        # Check if log is used
        self.connect(self.butt_linear, QtCore.SIGNAL('toggled(bool)'),
                     self.set_linear);
        self.butt_linear.click();


        # Selection of output format:
        self.connect(self.select_OutputFormat, QtCore.SIGNAL('activated(int)'),
                     self.set_OutputFormat);
        self.set_FormatOptions();
        self.set_OutputFormat(0); # x-window as default


         # Selection of contour format
        self.connect(self.butt_showCells, QtCore.SIGNAL('toggled(bool)'),
                     self.set_contourFormat);
        self.butt_showCells.click();


        # Selection of geometry type
        self.connect(self.butt_CoordCart, QtCore.SIGNAL('toggled(bool)'),
                     self.set_setGeomCartesian);
        self.connect(self.butt_CoordCyl, QtCore.SIGNAL('toggled(bool)'),
                     self.set_setGeomCylindrical);
        self.connect(self.butt_CoordSph, QtCore.SIGNAL('toggled(bool)'),
                     self.set_setGeomSpherical);
        self.butt_CoordCart.click();

        # available color tables
        self.available_colorTables = [m for m in plt.cm.datad if not m.endswith("_r")]

        # Selection of color table:
        self.spinBox_colorTable.setRange(0, len(self.available_colorTables) - 1);
        self.connect(self.spinBox_colorTable,
                     QtCore.SIGNAL('valueChanged(int)'), self.set_colorTable)
        self.spinBox_colorTable.setValue(self.available_colorTables.index('seismic'))


        # possibility to suppress all annotations
        self.connect(self.butt_noAnnotations2D, QtCore.SIGNAL('clicked()'),
                     self.set_noAnnotations2D)
        self.annotations2D = True
        self.connect(self.butt_noColorbar, QtCore.SIGNAL('clicked()'),
                     self.set_noColorbar)
        self.use_Colorbar = True

        
        # Possibility to set user data range:
        self.connect(self.butt_rangeCustom2D, QtCore.SIGNAL('toggled(bool)'),
                     self.toggle_rangeCustom2D)
        self.connect(self.line_zMin, QtCore.SIGNAL('editingFinished()'),
                     self.set_zMin);
        self.connect(self.line_zMax, QtCore.SIGNAL('editingFinished()'),
                     self.set_zMax);
        self.butt_rangeCustom2D.click();
        self.butt_rangeCustom2D.click();


        # Do a countour plot:
        self.connect(self.butt_plotContour, QtCore.SIGNAL('clicked()'),
                     self.do_plotContour);
        # Plot next Countour
        self.connect(self.butt_PlotCountourNext, QtCore.SIGNAL('clicked()'),
                     self.do_next_plotContour)
        # Do a flow plot:
        self.connect(self.butt_plotFlow, QtCore.SIGNAL('clicked()'),
                     self.do_plotFlow);
        

        # Initial scan:
        self.scanProjs();

        # default values
        self.i_slice = 0
        self.i_plotPlane = 0
        self.has_figure = False
        self.use_ownFormatCode = True

        # switch off some options
        self.butt_plane_xy.setEnabled(False);
        self.butt_plane_xz.setEnabled(False);
        self.butt_plane_yz.setEnabled(False);
        self.butt_plotFlow.setEnabled(False);

        self.select_vecField.setEnabled(False);


        # Buttons on line plot tab
        # Choice of direction
        self.connect(self.butt_Dir1D_x, QtCore.SIGNAL('clicked()'),
                     self.set_Dir1D_x);
        self.connect(self.butt_Dir1D_y, QtCore.SIGNAL('clicked()'),
                     self.set_Dir1D_y);
        self.connect(self.butt_Dir1D_z, QtCore.SIGNAL('clicked()'),
                     self.set_Dir1D_z);
        self.butt_Dir1D_x.click()
 #       self.i_lineDir = 0

        self.butt_Dir1D_x.setEnabled(False);
        self.butt_Dir1D_y.setEnabled(False);
        self.butt_Dir1D_z.setEnabled(False);

        # Choice of perpendicular position
        self.spinBox_slice1DPerp1.setRange(0, 0);
        self.slider_slice1DPerp1.setRange(0, 0);
        self.connect(self.spinBox_slice1DPerp1,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_slice1DPerp1, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_slice1DPerp1,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_slice1DPerp1, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_slice1DPerp1,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_perpPos1);
        self.spinBox_slice1DPerp1.setEnabled(False);
        self.slider_slice1DPerp1.setEnabled(False);

        self.spinBox_slice1DPerp2.setRange(0, 0);
        self.slider_slice1DPerp2.setRange(0, 0);
        self.connect(self.spinBox_slice1DPerp2,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_slice1DPerp2, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_slice1DPerp2,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_slice1DPerp2, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_slice1DPerp2,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_perpPos2);
        self.spinBox_slice1DPerp2.setEnabled(False);
        self.slider_slice1DPerp2.setEnabled(False);


        # Choice of plot range
        self.spinBox_xMin_1D.setRange(0, 0);
        self.slider_xMin_1D.setRange(0, 0);
        self.connect(self.spinBox_xMin_1D,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_xMin_1D, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_xMin_1D,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_xMin_1D, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_xMin_1D,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_xMin_1D);
        self.spinBox_xMin_1D.setEnabled(False);
        self.slider_xMin_1D.setEnabled(False);
        self.i_xMin_1D = 0

        self.spinBox_xMax_1D.setRange(0, 0);
        self.slider_xMax_1D.setRange(0, 0);
        self.connect(self.spinBox_xMax_1D,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_xMax_1D, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_xMax_1D,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_xMax_1D, QtCore.SLOT('setValue(int)'));
        self.connect(self.slider_xMax_1D,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_xMax_1D);
        self.spinBox_xMax_1D.setEnabled(False);
        self.slider_xMax_1D.setEnabled(False);
        self.i_xMax_1D = 0


        # Choice of line type
        self.connect(self.butt_1DPlotLine, QtCore.SIGNAL('toggled(bool)'),
                     self.set_1DPlotLine);
        self.connect(self.butt_1DPlotSymb, QtCore.SIGNAL('toggled(bool)'),
                     self.set_1DPlotSymb);
        self.connect(self.butt_1DPlotBoth, QtCore.SIGNAL('toggled(bool)'),
                     self.set_1DPlotBoth);

        self.butt_1DPlotLine.click()

        self.butt_plane_xy.click()



        self.i_posPerp = [0,0]

        # show analytical solution
        self.connect(self.butt_solutionTest, QtCore.SIGNAL('clicked()'),
                     self.set_showAnalytical1D)
        self.showAnalytical1D = False

        # use normalisation
        self.connect(self.butt_useSI, QtCore.SIGNAL('clicked()'),
                     self.butt_useSI2D, QtCore.SLOT('toggle()'));
        self.connect(self.butt_useSI2D, QtCore.SIGNAL('clicked()'),
                     self.butt_useSI, QtCore.SLOT('toggle()'));
        self.connect(self.butt_useSI, QtCore.SIGNAL('toggled(bool)'),
                     self.set_useSI)

        self.use_SI = False

        # use logarithmic x-axis
        self.connect(self.butt_logXAxis, QtCore.SIGNAL('clicked()'),
                     self.set_logXAxis)
        self.use_logXAxis = False

        # show gridlines
        self.connect(self.butt_showGrid, QtCore.SIGNAL('clicked()'),
                     self.set_showGrid)
        self.showGrid = False

        # show derivative
        self.connect(self.butt_useDerivative, QtCore.SIGNAL('clicked()'),
                     self.set_useDerivative)
        self.use_derivative = False

        # use powerlaw scale
        self.connect(self.butt_usePowerlawScale, QtCore.SIGNAL('clicked()'),
                     self.toggle_powerlawScale)
        self.connect(self.line_powerlawScaleExponent, QtCore.SIGNAL('editingFinished()'),
                     self.set_powerlawScaleExponent)
        self.toggle_powerlawScale()
        self.powerlawScaleExponent = 0

        # possibility to suppress all annotations
        self.connect(self.butt_noAnnotations1D, QtCore.SIGNAL('clicked()'),
                     self.set_noAnnotations1D)
        self.annotations1D = True


        # Do a stash plot
        self.connect(self.butt_addTo1DStash, QtCore.SIGNAL('clicked()'),
                     self.add_dataTo1DStash)
        self.connect(self.butt_clear1DStash, QtCore.SIGNAL('clicked()'),
                     self.clear1DStash)
        self.connect(self.butt_plot1DStash, QtCore.SIGNAL('clicked()'),
                     self.do_plotLineStash);
        self.connect(self.butt_plot1DRelStash, QtCore.SIGNAL('clicked()'),
                     self.do_1DstashRelPlot);
        self.stash1D = []


        # Do a single line plot:
        self.connect(self.butt_LinePlot, QtCore.SIGNAL('clicked()'),
                     self.do_plotLine);
        self.connect(self.butt_LinePlotNext, QtCore.SIGNAL('clicked()'),
                     self.do_plotLineNext);


        # combined picture
        self.connect(self.butt_combine_opt_add, QtCore.SIGNAL('clicked()'),
                     self.set_combineOpt)
        self.connect(self.butt_combine_opt_sub, QtCore.SIGNAL('clicked()'),
                     self.set_combineOpt)
        self.connect(self.butt_combine_opt_div, QtCore.SIGNAL('clicked()'),
                     self.set_combineOpt)
        self.connect(self.butt_combine_opt_mult, QtCore.SIGNAL('clicked()'),
                     self.set_combineOpt)
        self.connect(self.butt_combine_opt_rel, QtCore.SIGNAL('clicked()'),
                     self.set_combineOpt)

        self.combineOpt = "div"

        self.connect(self.butt_selectA, QtCore.SIGNAL('clicked()'),
                     self.set_combineImg_A)
        self.connect(self.butt_selectB, QtCore.SIGNAL('clicked()'),
                     self.set_combineImg_B)
        self.connect(self.butt_plotCombined, QtCore.SIGNAL('clicked()'),
                     self.do_plotContourCombine)

        self.stash2D_multi = [None, None]



        #####################################
        #                                   #
        #           SPECTRUM TAB            #
        #                                   #
        #####################################

        # Choice of position
        self.i_SpecPos = [-1, -1, -1]

        ## x - axis
        self.spinBox_spectrum_point_x.setRange(0, 0)
        self.slider_spectrum_point_x.setRange(0, 0)

        self.spinBox_spectrum_point_x.setEnabled(False)
        self.slider_spectrum_point_x.setEnabled(False)

        self.connect(self.spinBox_spectrum_point_x,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_spectrum_point_x, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_point_x,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_spectrum_point_x, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_point_x,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_spectrum_position_x)

        ## y - axis
        self.spinBox_spectrum_point_y.setRange(0, 0)
        self.slider_spectrum_point_y.setRange(0, 0)

        self.spinBox_spectrum_point_y.setEnabled(False)
        self.slider_spectrum_point_y.setEnabled(False)

        self.connect(self.spinBox_spectrum_point_y,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_spectrum_point_y, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_point_y,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_spectrum_point_y, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_point_y,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_spectrum_position_y)

        ## z - axis
        self.spinBox_spectrum_point_z.setRange(-1, -1)
        self.slider_spectrum_point_z.setRange(-1, -1)

        self.spinBox_spectrum_point_z.setEnabled(False)
        self.slider_spectrum_point_z.setEnabled(False)

        self.connect(self.spinBox_spectrum_point_z,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_spectrum_point_z, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_point_z,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_spectrum_point_z, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_point_z,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_spectrum_position_z)


        # Choice of plot range
        self.i_SpecEmin = 0
        self.i_SpecEmax = 0

        ## Emin
        self.spinBox_spectrum_Emin.setRange(0, 0)
        self.slider_spectrum_Emin.setRange(0, 0)

        self.spinBox_spectrum_Emin.setEnabled(False)
        self.slider_spectrum_Emin.setEnabled(False)

        self.connect(self.spinBox_spectrum_Emin,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_spectrum_Emin, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_Emin,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_spectrum_Emin, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_Emin,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_SpecEmin)

        ## Emax
        self.spinBox_spectrum_Emax.setRange(0, 0)
        self.slider_spectrum_Emax.setRange(0, 0)

        self.spinBox_spectrum_Emax.setEnabled(False)
        self.slider_spectrum_Emax.setEnabled(False)

        self.connect(self.spinBox_spectrum_Emax,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.slider_spectrum_Emax, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_Emax,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.spinBox_spectrum_Emax, QtCore.SLOT('setValue(int)'))
        self.connect(self.slider_spectrum_Emax,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.set_SpecEmax)

        # particle selector
        self.connect(self.select_particle, QtCore.SIGNAL('activated(int)'),
                     self.set_iParticle)
        self.i_particle = 0


        # plot modifiers
        ## use logarithmic x-axis
        self.connect(self.butt_spectrum_logXAxis, QtCore.SIGNAL('clicked()'),
                     self.set_spectrum_logXAxis)
        self.use_spectrum_logXAxis = False

        ## show gridlines
        self.connect(self.butt_spectrum_showGrid, QtCore.SIGNAL('clicked()'),
                     self.set_spectrum_showGrid)
        self.spectrum_showGrid = False


        # Choice of line type
        self.connect(self.butt_spectrum_PlotLine, QtCore.SIGNAL('toggled(bool)'),
                     self.set_spectrum_PlotLine)
        self.connect(self.butt_spectrum_PlotSymb, QtCore.SIGNAL('toggled(bool)'),
                     self.set_spectrum_PlotSymb)
        self.connect(self.butt_spectrum_PlotBoth, QtCore.SIGNAL('toggled(bool)'),
                     self.set_spectrum_PlotBoth)

        self.butt_spectrum_PlotLine.click()



        # Do a stash plot
        self.specStash = []
        self.connect(self.butt_addTo1DStash_spectrum, QtCore.SIGNAL('clicked()'),
                     self.add_spectrumToStash)
        self.connect(self.butt_clear1DStash_spectrum, QtCore.SIGNAL('clicked()'),
                     self.clear_spectrum_Stash)
        self.connect(self.butt_plot1DStash_spectrum, QtCore.SIGNAL('clicked()'),
                     self.do_spectrum_plotLineStash)
        self.connect(self.butt_plot1DRelStash_spectrum, QtCore.SIGNAL('clicked()'),
                     self.do_spectrum_stashRelPlot)


        # Do a single line plot:
        self.connect(self.butt_LinePlot_spectrum, QtCore.SIGNAL('clicked()'),
                     self.do_spectrum_plotLine)
        self.connect(self.butt_LinePlotNext_spectrum, QtCore.SIGNAL('clicked()'),
                     self.do_spectrum_plotLineNext)




    ################################################
    # Begin -> Spectrum

    def set_spectrum_position_x(self, i_SpecPosX):
        print(" setting spectum position x " + str(i_SpecPosX))
        self.i_SpecPos[0] = i_SpecPosX

        # Obtain corresponding position:
        # Now write a string holding the position

        message = " x = "
        pos = self.ProjHandler.get_pos(2, self.i_SpecPos[0])
        message += str(pos)
        self.label_spectrum_point_x.setText(message)

    def set_spectrum_position_y(self, i_SpecPosY):
        print(" setting spectum position y " + str(i_SpecPosY))
        self.i_SpecPos[1] = i_SpecPosY

        # Obtain corresponding position:
        # Now write a string holding the position

        message = " y = "
        pos = self.ProjHandler.get_pos(1, self.i_SpecPos[1])
        message += str(pos)
        self.label_spectrum_point_y.setText(message)

    def set_spectrum_position_z(self, i_SpecPosZ) :
        print(" setting spectum position z " + str(i_SpecPosZ))
        self.i_SpecPos[2] = i_SpecPosZ

        # Obtain corresponding position:
        # Now write a string holding the position

        message = " z = "
        pos = self.ProjHandler.get_pos(0, self.i_SpecPos[2])
        message += str(pos)
        self.label_spectrum_point_z.setText(message)

    def set_SpecEmin(self, i_SpecEmin):
        self.i_SpecEmin = i_SpecEmin

        # Obtain corresponding position:
        message = " Emin = "
        energy = self.ProjHandler.get_MomentumBinCenter(self.i_particle, self.i_SpecEmin)
        message += str(energy)
        self.label_spectrum_Emin.setText(message)

        if (self.i_SpecEmax < self.i_SpecEmin):
            self.spinBox_spectrum_Emax.setValue(self.i_SpecEmin)

    def set_SpecEmax(self, i_SpecEmax):
        self.i_SpecEmax = i_SpecEmax

        # Obtain corresponding position:
        message = " Emax = "
        energy = self.ProjHandler.get_MomentumBinCenter(self.i_particle, self.i_SpecEmax)
        message += str(energy)
        self.label_spectrum_Emax.setText(message)

        if (self.i_SpecEmax < self.i_SpecEmin):
            self.spinBox_spectrum_Emin.setValue(self.i_SpecEmax)


    # Plot Type

    def set_spectrum_PlotLine(self, checked):
        self.spectrum_PlotStyle = 0

    def set_spectrum_PlotSymb(self, checked):
        self.spectrum_PlotStyle = 1

    def set_spectrum_PlotBoth(self, checked):
        self.spectrum_PlotStyle = 2


    # Plot Modifiers

    def set_spectrum_logXAxis(self) :
        self.use_spectrum_logXAxis = self.butt_spectrum_logXAxis.isChecked()

    def set_spectrum_showGrid(self) :
        self.spectrum_showGrid = self.butt_spectrum_showGrid.isChecked()


    # Do Plots

    def do_spectrum_plotLine(self):
        self.WriteInfo(" Doing a line plot ")

        specStashSingle = []
        specStashSingle.append(self.LoadSpectrumData())

        # Now plot the data
        self.Plot1DStash(specStashSingle,
                         specStashSingle[0].xAxisName(),
                         specStashSingle[0].yAxisName(), True,
                    annotations1D=self.annotations1D,
                    use_powerlawScale=False,
                    powerlawScaleExponent=0.,
                    PlotStyle1D=self.spectrum_PlotStyle,
                    i_xMin=self.i_SpecEmin,
                    i_xMax=self.i_SpecEmax,
                    use_derivative=False,
                    showAnalytical1D=False,
                    test_case=None,
                    lin_plot=self.lin_plot,
                    use_logXAxis=self.use_spectrum_logXAxis,
                    showGrid=self.spectrum_showGrid
            )
        self.WriteInfo(" done (line plot) ")

    def do_spectrum_plotLineNext(self):
        i_time = self.slider_timestep.value()
        i_max_time = self.slider_timestep.maximum()

        i_time = i_time + 1
        if i_time > i_max_time: return
        self.slider_timestep.setValue(i_time)
        self.do_spectrum_plotLine()

    def do_spectrum_plotLineStash(self):
        self.WriteInfo(" Doing a line plot from stash ")
        self.Plot1DStash(self.specStash, self.specStash[0].xAxisName(),
                         self.specStash[0].yAxisName(), True,
                    annotations1D=self.annotations1D,
                    use_powerlawScale=False,
                    powerlawScaleExponent=0.,
                    PlotStyle1D=self.spectrum_PlotStyle,
                    i_xMin=self.i_SpecEmin,
                    i_xMax=self.i_SpecEmax,
                    use_derivative=False,
                    showAnalytical1D=False,
                    test_case=None,
                    lin_plot=self.lin_plot,
                    use_logXAxis=self.use_spectrum_logXAxis,
                    showGrid=self.spectrum_showGrid
            )
        self.WriteInfo(" done (line plots) ")

    def do_spectrum_stashRelPlot(self):
        self.WriteInfo(" Doing a relative plot from stash ")
        self.Plot1DStash(self.specStash, self.specStash[0].xAxisName(),
                         self.specStash[0].yAxisName(), False,
                    annotations1D=self.annotations1D,
                    use_powerlawScale=False,
                    powerlawScaleExponent=0.,
                    PlotStyle1D=self.spectrum_PlotStyle,
                    i_xMin=self.i_SpecEmin,
                    i_xMax=self.i_SpecEmax,
                    use_derivative=False,
                    showAnalytical1D=False,
                    test_case=None,
                    lin_plot=self.lin_plot,
                    use_logXAxis=self.use_spectrum_logXAxis,
                    showGrid=self.spectrum_showGrid
            )
        self.WriteInfo(" done (line plots) ")

    def add_spectrumToStash(self):
        self.WriteInfo(" Adding to spectrum stash ")
        self.specStash.append(self.LoadSpectrumData())
        self.WriteInfo(" done (add to spectrum stash)")

    def clear_spectrum_Stash(self):
        self.WriteInfo(" Clearing the stash ")
        self.specStash = []
        self.WriteInfo(" Done (Clearing the stash) ")


    # Loading Spectral Data


    def LoadSpectrumData(self):
        # Use just a single field
        print
        " Reading "
        data = self.ProjHandler.Load_Spectrum_Data(self.i_proj, self.i_time,
                                            self.i_particle,
                                            self.i_SpecPos,
                                            self.i_fluid)
        PlotName = "%s @ %s" % (self.ProjHandler.get_ParticleName(self.i_particle), self.i_SpecPos)

        # Todo Units Handling for spectra
        #if (self.use_SI):
        #    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
        #                                                self.i_dataset, self.i_fluid);
        #    data *= self.normValue

        # Now obtain position array:
        xPos = self.ProjHandler.get_MomentumBinCenters(self.i_particle)
        xName = " Momentum "
        yName = " Data "
        PlotName = " Line Plot "

        print("Plogetting datatting")

        # Now store stuff in stash
        Stash1D = PlotDataStorage.PlotData1D(xPos, data, xName,
                                             yName, PlotName)

        print("done")
        return Stash1D

    def set_iParticle(self, i_particle) :
        self.i_particle = i_particle
        print(" My Particle: " + str(self.i_particle))
        print(" corresponds to " + self.particleNames[self.i_particle])

        self.nParticleSpectrumBins = self.ProjHandler.get_nParticleSpectrumBins(self.i_particle)

        self.slider_spectrum_Emin.setRange(0, self.nParticleSpectrumBins - 1)
        self.spinBox_spectrum_Emin.setRange(0, self.nParticleSpectrumBins - 1)
        self.spinBox_spectrum_Emin.setValue(0)
        self.slider_spectrum_Emin.setEnabled(True)
        self.spinBox_spectrum_Emin.setEnabled(True)

        self.slider_spectrum_Emax.setRange(0, self.nParticleSpectrumBins - 1)
        self.spinBox_spectrum_Emax.setRange(0, self.nParticleSpectrumBins - 1)
        self.spinBox_spectrum_Emax.setValue(self.nParticleSpectrumBins - 1)
        self.slider_spectrum_Emax.setEnabled(True)
        self.spinBox_spectrum_Emax.setEnabled(True)

        self.set_SpecEmin(0)
        self.set_SpecEmax(self.nParticleSpectrumBins - 1)

    def setup_spectrumPosSlider(self):
        self.slider_spectrum_point_x.setRange(0, self.Nx[0] - 1)
        self.spinBox_spectrum_point_x.setRange(0, self.Nx[0] - 1)
        self.slider_spectrum_point_x.setEnabled(True)
        self.spinBox_spectrum_point_x.setEnabled(True)

        self.slider_spectrum_point_y.setRange(0, self.Nx[1] - 1)
        self.spinBox_spectrum_point_y.setRange(0, self.Nx[1] - 1)
        self.slider_spectrum_point_y.setEnabled(True)
        self.spinBox_spectrum_point_y.setEnabled(True)

        self.slider_spectrum_point_z.setRange(0, self.Nx[2] - 1)
        self.spinBox_spectrum_point_z.setRange(0, self.Nx[2] - 1)
        self.slider_spectrum_point_z.setEnabled(True)
        self.spinBox_spectrum_point_z.setEnabled(True)

        if self.i_SpecPos[0] < self.Nx[0] and self.i_SpecPos[0] > 0:
            self.spinBox_spectrum_point_x.setValue(self.i_SpecPos[0])
        else:
            self.spinBox_spectrum_point_x.setValue(self.Nx[0] / 2)

        if self.i_SpecPos[1] < self.Nx[1] and self.i_SpecPos[1] > 0:
            self.spinBox_spectrum_point_y.setValue(self.i_SpecPos[1])
        else:
            self.spinBox_spectrum_point_y.setValue(self.Nx[1] / 2)

        if self.i_SpecPos[2] < self.Nx[2] and self.i_SpecPos[2] > 0:
            self.spinBox_spectrum_point_z.setValue(self.i_SpecPos[2])
        else:
            self.spinBox_spectrum_point_z.setValue(self.Nx[2] / 2)




    def scan_particles(self):
        self.numParticles = self.ProjHandler.get_numParticles()
        self.particleNames = self.ProjHandler.get_ParticleNames()
        print("partNames", self.particleNames)
        self.select_particle.clear()

        if self.numParticles > 0:
            for name in self.particleNames:
                self.select_particle.addItem(name)

            if self.i_particle < self.numParticles:
                self.set_iParticle(self.i_particle)
            else:
                self.set_iParticle(0)

            self.select_particle.setCurrentIndex(self.i_particle)




    # END -> Spectrum
    ################################################


    def setDataDir(self):
        newDataDir = str(QtGui.QFileDialog.getExistingDirectory(self, "Select new data directory"))
        if newDataDir is None or len(newDataDir) == 0:
            return
        else:
            if not os.path.exists(newDataDir) : 
                self.lblStatus.setText("Directory not found")
                return
            else:
                # Set new datadir
                self._datadir = newDataDir
                self.ProjHandler.set_dataDir(newDataDir)
                self.scanProjs()



    def scanProjs(self) :
        self.WriteInfo(" Scanning for projects ...")
        self.n_proj = self.ProjHandler.scan_Projs();
        
        MaxFrames = 0
        for iproj in range(self.n_proj) :
            MaxFrames = max(MaxFrames, self.ProjHandler.get_nFrames(iproj));

        print " Max number of frames " + str(MaxFrames)
        self.lblStatus.setText("Status: Scanning ... ")
        self.prgScanProgress.setValue(0)
        self.prgScanProgress.setMaximum(self.n_proj)

        proj_list = []
        for iproj in range(self.n_proj) :
            self.prgScanProgress.setValue(iproj)
            list_entry = str(self.ProjHandler.get_nFrames(iproj))
            list_entry += "   "
            list_entry += self.ProjHandler.get_pname(iproj)
            proj_list.append(list_entry)
        self.lblStatus.setText("Scan complete (" + str(self.n_proj) + " projects)")
        self.prgScanProgress.setMaximum(100)
        self.prgScanProgress.setValue(0)
#        self.prgScanProgress.setValue(self.n_proj)

        # Add projects to the list
        self.list_projs.clear()
        for iproj in range(self.n_proj) :
            self.list_projs.addItem(proj_list[iproj])

        self.butt_ScanFrames.setEnabled(False);
        self.WriteInfo("...done (scan projs)");
        
        
    def scanFluids(self) :
        def progress(prg_value, prg_max):
            self.prgScanProgress.setValue(prg_value)
            self.prgScanProgress.setMaximum(prg_max)
            
        self.WriteInfo("Scanning for fluids ...");
        self.lblStatus.setText("Scanning for fluids ... ")
        # identify all available groups
        self.numFluids = self.ProjHandler.scan_DataGroups(self.i_proj, 0)
        # Set dafault fluid
        if(self.numFluids == 1) :
            self.i_fluid = 0;
        else :
            self.i_fluid = 1

            
        self.lblStatus.setText("Adding fluids ... ")
        self.select_fluid.clear();
        fluid_list = []
        if(self.numFluids > 1) :
            for i_fluid in range(1,self.numFluids) :
                list_entry = self.ProjHandler.get_groupName(i_fluid)
                # Add fluids to the list
                self.select_fluid.insertItem(i_fluid-1, list_entry);
                print " Adding: " + str(i_fluid-1) + " " + list_entry + " " + str(self.numFluids)
                print " "
                print " "

        else :
            list_entry = self.ProjHandler.get_groupName(0)
            self.select_fluid.insertItem(0, list_entry);


            #        self.select_fluid.setCurrentIndex(self.i_fluid)
            self.set_iFluid(0)


        # Now identify all available fields (use fluid 0 as standard)
        self.lblStatus.setText("Scanning data fields ... ")
        self.numFields = self.ProjHandler.scan_DataFields(self.i_proj, 0, self.i_fluid);
        print " Number of fields: " + str(self.numFields)

        # identify fluid type
        self.fluidType = self.ProjHandler.get_FluidType()
        print(" Fluid Type is", self.fluidType)


        # now check whether this is a test case
        self.test_case = self.ProjHandler.scan_TestCase(self.i_proj, 0);
        info = " Has test: " + str(self.test_case)
        if(self.test_case != 'none') :
            self.get_TestPars()
        self.WriteInfo(info)

        self.WriteInfo("...done (Scanning for fluids)");



    def get_TestPars(self) :
        info = " My case: " + str(self.test_case)
        self.WriteInfo(info)
        if(self.test_case == "isothermal Parker wind") :
            # need two parameters in this case:
            group_name = "ParkerPars"

            attr_name = "CriticalRadius"
            self.RCrit = self.ProjHandler.get_Attribute(self.i_proj, 0,
                                                        group_name, attr_name)

            attr_name = "cs"
            self.cs = self.ProjHandler.get_Attribute(self.i_proj, 0,
                                                     group_name, attr_name)
            print " Parker Pars: " + str(self.RCrit) + " " + str(self.cs)

            # prepare the test handler
            self.TestHandler = Parker.ParkerIso()
            self.TestHandler.set_pars(self.RCrit, self.cs)
        elif(self.test_case == "isothermal CAK wind") :
            # need three parameters in this case:
            group_name = "CAKPars"

            attr_name = "vterminal"
            self.vterm = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                        group_name, attr_name)
#            self.vterm = 436

            attr_name = "betaValue"
            self.betaValue = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                            group_name, attr_name)
            #            self.betaValue = 0.83
#            self.betaValue = 0.72
            print " Beta: " + str(self.betaValue)
            attr_name = "RStar"
            self.RStar = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                        group_name, attr_name)

            # prepare the test handler
            self.TestHandler = BetaLaw.BetaLawCAK()
            self.TestHandler.set_pars(self.vterm, self.betaValue, self.RStar)

            message = " Have value " + str(self.vterm)
            self.WriteInfo(message)
        elif(self.test_case == "hydrodynamical shock tube test") :
            # load parameters
            group_name = "SodPars"

            self.Sod_rhoL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                           group_name, "rhoL")
            self.Sod_uL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "uL")
            self.Sod_pL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "pL")

            self.Sod_rhoR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                           group_name, "rhoR")
            self.Sod_uR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "uR")
            self.Sod_pR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "pR")

            self.Sod_gamma = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                            group_name, "gamma")

            self.TestHandler = Sod.SodTest()
            self.TestHandler.set_pars(self.Sod_rhoL, self.Sod_pL, self.Sod_uL,
                                      self.Sod_rhoR, self.Sod_pR, self.Sod_uR, self.Sod_gamma, 0.)
        elif(self.test_case == "rhd 1d riemann problem") :
            # load parameters
            group_name = "RHDPars"

            self.RHD_rhoL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                           group_name, "rhoL")
            self.RHD_uxL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "uxL")
            self.RHD_utL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "utL")
            self.RHD_pL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "pL")

            self.RHD_rhoR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                           group_name, "rhoR")
            self.RHD_uxR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "uxR")
            self.RHD_utR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "utR")
            self.RHD_pR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "pR")

            self.RHD_gamma = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                            group_name, "gamma")
            if (self.RHD_uxL == -1):
				self.RHD_uxL = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "uL")
				self.RHD_utL = 0.

            if (self.RHD_uxR == -1):
				self.RHD_uxR = self.ProjHandler.get_Attribute(self.i_proj, self.i_time,
                                                         group_name, "uR")
				self.RHD_utR = 0.

            self.TestHandler = SRHD_Riemann.SRHD_Riemann()
            self.TestHandler.set_pars(self.RHD_rhoL, self.RHD_pL, self.RHD_uxL, self.RHD_utL,
                                      self.RHD_rhoR, self.RHD_pR, self.RHD_uxR, self.RHD_utR,
                                      self.RHD_gamma, True)

            
    

    def scanTSteps(self) :
        self.WriteInfo("Scanning for timesteps...");
        # Get chosen project:
        pname = self.ProjHandler.get_pname(self.i_proj);
        print " Using pname " + pname

        print " getting steps "

        def progress(prg_value, prg_max):
            self.prgScanProgress.setValue(prg_value)
            self.prgScanProgress.setMaximum(prg_max)

        self.lblStatus.setText("Status: Scanning ... ")
        self.n_times = self.ProjHandler.scan_tsteps(self.i_proj, progress);
        print " Got steps "
        # Fixing max number of timesteps for slider:
        self.spinBox_timestep.setRange(0, self.n_times-1);
        self.slider_timestep.setRange(0, self.n_times-1);
        # Set step=0 as default:
        self.is_NewProj = True
        if self.i_time < self.n_times:
            self.set_iTime(self.i_time)
        else:
            self.set_iTime(0)

        # Now identify all available fluids (using timestep 0 as standard)
        self.lblStatus.setText("Status: Scanning for fluids ... ")
        self.scanFluids();
        self.lblStatus.setText("Status: Scan complete ")
        self.prgScanProgress.setMaximum(100)
        self.prgScanProgress.setValue(100)
        self.WriteInfo("...done (Scanning timesteps)");
        

        

    def set_iproj(self) :
        # Get selection of project when clicked
        self.i_proj = self.list_projs.currentRow()
        if (self.i_proj >= self.n_proj):
            self.i_proj = 0
            return
        message = " Project selected: " + str(self.i_proj)
        self.WriteInfo(message);
        message = " chosen project: ";
        message += self.ProjHandler.get_pname(self.i_proj);
        self.butt_ScanFrames.setEnabled(True);
        # Do initial scan:
        self.scanTSteps()
        self.WriteInfo(message);


    def set_iTime(self, i_Time) :
        self.WriteInfo("Setting timestep to" + str(i_Time));
        self.i_time = i_Time

        # Obtain corresponding time:
        self.time_t = self.ProjHandler.get_time(self.i_time);
        self.timestep = self.ProjHandler.get_step(self.i_time)
        print " time is: " + str(self.time_t);

        # Now write a string holding the time:
        message = " Time: " + str(self.time_t)
        message += "  at step: " + str(self.timestep)

        print message
        self.info_timestep.setText(message)

        # Now identify all available fluids
        self.scanFluids();

 #       # Now identify all available files (use timestep 0 as standard)
 #       numFields = self.ProjHandler.scan_DataFields(self.i_proj, self.i_time, self.i_fluid);

        # Make list of all sensible fields:
        self.make_QuantList();

        print " New proj? " + str(self.is_NewProj)
        
        # Set default quantity:
        if(self.is_NewProj) :
            if self.i_quant < len(self.AllFields):
                self.set_iQuant(self.i_quant)
            else:
                print(" Reset iQuant ")
                self.set_iQuant(0)

        # Set direction
#        print " clicking on " + str(self.i_lineDir)
        if(self.i_lineDir == 0) :
            self.butt_Dir1D_x.click()
        elif(self.i_lineDir == 1) :
            self.butt_Dir1D_y.click()
        else :
            self.butt_Dir1D_z.click()

        self.setup_spectrumPosSlider()
        self.scan_particles()

        self.is_NewProj = False



    def set_iQuant(self, i_Quant) :
        #! Select quantity to plot (like, i.e., velocity
        self.i_quant = i_Quant;
        print " My Quantity: " + str(self.i_quant);
        print " corresponds to " + self.AllFields[self.i_quant]

        # Now determine possible sub-options:
        self.make_ModList()
        if hasattr(self, "i_mod") and self.i_mod < len(self.Mods):
            self.set_iMod(self.i_mod)
        else:
            print(" Reset iMod ")
            self.set_iMod(0)


    def set_iFluid(self, i_Fluid) :
        #! Select fluid to be used for plotting (for multifluid simulations)
        self.i_fluid = i_Fluid
        if(self.numFluids > 1) :
            self.i_fluid += 1
        print " Chose fluid is: " + str(self.i_fluid)

        # Now scan for available data fields:
        numFields = self.ProjHandler.scan_DataFields(self.i_proj, self.i_time, self.i_fluid);

        # Make list of all sensible fields:
        self.make_QuantList();


    def set_iMod(self, i_Mod) :
        self.i_mod = i_Mod;
        print " my mod: " + str(self.i_mod) + " " + self.Mods[self.i_mod]

        # Compute arr-index
        # Start with vectorial quantities:
        
        if(self.string_has_patt(self.AllFields[self.i_quant],"Velocity")) :
            if(self.string_has_patt(self.Mods[self.i_mod],"v_y")) :
                self.i_dataset = self.NameIDs["v_y"];
            elif (self.string_has_patt(self.Mods[self.i_mod],"v_z")) :
                self.i_dataset = self.NameIDs["v_z"];
            else :
                self.i_dataset = self.NameIDs["v_x"];
        elif(self.string_has_patt(self.AllFields[self.i_quant],"Momentum")) :
            if(self.string_has_patt(self.Mods[self.i_mod],"s_y")) :
                self.i_dataset = self.NameIDs["v_y"];
            elif (self.string_has_patt(self.Mods[self.i_mod],"s_z")) :
                self.i_dataset = self.NameIDs["v_z"];
            else :
                self.i_dataset = self.NameIDs["v_x"];
        elif(self.string_has_patt(self.AllFields[self.i_quant],"Mag. field")) :
            if(self.string_has_patt(self.Mods[self.i_mod],"B_y")) :
                self.i_dataset = self.NameIDs["B_y"];
            elif (self.string_has_patt(self.Mods[self.i_mod],"B_z")) :
                self.i_dataset = self.NameIDs["B_z"];
            else :
                self.i_dataset = self.NameIDs["B_x"];
        else :
            my_name = self.AllFields[self.i_quant];
            self.i_dataset = self.NameIDs[my_name];
        

        # Test if id is correctly identified
        print " You chose (at least): " + self.ProjHandler.get_FieldName(self.i_dataset);

        self.Nx = self.ProjHandler.get_ArraySizes(self.i_proj, self.i_time,
                                                  self.i_dataset, self.i_fluid);

        geom = self.ProjHandler.get_Geometry(self.i_proj, self.i_time);
        self.Proj_Geometry = geom

        # Enable plot choices
        self.butt_plane_xy.setEnabled(True);
        self.butt_plane_xz.setEnabled(True);
        self.butt_plane_yz.setEnabled(True);
        self.spinBox_slice.setEnabled(True);
        self.slider_slice.setEnabled(True);

        self.butt_Dir1D_x.setEnabled(True);
        self.butt_Dir1D_y.setEnabled(True);
        self.butt_Dir1D_z.setEnabled(True);
        self.spinBox_slice1DPerp1.setEnabled(True);
        self.slider_slice1DPerp1.setEnabled(True);
        self.spinBox_slice1DPerp2.setEnabled(True);
        self.slider_slice1DPerp2.setEnabled(True);

        self.slider_xMin_1D.setEnabled(True);
        self.spinBox_xMin_1D.setEnabled(True);
        self.slider_xMax_1D.setEnabled(True);
        self.spinBox_xMax_1D.setEnabled(True);

        # Set corresponding default values:
        self.set_iSlice(self.i_slice);
        self.spinBox_slice.setValue(self.i_slice);


        # Enable flow plot when chosen:
        if(self.Mods[self.i_mod] == " Flow ") :
            self.butt_plotFlow.setEnabled(True);
        else :
            self.butt_plotFlow.setEnabled(False);

        print " The number of gridpoints: " + str(self.Nx[0]) + " " + str(self.Nx[1]) + " " + str(self.Nx[2]);
#        print " slider select " + str(self.butt_plane_xz.isChecked())


        # Write data info into list widget:
        self.show_DataProperties();

        # Select geometry
        if self.Proj_Geometry == 1:     ## Cartesian
            self.butt_CoordCart.setChecked(True)
        elif self.Proj_Geometry == 2:   ## Cylindrical
            self.butt_CoordCyl.setChecked(True)
        elif self.Proj_Geometry == 3:   ## Spherical
            self.butt_CoordSph.setChecked(True)

        # Set standard extent for slider (standard is x-y plane):
        if(self.butt_plane_xy.isChecked()) :
            self.spinBox_slice.setRange(0,self.Nx[2]-1);
            self.slider_slice.setRange(0,self.Nx[2]-1);
        elif (self.butt_plane_xz.isChecked()) :
            self.spinBox_slice.setRange(0,self.Nx[1]-1);
            self.slider_slice.setRange(0,self.Nx[1]-1);
        else :
            self.spinBox_slice.setRange(0,self.Nx[0]-1);
            self.slider_slice.setRange(0,self.Nx[0]-1);
        

    def set_iVecField(self, i_VecField) :
        self.i_vecField = i_VecField;
        print " Chosen vector field: " + str(self.i_vecField) + " " + self.list_vecFields[self.i_vecField]

        if(self.i_vecField != 0) :
            self.butt_plotFlow.setEnabled(True);

        # Store necessary datasets
        if(self.list_vecFields[self.i_vecField] == " Velocity ") :
            self.i_dataset_x = self.NameIDs["v_x"]
            self.i_dataset_y = self.NameIDs["v_y"]
            self.i_dataset_z = self.NameIDs["v_z"]
            print " field components are " 
            print str(self.i_dataset_x) + " " + str(self.i_dataset_y) + " " + str(self.i_dataset_z)
        elif(self.list_vecFields[self.i_vecField] == " Mag. field ") :
            self.i_dataset_x = self.NameIDs["B_x"]
            self.i_dataset_y = self.NameIDs["B_y"]
            self.i_dataset_z = self.NameIDs["B_z"]


    def set_iSlice(self, i_Slice) :
        print " Setting slice to " + str(i_Slice);
        self.i_slice = i_Slice;

        # Obtain corresponding position:
        pos = self.ProjHandler.get_pos(self.i_plotPlane, self.i_slice);

        # Now write a string holding the position
        message = " Position: " + str(pos)
        self.info_slice.setText(message);

    def set_xyPlane(self, checked) :
        #! Do plot for x-y plane
        print " setting xy plane "
        self.i_plotPlane = 0;
        if(self.butt_plane_xy.isEnabled()) :
            self.spinBox_slice.setRange(0,self.Nx[2]-1);
            self.slider_slice.setRange(0,self.Nx[2]-1);
            # Initial value for slice
            self.i_slice = self.Nx[2]/2;
            self.set_iSlice(self.i_slice);
            self.spinBox_slice.setValue(self.i_slice);

    def set_xzPlane(self, checked) :
        #! Do plot for x-z plane
        print " setting xz plane "
        self.i_plotPlane = 1;
        if(self.butt_plane_xz.isEnabled()) :
            self.spinBox_slice.setRange(0,self.Nx[1]-1);
            self.slider_slice.setRange(0,self.Nx[1]-1);
            # Initial value for slice
            self.i_slice = self.Nx[1]/2;
            self.set_iSlice(self.i_slice);
            self.spinBox_slice.setValue(self.i_slice);


    def set_yzPlane(self, checked) :
        #! Do plot for y-z plane
        self.i_plotPlane = 2;
        if(self.butt_plane_yz.isEnabled()) :
            self.spinBox_slice.setRange(0,self.Nx[0]-1);
            self.slider_slice.setRange(0,self.Nx[0]-1);
            # Initial value for slice
            self.i_slice = self.Nx[0]/2;
            self.set_iSlice(self.i_slice);
            self.spinBox_slice.setValue(self.i_slice);

    def set_contourFormat(self, checked) :
        # Distinguish two types of contour plotting:
        self.cont_asCells = checked;



    def set_setGeomCartesian(self, checked) :
        self.geometry = 0

    def set_setGeomCylindrical(self, checked) :
        self.geometry = 1

    def set_setGeomSpherical(self, checked) :
        self.geometry = 2

    def make_QuantList(self) :
        #! Make a list of plotable quantities
        #!  From the list of fields obtained by the data-handler we
        #  extract a new list of modified quantities
        #

        # Get number of fields
#        numFields = self.ProjHandler.get_DataFields(self.i_proj, self.i_time);
        numFields = self.ProjHandler.get_numFields();
        
        # Now obtain a list of all available fields:
        name_fields = [];
        is_used = []
        
        for i_field in range(0, numFields) :
            name_fields.append(self.ProjHandler.get_FieldName(i_field))
            is_used.append(False);
        
        # Set all indicators to false:
        self.has_Density = False;
        self.has_Velocity = False;
        self.has_MagField = False;
        self.has_Etherm = False;

        self.AllFields = []
        self.NameIDs = {}

        print " "
        print " Old quant " + str(self.i_quant)


        for i_field in range(0, numFields) :


            print " My field: " + name_fields[i_field]

#            if(self.string_has_patt(name_fields[i_field], ".*rho.*")) :
            if(name_fields[i_field] == "rho") :
                print " Found density: " + name_fields[i_field]
                if not(self.has_Density) :
                    self.AllFields.append("Density")
                    
                self.has_Density = True
                self.NameIDs["Density"] = i_field

            elif(self.string_has_patt(name_fields[i_field], "^v_\w$")) :
#                elif(self.string_has_patt(name_fields[i_field], ".*v_.*")) :
                print " Found a velocity component: " + name_fields[i_field]
                if not(self.has_Velocity) :
                    self.AllFields.append("Velocity")
                    
                self.has_Velocity = True
                if(self.string_has_patt(name_fields[i_field], ".*v_x.*")) :
                    self.NameIDs["v_x"] = i_field
                elif(self.string_has_patt(name_fields[i_field], ".*v_y.*")) :
                    self.NameIDs["v_y"] = i_field
                elif(self.string_has_patt(name_fields[i_field], ".*v_z.*")) :
                    self.NameIDs["v_z"] = i_field

            elif(self.string_has_patt(name_fields[i_field], ".*B_.*")) :
                print " Found a mag component: " + name_fields[i_field]
                if not(self.has_MagField) :
                    self.AllFields.append("Mag. field")
                    
                self.has_MagField = True
                if(self.string_has_patt(name_fields[i_field], ".*B_x.*")) :
                    self.NameIDs["B_x"] = i_field
                elif(self.string_has_patt(name_fields[i_field], ".*B_y.*")) :
                    self.NameIDs["B_y"] = i_field
                elif(self.string_has_patt(name_fields[i_field], ".*B_z.*")) :
                    self.NameIDs["B_z"] = i_field

            elif(self.string_has_patt(name_fields[i_field], ".*Etherm.*")) :
                print " Found energy: " + name_fields[i_field]
                if not(self.has_Etherm) :
                    self.AllFields.append("Etherm")
                    
                self.has_Etherm = True
                self.NameIDs["Etherm"] = i_field

            else :
                self.AllFields.append(name_fields[i_field])
                self.NameIDs[name_fields[i_field]] = i_field


        # if possible: append momentum
        if(self.has_Density and self.has_Velocity) :
            self.AllFields.append("Momentum")

        # if possible: append SpecificEnergy
        if(self.has_Density and self.has_Etherm) :
            self.AllFields.append("SpecificEnergy")


        # Write full list:
        list_entries = len(self.AllFields)
        for i_field in range(list_entries) :
            print self.AllFields[i_field]

        # Now use the list of fields to generate choosable entries
        self.select_fields.clear();

        print " "
        print " OldVar quant " + str(self.i_quant)

        for i_sel in range(list_entries) :
            field = self.AllFields[i_sel];
            self.select_fields.insertItem(i_sel, field);

        self.select_fields.setCurrentIndex(self.i_quant)

        # Make an additional list of possible vectorfields
        self.make_vecFieldList()
        print " " 
        print " My vec " + str(self.i_vecField)
        self.set_iVecField(self.i_vecField);
        

    def make_ModList(self) :
        #! Generate list of possible modifiers
        #! Possible modifiers are e.g. abs values
         

        self.Mods = []

        # Walk through options for vector fields -- the rest is standard:
        if(self.string_has_patt(self.AllFields[self.i_quant],".*Velocity.*")) :
            # Test if all components are available
            allComp = False;
            if(self.NameIDs.has_key("v_x") and self.NameIDs.has_key("v_y") and
               self.NameIDs.has_key("v_z")) :
                allComp = True;

            if (self.fluidType == 'rhd'):
                if (allComp):
                    self.Mods.append(" Abs v ")
                    self.Mods.append(" Abs u ")
                    self.Mods.append(" Lorentz ")

                if (self.NameIDs.has_key("v_x")):
                    self.Mods.append(" v_x ")
                    self.Mods.append(" u_x ")

                if (self.NameIDs.has_key("v_y")):
                    self.Mods.append(" v_y ")
                    self.Mods.append(" u_y ")

                if (self.NameIDs.has_key("v_z")):
                    self.Mods.append(" v_z ")
                    self.Mods.append(" u_z ")

            else:
                if(allComp) :
                    self.Mods.append(" Abs ");

                if(self.NameIDs.has_key("v_x")) :
                    self.Mods.append(" v_x ");

                if(self.NameIDs.has_key("v_y")) :
                    self.Mods.append(" v_y ");

                if(self.NameIDs.has_key("v_z")) :
                    self.Mods.append(" v_z ");

                if(allComp) :
                    self.Mods.append(" Div ");
                    self.Mods.append(" Curl ");
                    self.Mods.append(" Flow ");

        elif(self.string_has_patt(self.AllFields[self.i_quant],".*Momentum.*")) :
            # Test if all components are available
            allComp = False;
            if(self.NameIDs.has_key("v_x") and self.NameIDs.has_key("v_y") and
               self.NameIDs.has_key("v_z") and self.NameIDs.has_key("Density")) :
                allComp = True;
            if(allComp) :
                self.Mods.append(" Abs ");

            if(self.NameIDs.has_key("v_x") and self.NameIDs.has_key("Density")) :
                self.Mods.append(" s_x ");

            if(self.NameIDs.has_key("v_y") and self.NameIDs.has_key("Density")) :
                self.Mods.append(" s_y ");
                
            if(self.NameIDs.has_key("v_z") and self.NameIDs.has_key("Density")) :
                self.Mods.append(" s_z ");
                

        elif (self.string_has_patt(self.AllFields[self.i_quant],".*Mag. field.*")) : # Mag. field
            # Test if all components are available
            allComp = False;
            if(self.NameIDs.has_key("B_x") and self.NameIDs.has_key("B_y") and
               self.NameIDs.has_key("B_z")) :
                allComp = True;
                
            if(allComp) :
                self.Mods.append(" Abs ");
                
            if(self.NameIDs.has_key("B_x")) :
                self.Mods.append(" B_x ");
                
            if(self.NameIDs.has_key("B_y")) :
                self.Mods.append(" B_y ");
                
            if(self.NameIDs.has_key("B_z")) :
                self.Mods.append(" B_z ");
                
            if(allComp) :
                self.Mods.append(" Div ");
                self.Mods.append(" Curl ");
                self.Mods.append(" Flow ");

                
        else : # others
            self.Mods.append(" None ");
            self.Mods.append(" Abs ");
        
        
        # Now use the list of options to generate choosable entries
        self.select_mod.clear();

        for i_sel in range(len(self.Mods)) :
            field = self.Mods[i_sel];
            self.select_mod.insertItem(i_sel, field);
        
    def make_vecFieldList(self) :
        #! Generate list of selectable vector fields:

        self.list_vecFields = []
        self.list_vecFields.append(" none ");
        
        # Test if all velocity components are available
        if(self.NameIDs.has_key("v_x") and self.NameIDs.has_key("v_y") and
           self.NameIDs.has_key("v_z")) :
            self.list_vecFields.append(" Velocity ");

        # Walk through options for vector fields -- the rest is standard:
        if(self.NameIDs.has_key("B_x") and self.NameIDs.has_key("B_y") and
           self.NameIDs.has_key("B_z")) :
            self.list_vecFields.append(" Mag. field ");

        # Now use the list of options to generate choosable entries
        self.select_vecField.clear();
        for i_field in range(len(self.list_vecFields)) :
            field_name = self.list_vecFields[i_field]
            self.select_vecField.insertItem(i_field, field_name);

        self.select_vecField.setEnabled(True);
        self.select_vecField.setCurrentIndex(self.i_vecField)

    def set_linear(self, checked) :
        self.lin_plot = checked;
        print " Doing a linear plot: " + str(self.lin_plot)


    def set_FormatOptions(self) :
        #! Set available output format options
        self.select_OutputFormat.insertItem(0, "x-window");
        self.select_OutputFormat.insertItem(1, "Postscript");
        self.select_OutputFormat.insertItem(2, "PNG");
        self.select_OutputFormat.insertItem(3, "PDF");
        self.select_OutputFormat.insertItem(4, "JPG");
        self.select_OutputFormat.insertItem(5, "SVG");

    def set_OutputFormat(self, outOption) :
         
        if(outOption==0) :
            self.fileFormatContour = "display";
        elif(outOption==1) :
            self.fileFormatContour = "ps";
        elif(outOption==2) :
            self.fileFormatContour = "png";
        elif(outOption==3) :
            self.fileFormatContour = "pdf";
        elif(outOption==4) :
            self.fileFormatContour = "jpeg";
        elif(outOption==5) :
            self.fileFormatContour = "svg";
        self.WriteInfo(" Output format is: " + self.fileFormatContour)



    def set_colorTable(self, colorOption) :
        self.i_color = colorOption
#        print " My color " + self.select_colorTable.currentText()

        self.name_colorTable = self.available_colorTables[self.i_color]

        print " My color table " + self.name_colorTable

        message = self.name_colorTable
        self.lineEdit_colorTable.clear();
        self.lineEdit_colorTable.insert(message);


    def LoadVecData(self, Stash2D) :
        #! Load vector data according to given field choice

        if(self.i_plotPlane==0) : # x-y plane
            xdata = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                 self.i_dataset_x,
                                                 self.i_plotPlane,
                                                 self.i_slice, self.i_fluid);
            ydata = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                 self.i_dataset_y,
                                                 self.i_plotPlane,
                                                 self.i_slice, self.i_fluid);
        elif(self.i_plotPlane==1) :
            xdata = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                 self.i_dataset_x,
                                                 self.i_plotPlane,
                                                 self.i_slice, self.i_fluid);
            ydata = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                 self.i_dataset_z,
                                                 self.i_plotPlane,
                                                 self.i_slice, self.i_fluid);
        else :
            xdata = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                 self.i_dataset_y,
                                                 self.i_plotPlane,
                                                 self.i_slice, self.i_fluid);
            ydata = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                 self.i_dataset_z,
                                                 self.i_plotPlane,
                                                 self.i_slice, self.i_fluid);
        # Now add vector data to stash:
        Stash2D.setVec(xdata, ydata)
            
            

    def LoadData(self, progress=None) :
        # Load chosen data files from data handler
        #! Here we load the data corresponding to the choices in the widget. In
        #  paricular for the vector-fields we need to load several arrays,
        #  e.g., for the abs value

        self.hasFlowVec = False
        xdata = 0
        ydata = 0
        PlotName = ""

        ## progress is a function taking two arguments
        ## If no progress is given we create a new function with no effect
        if progress is None : progress = lambda x,y : 0


        if(self.string_has_patt(self.AllFields[self.i_quant],"Velocity")) :
            if (self.fluidType == "rhd"):
                if (self.string_has_patt(self.Mods[self.i_mod], "Abs u") or
                        self.string_has_patt(self.Mods[self.i_mod], "Abs v") or
                        self.string_has_patt(self.Mods[self.i_mod], "Lorentz") or
                        self.string_has_patt(self.Mods[self.i_mod], "v_x") or
                        self.string_has_patt(self.Mods[self.i_mod], "v_y") or
                        self.string_has_patt(self.Mods[self.i_mod], "v_z")) :
                    progress(0, 3)
                    ux = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_x"],
                                                      self.i_plotPlane,
                                                      self.i_slice, self.i_fluid)
                    progress(1, 3)
                    uy = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_y"],
                                                      self.i_plotPlane,
                                                      self.i_slice, self.i_fluid)
                    progress(2, 3)
                    uz = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_z"],
                                                      self.i_plotPlane,
                                                      self.i_slice, self.i_fluid)
                    progress(3, 3)

                    u2 = ux * ux + uy * uy + uz * uz
                    if self.string_has_patt(self.Mods[self.i_mod], "Abs u"):
                        data = np.sqrt(u2)
                        PlotName = "u_Abs"
                    elif self.string_has_patt(self.Mods[self.i_mod], "Abs v"):
                        data = np.sqrt(u2 / (1 + u2))
                        PlotName = "v_Abs"
                    elif self.string_has_patt(self.Mods[self.i_mod], "Lorentz"):
                        data = np.sqrt(1 + u2)
                        PlotName = "Lorentz"
                    elif self.string_has_patt(self.Mods[self.i_mod], "v_x"):
                        data = ux / np.sqrt(1 + u2)
                        PlotName = "v_x"
                    elif self.string_has_patt(self.Mods[self.i_mod], "v_y"):
                        data = uy / np.sqrt(1 + u2)
                        PlotName = "v_y"
                    elif self.string_has_patt(self.Mods[self.i_mod], "v_z"):
                        data = uz / np.sqrt(1 + u2)
                        PlotName = "v_z"

                elif self.string_has_patt(self.Mods[self.i_mod], "u_x"):
                    data = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_x"],
                                                      self.i_plotPlane,
                                                      self.i_slice, self.i_fluid)
                    PlotName = "u_x"

                elif self.string_has_patt(self.Mods[self.i_mod], "u_y"):
                    data = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_y"],
                                                      self.i_plotPlane,
                                                      self.i_slice, self.i_fluid)
                    PlotName = "u_y"

                elif self.string_has_patt(self.Mods[self.i_mod], "u_z"):
                    data = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_z"],
                                                      self.i_plotPlane,
                                                      self.i_slice, self.i_fluid)
                    PlotName = "u_z"



            elif(self.string_has_patt(self.Mods[self.i_mod],"Abs") or
               self.string_has_patt(self.Mods[self.i_mod],"Div") or
               self.string_has_patt(self.Mods[self.i_mod],"Curl") or
               self.string_has_patt(self.Mods[self.i_mod],"Lorentz")) :

                progress(0,3)
                vx = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_x"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(1,3)
                vy = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_y"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(2,3)
                vz = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_z"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(3,3)
                data = np.copy(vx);


                if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :
                    data = np.sqrt(vx*vx + vy*vy + vz*vz)
                    PlotName = "v_Abs";
                elif(self.string_has_patt(self.Mods[self.i_mod],"Div")) :
                    if(self.i_plotPlane==1) :
                        vym = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                           self.NameIDs["v_y"],
                                                           self.i_plotPlane,
                                                           self.i_slice-1, self.i_fluid);
                        vyp = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                           self.NameIDs["v_y"],
                                                           self.i_plotPlane,
                                                           self.i_slice+1, self.i_fluid);
                        Nx1 = vx.shape[0]
                        Nz1 = vx.shape[0]

                        dx = self.ProjHandler.get_pos(0,1) - self.ProjHandler.get_pos(0,0)
                        dy = self.ProjHandler.get_pos(1,1) - self.ProjHandler.get_pos(1,0)
                        dz = self.ProjHandler.get_pos(2,1) - self.ProjHandler.get_pos(2,0)

                        for ix in range (1,Nx1-2) :
                            for iz in range (1,Nz1-2) :
                                dvxdx = (vx[iz,ix+1] - vx[iz,ix-1])/(2*dx)
                                dvydy = (vyp[iz,ix]  - vym[iz,ix])/(2*dy)
                                dvzdz = (vz[iz+1,ix] - vz[iz-1,ix])/(2*dz)


                                data[iz,ix] = dvxdx + dvydy + dvzdz

                        print Nx1
#                        data = np.sqrt(vx*vx + vy*vy + vz*vz)
                        PlotName = "div_v"
                    else :
                        print " Still working on this "
                        sys.exit()

                elif(self.string_has_patt(self.Mods[self.i_mod],"Div") or
                     self.string_has_patt(self.Mods[self.i_mod],"Curl")) :
                    print " Still working on this "
                    sys.exit()
                
                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.NameIDs["v_x"], self.i_fluid);
                    self.normUnit = self.ProjHandler.get_NormUnit()
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                    self.WriteInfo(message)

            elif(self.string_has_patt(self.Mods[self.i_mod],"Flow")) :
                progress(0,3)
                vx = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_x"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(1,3)
                vy = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_y"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(2,3)
                vz = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_z"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(3,3)

                data = np.sqrt(vx*vx + vy*vy + vz*vz)
                if(self.i_plotPlane==0) :
                    xdata = vx
                    ydata = vy
                elif(self.i_plotPlane==1) :
                    xdata = vx
                    ydata = vz
                else :
                    xdata = vy
                    ydata = vz

                PlotName = "Flow";
                self.hasFlowVec = True
                
            else :
                progress(0,1)
                # Use just a single field
                data = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                    self.i_dataset,
                                                    self.i_plotPlane,
                                                    self.i_slice, self.i_fluid);
                progress(1,1)
                PlotName = self.ProjHandler.get_FieldName(self.i_dataset);

                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.i_dataset, self.i_fluid);
                    self.normUnit = self.ProjHandler.get_NormUnit()
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                    self.WriteInfo(message)

        elif(self.string_has_patt(self.AllFields[self.i_quant],"Momentum")) :
            if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :

                progress(0,4)
                rho = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["Density"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(1,4)
                vx = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_x"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(2,4)
                vy = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_y"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(3,4)
                vz = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_z"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(4,4)
                data = np.copy(vx);


                if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :
                    data = np.sqrt(vx*vx + vy*vy + vz*vz)
                    data *= rho
                    PlotName = "s_Abs";
                elif(self.string_has_patt(self.Mods[self.i_mod],"Div") or
                     self.string_has_patt(self.Mods[self.i_mod],"Curl")) :
                    print " Still working on this "
                    sys.exit()
                
                if(self.use_SI) :
                    normValue1 = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                            self.NameIDs["Density"], self.i_fluid);
                    normUnit1 = self.ProjHandler.get_NormUnit()
                    normValue2 = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                            self.NameIDs["v_x"], self.i_fluid);
                    self.normValue = normValue1*normValue2
                    normUnit2 = self.ProjHandler.get_NormUnit()
                    self.normUnit = normUnit1 + " " + normUnit2
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                    self.WriteInfo(message)

            else :
                progress(0,2)
                # Use just a single field
                data = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                    self.i_dataset,
                                                    self.i_plotPlane,
                                                    self.i_slice, self.i_fluid);

                progress(1,2)
                rho = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["Density"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(2,2)
                data *= rho
                PlotName = self.ProjHandler.get_FieldName(self.i_dataset);

                if(self.use_SI) :
                    normValue1 = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                            self.NameIDs["Density"], self.i_fluid);
                    normUnit1 = self.ProjHandler.get_NormUnit()
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.i_dataset, self.i_fluid);
                    self.normValue *= normValue1
                    self.normUnit = self.ProjHandler.get_NormUnit()
                    self.normUni += " " + normUnit1
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                    self.WriteInfo(message)

        elif (self.string_has_patt(self.AllFields[self.i_quant], "SpecificEnergy")):

            progress(0, 2)
            rho = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                               self.NameIDs["Density"],
                                               self.i_plotPlane,
                                               self.i_slice, self.i_fluid)
            progress(1, 2)
            Etherm = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                              self.NameIDs["Etherm"],
                                              self.i_plotPlane,
                                              self.i_slice, self.i_fluid)
            progress(2, 2)

            data = Etherm / rho

            if (self.string_has_patt(self.Mods[self.i_mod], "Abs")):
                data = np.abs(data)

            if (self.use_SI):
                normValue1 = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                        self.NameIDs["Density"], self.i_fluid)
                normUnit1 = self.ProjHandler.get_NormUnit()
                normValue2 = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                        self.NameIDs["Etherm"], self.i_fluid)
                self.normValue = normValue2/normValue1
                normUnit2 = self.ProjHandler.get_NormUnit()
                self.normUnit = normUnit2 + "/ (" + normUnit1 + ")"
                data *= self.normValue
                message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                self.WriteInfo(message)




        elif(self.string_has_patt(self.AllFields[self.i_quant],"Mag. field")) :
            if(self.string_has_patt(self.Mods[self.i_mod],"Abs") or
               self.string_has_patt(self.Mods[self.i_mod],"Div") or
               self.string_has_patt(self.Mods[self.i_mod],"Curl")) :

                progress(0,3)
                Bx = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["B_x"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(1,3)
                By = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["B_y"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(2,3)
                Bz = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                  self.NameIDs["B_z"],
                                                  self.i_plotPlane,
                                                  self.i_slice, self.i_fluid);
                progress(3,3)
                data = np.copy(Bx);

                if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :
                    data = np.sqrt(Bx*Bx + By*By + Bz*Bz)
                    PlotName = "B_Abs";
                elif(self.string_has_patt(self.Mods[self.i_mod],"Div") or
                     self.string_has_patt(self.Mods[self.i_mod],"Curl")) :
                    print " Still working on this "
                    sys.exit()

                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.NameIDs["B_x"], self.i_fluid);
                    self.normUnit = self.ProjHandler.get_NormUnit()
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                    self.WriteInfo(message)

            else :
                # Use just a single field
                progress(0,1)
                data = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                    self.i_dataset,
                                                    self.i_plotPlane,
                                                    self.i_slice, self.i_fluid);
                progress(1,1)
                PlotName = self.ProjHandler.get_FieldName(self.i_dataset);

                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.i_dataset, self.i_fluid);
                    self.normUnit = self.ProjHandler.get_NormUnit()
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                    self.WriteInfo(message)
            
        else :
            # Use just a single field
            print " Reading "
            progress(0,1)
            data = self.ProjHandler.Load_2DData(self.i_proj, self.i_time,
                                                self.i_dataset,
                                                self.i_plotPlane,
                                                self.i_slice, self.i_fluid);

            if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :
               data = np.sqrt(data*data)

            progress(1,1)
            PlotName = self.ProjHandler.get_FieldName(self.i_dataset);

            if(self.use_SI) :
                self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                            self.i_dataset, self.i_fluid);
                self.normUnit = self.ProjHandler.get_NormUnit()
                if(self.AllFields[self.i_quant] == "Temp") :
                    self.normValue = 1.e4
                    self.normUnit = "K"
                data *= self.normValue
                message = " Using norm as " + str(self.normValue) + " " + str(self.normUnit)
                self.WriteInfo(message)
    


        # Start by getting the position
        if(self.i_plotPlane==0) :
            xPosL = self.ProjHandler.get_PosArrayLeft(0);
            yPosL = self.ProjHandler.get_PosArrayLeft(1);
            xPosC = self.ProjHandler.get_PosArrayCen(0);
            yPosC = self.ProjHandler.get_PosArrayCen(1);
            xName = "x";
            yName = "y";
        elif(self.i_plotPlane==1) :
            xPosL = self.ProjHandler.get_PosArrayLeft(0);
            yPosL = self.ProjHandler.get_PosArrayLeft(2);
            xPosC = self.ProjHandler.get_PosArrayCen(0);
            yPosC = self.ProjHandler.get_PosArrayCen(2);
            xName = "x";
            yName = "z";
        else :
            xPosL = self.ProjHandler.get_PosArrayLeft(1);
            yPosL = self.ProjHandler.get_PosArrayLeft(2);
            xPosC = self.ProjHandler.get_PosArrayCen(1);
            yPosC = self.ProjHandler.get_PosArrayCen(2);
            xName = "y";
            yName = "z";



        # Now store stuff in plot thingy
        Stash2D = PlotDataStorage.PlotData2D(xPosL, yPosL, xPosC, yPosC,
                                             data, xName,
                                             yName, PlotName,
                                             self.hasFlowVec, xdata, ydata)
        return Stash2D


    def do_next_plotContour(self):
        i_time = self.slider_timestep.value()
        i_max_time = self.slider_timestep.maximum()
        
        i_time = i_time + 1
        if i_time > i_max_time : return
        self.slider_timestep.setValue(i_time)
        self.do_plotContour()


    def do_plotContour(self) :
        # Finally make a countour plot
        
        self.WriteInfo(" Reading data ");
        

        self.lblStatus.setText("Status: Loading data ... ")
        stash2D = self.LoadData()

#        # if vector field is selected: load additional vector data:
#        if(self.i_vecField != 0) :
#            self.LoadVecData(stash2D);


#        print " data "
#        print stash2D.data

        # Now plot the stuff:
        self.lblStatus.setText("Status: Plotting ... ")
        self.Plot2DStash(stash2D, stash2D.xAxisName(), stash2D.yAxisName(), False)


        self.set_colorMap()
        self.lblStatus.setText("Status: Done ... ")


    def do_plotFlow(self) :
        # Plot a flow field

        self.WriteInfo(" Reading data ");
        
        self.lblStatus.setText("Status: Loading data ... ")
        stash2D = self.LoadData()

        # if vector field is selected: load additional vector data:
        if(self.i_vecField != 0) :
            self.LoadVecData(stash2D);

        self.lblStatus.setText("Status: Plotting ... ")
        # print flow pattern
#        self.Plot2DFlow(stash2D, stash2D.xAxisName(), stash2D.yAxisName())
        self.Plot2DStash(stash2D, stash2D.xAxisName(),
                         stash2D.yAxisName(), True)
        
        self.set_colorMap()
        self.lblStatus.setText("Status: Done ... ")


    def setup_polarAxes(self, fig, rect, rmin, rmax):
        # rotate theta=0 to top
        tr_rotate = Affine2D().translate(np.pi/2., 0)
#        tr_rotate = Affine2D().translate(0., 0)
        # invert theta axis
        tr_invert = Affine2D().from_values(-1,0,0,1,0,0);

        tr = tr_invert + tr_rotate + PolarAxes.PolarTransform()

        pi = np.pi
        # no ticks in theta direction
        angle_ticks = []

        grid_locator1 = FixedLocator([v for v, s in angle_ticks])
        tick_formatter1 = DictFormatter(dict(angle_ticks))

        grid_locator2 = MaxNLocator(4)

        gh = floating_axes.GridHelperCurveLinear(tr,
                                                 extremes=(0,pi, rmax,0),
                                                 grid_locator1=grid_locator1,
                                                 grid_locator2=grid_locator2,
                                                 tick_formatter1=tick_formatter1,
                                                 tick_formatter2=None,
                                             )

        print " rad " + str(rmax)

        ax_polar = floating_axes.FloatingSubplot(fig, rect, grid_helper=gh)

        fig.add_subplot(ax_polar)

        aux_ax = ax_polar.get_aux_axes(tr)

        aux_ax.patch = ax_polar.patch
        ax_polar.patch.zorder=0.9 

        return ax_polar, aux_ax


    def Plot2DStash(self, stash, xName, yName, add_vecField) :
        self.WriteInfo(" Doing actual 2D plot ")

        # Make sure to ignore nans and infs
        stash.data = np.ma.masked_invalid(stash.data)

        if (self.lin_plot == False):
            stash.data = np.ma.masked_where(stash.data == 0, stash.data)


#  http://scipy-lectures.github.io/intro/matplotlib/matplotlib.html#contour-plots

        if(self.geometry == 2 and (self.i_plotPlane == 0)) :
            self.use_polarAxes = True
        else :
            self.use_polarAxes = False


        if self.chkReplacePlot.isChecked() :
            if(self.has_figure) :
                plt.close()
                self.has_figure = False
        self.has_figure = True

        if(self.use_polarAxes) :
            self.fig = plt.figure()
        else :
            self.fig = plt.figure()
            plt.xlabel(xName)
            plt.ylabel(yName)
            plt.title("data")

        print " My xName: " + str(xName)


        num_pixels = len(stash.xAxisC)*len(stash.yAxisC)

#        if(num_pixels > 500000 and self.cont_asCells) :
        if(num_pixels > 2000000 and self.cont_asCells) :
            print " pixels: " + str(num_pixels)
            self.butt_showSmooth.click();



        self.set_colorMap()
        hasTicks = False

        if(self.geometry == 0) :
            is_polar = False
        elif(self.geometry == 1) :
            if(self.i_plotPlane == 0) :
                is_polar = True
            else :
                is_polar = False
        elif(self.geometry == 2) :
            if(self.i_plotPlane == 0 or self.i_plotPlane==1) :
                is_polar = True
            else :
                is_polar = False


#        ax = self.fig.add_subplot(111)
#        self.fig, ax = plt.subplots(subplot_kw=dict(projection='polar'))
        if not (is_polar) :
            self.fig, ax = plt.subplots()
            data_values = (stash.xAxisL, stash.yAxisL)
#            ax = self.fig.add_subplot(111)
        else :
            if(self.cont_asCells) :
                rad, theta = np.meshgrid(stash.xAxisL, stash.yAxisL)
#                rad, theta = np.meshgrid(stash.xAxisL, stash.yAxisL, indexing='ij')
                radC, thetaC = np.meshgrid(stash.xAxisC, stash.yAxisC)
                data_values = (stash.xAxisL, stash.yAxisL)
            else :
                rad, theta = np.meshgrid(stash.xAxisC, stash.yAxisC)
                radC, thetaC = np.meshgrid(stash.xAxisC, stash.yAxisC)
                data_values = (stash.xAxisC, stash.yAxisC)
            if(self.geometry == 2 and (self.i_plotPlane == 0)) :
#                self.fig, ax = plt.subplots()
                rmax = stash.xAxisC[len(stash.xAxisC)-1]
                rmin = 0.
                dummy, ax = self.setup_polarAxes(self.fig, 111,rmin,rmax)
            elif(self.geometry == 1 and (self.i_plotPlane == 0)) :
                print " Cylindrical setup "
                self.fig, ax = plt.subplots(subplot_kw=dict(projection='polar'))
                ax.set_theta_zero_location("E")
                ax.set_theta_direction(1)
#                rmax = stash.xAxisC[len(stash.xAxisC)-1]
#                rmin = 0.
#                dummy, ax = self.setup_polarAxes(self.fig, 111,rmin,rmax)
            else :
                self.fig, ax = plt.subplots(subplot_kw=dict(projection='polar'))
                ax.set_theta_zero_location("N")
                ax.set_theta_direction(-1)

        print " LenC, L " + str(len(stash.xAxisC)) + " " + str(len(stash.xAxisL))

#<<<<<<< .mine
        num_pixels = len(stash.xAxisC)*len(stash.yAxisC)


#        print " Value: " + str(stash.data[256,12]) + " " + str(stash.data[12,256])

#        if(num_pixels > 500000 and self.cont_asCells) :
#        if(num_pixels > 100000 and self.cont_asCells) :
#            print " pixels: " + str(num_pixels)
#            self.butt_showSmooth.click();

        if(add_vecField) :
            #            im = ax.streamplot(stash.xAxisC, stash.yAxisC, stash.xData,
            #                               stash.yData, color=stash.data, linewidth=2,
            #                               cmap=plt.cm.autumn)
            if not (is_polar) :
                im = ax.streamplot(stash.xAxisC, stash.yAxisC, stash.xData,
                                   stash.yData, color="w", linewidth=1,cmap=plt.cm.bone)
            else :
                print " shapes: "
#                print radC.shape
#                print thetaC.shape
#                print stash.xData.shape
#                print radC
#                print " old x "
#                print stash.xAxisC
#                im = ax.streamplot(stash.xAxisC, stash.yAxisC, stash.xData,
#                                   stash.yData, color="w", linewidth=1,cmap=plt.cm.bone)
#                im = ax.streamplot(thetaC, radC, stash.xData,
#                                   stash.yData, color="w", linewidth=1,cmap=plt.cm.bone)

        if(self.lin_plot == False) :
            message = " log data range: "
            self.WriteInfo(message)

            str_logmin = "  min: "
            if(stash.data.min() < 0.) :
                str_logmin += "invalid - using abs value "
                stash.data = np.absolute(stash.data)

            str_logmin += "%3.3f" % np.log10(stash.data.min())
            str_logmax = "; max: %3.3f" % np.log10(stash.data.max())

            message = str_logmin + str_logmax
            self.WriteInfo(message)

            extValsLoc = [stash.data.min(), stash.data.max()]

        else :
            extValsLoc = [stash.data.min(), stash.data.max()]

        extVals = [0,0]
        # if data range set by user -> use that one:
        if self.use_zMin_native == 0 :
            extVals[0] = self.zMin_custom
        elif self.use_zMin_native == 1 :
            extVals[0] = extValsLoc[0]
        elif self.use_zMin_native == 2 :
            extVals[0] = self.zMin_custom * extValsLoc[1]

        if self.use_zMax_native == 0 :
            extVals[1] = self.zMax_custom
        elif self.use_zMax_native == 1:
            extVals[1] = extValsLoc[1]
        elif self.use_zMax_native == 2:
            extVals[1] = self.zMax_custom * extValsLoc[0]

        print(" min: %s - max: %s " % (stash.data.min(), stash.data.max()))
        print(" plot range: %s - %s " % (extVals[0], extVals[1]))

        if(self.cont_asCells) :
            if(self.lin_plot) :
                if not (is_polar) :
                    print " Plotting (f) "
                    if(self.forFun and self.AllFields[self.i_quant] == 'tracer') :                        
                        im = ax.pcolormesh(stash.xAxisL, stash.yAxisL, stash.data,
                                           cmap=self.cmap, vmin=0., vmax=1.)
                    else :
#                        im = ax.pcolor(stash.xAxisL, stash.yAxisL, stash.data,
#                                       cmap=self.cmap,vmin=extVals[0], vmax=extVals[1])
                        im = ax.pcolorfast(stash.xAxisL, stash.yAxisL, stash.data,
                                       cmap=self.cmap,vmin=extVals[0], vmax=extVals[1])
#                        im = ax.pcolormesh(stash.xAxisL, stash.yAxisL, stash.data,
#                                       cmap=self.cmap,vmin=extVals[0], vmax=extVals[1])
                    print " shapes: "
                    print stash.data.shape
                else :
                    print " shapes: "
#                    print rad.shape
#                    print theta.shape
                    print stash.data.shape
                    print " Polar (a) "
#                    im = ax.pcolor(theta, rad, stash.data,
#                                   cmap=self.cmap,vmin=extVals[0], vmax=extVals[1])
                    im = ax.pcolorfast(theta, rad, stash.data,
                                   cmap=self.cmap,vmin=extVals[0], vmax=extVals[1])
                    #im = ax.pcolormesh(theta, rad, stash.data,
                    #               cmap=self.cmap)
                    #im = ax.pcolorfast(theta, rad, stash.data,
                    #                   cmap=self.cmap)
            else :
#                LogData = np.log(stash.data)
                if not (is_polar) :
#                    im = ax.pcolor(stash.xAxisL, stash.yAxisL, LogData,
#                                   cmap=self.cmap)
                    print " Plotting (a) "
#                    im = ax.pcolor(stash.xAxisL, stash.yAxisL, stash.data,
#                                   vmin=extVals[0], vmax=extVals[1], 
#                                   norm=LogNorm(vmin=extVals[0],
#                                                vmax=extVals[1]),
#                                   cmap=self.cmap)
                    im = ax.pcolorfast(stash.xAxisL, stash.yAxisL, stash.data,
                                       vmin=extVals[0], vmax=extVals[1],
                                       norm=LogNorm(vmin=extVals[0],
                                                    vmax=extVals[1]),
                                       cmap=self.cmap)
                else :
#                    im = ax.pcolor(theta, rad, LogData,
#                                   cmap=self.cmap)
                    print " Polar (b) "
                    im = ax.pcolorfast(theta, rad, stash.data,
                                   vmin=extVals[0], vmax=extVals[1], 
                                   norm=LogNorm(vmin=extVals[0],
                                                vmax=extVals[1]),
                                   cmap=self.cmap)
#                    im = ax.pcolor(theta, rad, stash.data,
#                                   vmin=extVals[0], vmax=extVals[1], 
#                                   norm=LogNorm(vmin=extVals[0],
#                                                vmax=extVals[1]),
#                                   cmap=self.cmap)
            if not (is_polar) :
                ax.axis([np.min(stash.xAxisC), np.max(stash.xAxisC),
                         np.min(stash.yAxisC), np.max(stash.yAxisC)])
            else :
#                ax.axis([0, 180, 0., np.max(rad)])
                print " Polar? " + str(self.use_polarAxes)
                if(not self.use_polarAxes) :
                    ax.set_rmax(np.max(rad))
        else :
            print " Line oder so " + str(self.lin_plot) + " " + str(is_polar)
            if(self.lin_plot) :
                if not (is_polar) :
                    print " Plotting (c) "
                    if(self.forFun and self.AllFields[self.i_quant] == 'tracer') :                        
                        myLevels = np.linspace(0,1., 100)
                        im = plt.contourf(stash.xAxisC,stash.yAxisC,stash.data,
                                          myLevels,interpolation='linear',cmap=self.cmap, extend="both")
                    else :
                        im = plt.contourf(stash.xAxisC,stash.yAxisC,stash.data,
                                          100,interpolation='linear',cmap=self.cmap)
                else :
                    print " Plotting (b) "
                    im = ax.contourf(theta,rad,stash.data,30,cmap=self.cmap)
#                ax.axis([-np.max(stash.xAxis), np.max(stash.xAxis),
#                         -np.max(stash.xAxis), np.max(stash.xAxis)])
            else :
                # set levels manually
#                zmax = np.ceil(np.log10(stash.data.max())+1)
#                zmin = np.floor(np.log10(stash.data.min())-1)
#                dz = (zmax - zmin)/100.
#                lev_exp = np.arange(zmin, zmax, dz)
#                levs = np.power(10, lev_exp)
#                cs = P.contourf(X, Y, z, levs, norm=colors.LogNorm())
#                im = plt.contourf(stash.xAxis,stash.yAxis,stash.data,
#                                  levs,interpolation='linear',cmap=self.cmap)
                LogData = np.log(stash.data)
                maxData = np.max(LogData)
                minData = maxData - 4.;
                myLevels = np.linspace(minData,maxData, 100)

                if not (is_polar) :

                    my_hack = False

                    if(my_hack) :
                        #                    im = plt.contourf(stash.xAxis,stash.yAxis,LogData,
#                                      100,interpolation='linear',cmap=self.cmap)
#                    im = plt.contourf(stash.xAxis,stash.yAxis,LogData, myLevels,
#                                      interpolation='linear',cmap=self.cmap, extend="both")

                        stash.data *= 10000

                        zmax = np.ceil(np.log10(stash.data.max())+1)
                        zmin = np.floor(np.log10(stash.data.min())-1)
                        dz = (zmax - zmin)/100.
                        lev_exp = np.arange(zmin, zmax, dz)
                        levs = np.power(10, lev_exp)
#                    lvls = np.logspace(np.log10(stash.data.min()),(np.log10(stash.data.max())),100)
#                    lvls = np.logspace(np.log10(stash.data.min()),np.log10(2.e8),100)
#                    lvls = np.logspace(np.log10(stash.data.min()),np.ceil(np.log10(stash.data.max())),100)
                        lvls = np.logspace(np.log10(stash.data.min()),(np.log10(stash.data.max())),100)

                        print lvls

                        im = plt.contourf(stash.xAxis,stash.yAxis,stash.data, lvls,
                                          interpolation='linear',cmap=self.cmap, norm=colors.LogNorm())
 
                        #                    delta = maxData - minData
                        #                    logDel = int(math.log10(delta))
                        #                    roundMin=round(minData,logDel+1)
                        #                    roundMax=round(maxData,logDel+1)
                        #                    contTicks=np.linspace(roundMin, roundMax, 9)
                        #                    hasTicks = True
#                        contTicks = np.logspace(np.log10(stash.data.min()),np.floor(np.log10(stash.data.max())),6)
                        contTicks = np.logspace(np.log10(stash.data.min()),np.log10(1.e8),5)
#                    contTicks = np.logspace(np.log10(stash.data.min()),np.ceil(np.log10(stash.data.max())),6)
 #                   self.fig.colorbar(im, ticks=myticks)
                        print contTicks
                        hasTicks = True

                    # plot crosses at stellar positions
                        posX = [-720, 720]
                        posY = [0,0]
                        plt.plot(posX, posY, color='w', marker='o',linestyle='');
                    else :
                        print " Plotting (d) "
#                        im = plt.contourf(stash.xAxisC,stash.yAxisC,LogData,
#                                          100,interpolation='linear',cmap=self.cmap)
                        lvls = np.logspace(np.log10(extVals[0]),np.log10(extVals[1]),200)           
                        im = plt.contourf(stash.xAxisC,stash.yAxisC, stash.data,
#                                          200,
                                          vmin=extVals[0], vmax=extVals[1], 
#                                          locator=ticker.LogLocator(numticks=60),
                                          norm=LogNorm(vmin=extVals[0],
                                                       vmax=extVals[1]),
#                                          norm=LogNorm(),
                                          levels=lvls,
#                                          levels=20,
                                          interpolation='linear',cmap=self.cmap)
#                        print " My levels: "
#                        print lvls

                else :
#                    im = plt.contourf(theta,rad,LogData,
#                                      100,interpolation='linear',cmap=self.cmap)
                    print " Plotting (h) "
                    lvls = np.logspace(np.log10(extVals[0]),np.log10(extVals[1]),30)          
                    print " levels "
                    print extVals[0] 
                    print stash.data.min()
                    print theta
                    im = ax.contourf(theta,rad,stash.data, levels=lvls,
                                      vmin=extVals[0], vmax=extVals[1],
                                      norm=LogNorm(vmin=extVals[0],
                                                   vmax=extVals[1]),
                                      cmap=self.cmap)
                    #                    im = plt.contourf(theta,rad,stash.data, 100,
                    #                                      vmin=extVals[0], vmax=extVals[1], 
                    #                                      norm=LogNorm(vmin=extVals[0],
                    #                                                   vmax=extVals[1]),
                    #                                      interpolation='linear',cmap=self.cmap)
                    #                im = plt.contourf(stash.xAxis,stash.yAxis,stash.data,
                    #                                  100,interpolation='linear',cmap=self.cmap,
                    #                                  norm=colors.LogNorm())


        if(not self.use_polarAxes) :
            ax.set_aspect('equal')

        delx = np.max(stash.xAxisC) - np.min(stash.xAxisC)
        dely = np.max(stash.yAxisC) - np.min(stash.yAxisC)

        my_fraction = 0.045
        if(self.use_polarAxes) :
            my_fraction = 0.03

        if(self.use_Colorbar) :
            print " Have ticks? " + str(hasTicks)
            if(delx/dely > 1.8 and not self.use_polarAxes) :
                if(hasTicks) :
                    cb = self.fig.colorbar(im, orientation='horizontal',
                                           ticks=contTicks,fraction=my_fraction)
                else :
                    if(self.lin_plot) :
                        print " Hier isser "
                        cb = self.fig.colorbar(im,fraction=my_fraction, orientation='horizontal',ticks=[0,2e6,4e6,6e6,8e6])
                    else :
                        cb = self.fig.colorbar(im,orientation='horizontal',fraction=my_fraction,format='%.2e')
            else :
                if(hasTicks) :
                    #                self.fig.colorbar(im,ticks=contTicks,format='%.2E')
                    cb = self.fig.colorbar(im,fraction=my_fraction,ticks=contTicks)
                    print " Ticks: "
                    print contTicks
                else :
                    if(self.lin_plot) :
                        print " Da isser "
                        contTicks = np.linspace(extVals[0],extVals[1],6)
                        print contTicks
                        cb = self.fig.colorbar(im,ticks=contTicks,fraction=my_fraction,format='%.2g')
#                        cb = self.fig.colorbar(im,ticks=[0.,2.e6,4.e6,6.e6,8.e6],fraction=my_fraction,format='%.2g')
                    else:
                        cb = self.fig.colorbar(im,fraction=my_fraction,format='%.2g')
#                        contTicks =  np.logspace(np.log10(extVals[0]),np.log10(extVals[1]),5)
#                        print contTicks
#                        cb = self.fig.colorbar(im,ticks=contTicks,fraction=my_fraction,format='%.2g')

            if(self.lin_plot) :
                tick_locator = ticker.MaxNLocator(nbins=5)
#                tick_locator = ticker.AutoLocator()
                cb.locator = tick_locator
                cb.update_ticks()
            else :
                #LogLocator(base=10.0, subs=[1.0], numdecs=4, numticks=15)
                tick_locator = ticker.LogLocator(base=10)
                cb.locator = tick_locator
                cb.update_ticks()

        ## ==== Set statusbar description ==== ##
#        rows = np.asarray(data_values[0])
        xpos = data_values[0]
        ypos = data_values[1]
        num_x = len(xpos)
        num_y = len(ypos)
        bounds = []
        bounds.append( (xpos[0], xpos[num_x-1]))
        bounds.append( (ypos[0], ypos[num_y-1]))
        dx = (float(bounds[0][1]) - float(bounds[0][0]))/float(num_x)
        dy = (float(bounds[1][1]) - float(bounds[1][0]))/float(num_y)
        plot_geom = 0
        if self.butt_plane_xy.isChecked() : plot_geom = 1
        elif self.butt_plane_xz.isChecked() : plot_geom = 2

        # check if grid is non-linear (for rows only so far)
        dx1 = xpos[1] - xpos[0]
        dx2 = xpos[2] - xpos[1]

        print " row stuff " + str(num_x) + " " + str(num_y)
#        print " x "
#        print xpos

        nonLin_x = False
        if(abs(dx1 - dx2) > 1.e-9*dx1) :
            print " Be careful: non-linear grid "
            nonLin_x = True
 #           print " Test: "
 #           entry = np.abs(rows - 42.3).argmin()
 #           print str(rows[entry]) + " " + str(rows[entry+1])
        
        def format_coords(x,y):
            # Handle the format string in a try block to ensure that this method does not crash the program
            try:
                label = ''
                coords = []

                # in polar coordinates - x and y are swappend (y is radius here)
                if(is_polar) :
                    coords.append(y - bounds[0][0])
                    coords.append(x - bounds[1][0])
                    coords.append(y)
                else :
                    coords.append(x - bounds[0][0])
                    coords.append(y - bounds[1][0])
                    coords.append(x)
#                print " pos: " + str(x) + " " + str(y)
                
                if is_polar :
                    ## Special handling for polar coordinates
                    m_pi = np.pi
                    
                    if plot_geom == 1 :
                        while (    coords[1] <= 0.0   ): coords[1] = coords[1] + m_pi
                        while (coords[1] >= 2.0 * m_pi): coords[1] = coords[1] - m_pi
                        
                        ix = int(coords[0] / dx)
                        iy = int(coords[1] / dy)

                        if(nonLin_x) :
                            ix = np.abs(xpos - coords[2]).argmin()
#                        print " radius " + str(ix) + " " + str(xpos[ix]) + " " + str(coords[2])

                    elif plot_geom == 2 :
                        while (    coords[1] < 0.0   ): coords[1] = coords[1] + m_pi
                        while (coords[1] >= 2.0 * m_pi): coords[1] = coords[1] - m_pi

                        ix = int(coords[0] / dx)
                        iy = int(coords[1] / dy)
                    
                    else:
                        ix = int((x - bounds[1][0]) / dx)
                        iy = int((y - bounds[0][0]) / dy)
                    
                    
                else:
                    ix = int((x - bounds[0][0]) / dx)
                    iy = int((y - bounds[1][0]) / dy)

#                    print " vals: " + str(ix) + " " + str(iy) + " " + str(x) + " " + str(y)
#                    print "       " + str(numcols) + " " + str(numrows)

                #                if(nonLin_x) :
                #                    col = np.abs(cols - y).argmin()

#                print " cols: " + str(col) + " " + str(numcols) + " " + str(row)
                if ix>=0 and ix<num_x and iy>=0 and iy<num_y:
                    value = stash.data[iy][ix]
                    #                    label = '(%1.4f, %1.4f) = %1.4f'%(x, y, value)
                    if(is_polar) :
                        label = 'gridpoints: (%d, %d); position: (%1.4f (%1.4f), %1.4f (%1.4g)) = %1.4f'%(ix, iy, coords[0] + bounds[0][0], xpos[ix],
                                                                          coords[1] + bounds[1][0],
                                                                          (coords[1] + bounds[1][0])*180/m_pi, value)
                    else :
                        label = 'gridpoints: (%d, %d); position: (%1.4f, %1.4f) = %1.4g'%(ix, iy, coords[0] + bounds[0][0],
                                                                    coords[1] + bounds[1][0], value)
                    
                    #print "(" + str(row) + "," + str(col) + ") = " + str(value)
                else:
                    if(is_polar) :
                        label = 'gridpoints: (%d, %d); position: (%1.4f, %1.4f (%1.4g)) = ?'%(ix, iy,
                                                                                              coords[0] + bounds[0][0],coords[1] + bounds[1][0],
                                                                                              (coords[1] + bounds[1][0])*180/m_pi)
                    else :
                        label = 'gridpoints: (%d, %d); position: (%1.4f, %1.4f) = ?'%(ix, iy, coords[0] + bounds[0][0], coords[1] + bounds[1][0])
                return label
            
            except Exception as e:
                label = '(%1.4f, %1.4f) = '%(x,y)
                label = label +  str(e)
                return label
        if(self.use_ownFormatCode) :
            ax.format_coord = format_coords
        ## ==== End of Statusbar description ==== ##



        self.show_Fig("Plot2D");
        self.WriteInfo(" done(actual 2D plot) ")



    def show_Fig(self, filename) :

        print self.fileFormatContour

        if(self.fileFormatContour == 'display') :
            self.WriteInfo(" Plotting on screen ")

            field_name = str(self.AllFields[self.i_quant])
            mod_name = str(self.Mods[self.i_mod])
            projection_name = ""
            geom = self.geometry
            if geom == 0:
                # Cartesian
                if self.butt_plane_xy.isChecked() : projection_name = "XY"
                elif self.butt_plane_xz.isChecked() : projection_name = "XZ"
                elif self.butt_plane_yz.isChecked() : projection_name = "YZ"
            elif geom == 1:
                # Cylindrical
                if self.butt_plane_xy.isChecked() : projection_name = "R-Theta"
                elif self.butt_plane_xz.isChecked() : projection_name = "R-Z"
                elif self.butt_plane_yz.isChecked() : projection_name = "Theta-Z"
            elif geom == 2:
                # Spherical
                if self.butt_plane_xy.isChecked() : projection_name = "R-Theta"
                elif self.butt_plane_xz.isChecked() : projection_name = "R-Phi"
                elif self.butt_plane_yz.isChecked() : projection_name = "Theta-Phi"
                
            else:
                # Unknwon
                projection_name = ""
            
            self.fig.canvas.set_window_title('[Plot_' + str(projection_name) + '] ' + field_name + " at timestep = " + str(self.timestep) + " (t = " + str(self.time_t) + ")")
            
            title = field_name + '_' + mod_name + "_" + str(projection_name) + ", t = " + str(round(self.time_t, 4)) + " at timestep " + str(self.timestep)
            if(self.annotations2D) : self.fig.suptitle(title)
            
            self.fig.show()
        else :
            print " in other plot "
            if(self.fileFormatContour == 'ps') :
                filename += '.ps'
            elif (self.fileFormatContour == 'png') :
                filename += '.png'
            elif (self.fileFormatContour == 'pdf') :
                filename += '.pdf'
            elif (self.fileFormatContour == 'jpeg') :
                filename += '.jpeg'
            elif (self.fileFormatContour == 'svg') :
                filename += '.svg'
                
            self.WriteInfo(" Saving the file " + filename)
            self.fig.tight_layout()
            if (self.fileFormatContour == 'png') :
                self.fig.savefig(filename,dpi=400)
            elif (self.fileFormatContour == 'jpeg') :
                self.fig.savefig(filename,dpi=400)
            else :
                self.fig.savefig(filename)
        

    def set_colorMap(self) :
        # set color map (still ugly):
        self.cmap = plt.get_cmap(self.name_colorTable)
        self.cmap.set_bad(color='gray', alpha=0.2)



    # now the stuff for the line tab:
#    def set_Dir1D_x(self, checked) :
    def set_Dir1D_x(self) :
        print " Dir was set "
        self.i_lineDir = 0
        self.label_slice1D_1.setText("y:");
        self.label_slice1D_2.setText("z:");
        self.label_xMin_1D.setText("xMin:");
        self.label_xMax_1D.setText("xMax:");

        if(self.butt_Dir1D_x.isEnabled()) :

            self.spinBox_slice1DPerp1.setRange(0,self.Nx[1]-1);
            self.spinBox_slice1DPerp2.setRange(0,self.Nx[2]-1);
            self.slider_slice1DPerp1.setRange(0,self.Nx[1]-1);
            self.slider_slice1DPerp2.setRange(0,self.Nx[2]-1);

            self.dir_Perp1 = 1
            self.dir_Perp2 = 2

            self.set_perpPos1(self.Nx[1]/2)
            self.set_perpPos2(self.Nx[2]/2)
            self.spinBox_slice1DPerp1.setValue(self.Nx[1]/2)
            self.spinBox_slice1DPerp2.setValue(self.Nx[2]/2)

            print " setting range up to " + str(self.Nx[0])
            self.spinBox_xMin_1D.setRange(0,self.Nx[0]-1);
            self.spinBox_xMax_1D.setRange(0,self.Nx[0]-1);
            self.slider_xMin_1D.setRange(0,self.Nx[0]-1);
            self.slider_xMax_1D.setRange(0,self.Nx[0]-1);
            self.set_xMin_1D(0)
            self.set_xMax_1D(self.Nx[0]-1)
            self.spinBox_xMin_1D.setValue(0)
            self.spinBox_xMax_1D.setValue(self.Nx[0]-1)


#    def set_Dir1D_y(self, checked) :
    def set_Dir1D_y(self) :
        self.i_lineDir = 1
        self.label_slice1D_1.setText("x:");
        self.label_slice1D_2.setText("z:");
        self.label_xMin_1D.setText("yMin:");
        self.label_xMax_1D.setText("yMax:");

        if(self.butt_Dir1D_y.isEnabled()) :

            self.spinBox_slice1DPerp1.setRange(0,self.Nx[0]-1);
            self.spinBox_slice1DPerp2.setRange(0,self.Nx[2]-1);
            self.slider_slice1DPerp1.setRange(0,self.Nx[0]-1);
            self.slider_slice1DPerp2.setRange(0,self.Nx[2]-1);
            
            self.dir_Perp1 = 0
            self.dir_Perp2 = 2
            
            self.set_perpPos1(self.Nx[0]/2)
            self.set_perpPos2(self.Nx[2]/2)
            self.spinBox_slice1DPerp1.setValue(self.Nx[0]/2)
            self.spinBox_slice1DPerp2.setValue(self.Nx[2]/2)
            
            self.spinBox_xMin_1D.setRange(0,self.Nx[1]-1);
            self.spinBox_xMax_1D.setRange(0,self.Nx[1]-1);
            self.slider_xMin_1D.setRange(0,self.Nx[1]-1);
            self.slider_xMax_1D.setRange(0,self.Nx[1]-1);
            self.set_xMin_1D(0)
            self.set_xMax_1D(self.Nx[1]-1)
            self.spinBox_xMin_1D.setValue(0)
            self.spinBox_xMax_1D.setValue(self.Nx[1]-1)

#    def set_Dir1D_z(self, checked) :
    def set_Dir1D_z(self) :
        self.i_lineDir = 2
        self.label_slice1D_1.setText("x:");
        self.label_slice1D_2.setText("y:");
        self.label_xMin_1D.setText("zMin:");
        self.label_xMax_1D.setText("zMax:");


        if(self.butt_Dir1D_z.isEnabled()) :

            self.spinBox_slice1DPerp1.setRange(0,self.Nx[0]-1);
            self.spinBox_slice1DPerp2.setRange(0,self.Nx[1]-1);
            self.slider_slice1DPerp1.setRange(0,self.Nx[0]-1);
            self.slider_slice1DPerp2.setRange(0,self.Nx[1]-1);

            self.dir_Perp1 = 0
            self.dir_Perp2 = 1

            self.set_perpPos1(self.Nx[0]/2)
            self.set_perpPos2(self.Nx[1]/2)
            self.spinBox_slice1DPerp1.setValue(self.Nx[0]/2)
            self.spinBox_slice1DPerp2.setValue(self.Nx[1]/2)

            self.spinBox_xMin_1D.setRange(0,self.Nx[2]-1);
            self.spinBox_xMax_1D.setRange(0,self.Nx[2]-1);
            self.slider_xMin_1D.setRange(0,self.Nx[2]-1);
            self.slider_xMax_1D.setRange(0,self.Nx[2]-1);
            self.set_xMin_1D(0)
            self.set_xMax_1D(self.Nx[2]-1)
            self.spinBox_xMin_1D.setValue(0)
            self.spinBox_xMax_1D.setValue(self.Nx[2]-1)

    def set_perpPos1(self, i_PosPerp1) :
        print " setting perp pos " + str(i_PosPerp1)
        self.i_posPerp[0] = i_PosPerp1

        # Obtain corresponding position:
        # Now write a string holding the position

        if(self.i_lineDir == 0) :
            message = "y = "
            pos = self.ProjHandler.get_pos(1, self.i_posPerp[0]);
        else :
            message = "x = "
            pos = self.ProjHandler.get_pos(2, self.i_posPerp[0]);

        message += str(pos)
        self.label_slice1D_1.setText(message);

    def set_perpPos2(self, i_PosPerp2) :
        self.i_posPerp[1] = i_PosPerp2


        # Obtain corresponding position:
        if(self.i_lineDir == 2) :
            message = "y = "
            pos = self.ProjHandler.get_pos(1, self.i_posPerp[1]);
        else :
            message = "z = "
            pos = self.ProjHandler.get_pos(0, self.i_posPerp[1]);

        message += str(pos)
        self.label_slice1D_2.setText(message);


    def set_xMin_1D(self, _i_xMin_1D) :
        self.i_xMin_1D = _i_xMin_1D

        # Obtain corresponding position:
        if(self.i_lineDir == 0) :
            message = "xMin = "
            pos = self.ProjHandler.get_pos(2, self.i_xMin_1D);
        elif(self.i_lineDir == 1) :
            message = "yMin = "
            pos = self.ProjHandler.get_pos(1, self.i_xMin_1D);
        else :
            message = "zMin = "
            pos = self.ProjHandler.get_pos(0, self.i_xMin_1D);

        message += str(pos)
        self.label_xMin_1D.setText(message);

        if(self.i_xMax_1D < self.i_xMin_1D) :
            self.spinBox_xMax_1D.setValue(self.i_xMin_1D)


    def set_xMax_1D(self, _i_xMax_1D) :
        self.i_xMax_1D = _i_xMax_1D

        # Obtain corresponding position:
        if(self.i_lineDir == 0) :
            message = "xMax = "
#           print " Trying " + str(self.i_xMax_1D);
            pos = self.ProjHandler.get_pos(2, self.i_xMax_1D);
        elif(self.i_lineDir == 1) :
            message = "yMax = "
            pos = self.ProjHandler.get_pos(1, self.i_xMax_1D);
        else :
            message = "zMax = "
            pos = self.ProjHandler.get_pos(0, self.i_xMax_1D);

        message += str(pos)
        self.label_xMax_1D.setText(message);

        if(self.i_xMax_1D < self.i_xMin_1D) :
            self.spinBox_xMin_1D.setValue(self.i_xMax_1D)


    def set_1DPlotLine(self, checked) :
        self.PlotStyle1D = 0

    def set_1DPlotSymb(self, checked) :
        self.PlotStyle1D = 1

    def set_1DPlotBoth(self, checked) :
        self.PlotStyle1D = 2
        

    def LoadData1D(self) :

        self.normValue = 1.
        self.normUnit = 'nil'

        if(self.string_has_patt(self.AllFields[self.i_quant],"Velocity")) :

            if (self.fluidType == "rhd"):
                if (self.string_has_patt(self.Mods[self.i_mod], "Abs u") or
                        self.string_has_patt(self.Mods[self.i_mod], "Abs v") or
                        self.string_has_patt(self.Mods[self.i_mod], "Lorentz") or
                        self.string_has_patt(self.Mods[self.i_mod], "v_x") or
                        self.string_has_patt(self.Mods[self.i_mod], "v_y") or
                        self.string_has_patt(self.Mods[self.i_mod], "v_z")) :

                    ux = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_x"],
                                                      self.i_lineDir,
                                                      self.i_posPerp, self.i_fluid)
                    uy = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_y"],
                                                      self.i_lineDir,
                                                      self.i_posPerp, self.i_fluid)
                    uz = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_z"],
                                                      self.i_lineDir,
                                                      self.i_posPerp, self.i_fluid)

                    u2 = ux * ux + uy * uy + uz * uz
                    if self.string_has_patt(self.Mods[self.i_mod], "Abs u"):
                        data = np.sqrt(u2)
                        PlotName = "u_Abs"
                    elif self.string_has_patt(self.Mods[self.i_mod], "Abs v"):
                        data = np.sqrt(u2 / (1 + u2))
                        PlotName = "v_Abs"
                    elif self.string_has_patt(self.Mods[self.i_mod], "Lorentz"):
                        data = np.sqrt(1 + u2)
                        PlotName = "Lorentz"
                    elif self.string_has_patt(self.Mods[self.i_mod], "v_x"):
                        data = ux / np.sqrt(1 + u2)
                        PlotName = "v_x"
                    elif self.string_has_patt(self.Mods[self.i_mod], "v_y"):
                        data = uy / np.sqrt(1 + u2)
                        PlotName = "v_y"
                    elif self.string_has_patt(self.Mods[self.i_mod], "v_z"):
                        data = uz / np.sqrt(1 + u2)
                        PlotName = "v_z"

                elif self.string_has_patt(self.Mods[self.i_mod], "u_x"):
                    data = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_x"],
                                                      self.i_lineDir,
                                                      self.i_posPerp, self.i_fluid)
                    PlotName = "u_x"

                elif self.string_has_patt(self.Mods[self.i_mod], "u_y"):
                    data = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_y"],
                                                      self.i_lineDir,
                                                      self.i_posPerp, self.i_fluid)
                    PlotName = "u_y"

                elif self.string_has_patt(self.Mods[self.i_mod], "u_z"):
                    data = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                      self.NameIDs["v_z"],
                                                      self.i_lineDir,
                                                      self.i_posPerp, self.i_fluid)
                    PlotName = "u_z"


            elif(self.string_has_patt(self.Mods[self.i_mod],"Abs") or
               self.string_has_patt(self.Mods[self.i_mod],"Div") or
               self.string_has_patt(self.Mods[self.i_mod],"Curl")) :

                vx = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_x"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                vy = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_y"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                vz = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_z"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                data = np.copy(vx);

                if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :
                    data = np.sqrt(vx*vx + vy*vy + vz*vz)
                    PlotName = "v_Abs";
                elif(self.string_has_patt(self.Mods[self.i_mod],"Div") or
                     self.string_has_patt(self.Mods[self.i_mod],"Curl")) :
                    print " Still working on this "
                    sys.exit()


                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.NameIDs["v_x"], self.i_fluid);
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue)
                    self.WriteInfo(message)

            else :
                # Use just a single field
                data = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                    self.i_dataset,
                                                    self.i_lineDir,
                                                    self.i_posPerp, self.i_fluid);
                PlotName = self.ProjHandler.get_FieldName(self.i_dataset);
                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.i_dataset, self.i_fluid);
                    data *= self.normValue

        elif(self.string_has_patt(self.AllFields[self.i_quant],"Momentum")) :
            if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :

                rho = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                   self.NameIDs["Density"],
                                                   self.i_lineDir,
                                                   self.i_posPerp, self.i_fluid);

                vx = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_x"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                vy = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_y"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                vz = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["v_z"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                data = np.copy(vx);

                if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :
                    data = np.sqrt(vx*vx + vy*vy + vz*vz)
                    data *= rho
                    PlotName = "s_Abs";
                elif(self.string_has_patt(self.Mods[self.i_mod],"Div") or
                     self.string_has_patt(self.Mods[self.i_mod],"Curl")) :
                    print " Still working on this "
                    sys.exit()


                if(self.use_SI) :
                    normValue1 = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                            self.NameIDs["Density"], self.i_fluid);
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.NameIDs["v_x"], self.i_fluid);
                    self.normValue *= normValue1
                    data *= self.normValue
                    message = " Using norm as " + str(self.normValue)
                    self.WriteInfo(message)

            else :
                print  self.NameIDs
                # Use just a single field
                rho = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                   self.NameIDs["Density"],
                                                   self.i_lineDir,
                                                   self.i_posPerp, self.i_fluid);

                data = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                    self.i_dataset,
                                                    self.i_lineDir,
                                                    self.i_posPerp, self.i_fluid);
                data *= rho
                PlotName = self.ProjHandler.get_FieldName(self.i_dataset);
                if(self.use_SI) :
                    normValue1 = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                            self.NameIDs["Density"], self.i_fluid);
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.i_dataset, self.i_fluid);
                    self.normValue *= normValue1
                    data *= self.normValue

        elif(self.string_has_patt(self.AllFields[self.i_quant],"Mag. field")) :
            if(self.string_has_patt(self.Mods[self.i_mod],"Abs") or
               self.string_has_patt(self.Mods[self.i_mod],"Div") or
               self.string_has_patt(self.Mods[self.i_mod],"Curl")) :

                Bx = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["B_x"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                By = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["B_y"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                Bz = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                  self.NameIDs["B_z"],
                                                  self.i_lineDir,
                                                  self.i_posPerp, self.i_fluid);
                data = np.copy(Bx);

                if(self.string_has_patt(self.Mods[self.i_mod],"Abs")) :
                    data = np.sqrt(Bx*Bx + By*By + Bz*Bz)
                    PlotName = "B_Abs";
                elif(self.string_has_patt(self.Mods[self.i_mod],"Div") or
                     self.string_has_patt(self.Mods[self.i_mod],"Curl")) :
                    print " Still working on this "
                    sys.exit()

                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                                self.NameIDs["B_x"], self.i_fluid);
                    data *= self.normValue
            else :
                # Use just a single field
                data = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                    self.i_dataset,
                                                    self.i_lineDir,
                                                    self.i_posPerp, self.i_fluid); 
                PlotName = self.ProjHandler.get_FieldName(self.i_dataset);
                if(self.use_SI) :
                    self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time, 
                                                                self.i_dataset, self.i_fluid);
                    data *= self.normValue
            
        else :
            # Use just a single field
            print " Reading "
            data = self.ProjHandler.Load_1DData(self.i_proj, self.i_time,
                                                self.i_dataset,
                                                self.i_lineDir,
                                                self.i_posPerp, self.i_fluid); 
            PlotName = self.ProjHandler.get_FieldName(self.i_dataset);
            if(self.use_SI) :
                self.normValue = self.ProjHandler.scan_Norm(self.i_proj, self.i_time,
                                                            self.i_dataset, self.i_fluid);
                data *= self.normValue

        # Now obtain position array:
        if(self.i_lineDir==0) :
            xPos = self.ProjHandler.get_PosArrayCen(0);
            xName = "x";
        elif(self.i_lineDir==1) :
            xPos = self.ProjHandler.get_PosArrayCen(1);
            xName = "y";
        else :
            xPos = self.ProjHandler.get_PosArrayCen(2);
            xName = "z";

        yName = " Data "
        PlotName = " Line Plot "


        # Now store stuff in stash
        Stash1D = PlotDataStorage.PlotData1D(xPos, data, xName,
                                             yName, PlotName)

        return Stash1D

    def on_Exit(self):
        self.close_plots()

    def close_plots(self):
        matplotlib.pyplot.close("all")
        
        
    def set_showAnalytical1D(self) :
        self.showAnalytical1D = self.butt_solutionTest.isChecked()

    def set_useSI(self, checked) :
        self.use_SI = self.butt_useSI.isChecked()
        if(self.use_SI) :
            message = " Using SI units "
            self.WriteInfo(message)


    def set_logXAxis(self) :
        self.use_logXAxis = self.butt_logXAxis.isChecked()

    def set_showGrid(self) :
        self.showGrid = self.butt_showGrid.isChecked()

    def toggle_powerlawScale(self):
        self.use_powerlawScale = self.butt_usePowerlawScale.isChecked()
        self.line_powerlawScaleExponent.setEnabled(self.use_powerlawScale)
        if self.use_powerlawScale:
            print(" apply powerlaw rescaling to data ")
        else:
            print(" no powerlaw rescaling ")

    def set_powerlawScaleExponent(self):
        self.powerlawScaleExponent = float(self.line_powerlawScaleExponent.text())
        print(" User Powerlaw Scale Exponent: %s" % self.powerlawScaleExponent)

    def set_useDerivative(self) :
        self.use_derivative = self.butt_useDerivative.isChecked()

    def do_plotLineNext(self) :

        i_time = self.slider_timestep.value()
        i_max_time = self.slider_timestep.maximum()
        
        i_time = i_time + 1
        if i_time > i_max_time : return
        self.slider_timestep.setValue(i_time)
        self.do_plotLine()

    def do_plotLine(self) :
        self.WriteInfo(" Doing a line plot ");

        stash1DSingle = []
        stash1DSingle.append(self.LoadData1D())
        
        # Now plot the data
        self.Plot1DStash(stash1DSingle,
                         stash1DSingle[0].xAxisName(),
                         stash1DSingle[0].yAxisName(), True)

        self.WriteInfo(" done (line plot) ");


    def do_plotLineStash(self) :
        self.WriteInfo(" Doing a line plot from stash ");

        self.Plot1DStash(self.stash1D, self.stash1D[0].xAxisName(),
                         self.stash1D[0].yAxisName(), True)

        self.WriteInfo(" done (line plots) ");


    def do_1DstashRelPlot(self) :
        self.WriteInfo(" Doing a relative plot from stash ");

        self.Plot1DStash(self.stash1D, self.stash1D[0].xAxisName(),
                         self.stash1D[0].yAxisName(), False)

        self.WriteInfo(" done (line plots) ");


    def add_dataTo1DStash(self) :
        self.WriteInfo(" Adding to 1D stash ");

        self.stash1D.append(self.LoadData1D())

        self.WriteInfo(" done (add to 1D stash)");

    def clear1DStash(self) :
        self.WriteInfo(" Clearing the stash ")
        self.stash1D = []
        self.WriteInfo(" Done (Clearing the stash) ")


    def Plot1DStash(self, stash1D, xName, yName, normalPlot,
                    annotations1D = None,
                    use_powerlawScale = None,
                    powerlawScaleExponent = None,
                    PlotStyle1D = None,
                    i_xMin = None,
                    i_xMax = None,
                    use_derivative = None,
                    showAnalytical1D = None,
                    test_case = None,
                    lin_plot = None,
                    use_logXAxis = None,
                    showGrid = None
                    ) :
        if annotations1D is None:
            annotations1D = self.annotations1D

        if use_powerlawScale is None:
            use_powerlawScale = self.use_powerlawScale

        if powerlawScaleExponent is None:
            powerlawScaleExponent = self.powerlawScaleExponent

        if PlotStyle1D is None:
            PlotStyle1D = self.PlotStyle1D

        if i_xMin is None:
            i_xMin = self.i_xMin_1D

        if i_xMax is None:
            i_xMax = self.i_xMax_1D

        if use_derivative is None:
            use_derivative = self.use_derivative

        if showAnalytical1D is None:
            showAnalytical1D = self.showAnalytical1D

        if test_case is None:
            test_case = self.test_case

        if lin_plot is None:
            lin_plot = self.lin_plot

        if use_logXAxis is None:
            use_logXAxis = self.use_logXAxis

        if showGrid is None:
            showGrid = self.showGrid



        self.WriteInfo(" Doing actual line plot ")

        # get length of stash
        numPlots = len(stash1D)


        if(self.has_figure) :
            plt.close()
            self.has_figure = False
        self.fig = plt.figure()
        self.has_figure = True

        if(annotations1D == True) :
            plt.xlabel(xName)
            if use_powerlawScale:
                plt.ylabel("%s * %s^%s" % (yName, xName, powerlawScaleExponent))
            else:
                plt.ylabel(yName)
            plt.title("line plot")

        ax = self.fig.add_subplot(111)


        markers = ["o","v","s","D","^"]
        markercycler = cycle(markers)

        lines = ["-", "--", ":", "-."]
        linecycler = cycle(lines)
        my_linewidth = 1
        if(self.fileFormatContour == 'display') :
            my_linewidth = 2

        thecolors = ["r","g","b","c","m","y"]
        colcycler = cycle(thecolors)

        yMin, yMax = np.inf, -np.inf

        for iPlot in range(numPlots) :
            my_color = next(colcycler);
            if(PlotStyle1D != 1) :
                my_linestyle=next(linecycler)
            else :
                my_linestyle=""
            
            if(PlotStyle1D != 0) :
                my_marker=next(markercycler)
            else :
                my_marker=""

            xLoc = stash1D[iPlot].xAxis[i_xMin: i_xMax + 1]
            yLoc = stash1D[iPlot].data[i_xMin: i_xMax + 1]

            if use_derivative :
                delx = xLoc[1:] - xLoc[:-1]
                dely = yLoc[1:] - yLoc[:-1]
                xLoc = xLoc[:-1]
                yLoc = dely / delx

            if(normalPlot != True) :
                yLoc /= stash1D[0].data[i_xMin: i_xMax + 1]

            if use_powerlawScale:
                yLoc = yLoc * np.power(xLoc, powerlawScaleExponent)

            if(PlotStyle1D != 1) :
                ax.plot(xLoc, yLoc,
                        my_linestyle, linewidth=my_linewidth,
                        color=my_color) #  ro means - red circles
            if(PlotStyle1D != 0) :
                ax.plot(xLoc, yLoc,
                        my_marker, color=my_color) #  ro means - red circles

            # get Y axis extreme values:
            yMin = min(yMin, yLoc.min())
            yMax = max(yMax, yLoc.max())

        # get X axis extreme values:
        xMin = xLoc.min()
        xMax = xLoc.max()

        print(" xrange: %s %s" % (xMin, xMax))
        print(" yrange: %s %s" % (yMin, yMax))



        # show analytical solution if desired:
        if(showAnalytical1D == True and test_case != 'none') :
            print " Plotting analytical solution "
            self.plot_Analytical1D(ax, stash1D[0].xAxis, my_linewidth)

            self.l2_Analytical(stash1D[0].xAxis, stash1D[0].data) 


        if(lin_plot == False) :
            ax.set_yscale('log')

        if(use_logXAxis == True) :
            ax.set_xscale('log')
            

        ax.grid(showGrid)

        self.show_Fig("LinePlot")

    def set_noAnnotations2D(self) :
        self.annotations2D = not self.annotations2D 

    def set_noAnnotations1D(self) :
        self.annotations1D = not self.annotations1D 

    def set_noColorbar(self) :
        self.use_Colorbar = not self.use_Colorbar

    def plot_Analytical1D(self, ax, xData, my_linewidth) :
        # Plot the different analytical solutions

        # generate data from test handler
        if(self.test_case == "hydrodynamical shock tube test") :
            # need to reset time
            self.TestHandler.reset_time(self.time_t)
            #            self.TestHandler.set_field(self.ProjHandler.get_FieldName(self.i_dataset))
            self.TestHandler.set_field(self.AllFields[self.i_quant])

            anaData = self.TestHandler.get_yvals(xData)
        elif(self.test_case == "rhd 1d riemann problem") :
            # need to reset time
            self.TestHandler.reset_time(self.time_t)
            self.TestHandler.set_field(self.AllFields[self.i_quant])
            self.TestHandler.set_mod(self.Mods[self.i_mod])

            anaData = self.TestHandler.get_yvals(xData)

        else :
            anaData = self.TestHandler.get_yvals(xData)
        
        my_lineStyle ="--"
        my_lineStyle ="-"
        my_color = '0.3'
        # Now plot the data
        ax.plot(xData, anaData, 
                my_lineStyle, color=my_color, linewidth=0.75*my_linewidth)

    def l2_Analytical(self, xData, numData) :

        # for Parker-wind: compute the error:
        if(self.test_case == "isothermal Parker wind") :

            anaData = self.TestHandler.get_yvals(xData)

            num = len(xData)
            l2Val = 0.
            for ix in xrange(10, num-2,1) :
                l2Val += (anaData[ix] - numData[ix])**2
            l2Val = np.sqrt(l2Val)

            message = " l2 error (Parker Wind): " + str(l2Val)
            print message
            self.WriteInfo(message)

            


    def WriteInfo(self, message) :
        self.info_list.addItem(message)
        self.info_list.scrollToBottom()
            

    def show_DataProperties(self) :

        self.list_dataProperties.clear();

        infoLine = "Resolution: ";
        self.list_dataProperties.insertItem(0, infoLine);
        infoLine = "  [" + str(self.Nx[0]) + " x " + str(self.Nx[1])
        infoLine += " x " + str(self.Nx[2]) + "]";
        self.list_dataProperties.insertItem(1, infoLine);

        xPos = self.ProjHandler.get_PosArrayCen(0);
        yPos = self.ProjHandler.get_PosArrayCen(1);
        zPos = self.ProjHandler.get_PosArrayCen(2);

        infoLine = "Extent: ";
        self.list_dataProperties.insertItem(3, infoLine);
        
        infoLine = " [" + str(xPos[0]) + "x" + str(xPos[self.Nx[0]-1]) + "]"
        self.list_dataProperties.insertItem(4, infoLine);
        infoLine = " [" + str(yPos[0]) + "x" + str(yPos[self.Nx[1]-1]) + "]"
        self.list_dataProperties.insertItem(5, infoLine);
        infoLine = " [" + str(zPos[0]) + "x" + str(zPos[self.Nx[2]-1]) + "]"
        self.list_dataProperties.insertItem(6, infoLine);

        geometry = self.Proj_Geometry
        if geometry == 1:
            geometry = "Cartesian (" + str(geometry) + ")"
        elif geometry == 2:
            geometry = "Cylindrical (" + str(geometry) + ")"
        elif geometry == 3:
            geometry = "Spherical (" + str(geometry) + ")"
        else:
            geometry = "Unknown (" + str(geometry) + ")"
        
        self.list_dataProperties.insertItem(6, "Geometry: " + geometry);

        
        #        if(self.butt_plane_xy.isEnabled()) :
        if(self.butt_plane_xy.isChecked()) :
            self.spinBox_slice.setRange(0,self.Nx[2]-1);
            self.slider_slice.setRange(0,self.Nx[2]-1);
            # Initial value for slice
            self.i_slice = self.Nx[2]/2;
            self.set_iSlice(self.i_slice);
            self.spinBox_slice.setValue(self.i_slice);

        #        elif(self.butt_plane_xz.isEnabled()) :
        elif(self.butt_plane_xz.isChecked()) :
            self.spinBox_slice.setRange(0,self.Nx[1]-1);
            self.slider_slice.setRange(0,self.Nx[1]-1);
            # Initial value for slice
            self.i_slice = self.Nx[1]/2;
            self.set_iSlice(self.i_slice);
            self.spinBox_slice.setValue(self.i_slice);

        #        elif(self.butt_plane_yz.isEnabled()) :
        elif(self.butt_plane_yz.isChecked()) :
            self.spinBox_slice.setRange(0,self.Nx[0]-1);
            self.slider_slice.setRange(0,self.Nx[0]-1);
            # Initial value for slice
            self.i_slice = self.Nx[0]/2;
            self.set_iSlice(self.i_slice);
            self.spinBox_slice.setValue(self.i_slice);


    def string_has_regexp(self, string, regexp):

        number = len(regexp.findall(string))
        if number == 1:
            return True
        else:
            return False


    def string_has_patt(self, string, patt):

        regexp = re.compile(patt)
        number = len(regexp.findall(string))
        if number == 1:
            return True
        else:
            return False

        
    def toggle_rangeCustom2D(self, is_set) :
        if(not is_set) :
            print " data Range set exact "
            self.line_zMax.setEnabled(False)
            self.line_zMin.setEnabled(False)
            self.use_zMin_native = 1
            self.use_zMax_native = 1
            
        else :
            print " x Range set to custom "
            self.line_zMax.setEnabled(True)
            self.line_zMin.setEnabled(True)
            self.set_zMin()
            self.set_zMax()

    def set_zMin(self) :
        #zmin = float(self.line_zMin.text())
        #self.zMin_custom = float(zmin)
        #print " User zmin: " + str(self.zMin_custom)
        #self.use_zMin_native = False
        zmin = self.line_zMin.text()
        if zmin[-1] == "*":
            self.use_zMin_native = 2
            self.zMin_custom = float(zmin[:-1])
        else:
            self.use_zMin_native = 0
            self.zMin_custom = float(zmin)

        print " User zmin: " + str(self.zMin_custom)


    def set_zMax(self) :
        #zmax = float(self.line_zMax.text())
        #self.zMax_custom = float(zmax)
        #print " User zmax: " + str(self.zMax_custom)
        #self.use_zMax_native = False
        zmax = self.line_zMax.text()
        if zmax[-1] == "*":
            self.use_zMax_native = 2
            self.zMax_custom = float(zmax[:-1])
        else:
            self.use_zMax_native = 2
            self.zMax_custom = float(zmax)

        print " User zmax: " + str(self.zMax_custom)


    def set_combineOpt(self) :
        print " setting combine opt "
        if (self.butt_combine_opt_add.isChecked()):
            self.combineOpt = "add"
        elif (self.butt_combine_opt_sub.isChecked()):
            self.combineOpt = "sub"
        elif (self.butt_combine_opt_div.isChecked()):
            self.combineOpt = "div"
        elif (self.butt_combine_opt_mult.isChecked()):
            self.combineOpt = "mult"
        elif (self.butt_combine_opt_rel.isChecked()):
            self.combineOpt = "rel"

        print(" --> %s " % self.combineOpt)

    def set_combineImg_A(self):
        self.WriteInfo(" Reading data ")
        self.lblStatus.setText("Status: Loading data ... ")
        self.stash2D_multi[0] = self.LoadData()

    def set_combineImg_B(self):
        self.WriteInfo(" Reading data ")
        self.lblStatus.setText("Status: Loading data ... ")
        self.stash2D_multi[1] = self.LoadData()

    def do_plotContourCombine(self):
        stash2D = None

        if (self.combineOpt == 'add'):
            stash2D = self.stash2D_multi[0] + self.stash2D_multi[1]
        elif (self.combineOpt == 'sub'):
            stash2D = self.stash2D_multi[0] - self.stash2D_multi[1]
        elif (self.combineOpt == 'div'):
            stash2D = self.stash2D_multi[0] / self.stash2D_multi[1]
        elif (self.combineOpt == 'mult'):
            stash2D = self.stash2D_multi[0] * self.stash2D_multi[1]
        elif (self.combineOpt == 'rel'):
            stash2D = (self.stash2D_multi[0] - self.stash2D_multi[1]) / self.stash2D_multi[1]

        # Now plot the stuff:
        self.lblStatus.setText("Status: Plotting ... ")
        self.Plot2DStash(stash2D, stash2D.xAxisName(), stash2D.yAxisName(), False)

        self.set_colorMap()
        self.lblStatus.setText("Status: Done ... ")
