import math
import numpy

class SodTest(object) :
    def set_pars(self, _rhoL, _pL, _vL, _rhoR, _pR, _vR, _gamma, _time) :
        self.rhoL = _rhoL
        self.pL = _pL
        self.uL = _vL
        self.rhoR = _rhoR
        self.pR = _pR
        self.uR = _vR
        self.gamma = _gamma
        self.time = _time

        self.csL = math.sqrt(self.gamma*self.pL/self.rhoL)
        self.csR = math.sqrt(self.gamma*self.pR/self.rhoR)

        self.G1 = (self.gamma - 1.0)/(2.0*self.gamma)
        self.G2 = (self.gamma + 1.0)/(2.0*self.gamma)
        self.G3 = 2.0*self.gamma/(self.gamma - 1.0)
        self.G4 = 2.0/(self.gamma - 1.0)
        self.G5 = 2.0/(self.gamma + 1.0)
        self.G6 = (self.gamma - 1.0)/(self.gamma + 1.0)
        self.G7 = (self.gamma - 1.0)/2.0
        self.G8 = self.gamma - 1.

        if(2./(self.gamma - 1.)*(self.csL + self.csR) < self.uR - self.uL) :
            print " Error: problem with pressure positivity: "
            print 2./(self.gamma - 1.)*(self.csL - self.csR)
            print self.uR - self.uL

        #        self.get_solution()
        fu = self.get_pStar()
        self.fieldName = "rho"
        print " PStar: " + str(fu)

    def reset_time(self, _time) :
        self.time = _time

    def set_field(self, _fieldName) :
        self.fieldName = _fieldName

    def get_pStar(self) :
        pStart = self.get_guess_pStar()
 #       pStart = 1.e-5

        del_u = self.uR - self.uL
        pOld = pStart

        p_eps = 1.e-8

        self.pS = pStart
        pold = 0.

        while (abs((self.pS - pold)/(0.5*(self.pS + pold))) > p_eps) :

            print " ps: " + str(self.pS)
            vals = self.get_auxFuncs(self.pS, self.rhoL, self.pL, self.csL)
            fL = vals[0]
            dfLdp = vals[1]

            vals = self.get_auxFuncs(self.pS, self.rhoR, self.pR, self.csR)
            fR = vals[0]
            dfRdp = vals[1]

            frac = (fL + fR + del_u)/(dfLdp + dfRdp)

            pold = self.pS
            self.pS = self.pS - frac

            if(self.pS < 0.) :
                self.pS = p_eps
        
        
        self.uS = 0.5*(self.uL + self.uR + fR - fL)

        return self.pS
        
    def get_auxFuncs(self, pVal, rhoRef, pRef, csRef) :
        # get auxilary functions (as PREFUN in Toro)
        if(pVal < pRef) :
            pRatio = pVal/pRef
            fVal = self.G4*csRef*(pRatio**self.G1 - 1.0)
            fDeriv = (1.0/(rhoRef*csRef))*pRatio**(-self.G2)
        else :
            aK = self.G5/rhoRef
            bK = self.G6*pRef
            qRT = math.sqrt(aK/(bK + pVal))
            fVal = (pVal - pRef)*qRT
            fDeriv = (1. - 0.5*(pVal - pRef)/(bK + pVal))*qRT

        return fVal, fDeriv

    def get_guess_pStar(self) :
        print " Drin " 
        csUP = 0.25*(self.rhoL + self.rhoR)*(self.csL + self.csR)
        pPV = 0.5*(self.pL + self.pR) + 0.5*(self.uL - self.uR)*csUP
        if(pPV < 0.) :
            pPV = 0.
        pMin = min(self.pL, self.pR)
        pMax = max(self.pL, self.pR)
        qMax = pMax/pMin
        pM = 0.

        if(qMax <= 2. and (pMin <= pPV and pPV < pMax)) :
           pM = pPV
        else :
            if(pPV < pMin) :
                pQ = (self.pL/self.pR)**((self.gamma - 1.)/(2.*self.gamma))
                uM = (pQ*self.uL/self.csL + self.uR/self.csR + 2./(self.gamma - 1.)*(pQ - 1.))/(pQ/self.csL + 1./self.csR)
                # weiter wie auf p 157
                pTL = 1.0 + 0.5*(self.gamma-1.)*(self.uL - uM)/self.csL
                pTR = 1.0 + 0.5*(self.gamma-1.)*(uM - self.uR)/self.csR
                pM = 0.5*(self.pL*pTL**(2.*self.gamma/(self.gamma - 1.)) +
                          self.pR*pTR**(2.*self.gamma/(self.gamma - 1.)))
            else :
                gEL = math.sqrt((self.G5/self.rhoL)/(self.G6*self.pL + pPV))
                gER = math.sqrt((self.G5/self.rhoR)/(self.G6*self.pR + pPV))
                pM = (gEL*self.pL + gER*self.pR - (self.uR - self.uL))/(gEL + gER)
        return pM



    def get_yvals(self, xData) :
        numPoints = len(xData)
        vals = numpy.array(range(numPoints), dtype='f')

        print " Name: " + self.fieldName

        if(self.fieldName == "Density") :
            for ix in range(numPoints) :
                print " Los: "
                print xData[ix]
                vals[ix] = self.get_rho(xData[ix])
        elif(self.fieldName == "Velocity") :
            for ix in range(numPoints) :
                vals[ix] = self.get_vel(xData[ix])
        elif(self.fieldName == "Etherm") :
            for ix in range(numPoints) :
                vals[ix] = self.get_Etherm(xData[ix])
        else :
            for ix in range(numPoints) :
                vals[ix] = self.get_rho(xData[ix])
        return vals

    def get_rho(self, xPos) :
        rho = 0.

        rho, vel, pres = self.get_allValues(xPos)

        return rho

    def get_vel(self, xPos) :
        rho = 0.

        rho, vel, pres = self.get_allValues(xPos)

        return vel


    def get_pres(self, xPos) :
        rho = 0.

        rho, vel, pres = self.get_allValues(xPos)

        return pres

    def get_Etherm(self, xPos) :
        rho = 0.

        rho, vel, pres = self.get_allValues(xPos)

        return pres/(self.gamma - 1)

    def get_Temp(self, xPos) :
#        self.get_solution()
        Temp = 0.

        rho, vel, pres = self.get_allValues(xPos)

        return pres/rho



    def get_allValues(self, xPos) :
        xRel = xPos - 0.5

#        print " xRel: "
#        print xRel
#        print " time: "
#        print time

        if(xRel <= self.uS*self.time) :
            # point left of contact discontinuity

            if(self.pS <= self.pL) :

                # left rarefaction
                shL = self.uL - self.csL

                if(xRel <= shL*self.time) :
                    # undisturbed left state
                    #print " left state " + str(xRel)
                    rho = self.rhoL
                    vel = self.uL
                    pres = self.pL
                else :
                    cSL = self.csL*(self.pS/self.pL)**self.G1
                    stL = self.uS - cSL

                    if(xRel > stL*self.time) :
                        #print " left star state " + str(xRel) + " " + str(stL*self.time)
                        # left star state
                        rho = self.rhoL*(self.pS/self.pL)**(1./self.gamma)
                        vel = self.uS
                        pres = self.pS
                    else :
                        # inside left fan
                        #print " left fan " + str(xPos)
                        vel = self.G5*(self.csL + self.G7*self.uL + xRel/self.time)
                        csLoc = self.G5*(self.csL + self.G7*(self.uL - xRel/self.time))

                        rho = self.rhoL*(csLoc/self.csL)**self.G4
                        pres = self.pL*(csLoc/self.csL)**self.G3
                        
            else :
                # left shock wave
                pSL = self.pS/self.pL
                vL = self.uL - self.csL*math.sqrt(self.G2*pSL + self.G1)

                if(xRel <= vL*self.time) :
                    #print " left state " + str(xRel)
                    # undisturbed left state
                    rho = self.rhoL
                    vel = self.uL
                    pres = self.pL
                else :
                    # left star state
                    #print " left star state " + str(xRel) + " " + str(vL*self.time)
                    rho = self.rhoL*(pSL + self.G6)/(pSL*self.G6 + 1.)
                    vel = self.uS
                    pres = self.pS
        else :
            # point right of contact discontinuity
            if(self.pS > self.pR) :
                # right shock
                pSR = self.pS/self.pR
                vR = self.uR + self.csR*math.sqrt(self.G2*pSR + self.G1)

                if(xRel >= vR*self.time) :
                    #print " first right state " + str(xRel)
                    # undisturbed right state
                    rho = self.rhoR
                    vel = self.uR
                    pres = self.pR
                else :
                    # right star state
                    #print " first right star state " + str(xRel)
                    rho = self.rhoR*(pSR + self.G6)/(pSR*self.G6 + 1.)
                    vel = self.uS
                    pres = self.pS
            else :
                # Right rarefaction:
                vhR = self.uR + self.csR

                if(xRel >= vhR*self.time) :
                    #print " second right state " + str(xRel)
                    # undisturbed right state
                    rho = self.rhoR
                    vel = self.uR
                    pres = self.pR
                else :
                    csSR = self.csR*(self.pS/self.pR)**self.G1
                    stR = self.uS + csSR

                    if(xRel <= stR*self.time) :
                        # left star state
                        #print " second right star state " + str(xRel)
                        rho = self.rhoR*(self.pS/self.pR)**(1./self.gamma)
                        vel = self.uS
                        pres = self.pS
                    else :
                        # inside left fan
                        #print " right left fan " + str(xRel)
                        vel = self.G5*(-self.csR + self.G7*self.uR + xRel/self.time)
                        csLoc = self.G5*(self.csR - self.G7*(self.uR - xRel/self.time))

                        rho = self.rhoR*(csLoc/self.csR)**self.G4
                        pres = self.pR*(csLoc/self.csR)**self.G3
            
        return rho, vel, pres





 
