#ifndef CRONOS_DUMMY_EMISSION_PROCESS_H
#define CRONOS_DUMMY_EMISSION_PROCESS_H

#include "RadiationField.H"
#include <gsl/gsl_sf.h>
#include "RadiativeLossProcess.H"

/**
 * @brief Constant Dummy Emission
 * @details This evaluates parts of an aggregated instance of InverseComptonIntegrals
 */
class DummyEmission: public EmissionField {
public:
	DummyEmission(
		normalisation * normHD,
		normalisation * normParticle,
		REAL differentialEmissivity
	);

	~DummyEmission(
	);

	REAL get_photonEmissivity_LAB(
		Data & gdata,
		int ix, int iy, int iz,
		const REAL photonEnergy_LAB,
		const REAL photonDirection_LAB[3]
	);

private:
	REAL differentialEmissivity;
};

/**
 * @brief Inverse compton loss process
 * @details This evaluates parts of an aggregated instance of InverseComptonIntegrals
 */
class DummyEmissionProcess: public RadiativeLossProcess {
public:
	DummyEmissionProcess(
			normalisation * normHD,
			normalisation * normParticle,
			Quantity differentialEmissivity
		);

	~DummyEmissionProcess(
		);

	void get_momentumLoss(
			REAL momentum,
			REAL & loss
	) final;

	std::string get_type(
	) const final;

private:
	DummyEmission * dummyEmission;
};




#endif
