#ifndef CRONOS_DUMMYSPECIFIC_H
#define CRONOS_DUMMYSPECIFIC_H = 1


#include "Hdf5File_cbase.H"
#include "solver.H"
#include <time.h>
#include <string>
#include <vector>
#include <sys/time.h>
#include "vector.H"
#include "CException.H"
#include "problem.H"
// #include "H5Cpp.h"


// #ifndef H5_NO_NAMESPACE
// using namespace H5;
// #endif

class DummyHyperbolicSolver: public rksolver {
public:
	DummyHyperbolicSolver(Data &, ProblemType &);
	~DummyHyperbolicSolver();
	virtual void init_constants(Data &);
	virtual void addConstants_toH5(Data &, Hdf5Stream &h5out);
	virtual void getConstants_fromH5(Data &, Hdf5iStream &h5in);
	virtual void init(Data &, gridFunc &, ProblemType &);
	virtual void restart(Data &, gridFunc &, ProblemType &);

private:
	void init_general(Data &, gridFunc &, ProblemType &);
	void CheckNan(NumMatrix<double,3> &, int, int, int, string);
	REAL power(double, int);

	void gen_MagBoundValues(Data &, gridFunc &, ProblemType &);
	void compute_B(Data &, gridFunc &, ProblemType &);
	void TransPot2Mag(Data &, gridFunc &, ProblemType &);

	//	int Fluid_Type;

	REAL yama;
	REAL rho0;
	REAL c2_iso, denomPres;
	REAL Mass0, Eges0, mv0[3], BAve0[3];
	REAL rhomin, Ethmin, vmax;
	REAL Ekin_tave, Emag_tave, Etherm_tave, Eges_tave;
	REAL Ekfluc_tave, Ebfluc_tave, Vorticity_tave, Enstrophy_tave;
	REAL cs_tave, v2ave_tave;
	REAL time_ave;

	int thermal;
	int TempNeg;
	int debug;
	int q_rho, q_sx, q_sy, q_sz, q_Bx, q_By, q_Bz, q_Eges, q_Eadd;

	bool IntegrateA, bcVecPotResized;
	bool is_OldRestart;

	clock_t cstart, cstep, cend;
	timeval tick,tock,tock2;
	timeval tstep;
	timeval tin1, tin2, tin3, tin4, tin5, tin6, tin7, tin8, tin9, tin10;
	// For the new solver:
	fields_1D *fieldsX, *fieldsY, *fieldsZ;

	int n_om, n_omInt, n_omIntAll, n_Omega;
	int n_omUser, n_omIntUser, n_OmegaUser;
	bool with_mag;
#if (FLUID_TYPE == CRONOS_MULTIFLUID)
//	RiemannSolver *RiemannSolversX, *RiemannSolversY, *RiemannSolversZ;
//	std::vector<HLLSolver>  RiemannSolversX, RiemannSolversY, RiemannSolversZ;
	std::vector<RiemannSolver*>  RiemannSolversX, RiemannSolversY, RiemannSolversZ;
//	RiemannSolver *RiemannSolversY, *RiemannSolversZ;
//	fields_1D *fieldsMultiX, *fieldsMultiY, *fieldsMultiZ;
	std::vector<fields_1D*> fieldsMultiX, fieldsMultiY, fieldsMultiZ;
//	phys_fields_1D *physValMultixL, *physValMultiyL, *physValMultizL;
//	phys_fields_1D *physValMultixR, *physValMultiyR, *physValMultizR;
	std::vector<phys_fields_1D*> physValMultixL, physValMultiyL, physValMultizL;
	std::vector<phys_fields_1D*> physValMultixR, physValMultiyR, physValMultizR;
//	Reconstruction *ReconstMultiX, *ReconstMultiY, *ReconstMultiZ;
	std::vector<Reconstruction**> ReconstMultiX, ReconstMultiY, ReconstMultiZ;
//	PhysFluxes *PhysFluxMulti;
	std::vector<PhysFluxes*> PhysFluxMulti;
	std::vector<Transformations*> TrafoMulti;
	int numFluids;
//	int N_OM, N_OMINT, N_OMINT_ALL;
//	int N_OM_USER, N_OMINT_USER;
#else
#endif
};

#endif
