#ifndef CRONOS_EMISSION_HANDLER_H
#define CRONOS_EMISSION_HANFLER_H

#ifdef parallel
#include "mpi.h"
#endif

#include <vector>

#include "ParameterFileReader.H"

#include "EmissionSolver.H"
#include "ParticleTransport.H"
#include "data.H"
#include "normalisation.H"

class EmissionHandler {
 public:
  EmissionHandler(ParameterFileReader& parameters);

  /**
   * @brief Prepare the data and delegate the work to emission solvers
   */
  void work(Data&, ProblemType&, ParticleTransport&);

  /**
   * @brief Set the postprocess normalisations
   */
  void setupPostprocesses(ParticleTransport&);

  /**
   * @brief Do some data manipulation before the actual computation starts
   */
  void preprocess(Data&, ParticleTransport&) const;

  /**
   * @brief Compute auxiliary fields (e.g. lorentz factor) from the loaded data
   * should be computed after the preprocessing
   */
  void computeAuxFields(Data&, ParticleTransport&) const;

  /**
   * @brief Writes the data to an hdf5 file
   */
  void dataout(const Data&) const;

 private:
  ParameterFileReader& parameters;

  size_t nEmissions;
  std::vector<EmissionSolver> solvers;

  EmissionNormalisation normEmission;
};

#endif