#ifndef CRONOS_EMISSION_POST_PROCESS_H
#define CRONOS_EMISSION_POST_PROCESS_H

#include "data.H"
#include "RadiativeLossProcess.H"

/**
 * @brief Emission Post Process - Interface class
 */
class EmissionPostProcess {
public:
	EmissionPostProcess() {};
	virtual ~EmissionPostProcess() {};

	/**
	 * @brief These functions are used to set the current point in parameter space
	 * @details They can be used for the computation of intermediate results, in order
	 * to avoid iterated recomputation. The emission solver calls them in the following order.
	 * */

	virtual void set_position(
			Data & gdata,
			int ix, int iy, int iz
	) {};

	virtual void set_direction(
			Data & gdata,
			const REAL observerDirection[3]
	) {};

	virtual void set_energy(
			Data & gdata,
			REAL photonEnergy
	) {};

	virtual void set_radiativeLossProcess(
			Data & gdata,
			RadiativeLossProcess * radiativeLossProcess
	) {};

	virtual REAL compute(
			Data & gdata,
			int ix, int iy, int iz,
			REAL photonSolidFlux,
			REAL photonEnergy,
			const REAL observerDirection[3]
	) = 0;

	virtual std::string get_name(
	) const = 0;

	virtual void set_normalisation(
			normalisation & normHD,
			normalisation & normParticle,
			normalisation & normEmission
	) {};
};


#endif
