#include "Progress.H"

using namespace std;

Progress::Progress()
{
}

StepProgress::StepProgress(
		const int totalSteps
		):
	Progress(),
	totalSteps(totalSteps),
	currentStep(0)
{
}

void StepProgress::reset(
		void
		)
{
	currentStep = 0;
}

void StepProgress::reset(
		const int totalSteps
		)
{
	this->totalSteps = totalSteps;
	reset();
}

float StepProgress::get()
{
	return float(currentStep) / float(totalSteps);
}

StepProgress& StepProgress::operator--(
		void
		)
{
	--currentStep;
	return *this;
}

StepProgress& StepProgress::operator++(
		void
		)
{
	++currentStep;
	return *this;
}


#ifdef parallel

MPIProgress::MPIProgress(
		Progress & progress,
		float interval
		):
		Progress(),
		progress(&progress),
		interval(interval)
{
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	bufferSnd = 0;
	bufferRcv = 0;
	progressLocal = 0;
	progressMpi = 0;
	requestFlag = 1;
	request = MPI_REQUEST_NULL;
	MPI_Test(&request, &requestFlag, MPI_STATUS_IGNORE);
	cerr << "blavla " << requestFlag<< endl;
	MPI_Barrier(MPI_COMM_WORLD);
}

MPIProgress::~MPIProgress()
{
//	// No clue why this is not working....
//
//	// cleaning up
//	MPI_Status status;
//	MPI_Test(&request, &requestFlag, &status);
//	if (requestFlag == 0) {
//		MPI_Cancel(&request);
//		MPI_Wait(&request, &status);
//	}
//	MPI_Barrier(MPI_COMM_WORLD);
}

float MPIProgress::get()
{
	progressLocal = progress->get();

	if(progressLocal >= last_sync + interval && requestFlag == 1) {
		sync();
		last_sync = progressLocal;
	}

	MPI_Test(&request, &requestFlag, MPI_STATUS_IGNORE);

	if (requestFlag == 1) {
		progressMpi = bufferRcv/size;
	}

	return progressMpi;
}

void MPIProgress::sync()
{
	bufferSnd = progressLocal;
	MPI_Iallreduce(&bufferSnd, &bufferRcv, 1, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD, &request);
}

#endif


///////////////////////////////////////////////////////////////////////////////
//
//	Printers
//

ProgressPrinter::ProgressPrinter(
		)
{
}

void ProgressPrinter::print(
		std::ostream & ostream,
		Progress & progress
)
{
	ostream << "Progress: " << 100 * progress.get() << "%" << endl;
}

IntervalProgressPrinter::IntervalProgressPrinter(
		):
		ProgressPrinter()
{
	last_interval = 0;
}

void IntervalProgressPrinter::print(
		std::ostream & ostream,
		Progress & progress,
		float interval
)
{
	float progressNow = progress.get();
	if(progressNow >= last_interval + interval) {
		ProgressPrinter::print(ostream, progress);
		last_interval = progressNow;
	}
}

void IntervalProgressPrinter::reset(
)
{
	last_interval = 0;
}


TimeIntervalProgressPrinter::TimeIntervalProgressPrinter(
):
		ProgressPrinter()
{
	gettimeofday(&tick, 0);
}

void TimeIntervalProgressPrinter::print(
		std::ostream & ostream,
		Progress & progress,
		float time_interval
)
{
	timeval tock;
	gettimeofday(&tock, 0);
	float dt = (tock.tv_sec - tick.tv_sec) + 1e-6 * (tock.tv_usec - tick.tv_usec);

	if(dt >= time_interval) {
		ProgressPrinter::print(ostream, progress);
		tick = tock;
	}
}