#ifndef CRONOS_PROGRESS_H
#define CRONOS_PROGRESS_H

#include <iostream>
#include <sys/time.h>

#ifdef parallel
#include "mpi.h"
#endif


class Progress {
public:
	Progress();
	virtual float get() = 0;
};

class StepProgress: public Progress {
public:
	StepProgress(const int totalSteps);
	float get();

	void reset();
	void reset(const int totalSteps);

	StepProgress& operator--();
	StepProgress& operator++();

protected:

	int totalSteps;
	int currentStep;
};

#ifdef parallel

class MPIProgress: public Progress {
public:
	MPIProgress(Progress & progress, float interval = 0);
	~MPIProgress();

	float get();

private:
	void sync();

	Progress * progress;

	float progressLocal, progressMpi;

	int size;
	float bufferSnd, bufferRcv;
	float interval, last_sync;

	int requestFlag;
	MPI_Request request;
};
#endif

class ProgressPrinter {
public:
	ProgressPrinter(
	);

	virtual void print(
			std::ostream & ostream,
			Progress & progress
	);
};

class IntervalProgressPrinter: public ProgressPrinter {
public:
	IntervalProgressPrinter(
	);

	void print(
			std::ostream & ostream,
			Progress & progress,
			float interval
	);

	void reset();

private:
	float last_interval;
};



class TimeIntervalProgressPrinter: public ProgressPrinter {
public:
	TimeIntervalProgressPrinter(
	);

	void print(
			std::ostream & ostream,
			Progress & progress,
			float time_interval
	);

private:
	timeval tick;
};


#endif