#include "EmissionDirection.H"
#include "catch.hpp"

TEST_CASE("Conversion", "[DirectionList]") {
  DirectionList directions;
  directions.push_back(vector3D{0, 0, 0});
  directions.push_back(vector3D{0, 0, 0});
  directions.push_back(vector3D{0, 0, 0});

  directions[0][0] = 1;
  directions[1][1] = 1;
  directions[2][2] = 1;

  SECTION("access") {
    REQUIRE(directions[0] == vector3D{1, 0, 0});
    REQUIRE(directions[1] == vector3D{0, 1, 0});
    REQUIRE(directions[2] == vector3D{0, 0, 1});
  }

  SECTION("iterator") {
    int n = 0;
    for (const auto& dir : directions) {
      vector3D vec = {0, 0, 0};
      vec[n] = 1;
      REQUIRE(dir == vec);
      ++n;
    }
  }

  SECTION("conversion") {
    NumMatrix<double, 2> conv(Index::set(0, 0), Index::set(2, 2));
    conv.clear();
    conv(0, 0) = 1;
    conv(1, 1) = 1;
    conv(2, 2) = 1;
    REQUIRE(conv == directions.convert());
  }
}

TEST_CASE("Full Sky", "[SphericalDirections]") {
  const size_t nTheta = 3;
  const size_t nPhi = 9;
  const auto thetas = NumArray<double>::set_linear(0, M_PI, nTheta);
  const auto phis = NumArray<double>::set_linear(0, 2. * M_PI, nPhi);
  SphericalDirections list({0, -1, 0}, thetas, phis);

  for (int itheta = 0; itheta < nTheta; ++itheta) {
    for (int iphi = 0; iphi < nPhi; ++iphi) {
      const auto theta = thetas(itheta);
      const auto phi = phis(iphi);
      const auto dir = vector3D{cos(phi) * sin(theta), -cos(theta), sin(phi) * sin(theta)};
      REQUIRE(list[itheta * nPhi + iphi].norm() == Approx(1.));
      REQUIRE(list[itheta * nPhi + iphi] == dir);
    }
  }
}