#include "catch.hpp"
#include "marray.H"


TEST_CASE("Marray has math operators", "[marray]") {
  const marray<double, 2> x({4, -3});

  SECTION("list initialisation") {
    REQUIRE(x[0] == 4);
    REQUIRE(x[1] == -3);
  }

  const marray<double, 2> y = x * 2;

  SECTION("scalar multiplication") {
    REQUIRE(y[0] == 8);
    REQUIRE(y[1] == -6);
  }

  SECTION("scalar division") {
    marray<double, 2> z = y / 2;
    REQUIRE(z[0] == x[0]);
    REQUIRE(z[1] == x[1]);
  }

  SECTION("array addition") {
    marray<double, 2> z = x + y;
    REQUIRE(z[0] == 12);
    REQUIRE(z[1] == -9);
  }

  SECTION("array multiplication") {
    marray<double, 2> z = x * y;
    REQUIRE(z[0] == 32);
    REQUIRE(z[1] == 18);
  }

  SECTION("scalar division assignment") {
    marray<double, 2> z = y;
    z /= 2;
    REQUIRE(z[0] == x[0]);
    REQUIRE(z[1] == x[1]);
  }

  SECTION("comparison") {
    REQUIRE(x == x);
    REQUIRE_FALSE(x == y);
  }

  SECTION("norm computation") {
    SECTION("norm value") {
      REQUIRE(x.norm() == 5.);
      REQUIRE(y.norm() == 10.);
      REQUIRE(vector3D{1., 0., 1.}.norm() == sqrt(2.));
      REQUIRE(vector3D{-1. / sqrt(2.), 0., 1. / sqrt(2.)}.norm() == Approx(1.));
    }

    SECTION("normalisation") {
      marray<double, 2> tmp = x.normalised();
      REQUIRE(tmp[0] == 4. / 5.);
      REQUIRE(tmp[1] == -3. / 5.);

      REQUIRE(vector2D{1.33542, 523.345}.normalised().norm() == Approx(1.));
    }
  }
}

TEST_CASE("Cross product", "[marray]") {
  const vector3D x = {1, 0, 0};

  SECTION("parallel") {
    const vector3D& y = x;
    const vector3D z = x.cross(y);
    REQUIRE(z == vector3D{0., 0., 0.});
  }

  SECTION("normal") {
    const vector3D y = {0, 1, 0};
    const vector3D z = x.cross(y);
    REQUIRE(z == vector3D{0., 0., 1.});
  }

  SECTION("skew") {
    const vector3D y = {1, 1, 1};
    const vector3D z = x.cross(y);
    REQUIRE(z == vector3D{0., -1, 1.});
  }
}

TEST_CASE("Inner product", "[marray]") {
  const vector3D x = {1, 0, 1};

  SECTION("parallel") {
    const vector3D y = x * 2.;
    REQUIRE(x.inner(y) == 4.);
  }

  SECTION("normal") {
    const vector3D y = {0, 1, 0};
    REQUIRE(x.inner(y) == 0.);
  }

  SECTION("skew") {
    const vector3D y = {0.5, 3., 1};
    REQUIRE(x.inner(y) == 1.5);
  }
}