#include "valencia.H"

#include "problem.H"
#include "data.H"
#include <exception>

using namespace std;


REAL CartesianFlatValencia::get_h0(REAL posx, REAL posy, REAL posz) {
	return 1;
}
REAL CartesianFlatValencia::get_h1(REAL posx, REAL posy, REAL posz) {
	return 1;
}
REAL CartesianFlatValencia::get_h2(REAL posx, REAL posy, REAL posz) {
	return 1;
}
REAL CartesianFlatValencia::get_alpha(REAL posx, REAL posy, REAL posz) {
	return 1;
}
REAL CartesianFlatValencia::get_betaCont(REAL posx, REAL posy, REAL posz, int dir) {
	return 0;
}

void CartesianFlatValencia::get_sources(Data &gdata, ProblemType &Problem, NumMatrix<REAL,3> nom[N_OMINT]) {
	return;
}

//----------------------------------------------------------------------------------------
// CYLINDRICAL VALENCIA
//----------------------------------------------------------------------------------------
void CylindricalFlatValencia::get_sources(Data &gdata, ProblemType &Problem, NumMatrix<REAL,3> nom[N_OMINT]) {
	// Set indices:
	int q_rho = gdata.fluid.get_q_rho();
	int q_sx = gdata.fluid.get_q_sx();
	int q_sy = gdata.fluid.get_q_sy();
	int q_sz = gdata.fluid.get_q_sz();
//	int q_Bx = fluid.get_q_Bx();
//	int q_By = fluid.get_q_By();
//	int q_Bz = fluid.get_q_Bz();
	int q_Eges = gdata.fluid.get_q_Eges();
//	int q_Eadd = fluid.get_q_Eadd();

//	int fluid_type = fluid.get_fluid_type();

	REAL r_inv(0.);
	REAL enthalpy(0.);
	REAL pressure(0.);
	REAL Etherm2Press((Problem.gamma - 1.));
	REAL sigma(Problem.gamma/Etherm2Press);


	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {
				r_inv=1./(gdata.getCen_x(ix));

				const REAL u_sqr = sqr(gdata.om[q_sx](ix, iy, iz)) +
														sqr(gdata.om[q_sy](ix, iy, iz)) +
														sqr(gdata.om[q_sz](ix, iy, iz));

				const REAL lorentz_sqr = 1. + u_sqr;
				pressure = Etherm2Press * gdata.om[q_Eges](ix, iy, iz);
				enthalpy = 1 + sigma * pressure/gdata.om[q_rho](ix, iy, iz);


				nom[q_sx](ix,iy,iz) -= (gdata.om[q_rho](ix, iy, iz) * enthalpy * sqr(gdata.om[q_sy](ix,iy,iz)) * lorentz_sqr
										+ pressure)*r_inv;


				nom[q_sy](ix,iy,iz) += (gdata.om[q_rho](ix, iy, iz) * enthalpy * gdata.om[q_sx](ix,iy,iz) *
										gdata.om[q_sy](ix,iy,iz) * lorentz_sqr )*r_inv;

			}
		}
	}
}



//----------------------------------------------------------------------------------------
// CARTESIAN COROTATING VALENCIA
//----------------------------------------------------------------------------------------

void CartesianCorotValencia::set_omega(REAL omega) {
	omega_corot = omega;
}

REAL CartesianCorotValencia::get_omega() const {
	return omega_corot;
}

REAL CartesianCorotValencia::get_alpha(REAL posx, REAL posy, REAL posz) {
	return 1;
}

REAL CartesianCorotValencia::get_betaCont(REAL posx, REAL posy, REAL posz, int dir) {
	switch(dir) {
		case 0 : return - posz * omega_corot;
		case 1 : return 0;
		case 2 : return posx * omega_corot;
		default: throw invalid_argument("Invalid direction: " + ToString(dir));
	}
}

void CartesianCorotValencia::get_sources(Data &gdata, ProblemType &Problem, NumMatrix<REAL,3> nom[N_OMINT]) {
	// Set indices:
	const int q_rho = gdata.fluid.get_q_rho();
	const int q_sx = gdata.fluid.get_q_sx();
	const int q_sy = gdata.fluid.get_q_sy();
	const int q_sz = gdata.fluid.get_q_sz();
	const int q_Eges = gdata.fluid.get_q_Eges();

	const REAL Etherm2Press((Problem.gamma - 1.));
	const REAL gamma_tilde(Problem.gamma/Etherm2Press);

	ccout << "  Computing corot valencia sources for omega = " << omega_corot << endl;

	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {
				const REAL xpos = gdata.getCen_x(ix);
				const REAL zpos = gdata.getCen_z(iz);

				const REAL ux = gdata.om[q_sx](ix,iy,iz);
				const REAL uz = gdata.om[q_sz](ix,iy,iz);

				const REAL u_sqr = sqr(gdata.om[q_sx](ix, iy, iz)) +
							sqr(gdata.om[q_sy](ix, iy, iz)) +
							sqr(gdata.om[q_sz](ix, iy, iz));

				const REAL lorentz_sqr = 1. + u_sqr;
				const REAL lorentz = sqrt(lorentz_sqr);
				const REAL pressure = Etherm2Press * gdata.om[q_Eges](ix, iy, iz);
				const REAL enthalpy = 1 + gamma_tilde * pressure/gdata.om[q_rho](ix, iy, iz);
				const REAL Dhg = gdata.om[q_rho](ix, iy, iz) * enthalpy * lorentz_sqr;

				// const REAL Wx = 1. / (1. - sqr(xpos * omega_corot));
				// const REAL Wz = 1. / (1. - sqr(zpos * omega_corot));

				const REAL vx = ux / lorentz;
				const REAL vz = uz / lorentz;

				const REAL S_sx = Dhg * omega_corot * vz;
				const REAL S_sz = -Dhg * omega_corot * vx;

				nom[q_sx](ix,iy,iz) -= S_sx;
				nom[q_sz](ix,iy,iz) -= S_sz;
			}
		}
	}
}

