#ifndef CRONOS_VALENCIA_H
#define CRONOS_VALENCIA_H = 1


#include "matrix.H"
#include "constants.H"

using namespace std;

class ProblemType;
class Data;

class Valencia {
public:
	virtual REAL get_h0(REAL, REAL, REAL) = 0;
	virtual REAL get_h1(REAL, REAL, REAL) = 0;
	virtual REAL get_h2(REAL, REAL, REAL) = 0;
	virtual REAL get_alpha(REAL, REAL, REAL) = 0;
	virtual REAL get_betaCont(REAL, REAL, REAL, int) = 0;
	virtual void get_sources(Data &, ProblemType &, NumMatrix<REAL,3> nom[]) = 0;
};


class CartesianFlatValencia: public Valencia {
public:
	virtual REAL get_h0(REAL, REAL, REAL);
	virtual REAL get_h1(REAL, REAL, REAL);
	virtual REAL get_h2(REAL, REAL, REAL);
	virtual REAL get_alpha(REAL, REAL, REAL);
	virtual REAL get_betaCont(REAL, REAL, REAL, int);
	virtual void get_sources(Data &, ProblemType &, NumMatrix<REAL,3> nom[]);
};


class UserValencia: public Valencia {
public:
	UserValencia(Data &, ProblemType &);
	virtual REAL get_h0(REAL, REAL, REAL);
	virtual REAL get_h1(REAL, REAL, REAL);
	virtual REAL get_h2(REAL, REAL, REAL);
	virtual REAL get_alpha(REAL, REAL, REAL);
	virtual REAL get_betaCont(REAL, REAL, REAL, int);
	virtual void get_sources(Data &, ProblemType &, NumMatrix<REAL,3> nom[]);
};


class CylindricalFlatValencia: public CartesianFlatValencia {
public:
	virtual void get_sources(Data &, ProblemType &, NumMatrix<REAL,3> nom[]);
};


class CartesianCorotValencia: public CartesianFlatValencia {
public:
	virtual REAL get_alpha(REAL, REAL, REAL);
	virtual REAL get_betaCont(REAL, REAL, REAL, int);
	virtual void get_sources(Data &, ProblemType &, NumMatrix<REAL,3> nom[]);
	virtual void set_omega(REAL omega);
	REAL get_omega() const;
private:
	REAL omega_corot = 0;
};

#endif
