#include "Adiabatic.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)

Adiabatic::Adiabatic(
		normalisation * normHD,
		normalisation * normParticle
		)
	: LossProcess(normHD, normParticle)
{
	adiabaticFac = 0.;
	adiabatic_c1_int = 1.;
	oneThird = 1. / 3.;
};

std::string Adiabatic::get_type(
) const
{
	return "AdiabaticLoss";
}

void Adiabatic::compute_constants(
		Data & gdata,
		int ix, int iy, int iz
	)
{
	LossProcess::compute_constants(gdata, ix, iy, iz);
	adiabaticFac = gdata.du_cov(ix, iy, iz) * oneThird;


	int q_rho = gdata.fluid.get_q_rho();
	int q_sx = gdata.fluid.get_q_sx();

	// temporal part
	adiabatic_c1_int = std::pow(gdata.om[q_rho](ix, iy, iz)/gdata.om_old[q_rho](ix, iy, iz), oneThird);

	// Spatial part
	int pos[3] = {ix, iy, iz};
	for (int dir = 0; dir < 3; ++dir) {
		int posp[3] = {ix, iy, iz};
		int posm[3] = {ix, iy, iz};
		posp[dir] += 1;
		posm[dir] -= 1;

		adiabatic_c1_int *= std::exp(
				(gdata.om[q_rho][posp] - gdata.om[q_rho][posm]) / (gdata.getCen(dir, posp[dir]) - gdata.getCen(dir, posm[dir]))
				* oneThird * 0.5 * gdata.dt * gdata.om[q_sx + dir][pos] / (gdata.lorentz[pos] * gdata.om[q_rho][pos])
																						  );

		adiabatic_c1_int *= std::exp(
				(gdata.om_old[q_rho][posp] - gdata.om_old[q_rho][posm]) / (gdata.getCen(dir, posp[dir]) - gdata.getCen(dir, posm[dir]))
				* oneThird * 0.5 * gdata.dt * gdata.om_old[q_sx + dir][pos] / (gdata.lorentz_old[pos] * gdata.om_old[q_rho][pos])
																						  );


	}
}

void Adiabatic::get_momentumLoss(
		REAL momentum,
		REAL & loss
	)
{
	loss -= momentum * adiabaticFac;
}

REAL Adiabatic::get_momentumLoss_c1_int(
		REAL momentum
	)
{
	return adiabatic_c1_int;
//	return 1.;
}


#endif
