#ifndef DIFF_CONV_ONEDIM
#define DIFF_CONV_ONEDIM

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

#include "Hdf5File_cbase.H"
#include "constants.H"
#include "solver.H"
#include "data.H"
#include "gridfunc.H"
#include "problem.H"
#include "SemilagrangeSolver.H"
#include "Functions.H"
#include "CronosTimer.H"
#include "ParticleTransport.H"


class MomentumLossFunction: public Function<REAL> {
public:

	MomentumLossFunction(
			ParticleTransport & particleTransport
			);

protected:
	REAL _get_value();

	ParticleTransport & particleTransport;
};


class DiffusionConvection1D : public eulersolver {
public:
	DiffusionConvection1D(
			Data & gdata,
			ProblemType & problem
			);

	~DiffusionConvection1D();

	void init(
			Data & gdata,
			gridFunc & gfunc,
			ProblemType & problem
			);

	void restart(
			Data & gdata,
			gridFunc & gfunc,
			ProblemType & problem
			);

	REAL singlestep(
			Data & gdata,
			gridFunc & gfunc,
			ProblemType & problem
			);

	void addConstants_toH5(
			Data &gdata,
			Hdf5Stream &h5out
			);

private:
	void compute_characteristica(
			Data & gdata,
			ParticleTransport & particleTransport,
			const NumMatrix<REAL,1> & momentumEdges,
			int iParticle,
			int ix, int iy, int iz
			);

	void do_linear_extrapolation(
			const NumMatrix<REAL,1> & cellCenters,
			NumMatrix<REAL,1> & values
			);

	void do_powerlaw_extrapolation(
			const NumMatrix<REAL,1> & cellCenters,
			NumMatrix<REAL,1> & values
			);

	BoundaryHandler1D boundaryHandler1D;
	std::vector<Grid1D> grids;
	SemilagrangeSolver semilagra;
	MomentumLossFunction momentumLossFunction;
	int nParticleSpecies;
	int lo, hi;

	/** Performance measures */
	CronosTimer timerFullStep;	 	/** start and ending timestamp for a complete step*/
	CronosTimer * timerParticle;	/** start and ending timestamp for each particle*/
	CronosTimer timerBoundaries;	/** start and ending timestamp for boundary conditions*/


	// Buffers
	// individual for each particles due to (possibly) different resolution
	// Initialized in DiffusionConvection1D::init

	// For characteristics
	std::vector<NumMatrix<REAL, 1> > momentumLoss_c1_int;
	std::vector<NumMatrix<REAL, 1> > momentumLoss_c2;
	std::vector<NumMatrix<REAL, 1> > momentumLoss_c2_old;
	std::vector<NumMatrix<REAL, 4> > momentumLoss_c2_old_save;



	// Todo this is just a hack
	ParticleTransport & particleTransport;

	bool initialized;

};

#endif
