#ifndef CRONOS_EMISSION_FIELD_H
#define CRONOS_EMISSION_FIELD_H

#include "data.H"
#include "normalisation.H"
#include "CException.H"
#include "LorentzTrafo.H"

#include <cassert>
#include <vector>

class EmissionField {
public:
	EmissionField(
			normalisation * normHD,
			normalisation * normParticle
		);

	virtual	~EmissionField(
		);

	// photon emissivity: photons per (time, volume, steradian, energy)

	// All arguments in LAB Frame
	virtual REAL get_photonEmissivity_LAB(
			Data & gdata,
			int ix, int iy, int iz,
			const REAL photonEnergy_LAB,
			const REAL photonDirection_LAB[3]
		);

	// All arguments in COM Frame
	virtual REAL get_photonEmissivity_COM(
			Data & gdata,
			int ix, int iy, int iz,
			const REAL photonEnergy_COM,
			const REAL photonDirection_COM[3]
		);

	// Todo is isotrope?

protected:
	normalisation * normHD;
	normalisation * normParticle;
};



class EmissionField_COM: public EmissionField {
public:
	EmissionField_COM(
			normalisation * normHD,
			normalisation * normParticle
		);

	virtual	~EmissionField_COM(
		);

	virtual REAL get_photonEmissivity_LAB(
			Data & gdata,
			int ix, int iy, int iz,
			const REAL photonEnergy_LAB,
			const REAL photonDirection_LAB[3]
		);

protected:
	LorentzTrafo lorentzTrafo;

};




#endif
