#ifndef DUMMY_GRID_MANAGER_H
#define DUMMY_GRID_MANAGER_H

#include <matrix.H>
#include <memory>

class Grid1D {
public:
	Grid1D(double xb, double xe, int Nx, int rim, int type);
	// Grid1D(const Grid1D &gm);
	Grid1D();
	void set_grid(NumMatrix<double, 1> Edges, int rim, int Nx);

	double get_xb() const;
	double get_xe() const;
	double get_Len() const;
	
	double get_xCen(int ix) const;
	double get_xL(int ix) const;
	double get_del(int ipos) const;

	virtual int get_ipos(const double pos) const;
	int get_mx() const;
	int get_Nx() const;
	int get_rim() const;

	const Index &getLow() const;
	const Index &getHigh() const;

	int getLo() const;
	int getHi() const;

	const NumMatrix<double, 1> &get_xCen() const;
	const NumMatrix<double, 1> &get_xL() const;
	const NumMatrix<double, 1> &get_del() const;

	/** Checks if a given cell is ghost cell */
	bool isGhost(int ix) const { return (ix < 0 || ix > mx); };

	virtual std::unique_ptr<Grid1D> clone() const;

private:
	void compute_depedentVars();
	void build_grid();
	bool is_set;
	double xb, xe, Len, del, idel;
	int mx, Nx, rim, type, ipos_get;
	NumMatrix<double, 1> cellCentres, cellWidths;
	NumMatrix<double, 1> cellEdges;

	Index lo, hi;
};

class LogarithmicGrid : public Grid1D
{
public:
	LogarithmicGrid(double xb, double xe, int Nx, int rim);
	LogarithmicGrid();
	int get_ipos(const double pos) const;

	double get_xC_log(const int ix) const;
	double get_xL_log(const int ix) const;

	virtual std::unique_ptr<Grid1D> clone() const override;

private:
	double fx, fxLog;
	NumMatrix<double, 1> cellEdgesLog, cellCentersLog;
};

class LinearGrid : public Grid1D
{
public:
	LinearGrid(double xb, double xe, int Nx, int rim);
	LinearGrid();
	int get_ipos(const double pos) const;

	virtual std::unique_ptr<Grid1D> clone() const override;

private:
	double dx;
};


#endif
