#include "LorentzTrafo.H"

using namespace std;

LorentzTrafo::LorentzTrafo(
	)
{
}

LorentzTrafo::~LorentzTrafo(
	)
{
}

void LorentzTrafo::set_boostVector(
		const REAL uboost[3],
		bool invert
		)
{
	REAL lorentz = sqrt(1.
			+ sqr(uboost[0])
			+ sqr(uboost[1])
			+ sqr(uboost[2])
			);
	set_boostVector(uboost, lorentz, invert);
}

void LorentzTrafo::set_boostVector(
		REAL uboostx,
		REAL uboosty,
		REAL uboostz,
		bool invert
		)
{
	REAL lorentz = sqrt(1.
			+ sqr(uboostx)
			+ sqr(uboosty)
			+ sqr(uboostz)
			);

	set_boostVector(uboostx, uboosty, uboostz, lorentz, invert);
}

void LorentzTrafo::set_boostVector(
		const REAL uboost[3],
		const REAL lorentz,
		bool invert
		)
{
	for (int iDir=0; iDir<3; ++iDir) {
		this->uboost[iDir] = uboost[iDir];
	}

	if (invert) {
		for (int iDir=0; iDir<3; ++iDir) {
			this->uboost[iDir] *= -1.;
		}
	}

	this->lorentz = lorentz;

}

void LorentzTrafo::set_boostVector(
		REAL uboostx,
		REAL uboosty,
		REAL uboostz,
		const REAL lorentz,
		bool invert
		)
{
	this->uboost[0] = uboostx;
	this->uboost[1] = uboosty;
	this->uboost[2] = uboostz;

	if (invert) {
		for (int iDir=0; iDir<3; ++iDir) {
			this->uboost[iDir] *= -1.;
		}
	}

	this->lorentz = lorentz;
}

REAL LorentzTrafo::get_matrixElement(
		int row,
		int col
		) const
{
	// Todo implementation?
}

void LorentzTrafo::nullLike_direction1(
		const REAL inDir[3],
		REAL outDir[3]
		) const
{
	REAL dopplerFactor = get_dopplerFactor(inDir);
	REAL fac = (dopplerFactor + 1.) / (lorentz + 1.);

	for (int iDir=0; iDir<3; ++iDir) {
		outDir[iDir] = dopplerFactor * inDir[iDir] - fac * uboost[iDir];
	}
}

void LorentzTrafo::nullLike_temporal1(
		const REAL inEnergy,
		const REAL inDir[3],
		REAL & outEnergy
		) const
{
	REAL dopplerFactor_inv = get_dopplerFactor_inv(inDir);
	outEnergy = inEnergy * dopplerFactor_inv;
}

void LorentzTrafo::nullLike_temporal1_iso(
		const REAL inEnergy,
		REAL & outEnergy
) const
{
	outEnergy = inEnergy * lorentz;
}

void LorentzTrafo::nullLike_temporal2(
		const REAL inEnergy,
		const REAL inDir[3],
		REAL & outEnergy
		) const
{
	REAL dopplerFactor_inv = get_dopplerFactor_inv(inDir);
	outEnergy = inEnergy * sqr(dopplerFactor_inv);
}

void LorentzTrafo::nullLike_temporal2_iso(
		const REAL inEnergy,
		REAL & outEnergy
) const
{
	outEnergy = inEnergy * (4. * sqr(lorentz) - 1.)/3.;
}

REAL LorentzTrafo::get_dopplerFactor(
		const REAL dir[3]
		) const
{
	REAL dopplerFactor = 1. / get_dopplerFactor_inv(dir);
	return dopplerFactor;
}

REAL LorentzTrafo::get_dopplerFactor_inv(
		const REAL dir[3]
		) const
{
	REAL ucos = 0.;
	for (int iDir=0; iDir<3; ++iDir) {
		ucos += uboost[iDir] * dir[iDir];
	}
	REAL dopplerFactor_inv = lorentz - ucos;
	return dopplerFactor_inv;
}


