#include "MagneticField.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)

MagneticField::MagneticField(
		normalisation * normHD
	)
{
	this->normHD = normHD;
}

MagneticField::~MagneticField(
	)
{
}

REAL MagneticField::get_magneticFieldStrength_LAB(
		Data & gdata,
		int ix, int iy, int iz
	)
{
	throw CException(" Not Implemented ", __PRETTY_FUNCTION__);
}

void MagneticField::get_magneticFieldDir_LAB(
		Data & gdata,
		int ix, int iy, int iz,
		REAL magneticFieldDir_LAB[3]
	)
{
	throw CException(" Not Implemented ", __PRETTY_FUNCTION__);
}

// All arguments in COM Frame
REAL MagneticField::get_magneticFieldStrength_COM(
		Data & gdata,
		int ix, int iy, int iz
	)
{
	throw CException(" Not Implemented ", __PRETTY_FUNCTION__);
}

void MagneticField::get_magneticFieldDir_COM(
		Data & gdata,
		int ix, int iy, int iz,
		REAL magneticFieldDir_COM[3]
	)
{
	throw CException(" Not Implemented ", __PRETTY_FUNCTION__);
}



MagneticField_idealMHD_LAB::MagneticField_idealMHD_LAB(
		normalisation * normHD
		):
	MagneticField(normHD)
{
}

MagneticField_idealMHD_LAB::~MagneticField_idealMHD_LAB(
	)
{
}

REAL MagneticField_idealMHD_LAB::get_magneticFieldStrength_COM(
		Data & gdata,
		int ix, int iy, int iz
	)
{
	assert(gdata.om[gdata.fluid.get_q_sx()].getName() == "v_x");
	assert(gdata.om[gdata.fluid.get_q_sy()].getName() == "v_y");
	assert(gdata.om[gdata.fluid.get_q_sz()].getName() == "v_z");

	REAL magneticFieldStrength_LAB = get_magneticFieldStrength_LAB(gdata, ix, iy, iz);
	REAL magneticFieldDir_LAB[3];
	this->get_magneticFieldDir_LAB(gdata, ix, iy, iz, magneticFieldDir_LAB);

	REAL ucos = 0.;
	for (int iDir=0; iDir < 3; ++iDir) {
		ucos += gdata.om[gdata.fluid.get_q_sx() + iDir](ix, iy, iz) * magneticFieldDir_LAB[iDir];
	}
	REAL magneticField_COM_sqr = (1 + sqr(ucos)) * sqr(magneticFieldStrength_LAB / gdata.lorentz(ix, iy, iz));
	REAL magneticFieldStrength_COM = sqrt(magneticField_COM_sqr);
	return magneticFieldStrength_COM;
}

void MagneticField_idealMHD_LAB::get_magneticFieldDir_COM(
		Data & gdata,
		int ix, int iy, int iz,
		REAL magneticFieldDir_COM[3]
	)
{
	assert(gdata.om[gdata.fluid.get_q_sx()].getName() == "v_x");
	assert(gdata.om[gdata.fluid.get_q_sy()].getName() == "v_y");
	assert(gdata.om[gdata.fluid.get_q_sz()].getName() == "v_z");

	// Todo not needed for now - maybe future work
	throw CException(" Not Implemented ", __PRETTY_FUNCTION__);
}



MagneticField_idealMHD_COM::MagneticField_idealMHD_COM(
		normalisation * normHD
):
		MagneticField(normHD)
{
}

MagneticField_idealMHD_COM::~MagneticField_idealMHD_COM(
)
{
}

REAL MagneticField_idealMHD_COM::get_magneticFieldStrength_LAB(
		Data & gdata,
		int ix, int iy, int iz
)
{
	assert(gdata.om[gdata.fluid.get_q_sx()].getName() == "v_x");
	assert(gdata.om[gdata.fluid.get_q_sy()].getName() == "v_y");
	assert(gdata.om[gdata.fluid.get_q_sz()].getName() == "v_z");

	REAL magneticFieldStrength_COM = get_magneticFieldStrength_COM(gdata, ix, iy, iz);
	REAL magneticFieldDir_COM[3];
	this->get_magneticFieldDir_COM(gdata, ix, iy, iz, magneticFieldDir_COM);

	REAL ucos = 0.;
	for (int iDir=0; iDir < 3; ++iDir) {
		ucos += gdata.om[gdata.fluid.get_q_sx() + iDir](ix, iy, iz) * magneticFieldDir_COM[iDir];
	}
	REAL magneticField_LAB_sqr = (sqr(gdata.lorentz(ix, iy, iz)) - sqr(ucos)) * sqr(magneticFieldStrength_COM);
	REAL magneticFieldStrength_LAB = sqrt(magneticField_LAB_sqr);
	return magneticFieldStrength_LAB;
}

void MagneticField_idealMHD_COM::get_magneticFieldDir_LAB(
		Data & gdata,
		int ix, int iy, int iz,
		REAL magneticFieldStrength_LAB[3]
)
{
	assert(gdata.om[gdata.fluid.get_q_sx()].getName() == "v_x");
	assert(gdata.om[gdata.fluid.get_q_sy()].getName() == "v_y");
	assert(gdata.om[gdata.fluid.get_q_sz()].getName() == "v_z");

	// Todo not needed for now - maybe future work
	throw CException(" Not Implemented ", __PRETTY_FUNCTION__);
}

#endif
