#ifndef CRONOS_MAGNETIC_FIELD_H
#define CRONOS_MAGNETIC_FIELD_H



#include "data.H"
#include "normalisation.H"
#include "CException.H"
#include "LorentzTrafo.H"

#include <cassert>
#include <vector>



class MagneticField {

	// This class is used to implements magnetic field models
	// if standard hydro (non-MHD) is used


public:
	MagneticField(
			normalisation * normHD
		);

	virtual	~MagneticField(
		);

	// All arguments in LAB Frame
	virtual REAL get_magneticFieldStrength_LAB(
			Data & gdata,
			int ix, int iy, int iz
		);

	virtual void get_magneticFieldDir_LAB(
			Data & gdata,
			int ix, int iy, int iz,
			REAL magneticFieldDir_LAB[3]
		);

	// All arguments in COM Frame
	virtual REAL get_magneticFieldStrength_COM(
			Data & gdata,
			int ix, int iy, int iz
		);

	virtual void get_magneticFieldDir_COM(
			Data & gdata,
			int ix, int iy, int iz,
			REAL magneticFieldDir_COM[3]
		);

protected:
	normalisation * normHD;

};


class MagneticField_idealMHD_LAB: public MagneticField {

	// This implements the LAB -> COM trafos under the ideal MHD assumption
	// i.e. infinite conductivity -> no electric Field in the COM frame

public:
	MagneticField_idealMHD_LAB(
			normalisation * normHD
		);

	virtual	~MagneticField_idealMHD_LAB(
		);

	// All arguments in COM Frame
	virtual REAL get_magneticFieldStrength_COM(
			Data & gdata,
			int ix, int iy, int iz
		);

	virtual void get_magneticFieldDir_COM(
			Data & gdata,
			int ix, int iy, int iz,
			REAL magneticFieldDir_COM[3]
		);

};

class MagneticField_idealMHD_COM: public MagneticField {

	// This implements the LAB -> COM trafos under the ideal MHD assumption
	// i.e. infinite conductivity -> no electric Field in the COM frame

public:
	MagneticField_idealMHD_COM(
			normalisation * normHD
	);

	virtual	~MagneticField_idealMHD_COM(
	);

	// All arguments in LAB Frame
	virtual REAL get_magneticFieldStrength_LAB(
			Data & gdata,
			int ix, int iy, int iz
	);

	virtual void get_magneticFieldDir_LAB(
			Data & gdata,
			int ix, int iy, int iz,
			REAL magneticFieldDir_LAB[3]
	);

};



#endif
