#ifndef CRONOS_PARTICLE_H
#define CRONOS_PARTICLE_H

#include <cmath>
#include <sstream>
#include "data.H"
#include "constants.H"
#include "CException.H"
#include "normalisation.H"
#include "Grid1D.H"

class Particle {
public:
	Particle(
			Data & gdata,
			normalisation * normParticle,
			const int particleID,
			const int omb
			);

	virtual ~Particle();

	const string get_particleType() const;
	const string get_particleName() const;

	const NumMatrix<REAL,1> & get_momentumCenter() const;
	const NumMatrix<REAL,1> & get_momentumEdge() const;
	const NumMatrix<REAL,1> & get_momentumWidth() const;

	const REAL get_momentumCenter(int ip) const;
	const REAL get_momentumEdge(int ip) const;
	const REAL get_momentumWidth(int ip) const;

	const int getLo() const;
	const int getHi() const;
	const int getNBins() const;
	const int getOmb() const;
	const int getOme() const;

	REAL & operator()(
		int iMomentum,
		int ix, int iy, int iz
	);

	const REAL & operator()(
			int iMomentum,
			int ix, int iy, int iz
	) const;

	void getSpectrum(
			int ix, int iy, int iz,
			NumMatrix<REAL, 1> & spectrum
	) const;

	int getIom(
			int iMomentum
	);

	void set_powerlaw(
			REAL powerlawConstant,
			REAL spectralIndex,
			int ix, int iy, int iz
	);

	void clear();

	REAL numberDensity(
			int ix, int iy, int iz,
			int method = 0
	) const;

	REAL energyDensity(
			int ix, int iy, int iz,
			int method = 0
	) const;

	const int particleID;

	REAL particleRestMass_num;
	REAL particleRestEnergy_num;
	REAL particleCharge_num;
	REAL particleChargeAbs_num;

	REAL injectionRatio;
	REAL injectionSpectrumParticleRatio, injectionSpectrumEnergyRatio, injectionSpectrumIndex, injectionAccelerationFactor;

	LogarithmicGrid grid;

protected:
	Data & gdata;

	normalisation * normParticle;

	int lo, hi;		// Lowest and highest index for the momentum Centers
	int omb, ome;	// Lowest and highest index for the particle in the user fields -> gdata.om_user[xx]
	int nBins;

	string particleType;
	string particleName;

	Quantity particleRestMass;
	Quantity particleCharge;
};

#if (PARTICLE_TRANSPORT == CRONOS_ON)

inline int Particle::getIom(
		int iMomentum
	)
{
	return omb + iMomentum;
}

inline REAL & Particle::operator()(
		int iMomentum,
		int ix, int iy, int iz
	)
{
	assert(iMomentum >= 0 && iMomentum < nBins);
	return gdata.om_user[omb + iMomentum](ix, iy, iz);
}

inline const REAL & Particle::operator()(
		int iMomentum,
		int ix, int iy, int iz
) const
{
	assert(iMomentum >= 0 && iMomentum < nBins);
	return gdata.om_user[omb + iMomentum](ix, iy, iz);
}


inline void Particle::getSpectrum(
		int ix, int iy, int iz,
		NumMatrix<REAL, 1> & spectrum
) const {
	spectrum.resize(grid.getLow(), grid.getHigh());
	for (int iMom = lo; iMom <= hi; ++iMom) {
		spectrum(iMom) = operator()(iMom, ix, iy, iz);
	}

	const int rim = grid.get_rim();
	for (int iRim = 1; iRim <= rim; ++iRim) {
    spectrum(lo - iRim) = 0;
    spectrum(hi + iRim) = 0;
  }
}



#endif


#endif
