#include "RadiationField.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)

REAL RadiationField::numberDensity(
		Data & gdata,
		int ix, int iy, int iz
)
{
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return numberDensity(gdata, pos);
}

REAL RadiationField::numberDensity_dEdOmega(
		Data &gdata,
		int ix, int iy, int iz,
		REAL energy,
		const REAL dir[3]
)
{
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return numberDensity_dEdOmega(gdata, pos, energy, dir);
}

bool RadiationField::isDefined(
		const Data & gdata,
		const REAL pos[3]
) const
{
	return true;
}

/* ----------------------------------------------------------------------------------------- */


SphericalSymmetry::SphericalSymmetry(
		const vector3D& center
	)
{
	set_center(center);
}

void SphericalSymmetry::get_pos(
		REAL _center [3]
) const
{
	for(int iDir=0; iDir < 3; ++iDir) {
		_center[iDir] = center[iDir];
	}
}

void SphericalSymmetry::get_ipos(
	Grid& grid,
	int ipos [3]
) const
{
	for (int iDir=0; iDir < 3; ++iDir) {
		ipos[iDir] = grid.get_idx(iDir, center[iDir], 0);
	}
}

void SphericalSymmetry::set_center(const vector3D& center) {
	this->center = center;
}

const vector3D& SphericalSymmetry::get_center() const {
	return center;
}


/* ----------------------------------------------------------------------------------------- */

RadiationFieldFactorized::RadiationFieldFactorized(
	)
{
	iRadiationFieldFactor_NumberDensity = NULL;
	iRadiationFieldFactor_Energy = NULL;
	iRadiationFieldFactor_Direction = NULL;
}

REAL RadiationFieldFactorized::numberDensity(
		Data &gdata,
		const REAL pos[3]
	)
{
	return iRadiationFieldFactor_NumberDensity->numberDensity(gdata, pos);
}

REAL RadiationFieldFactorized::numberDensity_dEdOmega(
		Data &gdata,
		const REAL pos[3],
		REAL energy,
		const REAL dir[3]
	)
{
	return iRadiationFieldFactor_NumberDensity->numberDensity(gdata, pos)
		   * iRadiationFieldFactor_Energy->fEnergy(gdata, pos, energy)
		   * iRadiationFieldFactor_Direction->fDirection(gdata, pos, dir);
}

void RadiationFieldFactorized::set_iRadiationFieldFactor_NumberDensity(
		std::shared_ptr<IRFFN> _iRadiationFieldFactor_NumberDensity
	)
{
	iRadiationFieldFactor_NumberDensity = _iRadiationFieldFactor_NumberDensity;
}

void RadiationFieldFactorized::set_iRadiationFieldFactor_Energy(
		std::shared_ptr<IRFFE> _iRadiationFieldFactor_Energy
	)
{
	iRadiationFieldFactor_Energy = _iRadiationFieldFactor_Energy;
}

void RadiationFieldFactorized::set_iRadiationFieldFactor_Direction(
		std::shared_ptr<IRFFD> _iRadiationFieldFactor_Direction
	)
{
	iRadiationFieldFactor_Direction = _iRadiationFieldFactor_Direction;
}


std::shared_ptr<IRFFN> RadiationFieldFactorized::get_iRadiationFieldFactor_NumberDensity(
	)
{
	return iRadiationFieldFactor_NumberDensity;
}

std::shared_ptr<IRFFE> RadiationFieldFactorized::get_iRadiationFieldFactor_Energy(
	)
{
	return iRadiationFieldFactor_Energy;
}

std::shared_ptr<IRFFD> RadiationFieldFactorized::get_iRadiationFieldFactor_Direction(
	)
{
	return iRadiationFieldFactor_Direction;
}


REAL IRadiationFieldFactor_NumberDensity::numberDensity(
		Data & gdata,
		int ix,int iy, int iz
	)
{
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return numberDensity(gdata, pos);
}

REAL IRadiationFieldFactor_Energy::fEnergy(
		Data & gdata,
		int ix,int iy, int iz,
		REAL energy
	)
{
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return fEnergy(gdata, pos, energy);
}

double IRadiationFieldFactor_Energy::getEMax(
		Data & gdata,
		int ix,int iy, int iz
) {
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return getEMax(gdata, pos);
}

double IRadiationFieldFactor_Energy::getEMin(
		Data & gdata,
		int ix,int iy, int iz
) {
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return getEMin(gdata, pos);
}

REAL IRadiationFieldFactor_Direction::fDirection(
		Data & gdata,
		int ix,int iy, int iz,
		const REAL beamDirection[3]
	)
{
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return fDirection(gdata, pos, beamDirection);
}

double IRadiationFieldFactor_Direction::getMuMin(Data & gdata, int ix, int iy, int iz, const double direction[3]) {
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);
	return getMuMin(gdata, pos, direction);
}

REAL IRadiationFieldFactor_Monochromatic::fEnergy(
		Data &gdata,
		const REAL pos[3],
		REAL energy
	)
{
	return 0;
}

REAL IRadiationFieldFactor_Monochromatic::get_photonEnergy(
		Data & gdata,
		int ix,int iy, int iz
	)
{
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return get_photonEnergy(gdata, pos);
}

double IRadiationFieldFactor_Monochromatic::getEMax(
		Data & gdata,
		const REAL pos[3]
) {
	return get_photonEnergy(gdata, pos);
}

double IRadiationFieldFactor_Monochromatic::getEMin(
		Data & gdata,
		const REAL pos[3]
) {
	return get_photonEnergy(gdata, pos);
}


REAL IRadiationFieldFactor_Beamed::fDirection(
		Data &gdata,
		const REAL pos[3],
		const REAL dir[3]
	)
{
	return 0;
}

void IRadiationFieldFactor_Beamed::get_beamDirection(
		Data & gdata,
		int ix,int iy, int iz,
		REAL beamDirection[3]
	)
{
	REAL pos[3];
	pos[0] = gdata.getCen_x(ix);
	pos[1] = gdata.getCen_y(iy);
	pos[2] = gdata.getCen_z(iz);

	return get_beamDirection(gdata, pos, beamDirection);
}

double IRadiationFieldFactor_Beamed::getMuMin(Data & gdata, const double pos[3], const double direction[3]) {
	double beamDirection[3];
	get_beamDirection(gdata, pos, beamDirection);
	double muMin = 0;
	for (int iDir = 0; iDir < 3; ++iDir) {
		muMin += direction[iDir] * beamDirection[iDir];
	}
	return muMin;
};

#endif


