#include "RadiationField.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)

RadiationField::RadiationField(
		normalisation * normHD,
		normalisation * normParticle
	)
{
	this->normHD = normHD;
	this->normParticle = normParticle;
}

RadiationField::~RadiationField(
		)
{
}

//REAL RadiationField::get_spectralEnergyDensity_LAB(
//		const Data & gdata,
//		int ix, int iy, int iz,
//		const REAL energy_LAB,
//		const REAL dir_LAB[3]
//	)
//{
//	throw CException(" Not Implemented ");
//}
//
//REAL RadiationField::get_spectralEnergyDensity_COM(
//		const Data & gdata,
//		int ix, int iy, int iz,
//		const REAL energy_COM,
//		const REAL dir_COM[3]
//	)
//{
//	throw CException(" Not Implemented ");
//}
//
//REAL RadiationField::get_spectralEnergyDensity_LAB(
//		const Data & gdata,
//		int ix, int iy, int iz,
//		const REAL energy_LAB
//)
//{
//	throw CException(" Not Implemented ");
//}
//
//REAL RadiationField::get_spectralEnergyDensity_COM(
//		const Data & gdata,
//		int ix, int iy, int iz,
//		const REAL energy_COM
//)
//{
//	throw CException(" Not Implemented ");
//}

//REAL RadiationField::get_deltaEnergy_COM(
//		const Data & gdata,
//		int ix,int iy, int iz,
//		const REAL dir[3]
//		)
//{
//	throw CException(" Not Implemented ");
//}
//
//void RadiationField::get_deltaDirection_LAB(
//		const Data & gdata,
//		int ix,int iy, int iz,
//		REAL dir[3]
//		)
//{
//	throw CException(" Not Implemented ");
//}
//
//void RadiationField::get_deltaDirection_COM(
//		const Data & gdata,
//		int ix,int iy, int iz,
//		REAL dir[3]
//		)
//{
//	throw CException(" Not Implemented ");
//}


REAL BeamedRadiationField::get_dE_dVdEdOmega_LAB(
		const Data &gdata,
		int ix, int iy, int iz,
		const REAL energy_LAB,
		const REAL dir_LAB[3]
)
{
	throw CException(" This method cannot be used for this subclass: Delta Distribution");
}

REAL BeamedRadiationField::get_dE_dVdEdOmega_COM(
		const Data &gdata,
		int ix, int iy, int iz,
		const REAL energy_COM,
		const REAL dir_COM[3]
)
{
	throw CException(" This method cannot be used for this subclass: Delta Distribution");
}

REAL IsotropeRadiationField_LAB::get_dE_dVdEdOmega_LAB(
		const Data &gdata,
		int ix, int iy, int iz,
		const REAL energy_LAB,
		const REAL dir_LAB[3]
)
{
	return get_dE_dVdE_LAB(gdata, ix, iy, iz, energy_LAB) * inverseSolidAngle;
}


REAL RadiationField_LAB::get_dE_dVdEdOmega_COM(
		const Data & gdata,
		int ix, int iy, int iz,
		const REAL energy_COM,
		const REAL dir_COM[3]
	)
{
	REAL dir_LAB[3];
	REAL energy_LAB;
	REAL spectralEnergyDensity_LAB;
	REAL spectralEnergyDensity_COM;

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sy()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sz()](ix,iy,iz),
			gdata.lorentz(ix, iy, iz),
			true
			);
	lorentzTrafo.nullLike_direction1(dir_COM, dir_LAB);
	lorentzTrafo.nullLike_temporal1(energy_COM, dir_COM, energy_LAB);

	spectralEnergyDensity_LAB = this->get_dE_dVdEdOmega_LAB(gdata, ix, iy, iz, energy_LAB, dir_LAB);

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sy()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sz()](ix,iy,iz),
			gdata.lorentz(ix, iy, iz)
			);

	lorentzTrafo.nullLike_temporal2(spectralEnergyDensity_LAB, dir_LAB, spectralEnergyDensity_COM);
	return spectralEnergyDensity_COM;
}


REAL IsotropeRadiationField_LAB::get_dE_dVdE_COM(
		const Data & gdata,
		int ix, int iy, int iz,
		const REAL energy_COM
) {
	REAL energy_LAB;
	REAL spectralEnergyDensity_LAB;
	REAL spectralEnergyDensity_COM;

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sy()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sz()](ix, iy, iz),
			gdata.lorentz(ix, iy, iz),
			true
	);
	lorentzTrafo.nullLike_temporal1_iso(energy_COM, energy_LAB);

	spectralEnergyDensity_LAB = this->get_dE_dVdE_LAB(gdata, ix, iy, iz, energy_LAB);

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sy()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sz()](ix, iy, iz),
			gdata.lorentz(ix, iy, iz)
	);

	lorentzTrafo.nullLike_temporal2_iso(spectralEnergyDensity_LAB, spectralEnergyDensity_COM);
	return spectralEnergyDensity_COM;
}



REAL BeamedRadiationField_LAB::get_dE_dVdE_COM(
		const Data & gdata,
		int ix, int iy, int iz,
		const REAL energy_COM
) {
	REAL energy_LAB;
	REAL spectralEnergyDensity_LAB;
	REAL spectralEnergyDensity_COM;

	REAL dir_LAB[3];
	REAL dir_COM[3];

	this->get_beamDirection_COM(gdata, ix, iy, iz, dir_LAB);
	this->get_beamDirection_COM(gdata, ix, iy, iz, dir_COM);

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sy()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sz()](ix, iy, iz),
			gdata.lorentz(ix, iy, iz),
			true
	);
	lorentzTrafo.nullLike_temporal1(energy_COM, dir_COM, energy_LAB);

	spectralEnergyDensity_LAB = this->get_dE_dVdE_LAB(gdata, ix, iy, iz, energy_LAB);

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sy()](ix, iy, iz),
			gdata.om[gdata.fluid.get_q_sz()](ix, iy, iz),
			gdata.lorentz(ix, iy, iz)
	);
	lorentzTrafo.nullLike_temporal2(spectralEnergyDensity_LAB, dir_LAB, spectralEnergyDensity_COM);
	return spectralEnergyDensity_COM;
}

void BeamedRadiationField_LAB::get_beamDirection_COM(
		const Data & gdata,
		int ix,int iy, int iz,
		REAL dir_COM[3]
)
{
	REAL dir_LAB[3];
	this->get_beamDirection_LAB(gdata, ix, iy, iz, dir_LAB);

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sy()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sz()](ix,iy,iz),
			gdata.lorentz(ix, iy, iz)
	);
	lorentzTrafo.nullLike_direction1(dir_LAB, dir_COM);
}

REAL MonochromaticRadiationField_LAB::get_photonEnergy_COM(
		const Data & gdata,
		int ix,int iy, int iz,
		REAL dir_COM[3]
	)
{

	REAL dir_LAB[3];
	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sy()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sz()](ix,iy,iz),
			gdata.lorentz(ix, iy, iz),
			true
	);
	lorentzTrafo.nullLike_direction1(dir_COM, dir_LAB);


	REAL energy_LAB = this->get_photonEnergy_LAB(gdata, ix, iy, iz);
	REAL energy_COM;

	lorentzTrafo.set_boostVector(
			gdata.om[gdata.fluid.get_q_sx()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sy()](ix,iy,iz),
			gdata.om[gdata.fluid.get_q_sz()](ix,iy,iz),
			gdata.lorentz(ix, iy, iz)
			);
	lorentzTrafo.nullLike_temporal1(energy_LAB, dir_LAB, energy_COM);

	return energy_COM;
}


#endif


