#ifndef CRONOS_SYNCHROTRON_H
#define CRONOS_SYNCHROTRON_H

#include <gsl/gsl_sf.h>
#include "EmissionField.H"
#include "EmissivitySpectrumIntegrator.H"
#include "MagneticField.H"
#include "RadiativeLossProcess.H"

class SynchrotronEmission : public EmissionField_COM {
 public:
  SynchrotronEmission(normalisation* normHD, normalisation* normParticle,
                      std::shared_ptr<MagneticField> magneticField);

  ~SynchrotronEmission();

  REAL get_photonEmissivity_COM(Data& gdata, int ix, int iy, int iz,
                                const REAL photonEnergy_COM,
                                const REAL photonDirection_COM[3]);

  void set_particle(Particle& particle);

 protected:
  std::shared_ptr<MagneticField> magneticField;
  Particle* particle;
  LorentzTrafo lorentzTrafo;

  REAL energyCritFac_num;
  REAL singleEmissivityFac_num;

  double compute_singleSpectrum(const double energy, const double lorentzParticle, const double perpMagneticFieldStrength) const;
  double compute_singleSpectrum(const REAL energy, const double lorentzParticle, const double magneticFieldStrength,
          const double magneticFieldDir_COM[3], const double photonDirection_COM[3]) const;

  // Differential Emissivity Integration

  enum class IntegrationType {
    BinnedCenter,
    GaussPowerlaw,
  };

  std::map<std::string, IntegrationType> catParameterMap = {
      {"BinnedConstant", IntegrationType::BinnedCenter},
      {"GaussPowerlaw", IntegrationType::GaussPowerlaw}};

  void setIntegrationType(const IntegrationType);
  NumMatrix<double, 1> particleSpectrumBuffer;
  std::unique_ptr<EmissivitySpectrumIntegrator> emissivitySpectrumIntegrator;
};

class SynchrotronProcess : public RadiativeLossProcess {
 public:
  SynchrotronProcess(normalisation* normHD, normalisation* normParticle,
                     std::shared_ptr<MagneticField> magneticField);

  ~SynchrotronProcess();

  void set_particle(Particle& particle);

  void compute_constants(Data& gdata, int ix, int iy, int iz);

  void get_momentumLoss(REAL momentum, REAL& loss);

  REAL get_momentumLoss_c2(REAL momentum);

  std::string get_type() const final;

 protected:
  std::shared_ptr<MagneticField> magneticField;
  SynchrotronEmission synchrotronEmission;

  REAL synchrotronUnit_num;
  REAL synchrotronFac_num;
};

#endif
