#include "PhysFluxesRHD.H"

PhysFluxesRHD::PhysFluxesRHD(const Data &gdata,
		const CronosFluid &fluid) : PhysFluxes(gdata, fluid) {

}

void PhysFluxesRHD::get_PhysFlux(Data &gdata,
                                ProblemType &Problem,
                                cronos::vector<REAL> &iPos,
                                phys_fields_1D &pf,
                                const int &dir, REAL shift, int iFluid){
  

	pf.flux[q_sx].clear();
	pf.flux[q_sy].clear();
	pf.flux[q_sz].clear();

	NumMatrix<REAL, 1> W(pf.uPri[0]);
	for (int i = -1; i <= gdata.mx[dir]+1; ++i){
			W(i) = sqrt(1. + sqr(pf.uPri[q_sx](i)) + sqr(pf.uPri[q_sy](i)) + sqr(pf.uPri[q_sz](i)));
		}

	// flux for the density
	for (int i = -1; i <= gdata.mx[dir]+1; ++i){
		pf.flux[q_rho](i) = pf.uPri[q_sx+dir](i)*pf.uPri[q_rho](i) ;
	}

	// flux for momentum vector

	for(int q=q_sx; q<=q_sz; ++q) {
		if(q == dir+1) {
			for (int i = -1; i <= gdata.mx[dir]+1; ++i){
				pf.flux[q](i) = pf.uCon[q](i) * pf.uPri[dir+q_sx](i) / W(i)	+ pf.ptherm(i);
			}
		} else {
			for (int i = -1; i <= gdata.mx[dir]+1; ++i){
				pf.flux[q](i) = pf.uCon[q](i) * pf.uPri[dir+q_sx](i) / W(i);
			}
		}
	}

   
	// flux for total energy
	if(ENERGETICS == FULL){
		for (int i = -1; i <= gdata.mx[dir]+1; ++i){
			pf.flux[q_Eges](i) = pf.uCon[dir+q_sx](i) - pf.flux[q_rho](i);
		}
	}

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	// Compute physical flux for entropy evolution -- flux is
	for (int i = -1; i <= gdata.mx[dir]+1; ++i){
		pf.flux[q_Eadd](i) = pf.uPri[q_Eadd](i)*pf.flux[q_rho](i);
	}
#endif



  
}


void PhysFluxesRHD::get_PhysFlux(Data &gdata,
        ProblemType &Problem, phys_fields_0D &fields,
        int ix, int iy, int iz, int face, int iFluid) {


}




