#ifndef CRONOS_RIEMANNSOLVERRHD_H
#define CRONOS_RIEMANNSOLVERRHD_H

#include "RiemannSolver.H"

class RiemannSolverRHD: public RiemannSolver {
public:
	RiemannSolverRHD(){}
	RiemannSolverRHD(const Data &gdata, int, int);
	void get_vChar(Data &, ProblemType &, cronos::vector<double> &,
			phys_fields_1D &, phys_fields_1D &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
//	               phys_fields_1D &, phys_fields_1D &, fields_1D &,
	               const int &, REAL &);
	virtual void get_vChar(Data &, ProblemType &, phys_fields_0D &, phys_fields_0D &,
				num_fields_0D &, int ix, int iy, int iz, int dir, REAL &cfl_lin);
//	void reset_Indices(ProblemType &Problem) {
	void reset_Indices(const CronosFluid &fluid) {
		//! Reset field indices (only necessary for multifluid simulations)
		this->q_rho = fluid.get_q_rho();
		this->q_sx = fluid.get_q_sx();
		this->q_sy = fluid.get_q_sy();
		this->q_sz = fluid.get_q_sz();
		this->q_Eges = fluid.get_q_Eges();
		this->q_Eadd = fluid.get_q_Eadd();
	};
protected:
	int q_rho, q_sx, q_sy, q_sz, q_Eges, q_Eadd;
};


// Generic HLL solver
class HLLSolver_RHD: public RiemannSolverRHD {
public:
	HLLSolver_RHD(const Data &, int, int);
	void get_NumFlux(Data &, cronos::vector<double> &, phys_fields_1D &,
	                 phys_fields_1D &, fields_1D &, const int &, int iFluid=0);
	virtual void get_NumFlux(Data &, phys_fields_0D &,
			phys_fields_0D &, num_fields_0D &, int dir, int iFluid=0);
	~HLLSolver_RHD(){};
private:
	REAL veps;
};


// HLLC Solver for Hydrodynamics
class HLLCSolver_RHD: public RiemannSolverRHD {
public:
	HLLCSolver_RHD(const Data &, int, int);
	void get_NumFlux(Data &, cronos::vector<double> &, phys_fields_1D &,
	                 phys_fields_1D &, fields_1D &, const int &, int iFluid=0);
	virtual void get_NumFlux(Data &, phys_fields_0D &,
			phys_fields_0D &, num_fields_0D &, int dir, int iFluid=0);
	~HLLCSolver_RHD(){};
private:
	int qvPar, qvP1, qvP2;
	REAL veps;
	REAL gamma;
};

#endif
