#include "old/rhd_implicit_trafo.H"
#include "old/rhd_implicit_trafo.C"

primState TransCons2Prim_RHD_1Dw(const consState & cons, const REAL & gamma_adiab)
{
	REAL Press2Etherm = 1./(gamma_adiab - 1.);
	REAL energy = cons.tau + cons.density;
	REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);

	NewtonMethod newtonSolver;
	RHD_TransCons2Wprime_implicit TransCons2Wprime_implicit_rhd;

	TransCons2Wprime_implicit_rhd.set_parameters(cons.density, momentum_sqr, cons.tau, gamma_adiab);
	newtonSolver.set_problem(TransCons2Wprime_implicit_rhd);
	newtonSolver.set_maxIter(20);

	REAL Wprime = newtonSolver.solve((2*energy + sqrt(4*sqr(energy)-3*momentum_sqr))/3. - cons.density, 1e-12);

	REAL pressure = Wprime - cons.tau;
	REAL u_sqr =  momentum_sqr / (sqr(Wprime + cons.density) - momentum_sqr);

	REAL gamma_rel = TransVelSqr2Gamma_RHD(u_sqr);
	REAL Dhgamma = Wprime + cons.density;

	// Overwrite the conservatives with the primitives
	REAL momentum_fac = gamma_rel / Dhgamma;

	primState prim;
	prim.rho = cons.density / gamma_rel;
	prim.Etherm = Press2Etherm*pressure;
	prim.ux = cons.momentum_x * momentum_fac;
	prim.uy = cons.momentum_y * momentum_fac;
	prim.uz = cons.momentum_z * momentum_fac;

	#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	prim.specific_entropy = (gamma_adiab-1) * prim.Etherm * pow(prim.rho, -gamma_adiab);
	#endif

	return prim;
}

bool print;
primState TransCons2Prim_RHD_numericalQuartic(const consState & cons, const REAL & gamma_adiab)
{
	REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
	REAL energy = cons.tau + cons.density;
	REAL gamma_tilde = gamma_adiab/(gamma_adiab-1.);

	REAL density_sqr = sqr(cons.density);
	REAL momentum = sqrt(momentum_sqr);
	REAL energy_sqr = sqr(energy);

	if ( energy_sqr - momentum_sqr < 1e-4 * energy_sqr ) {
		cerr << " Warning: The Numerical Quartic Inversion might yield wrong results. " << (energy_sqr - momentum_sqr) / energy_sqr << endl;
	}

	REAL coeff[5];
	coeff[4] = (momentum_sqr - energy_sqr) * sqr(gamma_tilde);
	coeff[3] = 2.*cons.density*energy*gamma_tilde;
	coeff[2] = gamma_tilde*(energy_sqr*gamma_tilde - 2.*momentum_sqr) - density_sqr;
	coeff[1] = -coeff[3];
	coeff[0] = density_sqr + momentum_sqr;

	PolynomialProblem<4> poly;
	poly.set_coefficients(coeff);

	NewtonMethod newtonSolver;
	newtonSolver.set_problem(poly);
	newtonSolver.set_maxIter(20);

	REAL Wlower = sqrt(
			(-2.*momentum_sqr + energy_sqr*gamma_tilde + energy*sqrt(energy_sqr*sqr(gamma_tilde) - 4.*momentum_sqr*(gamma_tilde-1.))) /
			(2.*(energy_sqr-momentum_sqr)*gamma_tilde)
			);
	REAL Wupper = 1./sqrt(1. - momentum_sqr/energy_sqr + 1e-15);
	Wupper += 1e-6;

	REAL W0 = std::max(1., 0.5*(Wlower + Wupper));
	REAL gamma_rel = newtonSolver.solve(W0, 1e-12);

	gamma_rel = std::max(1., gamma_rel);
	REAL u = sqrt(sqr(gamma_rel)-1.);

	// Overwrite the conservatives with the primitives
	REAL momentum_fac = u/(momentum + 1e-100);

	primState prim;
	prim.rho = cons.density / gamma_rel;
	prim.Etherm = energy - cons.density - momentum * u / gamma_rel;
	prim.ux = cons.momentum_x * momentum_fac;
	prim.uy = cons.momentum_y * momentum_fac;
	prim.uz = cons.momentum_z * momentum_fac;

	#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	prim.specific_entropy = (gamma_adiab-1) * prim.Etherm * pow(prim.rho, -gamma_adiab);
	#endif

	return prim;
}

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
primState TransEntropy2Prim_RHD(const consState & cons, const REAL & gamma_adiab)
{
	REAL Press2Etherm = 1./(gamma_adiab - 1.);
	REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
	REAL momentum = sqrt(momentum_sqr);
	REAL polytropic_K = cons.sigma/cons.density;

	NewtonMethod newtonSolver;
	RHD_TransEntropy2Lorentz_implicit TransEntropy2Lorentz_implicit_rhd;

	TransEntropy2Lorentz_implicit_rhd.set_parameters(cons.density, momentum_sqr, cons.sigma, gamma_adiab);
	newtonSolver.set_problem(TransEntropy2Lorentz_implicit_rhd);

	REAL lorentzUpper = sqrt(1. + momentum_sqr/sqr(cons.density));
	REAL lorentzLower = sqrt(1. + momentum_sqr/(sqr(cons.density + polytropic_K * gamma_adiab/(gamma_adiab - 1) * std::pow(cons.density, gamma_adiab))));
	REAL lorentz0 = lorentzUpper;

	REAL gamma_rel = newtonSolver.solve(lorentz0, 1e-12);

	gamma_rel = std::max(1., gamma_rel);
	REAL u = sqrt(sqr(gamma_rel)-1.);

	// Overwrite the conservatives with the primitives
	REAL Dh_inv = u/(momentum + 1e-100);

	primState prim;
	prim.specific_entropy = polytropic_K;
	prim.rho = cons.density / gamma_rel;

	prim.Etherm = Press2Etherm * prim.specific_entropy * std::pow(prim.rho, gamma_adiab);
	prim.ux = cons.momentum_x * Dh_inv;
	prim.uy = cons.momentum_y * Dh_inv;
	prim.uz = cons.momentum_z * Dh_inv;

	return prim;
}
#endif

primState TransCons2Prim_RHD_polytropic(const consState & cons, const REAL & gamma_adiab, const REAL & polytropic_K)
{
	REAL Press2Etherm = 1./(gamma_adiab - 1.);
	REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);

	NewtonMethod newtonSolver;
	RHD_TransCons2Prim_polytropic_implicit TransCons2Prim_polytropic;

	TransCons2Prim_polytropic.set_parameters(cons.density, momentum_sqr, gamma_adiab, polytropic_K);
	newtonSolver.set_problem(TransCons2Prim_polytropic);
	newtonSolver.set_maxIter(20);

	REAL rhoLower = cons.density / sqrt(1. + momentum_sqr/sqr(cons.density));
	REAL rhoUpper = cons.density / sqrt(1. + momentum_sqr/(sqr(cons.density + polytropic_K * gamma_adiab/(gamma_adiab - 1) * std::pow(cons.density, gamma_adiab))));
	REAL rho0 = sqrt(rhoLower*rhoUpper);
	// REAL rho = newtonSolver.solve(rho0, 1e-12);
	// rho = max(rhoLower, rho);
	// rho = min(rhoUpper, rho);

	auto fun = [](double rho, void * params) -> double {
		RHD_TransCons2Prim_polytropic_implicit & implicit_fun = * static_cast<RHD_TransCons2Prim_polytropic_implicit *>(params);
		implicit_fun.set_x(rho);
		return implicit_fun.get_value();
	};

	gsl_function F;
	F.function = fun;
	F.params = &TransCons2Prim_polytropic;

	static gsl_root_fsolver * gsl_root_f = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
	gsl_root_fsolver_set(gsl_root_f, &F, rhoLower, rhoUpper);

  int iter = 0, max_iter = 20;
	double rho, rho_lo, rho_hi;
  int status;
  do
    {
      iter++;
      status = gsl_root_fsolver_iterate (gsl_root_f);
      rho = gsl_root_fsolver_root(gsl_root_f);
      rho_lo = gsl_root_fsolver_x_lower(gsl_root_f);
      rho_hi = gsl_root_fsolver_x_upper(gsl_root_f);
      status = gsl_root_test_interval(rho_lo, rho_hi, 0, 1e-6);
    }
  while (status == GSL_CONTINUE && iter < max_iter);

	REAL pressure = polytropic_K * pow(rho, gamma_adiab);
	REAL Dh_inv = 1./(cons.density * (1. + gamma_adiab/(gamma_adiab - 1.) * pressure/rho));

	// Overwrite the conservatives with the primitives
	primState prim;
	prim.rho = rho;
	prim.Etherm = Press2Etherm * pressure;
	prim.ux = cons.momentum_x * Dh_inv;
	prim.uy = cons.momentum_y * Dh_inv;
	prim.uz = cons.momentum_z * Dh_inv;

	#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	prim.specific_entropy = polytropic_K;
	#endif

	return prim;

}