#ifndef RHD_IMPLICIT_TRAFO_H
#define RHD_IMPLICIT_TRAFO_H

#include "constants.H"
#include "root_finder.H"

class RHD_TransCons2P_implicit: public ExplicitSlopeProblem {
public:
	RHD_TransCons2P_implicit(): ExplicitSlopeProblem("RHD_TransCons2P_implicit") {};
	virtual void set_x(REAL);
	virtual REAL get_value();
	virtual REAL get_slope();
	void set_parameters(REAL, REAL, REAL, REAL);
private:
	REAL energy, density, momentum_sqr, gamma_tilde;
	REAL energy_tilde, gamma_rel, gamma_rel_sqr, pressure;
};


class RHD_TransCons2Wprime_implicit: public ExplicitSlopeProblem {
public:
	RHD_TransCons2Wprime_implicit(): ExplicitSlopeProblem("RHD_TransCons2Wprime_implicit") {};
	virtual void set_x(REAL);
	virtual REAL get_value();
	virtual REAL get_slope();
	REAL get_pressure();
	void set_parameters(REAL, REAL, REAL, REAL);
private:
	REAL tau, density, momentum_sqr, gamma_tilde;
	REAL u_sqr, gamma_rel, gamma_rel_sqr, Wprime, chi;
	REAL eos_dp_dchi, eos_dp_drho, dchi_dL, drho_dL;
};


class RHD_TransConsP2E_implicit: public ExplicitSlopeProblem {
public:
	RHD_TransConsP2E_implicit(): ExplicitSlopeProblem("RHD_TransConsP2E_implicit") {};
	virtual void set_x(REAL);
	virtual REAL get_value();
	virtual REAL get_slope();
	void set_parameters(REAL, REAL, REAL, REAL);
private:
	REAL pressure, density, momentum_sqr, gamma_tilde;
	REAL energy_tilde, gamma_rel, energy;
};

class RHD_TransEntropy2Lorentz_implicit: public ExplicitSlopeProblem {
public:
	RHD_TransEntropy2Lorentz_implicit(): ExplicitSlopeProblem("RHD_TransEntropy2Lorentz_implicit") {};
	virtual void set_x(REAL);
	virtual REAL get_value();
	virtual REAL get_slope();
	void set_parameters(REAL, REAL, REAL, REAL);
private:
	REAL density, momentum_sqr, sigma, gamma_tilde, gamma_adiab_red;
	REAL lorentz, Z, lorentz_sqr, Z_sqr;
};



class RHD_TransCons2Prim_polytropic_implicit: public ExplicitSlopeProblem {
public:
	RHD_TransCons2Prim_polytropic_implicit(): ExplicitSlopeProblem("RHD_TransCons2Prim_polytropic_implicit") {};
	virtual void set_x(REAL);
	virtual REAL get_value();
	virtual REAL get_slope();
	void set_parameters(REAL, REAL, REAL, REAL);
private:
	REAL density, momentum_sqr, polytropic_K, gamma_tilde, gamma_adiab;
	REAL rho, enthalpy, u_sqr, lorentz_sqr, lorentz, PoverRho;
};



#endif
