#include "specific.H"


template <typename... Args>
void printPack(std::ostream& out, Args&&... args)
{
    using expander = int[];
    (void)expander{0, (void(out << std::forward<Args>(args) << " "),0)...};
    out << endl;
}

template <typename... Args>
inline bool assert_print(bool condition, Args&&... args) {
	if (!condition) {
		printPack(cerr, " Failed Assertion Message: ", args...);
	}
	return condition;
}


void HyperbolicSolver::phystest_rhd(Data &gdata, gridFunc &gfunc,
                                ProblemType &Problem, int rkstep, int iFluid)
{

	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
				assert(assert_print(gdata.om[q_Eadd](ix, iy, iz) > 0., "PhysTest - s |", gdata.rank, ix, iy, iz, "->", gdata.om[q_Eadd](ix, iy, iz)));
#endif
				assert(assert_print(gdata.om[q_rho](ix, iy, iz) > 0., "PhysTest - rho |", gdata.rank, ix, iy, iz, "->", gdata.om[q_rho](ix, iy, iz)));
				assert(assert_print(gdata.om[q_Eges](ix, iy, iz) > 0., "PhysTest - Etherm |", gdata.rank, ix, iy, iz, "->", gdata.om[q_Eges](ix, iy, iz)));

#if (PARTICLE_TRANSPORT == CRONOS_ON)
				for (int ip = omb_particle; ip <= ome_particle; ip++)
					assert(assert_print(gdata.om_user[ip](ix, iy, iz) >= 0., "PhysTest - Particle ", ip, " |", gdata.rank, ix, iy, iz, "->", gdata.om_user[ip](ix, iy, iz)));
#endif

			}
		}
	}

	// Writing to screen -- if desired
	if((rkstep == TIME_SUBSTEPS-1)) {
		if (Problem.get_Info() && Problem.checkout(3)) {
			Problem.writePhystestInfo(gdata);
		}
	}






/*
	bool init(false);
	if(rkstep < 0) {
		init = true;
		rkstep = TIME_SUBSTEPS-1;
	}

	REAL dV(gdata.get_CellVolume(0,0,0));
	double Vol(gdata.get_Volume());
	double Mass(0.), Eges(0.);
	double mv[3] = {0., 0., 0.};
	double Ekin(0.), Emag(0.), Etherm(0.);
	double cs(0.);
	double Ekfluc(0.), Ebfluc(0.);
	double Vorticity(0.), Enstrophy(0.);
	double sqrdivv(0.), sqrdissv(0.);
	double v2ave(0.);
	REAL mvrms[3] = {0., 0., 0.};


#ifdef parallel
	gdata.MpiSum(Vol);
#endif

	// For multifluid case use only relevant range of fields
#if(FLUID_TYPE==CRONOS_MULTIFLUID)
//	cout << " Indizes: " << q_sx << " " << q_sy << endl;
	int q_min = gdata.fluids->fluids[iFluid].get_IndexGlobal(0);
	int q_Eges = gdata.fluids->fluids[iFluid].get_q_Eges_global();
	n_omInt = gdata.fluids->get_N_OMINT(iFluid);
	int q_rho = gdata.fluids->fluids[iFluid].get_q_rho_global();
	int q_sx = gdata.fluids->fluids[iFluid].get_q_sx_global();
	int q_sy = gdata.fluids->fluids[iFluid].get_q_sy_global();
	int q_sz = gdata.fluids->fluids[iFluid].get_q_sz_global();
	int q_Bx = gdata.fluids->fluids[iFluid].get_q_Bx_global();
	int q_By = gdata.fluids->fluids[iFluid].get_q_By_global();
	int q_Bz = gdata.fluids->fluids[iFluid].get_q_Bz_global();
#else
	int q_min = 0;
#endif

	// Enforce min and max values if desired by the user -- this is
	// done just at the beginning in order to compute volume
	// integrated quantities for the corrected values.
	for(int q=q_min; q<n_omInt; ++q) {
		if(Problem.force_min(q)) {
			gdata.om[q].set_min(Problem.min_Val(q));
		}
		if(Problem.force_max(q)) {
			gdata.om[q].set_max(Problem.max_Val(q));
		}
	}

	// Conservative quantities only to be evaluated, when output is
	// desired -- not anymore...
	//  if(Problem.get_AsciiOut() || Problem.get_Info()) {

	// Use conservative variables:
#if (ENERGETICS == FULL)
	// Check for primitives!

#if(FLUID_TYPE==CRONOS_MULTIFLUID)
	TrafoMulti[iFluid]->TransPrim2Cons(gdata, gfunc, Problem);
#else
	Trafo->TransPrim2Cons(gdata, gfunc, Problem);
#endif

#endif


	// Compute conservative quantities:
	Mass = gdata.computeInt(q_rho);

	mv[0] = gdata.computeInt(q_sx);
	mv[1] = gdata.computeInt(q_sy);
	mv[2] = gdata.computeInt(q_sz);


#if(ENERGETICS == FULL)
	Eges = gdata.computeInt(q_Eges);
#endif

	mvrms[0] = gdata.computeRMS(q_sx);
	mvrms[1] = gdata.computeRMS(q_sy);
	mvrms[2] = gdata.computeRMS(q_sz);

	REAL momentum_sqr(0.);
	bool physFlag;
	for (int iz = 0; iz <= gdata.mx[2]; iz++) {
		for (int iy = 0; iy <= gdata.mx[1]; iy++) {
			for (int ix = 0; ix <= gdata.mx[0]; ix++) {
				momentum_sqr = sqr(gdata.om[q_sx](ix, iy, iz)) +
							sqr(gdata.om[q_sy](ix, iy, iz)) +
							sqr(gdata.om[q_sz](ix, iy, iz));

				//physFlag = Trafo->isPhysicalState_RHD(gdata.om[q_rho](ix, iy, iz), momentum_sqr, gdata.om[q_Eges](ix, iy, iz));
				physFlag = (
							sqr(gdata.om[q_rho](ix, iy, iz)) + momentum_sqr <
							sqr(gdata.om[q_Eges](ix, iy, iz) + gdata.om[q_rho](ix, iy, iz))
							);
				if (!physFlag) {
					cout << "Phystest: Unphysical Conservative state for "
							<< gdata.rank << " " << ix << " " << iy << " " << iz << endl;
				}
			}
		}
	}


	// Making certain to use primitve Variables for the rest of the
	// computations

#if(FLUID_TYPE==CRONOS_MULTIFLUID)
	TrafoMulti[iFluid]->TransCons2Prim(gdata, gfunc, Problem);
#else
	Trafo->TransCons2Prim(gdata, gfunc, Problem);
#endif



	if((rkstep == TIME_SUBSTEPS-1)) {

		// Energetics only to be evaluated, when output is desired
		//    if((Problem.get_AsciiOut() || Problem.get_Info()) && !init) {
		// Now energetics is computed every time -- values are needed for averaging

		// Compute Energetics & speed of sound
		Vol = 0.;
		Ekin = Eges;

		for (int k = 0; k <= gdata.mx[2]; k++) {
			for (int j = 0; j <= gdata.mx[1]; j++) {
				for (int i = 0; i <= gdata.mx[0]; i++) {

					if(ENERGETICS == FULL) {
						Etherm += gdata.om[q_Eges](i,j,k)*dV;
						Ekin -= gdata.om[q_Eges](i,j,k)*dV;			// Cheat ?!?
						// cs += sqrt((yama - 1.)*gdata.om[q_Eges](i,j,k)/gdata.om[q_rho](i,j,k))*dV; warum gamma - 1 ?!?
						cs += sqrt(yama * (yama - 1) * gdata.om[q_Eges](i,j,k)/(gdata.om[q_rho](i,j,k) + yama * gdata.om[q_Eges](i,j,k)))*dV;
					} else {
						cs += Problem.c2_iso(gdata,i,j,k)*dV;
					}

					Vol += dV;
				}
			}
		}


#ifdef parallel
		gdata.MpiSum(Ekin);
		gdata.MpiSum(Etherm);
		gdata.MpiSum(cs);
		gdata.MpiSum(Vol);
#endif

		cs = cs/Vol;
		//     if(N_OMINT < 8) {
		//       Etherm = Mass*sqr(cs);
		//       if(yama != 1.) {
		// 	Etherm /= (yama - 1.);
		//       }
		//       Eges = Etherm + Ekin + Emag;
		//     }
		if(ENERGETICS != FULL) {
			Eges = Ekin;
		}


		// Set initial values at first step:
		if(init) {
			Mass0 = Mass;
			mv0[0] = mv[0];
			mv0[1] = mv[1];
			mv0[2] = mv[2];

			Eges0 = Eges;
		}

		// Saving time-averaged data:
		Ekin_tave += gdata.dt*Ekin;
		Etherm_tave += gdata.dt*Etherm;
		Eges_tave += gdata.dt*Eges;
		cs_tave += gdata.dt*cs;
		time_ave += gdata.dt;


		// Compute Velocity Stats:
		Vol = 0.;
		REAL hx[3]= {(gdata.hx[0]), (gdata.hx[1]), (gdata.hx[2])};

		for (int k = 0; k < gdata.mx[2]; k++) {
			for (int j = 0; j < gdata.mx[1]; j++) {
				for (int i = 0; i < gdata.mx[0]; i++) {

					double dxVx = (gdata.om[q_sx](i+1,j,k) -
					               gdata.om[q_sx](i-1,j,k))*hx[0];
					double dxVy = (gdata.om[q_sy](i+1,j,k) -
					               gdata.om[q_sy](i-1,j,k))*hx[0];
					double dxVz = (gdata.om[q_sz](i+1,j,k) -
					               gdata.om[q_sz](i-1,j,k))*hx[0];
					double dyVx = (gdata.om[q_sx](i,j+1,k) -
					               gdata.om[q_sx](i,j-1,k))*hx[1];
					double dyVy = (gdata.om[q_sy](i,j+1,k) -
					               gdata.om[q_sy](i,j-1,k))*hx[1];
					double dyVz = (gdata.om[q_sz](i,j+1,k) -
					               gdata.om[q_sz](i,j-1,k))*hx[1];
					double dzVx = (gdata.om[q_sx](i,j,k+1) -
					               gdata.om[q_sx](i,j,k-1))*hx[2];
					double dzVy = (gdata.om[q_sy](i,j,k+1) -
					               gdata.om[q_sy](i,j,k-1))*hx[2];
					double dzVz = (gdata.om[q_sz](i,j,k+1) -
					               gdata.om[q_sz](i,j,k-1))*hx[2];



					REAL curl_x = (dyVz - dzVy);
					REAL curl_y = (dzVx - dxVz);
					REAL curl_z = (dxVy - dyVx);

					REAL curl_val = (sqr(curl_x) + sqr(curl_y) + sqr(curl_z));

					Vorticity += sqrt(curl_val)*dV;
					Enstrophy += (curl_val)*dV;

					sqrdivv  += sqr(dxVx + dyVy + dzVz);
					sqrdissv += (sqr(dxVx) + sqr(dyVx) + sqr(dzVz) +
					             sqr(dxVy) + sqr(dyVy) + sqr(dzVy) +
					             sqr(dxVz) + sqr(dyVz) + sqr(dzVz));

					v2ave += (sqr(gdata.om[q_sx](i,j,k)) +
					          sqr(gdata.om[q_sy](i,j,k)) +
					          sqr(gdata.om[q_sz](i,j,k)))*dV;

					Vol += dV;

				}
			}
		}
#ifdef parallel
		gdata.MpiSum(Vorticity);
		gdata.MpiSum(Enstrophy);
		gdata.MpiSum(sqrdivv);
		gdata.MpiSum(sqrdissv);
		gdata.MpiSum(Vol);
#endif
		v2ave /= Vol;

		Problem.computeFluct(gdata, Ekfluc, Ebfluc);

#ifdef parallel
		gdata.MpiSum(Ekfluc);

		//     Ebfluc /= gdata.nproc[0]*gdata.nproc[1]*gdata.nproc[2];
		//     Ekfluc /= gdata.nproc[0]*gdata.nproc[1]*gdata.nproc[2];
#endif


		// Compute temporal averages (in case no output at every timestep
		Vorticity_tave += gdata.dt*Vorticity;
		Enstrophy_tave += gdata.dt*Enstrophy;
		v2ave_tave += gdata.dt*v2ave;
		Ekfluc_tave += gdata.dt*Ekfluc;

		// Compute quantities in derived class (if necessary)
		Problem.computePhystest(gdata);

		// Writing to screen -- if desired
		if(Problem.get_Info() && Problem.checkout(3)) {
			double MachAve = sqrt(v2ave)/cs;

			if(gdata.rank == 0) {
#if(FLUID_TYPE == CRONOS_MULTIFLUID)
				cout << " Fluid: " << iFluid << endl;
#endif
				cout << setiosflags(ios::scientific) << setprecision(3) << setw(12);
				cout << "-------Fluctuations-----------------------------------" << endl;
				cout << " E_kin:  " << setw(9) << Ekfluc;

				cout << " E_dyn:  " << setw(9) << Ekfluc + Ebfluc << endl;
				cout << "-------Energertics:-----------------------------------" << endl;
				cout << " E_ges:  " << setw(9) << Eges;
				if(ENERGETICS == FULL){
					cout << " E_th:   " << setw(9) << Etherm;
				}


				cout << endl;
				cout << "-------Dynamics:--------------------------------------" << endl;
				cout << " Vorti.: " << setw(9) << Vorticity;
				cout << " Enstr.: " << setw(9) << Enstrophy;
				cout << " M_rms:  " << setw(9) << MachAve << endl;
				cout << "-------Conserved quantities:--------------------------" << endl;
				cout << " Mass: " << setw(9) << Mass << " err: ";
				cout << setw(9) << Mass-Mass0 << endl;
				cout << " p_x:  " << setw(9) << mv[0] << " err: ";
				cout << setw(9)<< (mv[0]-mv0[0])/(mvrms[0]+1.e-20) << endl;
				cout << " p_y:  " << setw(9) << mv[1] << " err: ";
				cout << setw(9)<< (mv[1]-mv0[1])/(mvrms[1]+1.e-20) << endl;
				cout << " p_z:  " << setw(9) << mv[2] << " err: ";
				cout << setw(9) << (mv[2]-mv0[2])/(mvrms[2]+1.e-20) << endl;

				if(ENERGETICS == FULL) {
					cout << " Eges: " << setw(9) << Eges << " err: ";
					cout << setw(9) << Eges-Eges0 << endl;
				}
				// 	if(gdata.time == 0.) {
				// 	  cout << "------------------------------------------------------";
				// 	  cout << endl;
				// 	}
				cout << resetiosflags(ios::scientific) << setprecision(6);
			}
			Problem.writePhystestInfo(gdata);
		}



		// Writing to output file -- if desired
		// Only here we use the temporal average:


		if(Problem.get_AsciiOut()) {
			Problem.set_AsciiOut(false);

			Ekfluc = Ekfluc_tave/time_ave;

			Vorticity = Vorticity_tave/time_ave;
			Etherm = Etherm_tave/time_ave;

			char fileenerg[255];
			sprintf(fileenerg,"%s/%s_efluct",getenv("poub"),getenv("pname"));

			ofstream outfile;
			if(gdata.rank == 0) {
				if(gdata.time == 0. && init) {
					outfile.open(fileenerg, ios::trunc | ios::out);
				} else {
					outfile.open(fileenerg, ios::app | ios::out);
				}

				outfile << gdata.time << " " << Ekfluc << " " << Ebfluc;
				outfile << " " << Ekfluc+Ebfluc << " " << Vorticity << " " << Etherm;
			}
			Problem.writePhystest(gdata, outfile);
			if(gdata.rank == 0) {
				outfile << endl;
				outfile.close();
			}

			Ekin_tave = 0.;

			Etherm_tave = 0.;
			Eges_tave = 0.;
			cs_tave = 0.;
			Vorticity_tave = 0.;
			Enstrophy_tave = 0.;
			v2ave_tave = 0.;
			Ekfluc_tave = 0.;

			time_ave = 0.;

		}

	}

#if(ENERGETICS == FULL)
	if(!thermal) {
#if(FLUID_TYPE==CRONOS_MULTIFLUID)
		TrafoMulti[iFluid]->TransEth2T(gdata, gfunc, Problem);
#else
		Trafo->TransEth2T(gdata, gfunc, Problem);
#endif
	}
#endif



	// for(int q=0; q<N_OMINT; ++q) {
	// 	if(Problem.force_max(q)) {
	// 		gdata.om[q].set_min(Problem.min_Val(q));
	// 	}
	// 	if(Problem.force_min(q)) {
	// 		gdata.om[q].set_max(Problem.max_Val(q));
	// 	}
	// }

	// Errors will be evaluated every time
	TempNeg = 0;
	int erval(0), evval(0), epval(0);
	int e_nan[N_OMEGA] = {0};
	double Emin(1.e20);

	for (int k = 0; k <= gdata.mx[2]; k++) {
		for (int j = 0; j <= gdata.mx[1]; j++) {
			for (int i = 0; i <= gdata.mx[0]; i++) {

				REAL vabs = sqrt(sqr(gdata.om[q_sx](i,j,k)) +
				                 sqr(gdata.om[q_sy](i,j,k)) +
				                 sqr(gdata.om[q_sz](i,j,k)));

				if (gdata.om[q_rho](i,j,k) < rhomin) {
					cout << " Density below zero for " << gdata.rank << " " << i << " " << j << " " << k << " " << gdata.om[q_rho](i,j,k) << endl;
					exit(2);
					erval += 1;
				}

				if (vabs > vmax) {
					cout << gdata.rank << " " << i << " " << j << " ";
					cout << k << " " << vabs << " ";
					cout << gdata.om[q_sx](i,j,k) << " ";
					cout << gdata.om[q_sy](i,j,k) << " ";
					cout << gdata.om[q_sz](i,j,k) << endl;
					evval += 1;
				}

				// If dual energy is used the minimum thermal energy
				// is set to zero.
#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
				if(ENERGETICS == FULL && (gdata.om[q_Eges](i,j,k) < 0.)) {
#else
				if(ENERGETICS == FULL && (gdata.om[q_Eges](i,j,k) < Ethmin)) {
#endif
					TempNeg += 1;
					cout << " T < 0 for " << gdata.rank << " " << i << " " << j << " " << k << " " << gdata.om[q_Eges](i,j,k) << endl;
					cout << gdata.om[q_rho](i,j,k) << " ";
					cout << gdata.om[q_sx](i,j,k) << " ";
					cout << gdata.om[q_sy](i,j,k) << " ";
					cout << gdata.om[q_sz](i,j,k) << " ";
//					cout << gdata.om[q_Bx](i,j,k) << " ";
//					cout << gdata.om[q_By](i,j,k) << " ";
//					cout << gdata.om[q_Bz](i,j,k) << " ";
					cout << endl;
					epval += 1;
				}

				for (int q = 0; q < N_OMEGA; q++) {
					if (isnan(gdata.om[q](i,j,k))) {
						e_nan[q] += 1;
					}
				}

				if (ENERGETICS==FULL) {
					Emin = std::min(Emin, gdata.om[q_Eges](i,j,k));
				}
			}
		}
	}

	int errnum(0);
	errnum += erval;
	errnum += evval;
	errnum += epval;
	for (int q = 0; q < N_OMEGA; q++) {
		errnum += e_nan[q];
	}


	// Write error file only if errors did occur:
	if(errnum > 0) {

		char filecd[255];
		char filerr[255];

		sprintf(filecd,"%s",getenv("pname"));
#ifdef parallel
		sprintf(filerr,"%s.err.%2.2d",filecd,gdata.rank);
#else
		sprintf(filerr,"%s.err",filecd);
#endif


		FILE *fo;
		if ((fo = fopen(filerr,"w")) == NULL) {
			throw CException("cannot open error file!");
		}

		fprintf(fo,"phystest called with the following parameters:\n");
		fprintf(fo," rhomin = %g\n",rhomin);
		fprintf(fo," Ethmin = %g\n",Ethmin);
		fprintf(fo," vmax   = %g\n",vmax);
		fprintf(fo," time   = %g\n",gdata.time);

		if(debug > 0) {
			for (int k = 0; k <= gdata.mx[2]; k++) {
				double zz = gdata.getCen_z(k);
				for (int j = 0; j <= gdata.mx[1]; j++) {
					double yy = gdata.getCen_y(j);
					for (int i = 0; i <= gdata.mx[0]; i++) {
						double xx = gdata.getCen_x(i);

						REAL vabs = sqrt(sqr(gdata.om[q_sx](i,j,k)) +
						                 sqr(gdata.om[q_sy](i,j,k)) +
						                 sqr(gdata.om[q_sz](i,j,k)));

						if (gdata.om[q_rho](i,j,k) < rhomin) {
							fprintf(fo," *** rho < rhomin ***\n");
							fprintf(fo," -> (i,j,k) = (%3d,%3d,%3d)\n",i,j,k);
							fprintf(fo," -> (x,y,z) = (%5.2f,%5.2f,%5.2f)\n",xx,yy,zz);
							fprintf(fo," -> value   = %g\n",gdata.om[q_rho](i,j,k));
						}

						if (vabs > vmax) {
							fprintf(fo," *** v   > vmax ***\n");
							fprintf(fo," -> (i,j,k) = (%3d,%3d,%3d)\n",i,j,k);
							fprintf(fo," -> (x,y,z) = (%5.2f,%5.2f,%5.2f)\n",xx,yy,zz);
							fprintf(fo," -> value   = %g\n",vabs);
							fprintf(fo," -> rho*vx  = %g\n",gdata.om[q_sx](i,j,k));
							fprintf(fo," -> rho*vy  = %g\n",gdata.om[q_sy](i,j,k));
							fprintf(fo," -> rho*vz  = %g\n",gdata.om[q_sz](i,j,k));
							fprintf(fo," -> rho     = %g\n",gdata.om[q_rho](i,j,k));
						}

						for (int q = 0; q < N_OMEGA; q++) {
							if (isnan(gdata.om[q](i,j,k))) {
								fprintf(fo,"err: NaN @(ijk) = (%3d %3d %3d)",i,j,k);
								fprintf(fo," # q=%2d\n",q);
								fprintf(fo," -> (x,y,z) = (%5.2f,%5.2f,%5.2f)\n",xx,yy,zz);
							}
						}

						if(ENERGETICS == FULL && (gdata.om[q_Eges](i,j,k) < Ethmin)) {
							fprintf(fo," *** Etherm < Ethmin ***\n");
							fprintf(fo," -> (i,j,k) = (%3d,%3d,%3d)\n",i,j,k);
							fprintf(fo," -> (x,y,z) = (%5.2f,%5.2f,%5.2f)\n",xx,yy,zz);
							fprintf(fo," -> value   = %g\n",gdata.om[q_Eges](i,j,k));
						}

					}
				}
			}
		}
		fclose(fo);
	}

	bool abort(true);
	int nan_sum(0);
#ifdef parallel
	gdata.MpiSum(TempNeg);
	gdata.MpiSum(erval);
	gdata.MpiSum(evval);
	gdata.MpiSum(epval);

	int e_nanglobal[N_OMEGA];
	MPI_Reduce(&e_nan, &e_nanglobal, N_OMEGA, MPI_INT,
	           MPI_SUM, 0, gdata.comm3d);
	MPI_Bcast(&e_nanglobal, N_OMEGA, MPI_INT, 0, gdata.comm3d);
//	MPI_Barrier(gdata.comm3d);

	for (int q = 0; q < N_OMEGA; q++){
		e_nan[q] = e_nanglobal[q];
	}
#endif
	for (int q = 0; q < N_OMEGA; q++) {
		nan_sum += e_nan[q];  // Total NAN-count
	}
	// Total error count
	int esum = erval + epval + evval + nan_sum;



	if(erval > 0){
		if(gdata.rank == 0) {
			cout << " Negative densities: " << erval << endl;
		}
		abort = true;
	}

	if(ENERGETICS == FULL && TempNeg != 0 && gdata.rank == 0){
		cout << " Negative Temps: " << TempNeg << endl;
	}




	if (esum > 0) {
		if(gdata.rank==0) {
			cout << "phystest failed!" << endl;

			printf("\nat time = %f:\n",gdata.time);
			if (erval > 0) printf("rho    < rmin at %5d grid points\n",erval);
			if (epval > 0) printf("pgas   < pmin at %5d grid points\n",epval);
			if (evval > 0) printf("v      > vmax at %5d grid points\n",evval);
			for (int q = 0; q < N_OMEGA; q++) {
				if (e_nan[q] > 0) {
					printf("om[%2d] = NaN at %5d grid points\n", q, e_nan[q]);
				}
			}
		}
		abort = true;
	} else {
		abort = false;
	}

	if(abort && gdata.time == 0.) {
		throw CException(" Unphysical initial configuration ");
	} else if (abort) {
		throw CException(" Unphysical configuration ");
	} else {
		return;
	}

	*/
}

