#include "root_finder.H"
#include <iostream>

template<int degree>
REAL PolynomialProblem<degree>::get_value(){
	REAL val = coeff[degree];
	for (int i=degree - 1; i >= 0; i--){
		val *= x;
		val += coeff[i];
	}
	return val;
}

template<int degree>
REAL PolynomialProblem<degree>::get_slope(){
	REAL val = coeffDeriv[degree - 1];
	for (int i=degree - 2; i >= 0; i--){
		val *= x;
		val += coeffDeriv[i];
	}
	return val;
}


template<int degree>
void PolynomialProblem<degree>::set_coefficients(REAL * _coeff) {
	coeff[0] = _coeff[0];
	for (int i=1; i <= degree; i++){
		coeff[i] = _coeff[i];
		coeffDeriv[i-1] = _coeff[i]*i;
	}
}




template<class problemType>
RootFinder<problemType>::RootFinder(){
	nIter = 0;
	maxIter = -1;
	problem = NULL;
}

template<class problemType>
REAL RootFinder<problemType>::solve(REAL x0){
	return solve(x0, 1e-15);
}

template<class problemType>
REAL RootFinder<problemType>::solve(REAL x, REAL tol){
	REAL dx;
	REAL err = 0.;
	nIter = 0;

	if (maxIter > 0) {
		do {
			nIter++;
			dx = get_dx(x);
			x += dx;
			err = std::abs(dx / x);
		} while(err > tol && nIter < maxIter);

		if ( err > tol ) {
			std::cerr << " Root finder: not converged within max amount of iterations. " << problem->get_name() << " - " << err << std::endl;
		}

	} else {

		do {
			nIter++;
			dx = get_dx(x);
	//		std::cout << x <<"\t+\t" << dx << std::endl;
			x += dx;
			err = std::abs(dx / x);
	//		std::cout << "Iteration: " << nIter << " -> Rel Err " << std::abs(dx / x) << std::endl;
		} while(err > tol);
	}


	return x;
}

template<class problemType>
void RootFinder<problemType>::set_problem(problemType &problem_) {
	problem = &problem_;
}


template<class problemType>
int RootFinder<problemType>::get_nIter(){
	return nIter;
}


template<class problemType>
void RootFinder<problemType>::set_maxIter(int maxIter){
	this->maxIter = maxIter;
}


REAL NewtonMethod::get_dx(REAL x){
	problem->set_x(x);
	return - (problem->get_value())/problem->get_slope();
}



// ------------------------------------------------------------
//
// 			Some Inits

template class PolynomialProblem<2>;
template class PolynomialProblem<3>;
template class PolynomialProblem<4>;
template class PolynomialProblem<5>;

template class RootFinder<ExplicitSlopeProblem>;
