#ifndef CRONOS_ROOT_FINDER_H
#define CRONOS_ROOT_FINDER_H


#include <cmath>
#include <string>
#include "constants.H"

class RootFinderProblem {
public:
	RootFinderProblem() { this->problemName = "";}
	RootFinderProblem(std::string problemName) { this->problemName = problemName;}
	virtual void set_x(REAL x_) {x = x_;}
	virtual REAL get_value() = 0;
	std::string get_name() {return problemName;}

protected:
	REAL x;
	std::string problemName;
};

class ExplicitSlopeProblem: public RootFinderProblem {
public:
	ExplicitSlopeProblem(): RootFinderProblem() {};
	ExplicitSlopeProblem(std::string problemName): RootFinderProblem(problemName) {};
	virtual REAL get_slope() = 0;
};

template<int degree>
class PolynomialProblem: public ExplicitSlopeProblem {
public:
	PolynomialProblem(): ExplicitSlopeProblem("Polynomial Problem") {};
	virtual REAL get_value();
	virtual REAL get_slope();
	void set_coefficients(REAL *);
protected:
	// coefficents are set and stored
	// from the zero-th to the highest order
	REAL coeff[degree + 1];
	REAL coeffDeriv[degree];
};






template<class problemType>
class RootFinder {
public:
	RootFinder();
	void set_problem(problemType &problem);
	virtual REAL solve(REAL);
	virtual REAL solve(REAL, REAL);

	int get_nIter();
	void set_maxIter(int maxIter);


protected:
	virtual REAL get_dx(REAL) = 0;

	problemType* problem;
	int nIter;
	int maxIter;
};

class NewtonMethod: public RootFinder<ExplicitSlopeProblem> {
public:
	NewtonMethod(): RootFinder() {};
protected:
	virtual REAL get_dx(REAL);
};







#endif
