#include "transformations.H"
#include <cassert>
#include "state_correction.C"
#include "state.C"


#if (FLUID_TYPE == CRONOS_RHD)


template <typename... Args>
void stringstreamPack(std::stringstream& ss, Args&&... args)
{
	using expander = int[];
	(void)expander{0, (void(ss << std::forward<Args>(args) << " "),0)...};
}

template <typename... Args>
void printPack(std::ostream& out, Args&&... args)
{
	stringstream ss;
	stringstreamPack(ss, args...);
	out << ss.str() << endl;
}

template <typename... Args>
inline bool assert_print(bool condition, Args&&... args) {
	if (!condition) {
		printPack(cerr, " Failed Assertion Message: ", args...);
	}
	return condition;
}


inline REAL TransVelSqr2GammaSqr_RHD(const REAL & vel_sqr)
{
	return 1. + vel_sqr;
}

inline REAL TransVelSqr2Gamma_RHD(const REAL & vel_sqr)
{
	return sqrt(TransVelSqr2GammaSqr_RHD(vel_sqr));
}

inline REAL TransPrim2Enthalpy_RHD(const REAL & gamma_adiab, const REAL & rho, const REAL & pressure)
{
	return 1. + pressure/rho * gamma_adiab/(gamma_adiab - 1.);
}


static double _dummy_indicator;

inline bool isAdmissible(REAL density, REAL momentum_sqr, REAL tau, REAL & indicator) {
//	indicator = sqr(tau + density) - (sqr(density) + momentum_sqr);
//	return (indicator > 0. && density > 0.);

	indicator = sqr(tau + density)/(sqr(density) + momentum_sqr);
	return (indicator > 1.);
}

inline bool isAdmissible(REAL density, REAL mx, REAL my, REAL mz, REAL tau, REAL & indicator) {
	return isAdmissible(density, sqr(mx)+sqr(my)+sqr(mz), tau, indicator);
}

inline bool Transformations::isPhysicalState_RHD(Data &gdata, int ix, int iy, int iz) {
	return isPhysicalState_RHD(gdata, ix, iy, iz,_dummy_indicator);
}

inline bool Transformations::isPhysicalState_RHD(Data &gdata, int ix, int iy, int iz, REAL & indicator) {
	return isAdmissible(
			gdata.om[q_rho](ix, iy, iz),
			gdata.om[q_sx](ix, iy, iz),
			gdata.om[q_sy](ix, iy, iz),
			gdata.om[q_sz](ix, iy, iz),
			gdata.om[q_Eges](ix, iy, iz),
			indicator
			);
}

//#include "old/inversions.C"

#define APPLY_INVERTOR(invertor, success) { \
	exceptionRaised = false; \
	try { \
		prim = invertor.getPrim(cons, Problem);	\
	} catch (const InversionFailure & e) { \
		exceptionRaised = true; \
	} \
	if (exceptionRaised == false && prim.isOk() && !prim.isNan()) success\
}

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
int counterInvEntropyPolytropic, counterInvEntropyPluto, counterInvEntropyFallback; 
primState Transformations::TransStateCons2Prim_Entropy_RHD(const consState & cons, const Data & gdata, int ix, int iy, int iz, const ProblemType & Problem) {
	primState prim;
	bool exceptionRaised;

	if( gdata.entropyFlag(ix, iy, iz) > 0 ) {
		// prim = TransEntropy2Prim_RHD(cons, Problem.gamma);
		// prim = TransCons2Prim_RHD_polytropic(cons, Problem.gamma, cons.sigma/cons.density);
		// APPLY_INVERTOR(invertorEntropy);

		// printPack(cerr, " Entropy inversion failed for: ", gdata.rank, ix, iy, iz, "\n  ", cons, "->", prim, " ---> trying second");
		APPLY_INVERTOR(invertorEntropyPolytropic, {++counterInvEntropyPolytropic; return prim;});

		// printPack(cerr, " Entropy inversion failed for: ", gdata.rank, ix, iy, iz, "\n  ", cons, "->", prim, " ---> trying pluto");
		APPLY_INVERTOR(invertorEntropyPluto, {++counterInvEntropyPluto; return prim;});

		printPack(cerr, " Entropy inversion failed for: ", gdata.rank, ix, iy, iz, "\n  ", cons, "->", prim, " ---> trying standard inversion");

		++counterInvEntropyFallback;
		prim = TransStateCons2Prim_Standard_RHD(cons, gdata, ix, iy, iz, Problem);
	} else {
		prim = TransStateCons2Prim_Standard_RHD(cons, gdata, ix, iy, iz, Problem);
	}

	return prim;
}
#endif


int counterInv1Dw, counterInvPluto, counterInvQuartic; 
primState Transformations::TransStateCons2Prim_Standard_RHD(const consState & cons, const Data & gdata, int ix, int iy, int iz, const ProblemType & Problem) {
	primState prim;
	bool exceptionRaised;
// #if (RHD_TRAFO_STRATEGY == MIGNONE_1Dw)
	// prim = TransCons2Prim_RHD_1Dw(cons, Problem.gamma);

	APPLY_INVERTOR(invertor1Dw, {++counterInv1Dw; return prim;});

	printPack(cerr, " First inversion failed for: ", gdata.rank, ix, iy, iz, "\n  ", cons, "->", prim, " ---> trying pluto");
	APPLY_INVERTOR(invertorPluto, {++counterInvPluto; return prim;});
	// prim = TransCons2Prim_RHD_numericalQuartic(cons, Problem.gamma);

	printPack(cerr, " Second inversion failed for: ", gdata.rank, ix, iy, iz, "\n  ", cons, "->", prim, " ---> trying quartic");
	APPLY_INVERTOR(invertorQuartic, {++counterInvQuartic; return prim;});
	// prim = TransCons2Prim_RHD_numericalQuartic(cons, Problem.gamma);

// #elif (RHD_TRAFO_STRATEGY == QUARTIC)
// 	// prim = TransCons2Prim_RHD_numericalQuartic(cons, Problem.gamma);
// 	// cout << "1st: " << prim << endl;
// 	prim = invertorQuartic.getPrim(cons, Problem);
// 	// cout << "2nd: " << prim << endl;
// 	if (!prim.isOk() || prim.isNan()) {
// 		printPack(cerr, " First inversion failed for: ", gdata.rank, ix, iy, iz, "\n  ", cons, "->", prim, " ---> trying second");
// 		// prim = TransCons2Prim_RHD_1Dw(cons, Problem.gamma);
// 		prim = invertor1Dw.getPrim(cons, Problem);
// 	}
// #endif

	printPack(cerr, " Nothing works for: ", gdata.rank, ix, iy, iz, " ---> flushing");
	prim = pressureFlush(cons, Problem);

	return prim;
}


void Transformations::TransCons2Prim_RHD(Data &gdata, gridFunc &gfunc,
                                     ProblemType &Problem)
{
	if(gdata.om[q_rho].getName() == "rho") {
		throw CException(" Restmass density instead of Lab density ");
	}

	if(gdata.om[q_sx].getName() == "v_x" ||
	   gdata.om[q_sy].getName() == "v_y" ||
	   gdata.om[q_sz].getName() == "v_z") {
		throw CException(" Velocity instead of momentum ");
	}

	if(gdata.om[q_rho].getName() == "Etherm") {
		throw CException(" Etherm instead of total energy ");
	}
	


	REAL momentum_sqr(0.);
	REAL indicator;
	bool admissibleFlag;
	int numNonadmissible = 0;


#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	int numEntropy = 0;
	int negativeEntropy = 0;

	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {

				admissibleFlag = isPhysicalState_RHD(gdata, ix, iy, iz, indicator);

				if(gdata.write_admissibleIndicator) {
					gdata.admissibleIndicator(ix, iy, iz) = indicator - 1.;
				}

				/* activate the entropy conversion for non admissible states,
						even though this might not be allowed (due to shocks.) */
				if( !admissibleFlag || indicator < 1. + 1e-9)
				{
					gdata.entropyFlag(ix, iy, iz) = 2;

					numNonadmissible++;
					numEntropy++;

					/* manually correct the state if the entropy conversion is impossible */
					if (gdata.om[q_Eadd](ix, iy, iz) < 0) {
						printPack(cerr, " Negative Entropy at", gdata.rank, ix, iy, iz, gdata.om[q_Eadd](ix, iy, iz));
						consState cons = readCons(gdata, ix, iy, iz);
						cons = CorrectUnphysicalState_RHD(cons, Problem);
						writeCons(gdata, ix, iy, iz, cons);
						negativeEntropy++;
					}

				}
				else if( gdata.entropyFlag(ix, iy, iz) > 0 )
				{
					/* fall back to the standard conversion if the entropy is negative */
					if (gdata.om[q_Eadd](ix, iy, iz) < 0) {
						gdata.entropyFlag(ix, iy, iz) = -1;
						// printPack(cerr, " Warning: Negative Entropy at ", gdata.rank, ix, iy, iz, " using standart variable inversion ");
					} else {
						numEntropy++;
					}
				}
			}
		}
	}

#ifdef parallel
//	MPI_Barrier(gdata.comm3d);
	gdata.MpiSum(numNonadmissible);
	gdata.MpiSum(numEntropy);
//	MPI_Barrier(gdata.comm3d);
#endif
	if(gdata.rank == 0) {

		cout << " Using Energy correction in " << numEntropy << " of ";
		#ifdef parallel
		cout << ((gdata.global_mx[0]+1)*(gdata.global_mx[1]+1)*(gdata.global_mx[2]+1));
		#else
		cout << (gdata.mx[0]+1)*(gdata.mx[1]+1)*(gdata.mx[2]+1);
		#endif
		cout << " cells -- to correct " << numNonadmissible << " errors " << endl;

		#if (RHD_STATE_CORRECTION == CRONOS_OFF)
		if (negativeEntropy > 0) {
			throw CException(" Negative Entropy detected ");
		}
		#endif

	}


// #else

// 	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
// 		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
// 			for(int ix=0; ix<=gdata.mx[0]; ++ix) {
// 				admissibleFlag = isPhysicalState_RHD(gdata, ix, iy, iz, indicator);

// 				if(gdata.write_admissibleIndicator) {
// 					gdata.admissibleIndicator(ix, iy, iz) = indicator;
// 				}

// 				if (!admissibleFlag) {
// 					cout 	<< "Unphysical state at "
// 							<< ix << " " << iy << " " << iz
// 							<< " with " << " " << indicator << " " << gdata.om[q_rho](ix, iy, iz)
// 							<< endl;
// 					numNonadmissible++;
// 					CorrectUnphysicalState_RHD(gdata, gfunc, Problem, ix, iy, iz);
// 				}
// 			}
// 		}
// 	}

// #if(RHD_STATE_CORRECTION == CRONOS_OFF)
// 	if (numNonadmissible > 0) {
// 		exit(1);
// 	}
// #endif
#endif


	counterInv1Dw = 0;
	counterInvPluto = 0;
	counterInvQuartic = 0; 
#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	counterInvEntropyPolytropic = 0;
	counterInvEntropyPluto = 0;
	counterInvEntropyFallback = 0;
#endif

	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {
				// Do Transformation

				if (Problem.isInternalBoundary(gdata, ix, iy, iz)) {
					Problem.setInternalBoundary(gdata, ix, iy, iz);
					continue;
				}

				consState cons = readCons(gdata, ix, iy, iz);
				if (!cons.isOk() || cons.isNan()) {
					printPack(cerr, " Cons State Error ",
						gdata.rank, ix, iy, iz, " @ ",
						gdata.get_idx_global(0, ix), gdata.get_idx_global(1, iy), gdata.get_idx_global(2, iz),
						"|", cons);
					#if(RHD_STATE_CORRECTION == CRONOS_OFF)
						throw CException(" Cons State Error ");
					#else
						cons = CorrectUnphysicalState_RHD(cons, Problem);
					#endif
				}

				primState prim;

#if (EOS == IDEAL)

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
				prim = TransStateCons2Prim_Entropy_RHD(cons, gdata, ix, iy, iz, Problem);
#else
				prim = TransStateCons2Prim_Standard_RHD(cons, gdata, ix, iy, iz, Problem);
#endif

#elif (EOS == POLYTROPIC)
				prim = invertorPolytropic.getPrim(cons, Problem);
#endif

				writePrim(gdata, ix, iy, iz, prim);
#if (PARTICLE_TRANSPORT == TRUE)
				gdata.lorentz(ix, iy, iz) = prim.computeLorentz();
#endif

				// some checks
				assert(assert_print(prim.isOk() && !prim.isNan(), 
						gdata.rank, ix, iy, iz, " @ ",
						gdata.get_idx_global(0, ix), gdata.get_idx_global(1, iy), gdata.get_idx_global(2, iz),
						prim));
			}
		}
	}

#ifdef parallel
	gdata.MpiSum(counterInv1Dw);
	gdata.MpiSum(counterInvPluto);
	gdata.MpiSum(counterInvQuartic);

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	gdata.MpiSum(counterInvEntropyPolytropic);
	gdata.MpiSum(counterInvEntropyPluto);
	gdata.MpiSum(counterInvEntropyFallback);
#endif
#endif

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
ccout << " Entropy Inversion stats: " << counterInvEntropyPolytropic << " " << counterInvEntropyPluto << " " << counterInvEntropyFallback << endl;
#endif

ccout << " Energy Inversion stats: " << counterInv1Dw << " " << counterInvPluto << " " << counterInvQuartic << endl;

//	Store thermal pressure if necessary
	if(gdata.storePressure) {
		for(int k = 0; k<=gdata.mx[2]; ++k){
			for(int j = 0; j<=gdata.mx[1]; ++j){
				for(int i = 0; i<=gdata.mx[0]; ++i){
					gdata.pTherm(i,j,k) = (Problem.gamma-1)*gdata.om[q_Eges](i,j,k);
				}
			}
		}
		gfunc.BoundaryHandler.boundary(gdata, Problem, &gdata.pTherm, 1);
	}



	gdata.om[q_rho].rename("rho");
	gdata.om[q_sx].rename("v_x");
	gdata.om[q_sy].rename("v_y");
	gdata.om[q_sz].rename("v_z");
	gdata.om[q_Eges].rename("Etherm");
}


void Transformations::TransPrim2Cons_RHD(Data &gdata, gridFunc &gfunc,
                                     ProblemType &Problem)
{
	if(gdata.om[q_rho].getName() == "density") {
		throw CException(" Lab density instead of restmass density ");
	}

	if(gdata.om[q_sx].getName() == "s_x" ||
	   gdata.om[q_sy].getName() == "s_y" ||
	   gdata.om[q_sz].getName() == "s_z") {
		throw CException(" Momentum instead of velocity ");
	}

	if(gdata.om[q_rho].getName() == "Eges") {
		throw CException(" Total energy instead of thermal energy ");
	}

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	SetEntropySwitches_RHD(gdata, gfunc, Problem);
#endif

	const REAL Etherm2Press((Problem.gamma - 1.));


	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {

				const REAL u_sqr = sqr(gdata.om[q_sx](ix, iy, iz)) +
														sqr(gdata.om[q_sy](ix, iy, iz)) +
														sqr(gdata.om[q_sz](ix, iy, iz));

				const REAL lorentz = TransVelSqr2Gamma_RHD(u_sqr);
				const REAL pressure = Etherm2Press * gdata.om[q_Eges](ix, iy, iz);
				const REAL enthalpy = TransPrim2Enthalpy_RHD(Problem.gamma, gdata.om[q_rho](ix, iy, iz), pressure);
				const REAL D = gdata.om[q_rho](ix, iy, iz) * lorentz;
				const REAL Dh = D * enthalpy;

				gdata.om[q_rho](ix, iy, iz) = D;
				
				for (int q = q_sx; q <= q_sz; ++q) {
					gdata.om[q](ix, iy, iz) *= Dh;
				}

				gdata.om[q_Eges](ix, iy, iz) = D * (enthalpy * lorentz - 1.) - pressure;

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
				// Computation of entropy for dual energy corrections
				gdata.om[q_Eadd](ix, iy, iz) *= D;
#endif

				// some checks
				assert(assert_print(gdata.om[q_rho](ix, iy, iz) > 0., gdata.rank, ix, iy, iz, "|", "D =", gdata.om[q_rho](ix, iy, iz)));
				assert(assert_print(gdata.om[q_Eges](ix, iy, iz) > 0., gdata.rank, ix, iy, iz,  "|", "tau =", gdata.om[q_Eges](ix, iy, iz)));

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
				assert(assert_print(gdata.om[q_Eadd](ix, iy, iz) > 0., gdata.rank, ix, iy, iz,  "|", "S =", gdata.om[q_Eadd](ix, iy, iz)));
#endif
				// This can be violated by truncation errors
				// assert(isPhysicalState_RHD(gdata, ix, iy, iz));
			}
		}
	}

	gdata.om[q_rho].rename("density");
	gdata.om[q_sx].rename("s_x");
	gdata.om[q_sy].rename("s_y");
	gdata.om[q_sz].rename("s_z");
	gdata.om[q_Eges].rename("Eges");
}


#if (CRSWITCH_DUAL_ENERGY == CRONOS_ON)
void Transformations::SetEntropySwitches_RHD(Data &gdata, gridFunc &gfunc, ProblemType &Problem) {
	// Entropy correction according to Balsara & Spicer (1999) JCP 148, 133


	if(gdata.om[q_sx].getName() != "v_x" ||
	   gdata.om[q_sy].getName() != "v_y" ||
	   gdata.om[q_sz].getName() != "v_z") {
		throw CException(" Must be velocity ");
	}


//#if(FLUID_TYPE == CRONOS_MULTIFLUID)
//	int n_omIntAll = gdata.fluids->get_N_OMINT_ALL();
//#else
//	int n_omIntAll = N_OMINT;
//#endif
//
//	if(gdata.om[n_omIntAll].getName() != "Eges" ||
//	   gdata.om[q_Eges].getName() != "Etherm") {
//		throw CException(" Need overall and thermal energy for dual energy evolution ");
//	}


	static Pot & Etherm = gdata.om[q_Eges];
	static Pot & rho = gdata.om[q_rho];
	static Pot & ux = gdata.om[q_sx];
	static Pot & uy = gdata.om[q_sy];
	static Pot & uz = gdata.om[q_sz];

//	Pot h(gdata.mx);
//	Pot lorentz(gdata.mx);
	static Pot cs(gdata.mx);
	static Pot vx(gdata.mx);// = gdata.om[q_sx];
	static Pot vy(gdata.mx);// = gdata.om[q_sy];
	static Pot vz(gdata.mx);// = gdata.om[q_sz];

	// Compute characteristic speeds
	for (int k = -B; k <= gdata.mx[2]+B; ++k) {
	for (int j = -B; j <= gdata.mx[1]+B; ++j) {
	for (int i = -B; i <= gdata.mx[0]+B; ++i) {

		gdata.entropyFlag(i, j, k) = 1;


		REAL temp =  Problem.gamma * Etherm(i,j,k);
		cs(i,j,k) = sqrt((Problem.gamma - 1.) * temp /	(gdata.om[q_rho](i,j,k) + temp));

		REAL u2 = sqr(ux(i,j,k)) + sqr(uy(i,j,k)) + sqr(uz(i,j,k));
//		u(i,j,k) = sqrt(u2);
//		lorentz(i,j,k) = sqrt(1. + u2);

		REAL lorentz = sqrt(1. + u2);
		REAL lorentz_inv = 1./lorentz;

		vx(i,j,k) = gdata.om[q_sx](i,j,k) * lorentz_inv;
		vy(i,j,k) = gdata.om[q_sy](i,j,k) * lorentz_inv;
		vz(i,j,k) = gdata.om[q_sz](i,j,k) * lorentz_inv;

//		h(i,j,k) = 1. + Problem.gamma * Etherm(i,j,k) / rho(i,j,k);
	}}}

	for (int k = 0; k <= gdata.mx[2]; ++k) {
	for (int j = 0; j <= gdata.mx[1]; ++j) {
	for (int i = 0; i <= gdata.mx[0]; ++i) {
//	for (int k = -B + 1; k <= gdata.mx[2] + B - 1; ++k) {
//	for (int j = -B + 1; j <= gdata.mx[1] + B - 1; ++j) {
//	for (int i = -B + 1; i <= gdata.mx[0] + B - 1; ++i) {


#if (GEOM == CARTESIAN)
		REAL div_v = ((vx(i+1,j,k) - vx(i-1,j,k))/
					  (gdata.getCen_x(i+1) - gdata.getCen_x(i-1)) +
					  (vy(i,j+1,k) - vy(i,j-1,k))/
					  (gdata.getCen_y(j+1) - gdata.getCen_y(j-1)) +
					  (vz(i,j,k+1) - vz(i,j,k-1))/
					  (gdata.getCen_z(k+1) - gdata.getCen_z(k-1)));

		double del = std::min(gdata.getCen_dx(0,i),
							  std::min(gdata.getCen_dx(1,j),
									   gdata.getCen_dx(2,k)));

#elif (GEOM == CYLINDRICAL)
		Pot & v_rad = vx;
		Pot & v_phi = vy;
		Pot & v_z   = vz;
		REAL r_cyl = gdata.getCen_x(i);
		REAL div_v = (((v_rad(i+1,j,k) - v_rad(i-1,j,k))/
					   (gdata.getCen_x(i+1) - gdata.getCen_x(i-1)) +
					   (v_rad(i,j,k)/r_cyl)) +
					  ((v_phi(i,j+1,k) - v_phi(i,j-1,k))/
					   (r_cyl*(gdata.getCen_y(j+1) - gdata.getCen_y(j-1)))) +
					  ((v_z(i,j,k+1) - v_z(i,j,k-1))/
					   (gdata.getCen_z(k+1) - gdata.getCen_z(k-1))));


//		Pot & u_rad = ux;
//		Pot & u_phi = uy;
//		Pot & u_z   = uz;
//		REAL lorentz_inv = 1. / lorentz(i,j,k);
//
//		REAL div_v_u = ((
//						(u_rad(i+1,j,k) - u_rad(i-1,j,k))
//						- u_rad(i,j,k) * lorentz_inv * (lorentz(i+1,j,k) - lorentz(i-1,j,k))
//							)/
//					   (gdata.getCen_x(i+1) - gdata.getCen_x(i-1))) +
//						(u_rad(i,j,k)/r_cyl) +
//					  (((u_phi(i,j+1,k) - u_phi(i,j-1,k))
//						- u_phi(i,j,k) * lorentz_inv * (lorentz(i,j+1,k) - lorentz(i,j-1,k))
//							)/
//					   (r_cyl*(gdata.getCen_y(j+1) - gdata.getCen_y(j-1)))) +
//					  (((u_z(i,j,k+1) - u_z(i,j,k-1))
//						- u_z(i,j,k) * lorentz_inv * (lorentz(i,j,k+1) - lorentz(i,j,k-1))
//							)/
//					   (gdata.getCen_z(k+1) - gdata.getCen_z(k-1)));
//
//		div_v_u *= lorentz_inv;

		double del = std::min(gdata.getCen_dx(0,i),
							  std::min(r_cyl*gdata.getCen_dx(1,j),
									   gdata.getCen_dx(2,k)));

#elif (GEOM == SPHERICAL)
		Pot & v_rad = vx;
		Pot & v_tht = vy;
		Pot & v_phi = vz;
		REAL r_sph = gdata.getCen_x(i);
		REAL theta = gdata.getCen_y(j);
		REAL div_v = (((v_rad(i+1,j,k) - v_rad(i-1,j,k))/
					   (gdata.getCen_x(i+1) - gdata.getCen_x(i-1)) +
					   (2.*v_rad(i,j,k)/r_sph)) +
					  (((v_tht(i,j+1,k) - v_tht(i,j-1,k))/
						(gdata.getCen_y(j+1) - gdata.getCen_y(j-1)) +
						(v_tht(i,j,k)/tan(theta)))/r_sph) +
					  ((v_phi(i,j,k+1) - v_phi(i,j,k-1))/
					   (r_sph*sin(theta)*(gdata.getCen_z(k+1) -
										  gdata.getCen_z(k-1)))));

		double del = std::min(gdata.getCen_dx(0,i),
							  std::min(r_sph*gdata.getCen_dx(1,j),
									   r_sph*sin(theta)*gdata.getCen_dx(2,k)));
#endif



//		REAL us = cs(i,j,k) / (1. - sqr(cs(i,j,k))) * minMach;
//		REAL Vs = us / sqrt(1. + sqr(us));
//		REAL temp = h(i,j,k) * ( ux(i,j,k) * ux(i,j,k) + uy(i,j,k) * uy(i,j,k) + uz(i,j,k) * uz(i,j,k));
//
//		if ( ( Etherm(i+1,j,k) - Etherm(i,j,k) ) * (Problem.gamma  - 1.)  / (
//			rho(i,j,k) * (
//					temp
//					- h(i+1,j,k) * ( ux(i+1,j,k) * ux(i,j,k) + uy(i+1,j,k) * uy(i,j,k) + uz(i+1,j,k) * uz(i,j,k))
//				)) > alpha2 )
//		{
//			gdata.entropySwitch2(i,j,k) = false;
//		} else
//		if ( ( Etherm(i-1,j,k) - Etherm(i,j,k) ) * (Problem.gamma  - 1.)  / (
//			rho(i,j,k) * (
//					temp
//					- h(i-1,j,k) * ( ux(i-1,j,k) * ux(i,j,k) + uy(i-1,j,k) * uy(i,j,k) + uz(i-1,j,k) * uz(i,j,k))
//				)) > alpha2 )
//		{
//			gdata.entropySwitch2(i,j,k) = false;
//		} else
//		if ( ( Etherm(i,j+1,k) - Etherm(i,j,k) ) * (Problem.gamma  - 1.)  / (
//			rho(i,j,k) * (
//					temp
//					- h(i,j+1,k) * ( ux(i,j+1,k) * ux(i,j,k) + uy(i,j+1,k) * uy(i,j,k) + uz(i,j+1,k) * uz(i,j,k))
//				)) > alpha2 )
//		{
//			gdata.entropySwitch2(i,j,k) = false;
//		} else
//		if ( ( Etherm(i,j-1,k) - Etherm(i,j,k) ) * (Problem.gamma  - 1.)  / (
//			rho(i,j,k) * (
//					temp
//					- h(i,j-1,k) * ( ux(i,j-1,k) * ux(i,j,k) + uy(i,j-1,k) * uy(i,j,k) + uz(i,j-1,k) * uz(i,j,k))
//				)) > alpha2 )
//		{
//			gdata.entropySwitch2(i,j,k) = false;
//		} else
//		if ( ( Etherm(i,j,k+1) - Etherm(i,j,k) ) * (Problem.gamma  - 1.)  / (
//			rho(i,j,k) * (
//					temp
//					- h(i,j,k+1) * ( ux(i,j,k+1) * ux(i,j,k) + uy(i,j,k+1) * uy(i,j,k) + uz(i,j,k+1) * uz(i,j,k))
//				)) > alpha2 )
//		{
//			gdata.entropySwitch2(i,j,k) = false;
//		} else
//		if ( ( Etherm(i,j,k-1) - Etherm(i,j,k) ) * (Problem.gamma  - 1.)  / (
//			rho(i,j,k) * (
//					temp
//					- h(i,j,k-1) * ( ux(i,j,k-1) * ux(i,j,k) + uy(i,j,k-1) * uy(i,j,k) + uz(i,j,k-1) * uz(i,j,k))
//				)) > alpha2 )
//		{
//			gdata.entropySwitch2(i,j,k) = false;
//
//		} else
//		{
//			gdata.entropySwitch2(i,j,k) = true;
//		}
//
//		gdata.entropySwitch2(i,j,k) = true;


		REAL div_p = (std::abs(Etherm(i+1,j,k) - Etherm(i-1,j,k)) +
					   std::abs(Etherm(i,j+1,k) - Etherm(i,j-1,k)) +
					   std::abs(Etherm(i,j,k+1) - Etherm(i,j,k-1)));

		REAL min_p = min(Etherm(i  ,j,k  ),
						  Etherm(i+1,j,k), Etherm(i-1,j,k  ),
						  Etherm(i,j+1,k), Etherm(i,j-1,k  ),
						  Etherm(i,j,k+1), Etherm(i,j,k-1));

		REAL max_cs = max(	cs(i,j  ,k),
							cs(i+1,j,k), cs(i-1,j,k),
							cs(i,j+1,k), cs(i,j-1,k),
							cs(i,j,k+1), cs(i,j,k-1));


		//
		// Set the individual switches
		//
		if(div_p < entropy_flag_alpha2 * min_p) {
			gdata.entropySwitch2(i,j,k) = true;
		} else {
			gdata.entropySwitch2(i,j,k) = false;
		}

//		if(- shock_flag_alpha1 * max_cs < del * div_u * sqrt(1. - max_cs)) {
//			gdata.entropySwitch1(i,j,k) = true;
//		} else {
//			gdata.entropySwitch1(i,j,k) = false;
//		}


		if(- entropy_flag_alpha3 * max_cs < del * div_v) {
			gdata.entropySwitch3(i,j,k) = true;
		} else {
			gdata.entropySwitch3(i,j,k) = false;
		}


		// Set the entropy flag

		if (gdata.entropySwitch2(i,j,k) || gdata.entropySwitch3(i,j,k)) {
			gdata.entropyFlag(i, j, k) *= 1;
		} else {
			gdata.entropyFlag(i, j, k) *= 0;
			gdata.entropyFlag(i + 1, j, k) *= 0;
			gdata.entropyFlag(i - 1, j, k) *= 0;
			gdata.entropyFlag(i, j + 1, k) *= 0;
			gdata.entropyFlag(i, j - 1, k) *= 0;
			gdata.entropyFlag(i, j, k + 1) *= 0;
			gdata.entropyFlag(i, j, k - 1) *= 0;
		}

	}}}
}
#endif



void Transformations::get_Cons_RHD(Data &gdata,
                               ProblemType &Problem,
                               EquationOfState  &eos,
                               cronos::vector<REAL> &Pos,
                               phys_fields_1D &fields,
                               int dir, REAL shift){

	// Set / compute conservative variables:
	const REAL Etherm2Press(Problem.gamma - 1.);

	for (int i = -2; i <= gdata.mx[dir]+1; ++i){
		fields.ptherm(i) = Etherm2Press*fields.uPri[q_Eges_loc](i);

		const REAL u_sqr = sqr(fields.uPri[q_sx_loc](i))
				+ sqr(fields.uPri[q_sy_loc](i))
				+ sqr(fields.uPri[q_sz_loc](i));
		const REAL lorentz = TransVelSqr2Gamma_RHD(u_sqr);
		const REAL enthalpy = TransPrim2Enthalpy_RHD(Problem.gamma, fields.uPri[q_rho_loc](i), fields.ptherm(i));
		const REAL D = lorentz*fields.uPri[q_rho_loc](i);
		const REAL Dh = D * enthalpy;

		fields.uCon[q_rho_loc](i) = D;
		fields.uCon[q_sx_loc](i) = Dh*fields.uPri[q_sx_loc](i);
		fields.uCon[q_sy_loc](i) = Dh*fields.uPri[q_sy_loc](i);
		fields.uCon[q_sz_loc](i) = Dh*fields.uPri[q_sz_loc](i);
		fields.uCon[q_Eges_loc](i) = D * (enthalpy * lorentz - 1.) - fields.ptherm(i);

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
		// Computation of entropy for dual energy corrections
		// fields.uPri[q_Eadd_loc](i) = fields.ptherm(i) * pow(fields.uPri[q_rho_loc](i), - Problem.gamma);
		fields.uCon[q_Eadd_loc](i) = fields.uPri[q_Eadd_loc](i) * D;
#endif

		// some checks
		assert(assert_print(fields.uCon[q_rho_loc](i) > 0., gdata.rank, dir, i,  "|", "D =", fields.uCon[q_rho_loc](i)));
		assert(assert_print(fields.uCon[q_Eges_loc](i) > 0., gdata.rank, dir, i,  "|", "tau =", fields.uCon[q_Eges_loc](i)));

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
		assert(assert_print(fields.uCon[q_Eadd_loc](i) > 0., gdata.rank, dir, i,  "|", "S =", fields.uCon[q_Eadd_loc](i)));
#endif

	}

	// Total pressure:
	fields.ptotal = fields.ptherm;

}


void Transformations::get_Cons_RHD(Data &gdata, ProblemType &Problem,
		EquationOfState  &eos, phys_fields_0D &fields, int ix, int iy, int iz, int face)
{
	throw CException(" 0D Methods are not available for RHD yet ");
}



#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
void Transformations::computeEntropyFromE_RHD(Data &gdata, gridFunc &gfunc,
                                          ProblemType &Problem)
{
	if(gdata.om[q_Eges].getName() != "Etherm") {
		throw CException(" Need thermal energy for dual energy computation ");
	}
	// Only if dual energy is chosen
#if(AUX_ENERGY == ENTROPY)
	// Get entropy from thermal energy
	for(int k = -B; k <= gdata.mx[2]+B; ++k) {
		for (int j = -B; j <= gdata.mx[1]+B; ++j) {
			for (int i = -B; i <= gdata.mx[0]+B; ++i) {
				gdata.om[q_Eadd](i,j,k) = 
					gdata.om[q_Eges](i,j,k)*pow(gdata.om[q_rho](i,j,k),-Problem.gamma) * (Problem.gamma - 1.);
			}
		}
	}
#else
	// Save thermal energy
	#error "no clue what to do"
#endif
}
#endif


#endif
