#include "RiemannSolverRHD.H"

RiemannSolverRHD::RiemannSolverRHD(const Data &gdata, int dir, int Fluid_Type) : RiemannSolver(gdata, dir, Fluid_Type) {
#if(FLUID_TYPE==CRONOS_RHD)
	this->q_rho = gdata.fluid.get_q_rho();
	this->q_sx  = gdata.fluid.get_q_sx();
	this->q_sy  = gdata.fluid.get_q_sy();
	this->q_sz  = gdata.fluid.get_q_sz();
	this->q_Eges = gdata.fluid.get_q_Eges();
	this->q_Eadd = gdata.fluid.get_q_Eadd();
#endif
}

void RiemannSolverRHD::get_vChar(Data &gdata,
                                ProblemType &Problem,
                                cronos::vector<double> &iPos,
                                phys_fields_1D &pfL,
                                phys_fields_1D &pfR,
                                NumMatrix<REAL,1> &v_ch_m,
                                NumMatrix<REAL,1> &v_ch_p,
//                                fields_1D &fields,
                                const int &dir, REAL &cfl_lin) {
  
	for (int i = -1; i <= gdata.mx[dir]+1; ++i){

		iPos.set(dir, i-0.5);
    
		REAL rhoinv_p = 1./pfL.uPri[q_rho](i  );
		REAL rhoinv_m = 1./pfR.uPri[q_rho](i-1);

		// relativistic parameters
		REAL gamma_rel_sqr_p = 1.+sqr(pfL.uPri[q_sx](i  ))+sqr(pfL.uPri[q_sy](i  ))+sqr(pfL.uPri[q_sz](i  ));
		REAL gamma_rel_sqr_m = 1.+sqr(pfR.uPri[q_sx](i-1))+sqr(pfR.uPri[q_sy](i-1))+sqr(pfR.uPri[q_sz](i-1));
		REAL gamma_relinv_sqr_p = 1./gamma_rel_sqr_p;
		REAL gamma_relinv_sqr_m = 1./gamma_rel_sqr_m;
    
		// Flow velocity
		REAL u_p = pfL.uPri[dir+q_sx](i  ) * sqrt(gamma_relinv_sqr_p);
		REAL u_m = pfR.uPri[dir+q_sx](i-1) * sqrt(gamma_relinv_sqr_m);
    
		REAL pres_p = pfL.ptherm(i);
		REAL pres_m = pfR.ptherm(i-1);

		REAL enthalpy_p = 1. + pres_p*rhoinv_p * Problem.gamma / (Problem.gamma - 1.);
		REAL enthalpy_m = 1. + pres_m*rhoinv_m * Problem.gamma / (Problem.gamma - 1.);
    
		// Sound speed
		REAL cs_sqr_p   = Problem.gamma*pres_p*rhoinv_p/enthalpy_p;
		REAL cs_sqr_m   = Problem.gamma*pres_m*rhoinv_m/enthalpy_m;


		REAL sigs_p = cs_sqr_p*gamma_relinv_sqr_p/(1.-cs_sqr_p);
		REAL sigs_m = cs_sqr_m*gamma_relinv_sqr_m/(1.-cs_sqr_m);

		REAL lam_p_p = (u_p + sqrt(sigs_p*(1 - sqr(u_p) + sigs_p)))/(1.+ sigs_p);
		REAL lam_p_m = (u_m + sqrt(sigs_m*(1 - sqr(u_m) + sigs_m)))/(1.+ sigs_m);

		REAL lam_m_p = (-u_p + sqrt(sigs_p*(1 - sqr(u_p) + sigs_p)))/(1.+ sigs_p); // vorzeichen umgedreht wegen
		REAL lam_m_m = (-u_m + sqrt(sigs_m*(1 - sqr(u_m) + sigs_m)))/(1.+ sigs_m); // definition der Richtungen


		v_ch_p(i) = std::max(std::max(lam_p_m,lam_p_p),0.);
		v_ch_m(i) = std::max(std::max(lam_m_m, lam_m_p),0.);

//		if(i==399) {
//			cout << " vfp " << cs_p << " " << cs_m;
//			cout << endl;
//			cout << q_sx << endl;
//		}
    
		REAL vmax = std::max(v_ch_p(i), v_ch_m(i));

//		fields.v_ch_p(i) = std::max(std::max(cs_p+u_p,cs_m+u_m),0.);
//		fields.v_ch_m(i) = std::max(std::max(cs_p-u_p,cs_m-u_m),0.);
//
//		REAL vmax = std::max(fields.v_ch_p(i),fields.v_ch_m(i));

		// Local computation of cfl number
#if  (NON_LINEAR_GRID == CRONOS_OFF)
		REAL cfl_loc = vmax*gdata.idx[dir];
#else
		REAL cfl_loc = vmax*gdata.getCen_idx(dir, i);
#endif

#ifdef GEOM
#if GEOM != CARTESIAN
		if(dir==0) {
			cfl_loc /= gdata.getCen_h0(iPos.get(0), iPos.get(1), iPos.get(2));
		} else if (dir==1) {
			cfl_loc /= gdata.getCen_h1(iPos.get(0), iPos.get(1), iPos.get(2));
		} else {
			cfl_loc /= gdata.getCen_h2(iPos.get(0), iPos.get(1), iPos.get(2));
		}
#endif
#endif
		cfl_lin = std::max(cfl_lin, cfl_loc);
    
	}


}



void RiemannSolverRHD::get_vChar(Data &gdata, ProblemType &Problem,
		phys_fields_0D &pfL, phys_fields_0D &pfR, num_fields_0D &f_num,
		int ix, int iy, int iz, int dir, REAL &cfl_lin) {
	//! Compute characteristic velocities

	int shift_vec[3] = {0,0,0};
	shift_vec[dir] = -1;

	int iPos[3] = {ix, iy, iz};

	REAL rhoinv_p = 1./pfL.uCon[q_rho];
	REAL rhoinv_m = 1./pfR.uCon[q_rho];

	// Flow velocity
	REAL u_p = pfL.uPri[dir+q_sx];
	REAL u_m = pfR.uPri[dir+q_sx];

	REAL pres_p = pfL.ptherm;
	REAL pres_m = pfR.ptherm;

	// relativistic parameters
	REAL gamma_relinv_sqr_p = 1.-sqr(pfL.uPri[q_sx])-sqr(pfL.uPri[q_sy])-sqr(pfL.uPri[q_sz]);
	REAL gamma_relinv_sqr_m = 1.-sqr(pfR.uPri[q_sx])-sqr(pfR.uPri[q_sy])-sqr(pfR.uPri[q_sz]);
	REAL gamma_rel_sqr_p = 1./gamma_relinv_sqr_p;
	REAL gamma_rel_sqr_m = 1./gamma_relinv_sqr_m;

	REAL enthalpy_p = 1. + pres_p*rhoinv_p * Problem.gamma / (Problem.gamma - 1.);
	REAL enthalpy_m = 1. + pres_m*rhoinv_m * Problem.gamma / (Problem.gamma - 1.);

	// Sound speed
	REAL cs_sqr_p   = Problem.gamma*pres_p*rhoinv_p/enthalpy_p;
	REAL cs_sqr_m   = Problem.gamma*pres_m*rhoinv_m/enthalpy_m;


	REAL sigs_p = cs_sqr_p/(gamma_rel_sqr_p*(1.-cs_sqr_p));
	REAL sigs_m = cs_sqr_m/(gamma_rel_sqr_m*(1.-cs_sqr_m));

	REAL lam_p_p = (u_p + sqrt(sigs_p*(1 - sqr(u_p) + sigs_p)))/(1.+ sigs_p);
	REAL lam_p_m = (u_m + sqrt(sigs_m*(1 - sqr(u_m) + sigs_m)))/(1.+ sigs_m);

	REAL lam_m_p = (-u_p + sqrt(sigs_p*(1 - sqr(u_p) + sigs_p)))/(1.+ sigs_p); // vorzeichen umgedreht wegen
	REAL lam_m_m = (-u_m + sqrt(sigs_m*(1 - sqr(u_m) + sigs_m)))/(1.+ sigs_m); // definition der Richtungen


	REAL v_ch_p = std::max(std::max(lam_p_m,lam_p_p),0.);
	REAL v_ch_m = std::max(std::max(lam_m_m, lam_m_p),0.);

	f_num.v_ch_p = v_ch_p;
	f_num.v_ch_m = v_ch_m;

	REAL vmax = std::max(v_ch_p, v_ch_m);

	//		fields.v_ch_p(i) = std::max(std::max(cs_p+u_p,cs_m+u_m),0.);
	//		fields.v_ch_m(i) = std::max(std::max(cs_p-u_p,cs_m-u_m),0.);
	//
	//		REAL vmax = std::max(fields.v_ch_p(i),fields.v_ch_m(i));

	// Local computation of cfl number
#if  (NON_LINEAR_GRID == CRONOS_OFF)
	REAL cfl_loc = vmax*gdata.idx[dir];
#else
	REAL cfl_loc = vmax*gdata.getCen_idx(dir, iPos[dir]);
#endif

#ifdef GEOM
#if GEOM != CARTESIAN
	if(dir==0) {
		cfl_loc /= gdata.h0(ix, iy, iz, shift_vec[0],shift_vec[1],shift_vec[2]);
	} else if (dir==1) {
		cfl_loc /= gdata.h1(ix, iy, iz, shift_vec[0],shift_vec[1],shift_vec[2]);
	} else {
		cfl_loc /= gdata.h2(ix, iy, iz, shift_vec[0],shift_vec[1],shift_vec[2]);
	}
#endif
#endif
	cfl_lin = std::max(cfl_lin, cfl_loc);

}

