#include "fields_local.H"
#include "fieldLists.H"
#include <iostream>

using namespace std;

fields_1D::fields_1D(const Data &gdata, int dir, const CronosFluid &fluid){


	this->dir = dir;
	generic = true; // Code native stuff

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	fieldLists List(fluid, dir);
	this->num = fluid.get_N_OMINT();
//	fieldLists List(gdata.fluids, dir);
//	this->num = gdata.fluids->get_N_OMINT_ALL();
//	fieldLists List(gdata.fluids->fluids, gdata.fluids->get_fluidType(iFluid), dir);
#else
	fieldLists List(fluid, dir);
//	fieldLists List(gdata.fluid, dir);
	this->num = N_OMINT;
#endif

	omLoc = new NumMatrix<REAL,1> [num];
	deriv = new NumMatrix<REAL,1> [num];
	flux = new NumMatrix<REAL,1> [num];
	omLocP = new NumMatrix<REAL,1> [num];
	omLocM = new NumMatrix<REAL,1> [num];
	has_perp = true;

	for(int q=0; q<num; ++q) {
		// Identify fields that require different reconstruction
		if(List.isGenTypePerpendicular(q)) {
			omLocP[q].resize(Index::set(-3),Index::set(gdata.mx[dir]+2));
			omLocP[q].clear();
			omLocM[q].resize(Index::set(-3),Index::set(gdata.mx[dir]+2));
			omLocM[q].clear();
		} else {
			omLoc[q].resize(Index::set(-3),Index::set(gdata.mx[dir]+2));
			omLoc[q].clear();
		}
//		cout << " Loop: " << q << " " << num << " " << dir << " " << omLoc[q].getHigh(0) << endl << endl;
	}

	for(int q=0; q<num; ++q) {
		deriv[q].resize(Index::set(-3),Index::set(gdata.mx[dir]+2));
		deriv[q].clear();
	}

//#if(FLUID_TYPE == CRONOS_MULTIFLUID)
//	// In multifluid case make one vChar and one ptotal for each fluid
//	int numFluids = gdata.fluids->get_numFluids();
//
//	ptotal = new NumMatrix<REAL,1> [numFluids];
//	for(int iFluid; iFluid<numFluids; ++iFluid) {
//		ptotal[iFluid].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
//	}
//
//	v_ch_p = new NumMatrix<REAL,1> [numFluids];
//	v_ch_m = new NumMatrix<REAL,1> [numFluids];
//	for(int iFluid; iFluid<numFluids; ++iFluid) {
//		v_ch_p[iFluid].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
//		v_ch_m[iFluid].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
//	}
//#else
	ptotal.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	ptotal.clear();

	v_ch_p.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	v_ch_p.clear();
	v_ch_m.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	v_ch_m.clear();
//#endif

	if(gdata.use_carbuncleFlag) {
		carbuncle_flag.resize(Index::set(-2),Index::set(gdata.mx[dir]+2));
		carbuncle_flag.clear();
	}

	for(int q=0; q<num; ++q) {
		flux[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		flux[q].clear();
	}


}


fields_1D::fields_1D(const Data &gdata, int dir, int num, int iFluid){

	this->num = num;
	this->dir = dir;
	generic = false; // User dependent stuff

	omLoc = new NumMatrix<REAL,1> [num];
	deriv = new NumMatrix<REAL,1> [num];
	flux = new NumMatrix<REAL,1> [num];
	has_perp = false;

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	fieldLists List(gdata.fluids->fluids[iFluid], dir);
	//	fieldLists List(gdata.fluids->fluids, gdata.fluids->get_fluidType(iFluid), dir);
#else
	fieldLists List(gdata.fluid, dir);
#endif

	for(int q=0; q<num; ++q) {
		omLoc[q].resize(Index::set(-3),Index::set(gdata.mx[dir]+2));
		omLoc[q].clear();

		deriv[q].resize(Index::set(-3),Index::set(gdata.mx[dir]+2));
		deriv[q].clear();

		flux[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		flux[q].clear();
	}

//#if(FLUID_TYPE == CRONOS_MULTIFLUID)
//	// In multifluid case make one vChar and one ptotal for each fluid
//	int numFluids = gdata.fluids->get_numFluids();
//
//	ptotal = new NumMatrix<REAL,1> [numFluids];
//	for(int iFluid; iFluid<numFluids; ++iFluid) {
//		ptotal[iFluid].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
//	}
//
//	v_ch_p = new NumMatrix<REAL,1> [numFluids];
//	v_ch_m = new NumMatrix<REAL,1> [numFluids];
//	for(int iFluid; iFluid<numFluids; ++iFluid) {
//		v_ch_p[iFluid].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
//		v_ch_m[iFluid].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
//	}
//#else
	v_ch_p.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	v_ch_p.clear();
	v_ch_m.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	v_ch_m.clear();

	ptotal.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	ptotal.clear();
//#endif
}


fields_1D::~fields_1D() {
	delete [] omLoc;
	delete [] deriv;
	delete [] flux;
	if(has_perp) {
		delete [] omLocP;
		delete [] omLocM;
	}
//#if(FLUID_TYPE == CRONOS_MULTIFLUID)
//	delete [] v_ch_m;
//	delete [] v_ch_p;
//	delete [] ptotal;
//#endif
}


bool fields_1D::isgeneric() {
	return generic;
}

int fields_1D::get_num() {
	return num;
}


fields_2D::fields_2D(const Data &gdata, const int &dir) {

	assert(dir >= 0 && dir < DIM);
	if(dir == 0) {
		this->dir0 = 1;
		this->dir1 = 2;
	} else if (dir == 1) {
		this->dir0 = 2;
		this->dir1 = 0;
	} else {
		this->dir0 = 0;
		this->dir1 = 1;
	}

#if(FLUID_TYPE == CRONOS_MULTIFLUID)
	int iMagFluid = gdata.fluids->get_i_magFluid();
	int num = gdata.fluids->get_N_OMINT(iMagFluid);
	int q_Bx = gdata.fluids->get_q_Bx();
	int q_By = gdata.fluids->get_q_By();
	int q_Bz = gdata.fluids->get_q_Bz();
#else
	int num = N_OMINT;
	int q_Bx = gdata.fluid.get_q_Bx();
	int q_By = gdata.fluid.get_q_By();
	int q_Bz = gdata.fluid.get_q_Bz();
#endif

	fluxWE = new NumMatrix<REAL,2> [num];
	fluxSN = new NumMatrix<REAL,2> [num];
	fluxBT = new NumMatrix<REAL,2> [num];

#if (CT_TYPE == STONE)
	if(dir == 0) {
		Ex.resize(Index::set(-1, -1),
		          Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxSN[q_Bz].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxBT[q_By].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

#if (STONE_TYPE == STONE_CENTRE)
		fluxSN[0].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxBT[0].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

		dExdyjm1_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dExdyjm3_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dExdzkm1_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dExdzkm3_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
#endif

	} else if(dir == 1) {
		Ey.resize(Index::set(-1, -1),
		          Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxWE[q_Bz].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxBT[q_Bx].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

#if (STONE_TYPE == STONE_CENTRE)
		fluxWE[0].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxBT[0].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

		dEydxim1_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dEydxim3_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dEydzkm1_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dEydzkm3_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
#endif
	} else {
		Ez.resize(Index::set(-1, -1),
		          Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxWE[q_By].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxSN[q_Bx].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

#if (STONE_TYPE == STONE_CENTRE)
		fluxWE[0].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		fluxSN[0].resize(Index::set(-2, -2),
		                 Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

		dEzdxim1_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dEzdxim3_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dEzdyjm1_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dEzdyjm3_4.resize(Index::set(-2, -2),
		                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
#endif

	}
#endif

	if(dir0 == 0 || dir1 == 0) {
		v_x.resize(Index::set(-2, -2),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		B_x.resize(Index::set(-2, -2),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dBxdx.resize(Index::set(-2, -2),
		             Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

		for(int num=0; num<2; ++num) {
			dvxdx[num].resize(Index::set(-2, -2),
			                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		}
	}
   
	if(dir0 == 1 || dir1 == 1) {
		v_y.resize(Index::set(-2, -2),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		B_y.resize(Index::set(-2, -2),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dBydx.resize(Index::set(-2, -2),
		             Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

		for(int num=0; num<2; ++num) {
			dvydx[num].resize(Index::set(-2, -2),
			                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		}
	}

	if(dir0 == 2 || dir1 == 2) {
		v_z.resize(Index::set(-2, -2),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		B_z.resize(Index::set(-2, -2),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		dBzdx.resize(Index::set(-2, -2),
		             Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

		for(int num=0; num<2; ++num) {
			dvzdx[num].resize(Index::set(-2, -2),
			                  Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		}
	}

#if (CT_TYPE == CONSISTENT)
	for(int num=0; num<2; ++num) {

		v_ch_p[num].resize(Index::set(-2, -2),
		                   Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		v_ch_m[num].resize(Index::set(-2, -2),
		                   Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		v_cor_p[num].resize(Index::set(-2, -2),
		                    Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
		v_cor_m[num].resize(Index::set(-2, -2),
		                    Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

		v_ch_p[num].set_constVal(0.);
		v_ch_m[num].set_constVal(0.);
		v_cor_p[num].set_constVal(0.);
		v_cor_m[num].set_constVal(0.);

	}
#endif
	emf.resize(Index::set(-2, -2),
	           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));

}


fields_2D::~fields_2D() {
	//! Destructor
	delete [] fluxWE;
	delete [] fluxSN;
	delete [] fluxBT;
}

phys_fields_1D::phys_fields_1D(const Data &gdata, const int &dir) {

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	int n_omInt = gdata.fluids->get_N_OMINT_ALL();
#else
	int n_omInt = gdata.fluid.get_N_OMINT();
#endif

	uPri = new NumMatrix<REAL,1> [n_omInt];
	uCon = new NumMatrix<REAL,1> [n_omInt];
	flux = new NumMatrix<REAL,1> [n_omInt];

	ptherm.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	ptotal.resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	
#if (USE_COROTATION == CRONOS_ON)
	// Prepare fields holding velocity in inertial frame
	uInertial[0].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	uInertial[1].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
	uInertial[2].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
#endif

	for(int q=0; q<n_omInt; ++q) {
		uPri[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		uPri[q].clear();
		uCon[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		uCon[q].clear();

		flux[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		flux[q].clear();
	}
}


phys_fields_1D::phys_fields_1D(const Data &gdata, int dir, int num) {

	uPri = new NumMatrix<REAL,1> [num];
	uCon = new NumMatrix<REAL,1> [num];
	flux = new NumMatrix<REAL,1> [num];

	for(int q=0; q<num; ++q) {
		uPri[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		uPri[q].clear();
		uCon[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		uCon[q].clear();

		flux[q].resize(Index::set(-2),Index::set(gdata.mx[dir]+1));
		flux[q].clear();
	}

}

phys_fields_1D::~phys_fields_1D() {
	delete [] uPri;
	delete [] uCon;
	delete [] flux;
}

phys_fields_2D::phys_fields_2D(const Data &gdata, const int &dir) {

	assert(dir >= 0 && dir < DIM);
	if(dir == 0) {
		this->dir0 = 1;
		this->dir1 = 2;
	} else if (dir == 1) {
		this->dir0 = 2;
		this->dir1 = 0;
	} else {
		this->dir0 = 0;
		this->dir1 = 1;
	}

	if(dir == 1 || dir == 2) {
		v_x.resize(Index::set(-1, -1),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
	}
	if(dir == 0 || dir == 2) {
		v_y.resize(Index::set(-1, -1),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
	}
	if(dir == 0 || dir == 1) {
		v_z.resize(Index::set(-1, -1),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
	}

	if(dir == 1 || dir == 2) {
		B_x.resize(Index::set(-1, -1),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
	}
	if(dir == 0 || dir == 2) {
		B_y.resize(Index::set(-1, -1),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
	}
	if(dir == 0 || dir == 1) {
		B_z.resize(Index::set(-1, -1),
		           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
	}

	emf.resize(Index::set(-1, -1),
	           Index::set(gdata.mx[dir0]+1, gdata.mx[dir1]+1));
  
}


phys_fields_0D::phys_fields_0D(const Data &gdata, int face, const CronosFluid &fluid) {

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
//	fieldLists List(fluid, dir);
	this->num = fluid.get_N_OMINT();
#else
	this->num = N_OMINT;
#endif

	this->face = face;

	uPri.resize(num);
	uCon.resize(num);
	flux_phys.resize(num);
#if (USE_COROTATION == CRONOS_ON)
	uInertial.resize(num);
#endif

	ptotal = 0.;
	ptherm = 0.;
	carbuncle_flag = 0;

}

int phys_fields_0D::get_num() {
	return num;
}

num_fields_0D::num_fields_0D(const Data &, const CronosFluid &fluid) {
#if (FLUID_TYPE == CRONOS_MULTIFLUID)
//	fieldLists List(fluid, dir);
	this->num = fluid.get_N_OMINT();
#else
	this->num = N_OMINT;
#endif

	v_ch_p = 1.;
	v_ch_m = 1.;
	flux_num.resize(num);

}
