
#ifndef CRONOS_RECONST_2ND_H
#define CRONOS_RECONST_2ND_H = 1

#include "reconst.H"
#include "limiter.H"
#include "fields_local.H"
#include "data.H"
#include "matrix.H"
#include <vector>

class SingleReconstruction_2nd: public SingleReconstruction {
	/** Class for second order spatial reconstruction
	 * --> using different limiters
	*/
public:
	SingleReconstruction_2nd(){}
	SingleReconstruction_2nd(const Data &, int dir, int substep);
	SingleReconstruction_2nd(const Data &, const CronosFluid &fluid, int dir, int qReconst, int substep);

	virtual ~SingleReconstruction_2nd();

	/**
	 * Compute limited version of derivatives for case of 1D arrays
	 * */
	void getDeriv_limit(const Data &, const NumMatrix<REAL,1> &,
	              NumMatrix<REAL,1> &);

	/**
	 * Compute limited version of derivatives for block-structured version of code
	 * */
	void perpareDerivs(const Data &, int ix, int iy, int iz);
	void computeNormal(const Data &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
	                   NumMatrix<REAL,1> &, NumMatrix<REAL,1> &);
	void computePar(const Data &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
	                NumMatrix<REAL,1> &);
	void computePerp(const Data &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
	                 NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
	                 NumMatrix<REAL,1> &);
	void get_Vals_EW(const Data &, phys_fields_0D &xFieldsW,
			phys_fields_0D &xFieldsE, int, int, int);
	void get_Vals_SN(const Data &, phys_fields_0D &xFieldsS,
			phys_fields_0D &xFieldsN, int, int, int);
	void get_Vals_BT(const Data &, phys_fields_0D &xFieldsB,
			phys_fields_0D &xFieldsT, int, int, int);
private:
	limiter* Limiter;
};

class Reconstruction2D_2nd: public Reconstruction2D {
public:
	Reconstruction2D_2nd(const Data &, const int &);
	virtual ~Reconstruction2D_2nd(){}
	void computeNormal(const Data &,
	                   NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                   NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                   NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                   NumMatrix<REAL,2> &);
	void computePerp(const Data &, NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                 const int &,
	                 NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                 NumMatrix<REAL,2> &, NumMatrix<REAL,2> &);
	void getDerivs(const Data &, const NumMatrix<REAL,2> &,
	               NumMatrix<REAL,2> &, NumMatrix<REAL,2> &);
	void getDeriv_limit(const Data &, const NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	              int);
private:
 	limiter* Limiter;
};

#endif
