#include "gridgen.H"
#include <iomanip>
#include <cmath>
#include <vector>
#include "DissipationMHD.H"

using namespace std;

#if ( FLUID_TYPE != CRONOS_RHD )
#error
#endif


REAL HyperbolicSolver::singlestep(Data &gdata, gridFunc &gfunc,
                                  ProblemType &Problem, int n)
{

// ----------------------------------------------------------------
// Starting the clocks
// ----------------------------------------------------------------
	if( n == 0 ) {
		timerFullStep.start();
	}
	timerSubStep.start();


// ----------------------------------------------------------------
// Initializing
// ----------------------------------------------------------------

	if(eos == (EquationOfState*)0) {
		eos = new EquationOfState(Problem);
	}
	if(sources == NULL) {
		sources = new SourceTerms(gdata, Problem, IntegrateA, thermal);
	}
	
#ifdef USE_VALENCIA
	gdata.valencia = Problem.valencia;
#endif

	Save = new Saves(gdata);


	for (int q = 0; q < n_omInt; ++q) {
		gdata.nom[q].clear();
	}


#if (OMS_USER == TRUE)

	for (int q = 0; q < n_omIntUser; ++q) {
		gdata.nom_user[q].clear();
	}

#endif



// ----------------------------------------------------------------
// Save current state
// ----------------------------------------------------------------
#if (PARTICLE_TRANSPORT == CRONOS_ON)
	if (n == 0) {
		for (int q = 0; q < n_omInt; ++q) {
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
						gdata.om_old[q](ix, iy, iz) = gdata.om[q](ix, iy, iz);
					}
				}
			}
		}
		for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
			for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
				for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
					gdata.lorentz_old(ix, iy, iz) = gdata.lorentz(ix, iy, iz);
				}
			}
		}
	}
#endif


// ---------------------------------------------------------------	      
//      Aux transormations
//----------------------------------------------------------------



//---------------------------------------------------------------
//		Fill mass flux array with zeros
//---------------------------------------------------------------

#if(GEOM == SPHERICAL)
	gdata.massFlux.clear();
#endif

	gdata.CheckNan(1);

	// Compute the carbuncle-flag if necessary
	if(gdata.use_carbuncleFlag) {
		RiemannX->compute_carbuncleFlag(gdata);
	}

	REAL cfl_lin(0.), cfl_eta(0.);

	cronos::vector<REAL> pos(0,0,0);
	cronos::vector<REAL> ipos(0,0,0);

	// x-direction
	int dir=0;


	for(int k=-1; k<=gdata.mx[2]+1; ++k) {
		// Update of perp part of position:
		pos.set(2, gdata.getCen_z(k));
		ipos.set(2, k);
		for (int j=-1; j<=gdata.mx[1]+1; ++j){
			// Update of perp part of position:
			pos.set(1, gdata.getCen_y(j));
			ipos.set(1, j);

			// Assigning 1D array data
			for (int q = 0; q < n_omInt; ++q){
				if((FLUID_TYPE == CRONOS_MHD && (q<q_By || q>q_Bz)) ||
				   FLUID_TYPE != CRONOS_MHD) {
					for (int i = -3; i <= gdata.mx[0]+2; ++i){
						fieldsX->omLoc[q](i) = gdata.om[q](i,j,k);
					}
				}
			}

			// Assigning shifted fields for magnetic induction
#if (FLUID_TYPE == CRONOS_MHD)
			for (int i = -3; i <= gdata.mx[0]+2; ++i){
				fieldsX->omLocP[q_By](i) = gdata.om[q_By](i,j,k);
			}
			for (int i = -3; i <= gdata.mx[0]+2; ++i){
				fieldsX->omLocM[q_By](i) = gdata.om[q_By](i,j-1,k);
			}
			for (int i = -3; i <= gdata.mx[0]+2; ++i){
				fieldsX->omLocP[q_Bz](i) = gdata.om[q_Bz](i,j,k);
			}
			for (int i = -3; i <= gdata.mx[0]+2; ++i){
				fieldsX->omLocM[q_Bz](i) = gdata.om[q_Bz](i,j,k-1);
			}
#endif


			// Assigning carbuncle flag values if necessary
			if(gdata.use_carbuncleFlag) {
				for (int i = -2; i <= gdata.mx[0]+2; ++i){
					fieldsX->carbuncle_flag(i) = gdata.carbuncleFlag(i,j,k);
				}
			}

			// Doing the reconstruction:
			ReconstX[n]->compute(gdata, *fieldsX, *physValxL, *physValxR);

#if (OMS_USER == TRUE)
			// Reconstruction for User-data:
			ReconstXUser[n]->compute(gdata, *fieldsXUser,
			                      *physValxLUser, *physValxRUser);
#endif

			// Get conservative variables:
			// on LHS
			Trafo->get_Cons(gdata, Problem, *eos, ipos, *physValxL, dir, -0.5);
			// on RHS
			Trafo->get_Cons(gdata, Problem, *eos, ipos, *physValxR, dir,  0.5);


			// Compute physical fluxes:
			// on LHS
			PhysFlux->get_PhysFlux(gdata, Problem, ipos, *physValxL, dir, -0.5);
			// on RHS
			PhysFlux->get_PhysFlux(gdata, Problem, ipos, *physValxR, dir,  0.5);


			// Possible modification by user
			Problem.get_PhysFlux(gdata, ipos, *physValxL, dir, -0.5);

			Problem.get_PhysFlux(gdata, ipos, *physValxR, dir,  0.5);

			// Compute characteristic velocities:
			RiemannX->get_vChar(gdata, Problem, ipos, *physValxL, *physValxR,
					fieldsX->v_ch_m, fieldsX->v_ch_p, dir, cfl_lin);

			// Compute numerical fluxes:
			RiemannX->get_NumFlux(gdata, pos, *physValxL, *physValxR, *fieldsX, dir);

			// Compute changes (noms)
			if((j>=0 && j<=gdata.mx[1]) && (k>=0 && k<=gdata.mx[2])) {

				get_Changes(gdata, fieldsX->flux, gdata.nom, fieldsX->ptotal,
				            0, j, k, n_omInt, RiemannX->get_Fluid_Type());
			}

			// Save 1D data if necessary
			Save->save_vars(gdata, *fieldsX, 0, j, k);


#if (OMS_USER == TRUE)
			// Assing 1D array data for user fields:
			for (int q = 0; q < n_omIntUser; ++q){
				for (int i = -3; i <= gdata.mx[0]+2; ++i){
					fieldsXUser->omLoc[q](i) = gdata.om_user[q](i,j,k);
				}
			}

			// Reconstruction for User-data:
			ReconstXUser[n]->compute(gdata, *fieldsXUser, *physValxLUser, *physValxRUser);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
			Problem.particleTransport.get_Cons(gdata, ipos, *physValxL, *physValxLUser, dir, -0.5);
			Problem.particleTransport.get_Cons(gdata, ipos, *physValxR, *physValxRUser, dir,  0.5);

			Problem.particleTransport.get_PhysFlux(gdata, ipos, *physValxL, *physValxLUser, dir, -0.5);
			Problem.particleTransport.get_PhysFlux(gdata, ipos, *physValxR, *physValxRUser, dir,  0.5);
#endif

			// Get conservative variables (needed for Riemann solver)
			// -- default is just a copy
			// LHS
			Problem.get_Cons(gdata, ipos, *physValxL, *physValxLUser, dir, -0.5);
			// RHS
			Problem.get_Cons(gdata, ipos, *physValxR, *physValxRUser, dir,  0.5);

			// Compute physical fluxes for user fields:
			// on LHS
			Problem.get_PhysFluxUser(gdata, ipos, *physValxL, *physValxLUser, dir, -0.5);
			// on RHS
			Problem.get_PhysFluxUser(gdata, ipos, *physValxR, *physValxRUser, dir,  0.5);

			fieldsXUser->v_ch_p = fieldsX->v_ch_p;
			fieldsXUser->v_ch_m = fieldsX->v_ch_m;

			RiemannXUser->get_NumFlux(gdata, pos, *physValxLUser, *physValxRUser,
			                          *fieldsXUser, dir);


			// Compute changes (noms)
			if((j>=0 && j<=gdata.mx[1]) && (k>=0 && k<=gdata.mx[2])) {
				get_Changes(gdata, fieldsXUser->flux, gdata.nom_user,
				            fieldsXUser->ptotal,
				            0, j, k, n_omIntUser,
				            RiemannXUser->get_Fluid_Type());

			}
#endif
      
		}
	}


	// Compute sum of mass flux along phi and theta for MPI parallel case
#if (GEOM == SPHERICAL)
#ifdef parallel
	NumMatrix<double,1> massFlux_global;
	int len = gdata.massFlux.getHigh(0) - gdata.massFlux.getLow(0) + 1;
	massFlux_global.resize(gdata.massFlux);
	MPI_Reduce(gdata.massFlux, massFlux_global, len, MPI_DOUBLE,
			MPI_SUM, 0, gdata.comm_constx);
	MPI_Bcast(massFlux_global, len, MPI_DOUBLE, 0, gdata.comm_constx);
//	MPI_Barrier(gdata.comm_constx);
	gdata.massFlux = massFlux_global;
#endif
#endif


	// y-direction
	dir = 1;

	for (int k = -1; k <= gdata.mx[2]+1; ++k){
		// Update of perp part of position:
		pos.set(2, gdata.getCen_z(k));
		ipos.set(2, k);
		for (int i = -1; i <= gdata.mx[0]+1; ++i){
			// Update of perp part of position:
			pos.set(0, gdata.getCen_x(i));
			ipos.set(0, i);
      
			// Assigning 1D array data
			for (int q = 0; q < n_omInt; ++q){
				if((FLUID_TYPE == CRONOS_MHD && q!=q_Bx && q!=q_Bz) ||
				   FLUID_TYPE != CRONOS_MHD) {
					for (int j = -3; j <= gdata.mx[1]+2; ++j){
						fieldsY->omLoc[q](j) = gdata.om[q](i,j,k);
					}
				}
			}


			// Assigning shifted fields for magnetic induction
#if (FLUID_TYPE == CRONOS_MHD)			
			for (int j = -3; j <= gdata.mx[1]+2; ++j){
				fieldsY->omLocP[q_Bx](j) = gdata.om[q_Bx](i,j,k);
//				fieldsY->omLocP[q_Bx](j) = gdata.om[q_Bx](i,j,k) - Problem.get_FieldAnalytical(4,gdata.getEdgL_x(i+1),gdata.getCen_y(j),gdata.getCen_z(k));
			}
			for (int j = -3; j <= gdata.mx[1]+2; ++j){
				fieldsY->omLocM[q_Bx](j) = gdata.om[q_Bx](i-1,j,k);
//				fieldsY->omLocM[q_Bx](j) = gdata.om[q_Bx](i-1,j,k) - Problem.get_FieldAnalytical(4,gdata.getEdgL_x(i),gdata.getCen_y(j),gdata.getCen_z(k));
			}
			for (int j = -3; j <= gdata.mx[1]+2; ++j){
				fieldsY->omLocP[q_Bz](j) = gdata.om[q_Bz](i,j,k);
			}
			for (int j = -3; j <= gdata.mx[1]+2; ++j){
				fieldsY->omLocM[q_Bz](j) = gdata.om[q_Bz](i,j,k-1);
			}
#endif

			// Assigning carbuncle flag values if necessary
			if(gdata.use_carbuncleFlag) {
				for (int j = -2; j <= gdata.mx[1]+2; ++j){
					fieldsY->carbuncle_flag(j) = gdata.carbuncleFlag(i,j,k);
				}
			}

			// Doing the reconstruction:
			ReconstY[n]->compute(gdata, *fieldsY, *physValyL, *physValyR);

      
			// Get conservative variables:
			// on LHS
			Trafo->get_Cons(gdata, Problem, *eos, ipos, *physValyL, dir, -0.5);
			// on RHS
			Trafo->get_Cons(gdata, Problem, *eos, ipos, *physValyR, dir,  0.5);


			// Compute physical fluxes:
			// on LHS
			PhysFlux->get_PhysFlux(gdata, Problem, ipos, *physValyL, dir, -0.5);
			// on RHS
			PhysFlux->get_PhysFlux(gdata, Problem, ipos, *physValyR, dir,  0.5);


			// Possible modification by user
			Problem.get_PhysFlux(gdata, ipos, *physValyL, dir, -0.5);

			Problem.get_PhysFlux(gdata, ipos, *physValyR, dir,  0.5);


			// Compute characteristic velocities:
			RiemannY->get_vChar(gdata, Problem, ipos, *physValyL, *physValyR,
					fieldsY->v_ch_m, fieldsY->v_ch_p, dir, cfl_lin);
//			RiemannY->get_vChar(gdata, Problem, ipos, *physValyL, *physValyR, *fieldsY,
//			                   dir, cfl_lin);

			// Compute numerical flux:
			RiemannY->get_NumFlux(gdata, pos, *physValyL, *physValyR, *fieldsY, dir);


			// Compute changes (noms)
			if((i>=0 && i<=gdata.mx[0]) && (k>=0 && k<=gdata.mx[2])) {
				get_Changes(gdata, fieldsY->flux, gdata.nom, fieldsY->ptotal,
				            dir, i, k, n_omInt, RiemannY->get_Fluid_Type());
			}


			// Save 1D data if necessary
			Save->save_vars(gdata, *fieldsY, dir, i, k);



#if (OMS_USER == TRUE)
			for (int q = 0; q < n_omIntUser; ++q){
				for (int j = -3; j <= gdata.mx[dir]+2; ++j){
					fieldsYUser->omLoc[q](j) = gdata.om_user[q](i,j,k);
				}
			}


			// Reconstruction for User-data:
			ReconstYUser[n]->compute(gdata, *fieldsYUser, *physValyLUser, *physValyRUser);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
			Problem.particleTransport.get_Cons(gdata, ipos, *physValyL, *physValyLUser, dir, -0.5);
			Problem.particleTransport.get_Cons(gdata, ipos, *physValyR, *physValyRUser, dir, 0.5);
			Problem.particleTransport.get_PhysFlux(gdata, ipos, *physValyL, *physValyLUser, dir, -0.5);
			Problem.particleTransport.get_PhysFlux(gdata, ipos, *physValyR, *physValyRUser, dir,  0.5);
#endif


			// Get conservative variables (needed for Riemann solver)
			// -- just by copying for the time being
			// LHS
			Problem.get_Cons(gdata, ipos, *physValyL, *physValyLUser, dir, -0.5);
			// Trafo->get_ConsUser(gdata, *physValyLUser, n_omIntUser);
			// RHS
			Problem.get_Cons(gdata, ipos, *physValyR, *physValyRUser, dir, 0.5);
			// Trafo->get_ConsUser(gdata, *physValyRUser, n_omIntUser);
			

			// Compute physical fluxes for user fields:
			// on LHS
			Problem.get_PhysFluxUser(gdata, ipos, *physValyL, *physValyLUser, dir, -0.5);

			// on RHS
			Problem.get_PhysFluxUser(gdata, ipos, *physValyR, *physValyRUser, dir,  0.5);
			



			fieldsYUser->v_ch_p = fieldsY->v_ch_p;
			fieldsYUser->v_ch_m = fieldsY->v_ch_m;


			RiemannYUser->get_NumFlux(gdata, pos, *physValyLUser, *physValyRUser,
			                          *fieldsYUser, dir);


			if((i>=0 && i<=gdata.mx[0]) && (k>=0 && k<=gdata.mx[2])) {

				get_Changes(gdata, fieldsYUser->flux, gdata.nom_user,
				            fieldsYUser->ptotal,
				            dir, i, k, n_omIntUser,
				            RiemannYUser->get_Fluid_Type());
			}
				
#endif
			

		}
	}



	// z -direction
	dir = 2;

	for (int j = -1; j <= gdata.mx[1]+1; ++j){
		// Update of perp part of position:
		pos.set(1, gdata.getCen_y(j));
		ipos.set(1, j);
		for (int i = -1; i <= gdata.mx[0]+1; ++i){
			// Update of perp part of position:
			pos.set(0, gdata.getCen_x(i));
			ipos.set(0, i);

			// Assigning 1D array data
			for (int q = 0; q < n_omInt; ++q){
				if((FLUID_TYPE == CRONOS_MHD && (q<q_Bx || q>q_By)) ||
					FLUID_TYPE != CRONOS_MHD) {
					for (int k = -3; k <= gdata.mx[2]+2; ++k){
						fieldsZ->omLoc[q](k) = gdata.om[q](i,j,k);
					}
				}
			}
      
			// Assigning shifted fields for magnetic induction
#if (FLUID_TYPE == CRONOS_MHD)
			for (int k = -3; k <= gdata.mx[dir]+2; ++k){
				fieldsZ->omLocP[q_Bx](k) = gdata.om[q_Bx](i,j,k);
			}
			for (int k = -3; k <= gdata.mx[dir]+2; ++k){
				fieldsZ->omLocM[q_Bx](k) = gdata.om[q_Bx](i-1,j,k);
			}
			for (int k = -3; k <= gdata.mx[dir]+2; ++k){
				fieldsZ->omLocP[q_By](k) = gdata.om[q_By](i,j,k);
			}
			for (int k = -3; k <= gdata.mx[dir]+2; ++k){
				fieldsZ->omLocM[q_By](k) = gdata.om[q_By](i,j-1,k);
			}
#endif

			// Assigning carbuncle flag values if necessary
			if(gdata.use_carbuncleFlag) {
				for (int k = -2; k <= gdata.mx[2]+2; ++k){
					fieldsZ->carbuncle_flag(k) = gdata.carbuncleFlag(i,j,k);
				}
			}

			// Doing the reconstruction:
			ReconstZ[n]->compute(gdata, *fieldsZ, *physValzL, *physValzR);

      
			// Get conservative variables:
			// on LHS
			Trafo->get_Cons(gdata, Problem, *eos, ipos, *physValzL, dir, -0.5);
			// on RHS
			Trafo->get_Cons(gdata, Problem, *eos, ipos, *physValzR, dir,  0.5);


			// Compute physical fluxes:
			// on LHS
			PhysFlux->get_PhysFlux(gdata, Problem, ipos, *physValzL, dir, -0.5);
			// on RHS
			PhysFlux->get_PhysFlux(gdata, Problem, ipos, *physValzR, dir,  0.5);


			// Possible modification by user
			Problem.get_PhysFlux(gdata, ipos, *physValzL, dir, -0.5);

			Problem.get_PhysFlux(gdata, ipos, *physValzR, dir,  0.5);


			// Compute characteristic velocities:
			RiemannZ->get_vChar(gdata, Problem, ipos, *physValzL, *physValzR,
						fieldsZ->v_ch_m, fieldsZ->v_ch_p, dir, cfl_lin);
//			RiemannZ->get_vChar(gdata, Problem, ipos, *physValzL, *physValzR, *fieldsZ,
//			                   dir, cfl_lin);


			// Compute numerical flux:
			RiemannZ->get_NumFlux(gdata, pos, *physValzL, *physValzR, *fieldsZ, dir);


			// Compute changes (noms)
			if((i>=0 && i<=gdata.mx[0]) && (j>=0 && j<=gdata.mx[1])) {
				get_Changes(gdata, fieldsZ->flux, gdata.nom, fieldsZ->ptotal,
				            dir, i, j, n_omInt, RiemannZ->get_Fluid_Type());
			}


			// Save 1D data if necessary
			Save->save_vars(gdata, *fieldsZ, dir, i, j);



#if (OMS_USER == TRUE)
			for (int q = 0; q < n_omIntUser; ++q){
				for (int k = -3; k <= gdata.mx[dir]+2; ++k){
					fieldsZUser->omLoc[q](k) = gdata.om_user[q](i,j,k);
				}
			}


			// Reconstruction for User-data:
			ReconstZUser[n]->compute(gdata, *fieldsZUser, *physValzLUser, *physValzRUser);


#if (PARTICLE_TRANSPORT == CRONOS_ON)
			Problem.particleTransport.get_Cons(gdata, ipos, *physValzL, *physValzLUser, dir, -0.5);
			Problem.particleTransport.get_Cons(gdata, ipos, *physValzR, *physValzRUser, dir,  0.5);

			Problem.particleTransport.get_PhysFlux(gdata, ipos, *physValzL, *physValzLUser, dir, -0.5);
			Problem.particleTransport.get_PhysFlux(gdata, ipos, *physValzR, *physValzRUser, dir,  0.5);
#endif


			// Get conservative variables (needed for Riemann solver)
			// -- just by copying for the time being
			// LHS
			Problem.get_Cons(gdata, ipos, *physValzL, *physValzLUser, dir, -0.5);
			// Trafo->get_ConsUser(gdata, *physValzLUser, n_omIntUser);
			// RHS
			Problem.get_Cons(gdata, ipos, *physValzR, *physValzRUser, dir,  0.5);
			// Trafo->get_ConsUser(gdata, *physValzRUser, n_omIntUser);


			// Compute physical fluxes for user fields:
			// on LHS
			Problem.get_PhysFluxUser(gdata, ipos, *physValzL, *physValzLUser, dir, -0.5);
			// on RHS
			Problem.get_PhysFluxUser(gdata, ipos, *physValzR, *physValzRUser, dir,  0.5);


			fieldsZUser->v_ch_p = fieldsZ->v_ch_p;
			fieldsZUser->v_ch_m = fieldsZ->v_ch_m;

			RiemannZUser->get_NumFlux(gdata, pos, *physValzLUser, *physValzRUser, *fieldsZUser, dir);

			if((i>=0 && i<=gdata.mx[0]) && (j>=0 && j<=gdata.mx[1])) {
				get_Changes(gdata, fieldsZUser->flux, gdata.nom_user,
				            fieldsZUser->ptotal,
				            dir, i, j, n_omIntUser,
				            RiemannZUser->get_Fluid_Type());
			}
			
#endif


		}
	}

// ---------------------------------------------------------------
//   Evolution for EMF via Constrained Transport
// ----------------------------------------------------------------

	gettimeofday(&tin5, 0);

#if (FLUID_TYPE == CRONOS_MHD)

	if(gdata.tstep==0 && n==0) {
		if(IntegrateA) {
			gdata.nom[q_Bx].resize(Index::set(0,0,0),
					Index::set(gdata.mx[0],gdata.mx[1]+1,gdata.mx[2]+1));
			gdata.nom[q_By].resize(Index::set(0,0,0),
					Index::set(gdata.mx[0]+1,gdata.mx[1],gdata.mx[2]+1));
			gdata.nom[q_Bz].resize(Index::set(0,0,0),
					Index::set(gdata.mx[0]+1,gdata.mx[1]+1,gdata.mx[2]));
		} else {
			gdata.nom[q_Bx].resize(Index::set(-1,0,0),
					Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
			gdata.nom[q_By].resize(Index::set(0,-1,0),
					Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
			gdata.nom[q_Bz].resize(Index::set(0,0,-1),
					Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
		}
	}
	gdata.nom[q_Bx].clear();
	gdata.nom[q_By].clear();
	gdata.nom[q_Bz].clear();

//#if (USE_COROTATION == CRONOS_ON)
//	// E_x:
//	CTSolveX->get_NumEMFCT(gdata, Problem, *Save, *Trafo, nom);
//
//
//	// Ey:
//	CTSolveY->get_NumEMFCT(gdata, Problem, *Save, *Trafo, nom);
//
//
//	// Ez:
//	CTSolveZ->get_NumEMFCT(gdata, Problem, *Save, *Trafo, nom);
//#else
	// E_x:
	CTSolveX->get_NumEMFCT(gdata, Problem, *Save, gdata.nom);


	// Ey:
	CTSolveY->get_NumEMFCT(gdata, Problem, *Save, gdata.nom);


	// Ez:
	CTSolveZ->get_NumEMFCT(gdata, Problem, *Save, gdata.nom);
//#endif


#endif


	delete Save;


// ----------------------------------------------------------------
//   Check for errors:
// ----------------------------------------------------------------

	for(int q = 0; q<n_omInt; ++q) {
		CheckNan(gdata.nom[q],q, 0, 1,"nom");
	}

// ----------------------------------------------------------------
//   Compute Courant number
// ----------------------------------------------------------------

	REAL cfl = compute_cfl(gdata, Problem, cfl_eta, cfl_lin, n);
	
// ----------------------------------------------------------------
//   Geometrical source terms:
// ----------------------------------------------------------------

#ifdef USE_VALENCIA
	Problem.valencia->get_sources(gdata, Problem, gdata.nom);
#else

#if (GEOM != CARTESIAN)
	sources->src_Geom(gdata, Problem, gdata.nom);
#endif
#endif

// ----------------------------------------------------------------
//   User defined source terms:
// ----------------------------------------------------------------

#if (OMS_USER == TRUE)
	Problem.src_User(gdata, gdata.nom, gdata.nom_user);
#endif


// ----------------------------------------------------------------
//   Check for errors again before applying the changes:
// ----------------------------------------------------------------

	for(int q = 0; q<n_omInt; ++q) {
		CheckNan(gdata.nom[q],q, 0, 2,"nom");
	}

// ----------------------------------------------------------------
//   Transform to conservative variables:
// ----------------------------------------------------------------

	Trafo->TransPrim2Cons(gdata, gfunc, Problem);
	
#if (PARTICLE_TRANSPORT == CRONOS_ON)
	Problem.particleTransport.TransPrim2Cons(gdata);
#endif

	Problem.TransPrim2Cons(gdata);


// ----------------------------------------------------------------
//   Determine domain to be integrated and apply changes:
// ----------------------------------------------------------------

	for (int q = 0; q < n_omInt; ++q) {
		
		TimeIntegratorGeneric[q]->Substep(gdata, Problem, gdata.nom[q], gdata.om, n);

	}

#if (OMS_USER == TRUE)

	for (int q=0; q<n_omIntUser; ++q) {

		TimeIntegratorUser[q]->Substep(gdata, Problem, gdata.nom_user[q],
		                              gdata.om_user, n);
		
	}

#endif


// ----------------------------------------------------------------
//   Determine time at intermediate steps:
// ----------------------------------------------------------------

	// Time must come before boundary for time dependent boundaries
	gdata.time += TimeIntegratorGeneric[0]->get_dt(gdata, n);


// ----------------------------------------------------------------
//   Transformation to base variables and bcs
// ----------------------------------------------------------------

	Trafo->TransCons2Prim(gdata, gfunc, Problem);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	Problem.particleTransport.TransCons2Prim(gdata);
#endif

	Problem.TransCons2Prim(gdata);

	timerBoundaries.start();
	gfunc.boundary(gdata, Problem);
	timerBoundaries.stop();


// ----------------------------------------------------------------
//   Clean negative particles
// ----------------------------------------------------------------

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	for(int ipart = 0; ipart < N_PARTICLE_SPECIES; ++ipart) {
		Particle & particle = Problem.particleTransport.get_particle(ipart);
		for(int ip=particle.getLo(); ip<=particle.getHi(); ++ip) {
			for(int iz=0; iz<=gdata.mx[2]; ++iz) {
				for(int iy=0; iy<=gdata.mx[1]; ++iy) {
					for(int ix=0; ix<=gdata.mx[0]; ++ix) {
						if (particle(ip, ix, iy, iz) < 0) {
							particle(ip, ix, iy, iz) = 0;
						}
					}
				}
			}
		}
	}
#endif


// ----------------------------------------------------------------
//   Test physical state
// ----------------------------------------------------------------

#if (USE_COROTATION == CRONOS_ON)
	// Transform to inertial frame velocity for phystest
	Trafo->TransCorotToInert(gdata, gfunc, Problem);
#endif

	phystest(gdata, gfunc, Problem, n);

#if (USE_COROTATION == CRONOS_ON)
	// Transform back to co-rotating frame
	Trafo->TransInertToCorot(gdata, gfunc, Problem);
#endif

	// Computing div B only at end of timestep
	if(Problem.mag && n == TIME_SUBSTEPS-1){
		compute_divB(gdata, gfunc, Problem);
	}


// ----------------------------------------------------------------
//   Performance estimate
// ----------------------------------------------------------------

	timerSubStep.stop();

	timeval tstepOld = tstep;
	if(n == TIME_SUBSTEPS-1) {
		timerFullStep.stop();
	}

	if(gdata.rank == 0) {
		if(Problem.get_Info() && Problem.checkout(5)) {
			cout << "   Time needed for substep    " << n + 1 << ": " << timerSubStep.dtime() << endl;
			cout << "   Time needed for boundaries " << n + 1 << ": " << timerBoundaries.dtime()  << endl;

			if(n == TIME_SUBSTEPS-1) {
				cout << "               for full step   : " << timerFullStep.dtime() << endl;
			}
		}
	}



	return cfl;
}
