#ifndef CRONOS_TRANSFORMATIONS_H
#define CRONOS_TRANSFORMATIONS_H

#include "CException.H"
#include "data.H"
#include "gridfunc.H"
#include "problem.H"
#include "eos.H"
#include "vector.H"
#include "fields_local.H"

#if (FLUID_TYPE == CRONOS_RHD)
#include "state.H"
#include <gsl/gsl_roots.h>
#include "StateInvertors.H"
#endif

/*********************************************************\
 *
 * CLASS: Transformations:
 *
 * Supplies routines to convert to and fro between 
 * conservative and different kinds of characteristic 
 * variables
 *
\*********************************************************/

class Transformations {
public:
//	Transformations(ProblemType &, bool);
	Transformations(const CronosFluid&, ProblemType &, bool, int iFluid=0);
	void TransPrim2Cons(Data &gdata, gridFunc &gfunc,
	                    ProblemType &Problem);
	void TransCons2Prim(Data &gdata, gridFunc &gfunc,
	                    ProblemType &Problem);
	void TransE(Data &, gridFunc &, ProblemType &, string);
	void TransE2Eth(Data &, gridFunc &, ProblemType &, int, bool);
	void TransE2Eth(Data &, gridFunc &, ProblemType &);
	void TransEth2E(Data &, gridFunc &, ProblemType &);
	void TransT2Eth(Data &, gridFunc &, ProblemType &);
	REAL TransEth2T(Data &, gridFunc &, ProblemType &);
	void TransT2E(Data &, gridFunc &, ProblemType &);
	REAL TransE2T(Data &, gridFunc &, ProblemType &);
#if (USE_ANGULAR_MOMENTUM == TRUE)
	void TransAngMom2Vel(Data &, gridFunc &, ProblemType &);
	void TransAngMom2Momen(Data &, gridFunc &, ProblemType &);
	void TransVel2AngMom(Data &, gridFunc &, ProblemType &);
	void TransMomen2AngMom(Data &, gridFunc &, ProblemType &);
#endif
#if (USE_COROTATION == CRONOS_ON)
	void TransCorotToInert(Data &, gridFunc &, ProblemType &);
	void TransInertToCorot(Data &, gridFunc &, ProblemType &);
//	REAL TransCorotToInert(Data &, REAL vCorot, int dir, int ix, int iy, int iz);
//	REAL TransInertToCorot(Data &, REAL vInert, int dir, int ix, int iy, int iz);
	REAL TransCorotToInert_x(Data &, REAL vCorot, int ix, int iy, int iz);
	REAL TransCorotToInert_y(Data &, REAL vCorot, int ix, int iy, int iz);
	REAL TransCorotToInert_z(Data &, REAL vCorot, int ix, int iy, int iz);
	REAL TransInertToCorot_x(Data &, REAL vInert, int ix, int iy, int iz);
	REAL TransInertToCorot_y(Data &, REAL vInert, int ix, int iy, int iz);
	REAL TransInertToCorot_z(Data &, REAL vInert, int ix, int iy, int iz);
//	void setOmega(double OmegaZ);
	void src_Corotating(Data &, ProblemType &, NumMatrix<REAL, 3> nom[]);
	void store_uInert(Data &gdata, cronos::vector<REAL> &Pos, phys_fields_1D &fields, int dir);
	void store_uInert(Data &gdata, phys_fields_0D &fields, int ix, int iy, int iz);
#endif
	void TransMomen2Vel(Data &, gridFunc &, ProblemType &);
	void TransVel2Momen(Data &, gridFunc &, ProblemType &);
	void EntropyCorrection(Data &, gridFunc &, ProblemType &, int);
	void computeEntropyFromE(Data &, gridFunc &, ProblemType &);

	inline REAL TransEth2E(REAL, REAL, REAL, REAL);
//	inline REAL TransEth2E(REAL, REAL, REAL, REAL, int fluidType=0);
	inline REAL TransT2E(ProblemType &, REAL, REAL, REAL, REAL);
//	inline REAL TransT2E(ProblemType &, REAL, REAL, REAL, REAL, int fluidType=0);
	void get_Cons(Data &, ProblemType &, EquationOfState &,
	              cronos::vector<REAL> &, phys_fields_1D &, int, REAL);
	void get_Cons(Data &, ProblemType &, EquationOfState &,
			phys_fields_0D &, int ix, int iy, int iz, int face);
//	void get_Cons(Data &, ProblemType &, EquationOfState &,
//	              cronos::vector<REAL> &, phys_fields_1D &, int, REAL, int fluidType=0);
	// void get_Cons(Data &, ProblemType &, EquationOfState &,
	//               cronos::vector<REAL> &, phys_fields_1D &, int, REAL);
	void get_ConsUser(Data &, phys_fields_1D &, int);
	void set_thermal(bool thermal);
//	void reset_Indices(ProblemType &);
	void reset_Indices(CronosFluid &);

#if (FLUID_TYPE == CRONOS_RHD)
	bool isPhysicalState_RHD(Data &, int, int, int);
	bool isPhysicalState_RHD(Data &, int, int, int, double&);
	void TransCons2Prim_RHD(Data &, gridFunc &, ProblemType &);
	void TransPrim2Cons_RHD(Data &, gridFunc &, ProblemType &);
	consState CorrectUnphysicalState_RHD(const consState &, const ProblemType &);
	primState TransStateCons2Prim_Standard_RHD(const consState &, const Data & , int, int, int, const ProblemType &);

	Invertor1Dw invertor1Dw;
	InvertorQuartic invertorQuartic;
	InvertorPolytropic invertorPolytropic;
	InvertorPluto invertorPluto;

#if (CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	void computeEntropyFromE_RHD(Data &, gridFunc &, ProblemType &);
	void SetEntropySwitches_RHD(Data &, gridFunc &, ProblemType &);
	primState TransStateCons2Prim_Entropy_RHD(const consState &, const Data & , int, int, int, const ProblemType &);
	
	InvertorEntropyPolytropic invertorEntropyPolytropic;
	InvertorEntropy invertorEntropy;
	InvertorEntropyPluto invertorEntropyPluto;
#endif
	
	/* Implemented in state.C */
	consState readCons(const Data & gdata, int ix, int iy, int iz);
	void writePrim(const Data & gdata, int ix, int iy, int iz, const primState & prim);
	void writeCons(const Data & gdata, int ix, int iy, int iz, const consState & cons);
#endif
		
private:
	inline REAL TransEth2E_HD(REAL, REAL, REAL, REAL);
	inline REAL TransEth2E_MHD(REAL, REAL, REAL, REAL);
	inline REAL TransT2E_HD(ProblemType &, REAL, REAL, REAL, REAL);
	inline REAL TransT2E_MHD(ProblemType &, REAL, REAL, REAL, REAL);
	void get_Cons_HD(Data &, ProblemType &, EquationOfState &,
			cronos::vector<REAL> &, phys_fields_1D &, int, REAL);
	void get_Cons_MHD(Data &, ProblemType &, EquationOfState &,
			cronos::vector<REAL> &, phys_fields_1D &, int, REAL);
	void get_Cons_RHD(Data &, ProblemType &, EquationOfState &,
			cronos::vector<REAL> &, phys_fields_1D &, int, REAL);
	void get_Cons_HD(Data &, ProblemType &, EquationOfState &,
				phys_fields_0D &, int ix, int iy, int iz, int face);
	void get_Cons_MHD(Data &, ProblemType &, EquationOfState &,
				phys_fields_0D &, int ix, int iy, int iz, int face);
	void get_Cons_RHD(Data &, ProblemType &, EquationOfState &,
				phys_fields_0D &, int ix, int iy, int iz, int face);
	inline double get_EMag(const Data &, int ix, int iy, int iz);
	REAL max(const REAL &, const REAL &, const REAL &, const REAL &,
	         const REAL &, const REAL &, const REAL &);
	REAL min(const REAL &, const REAL &, const REAL &, const REAL &,
	         const REAL &, const REAL &, const REAL &);
	REAL TNorm;
#if (USE_COROTATION == CRONOS_ON)
	double omegaZ;
#endif
	int TempErr, thermal;
	int fluidType, iFluid;
	int q_rho, q_sx, q_sy, q_sz, q_Bx, q_By, q_Bz, q_Eges, q_Eadd;
	int q_rho_loc, q_sx_loc, q_sy_loc, q_sz_loc;
	int q_Bx_loc, q_By_loc, q_Bz_loc, q_Eges_loc, q_Eadd_loc;
	bool TPhys;
	bool magFluid;


#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	REAL entropy_flag_alpha2, entropy_flag_alpha3;
#endif
};


#endif
