#include "gridgen.H"
#include "DiffusionConvection1D.H"

void Environment::init_solvers(Data &gdata)
{
	// Specified user class
	RKSolver = new HyperbolicSolver(gdata, *Problem);
	try{
		RKSolver->init_constants(gdata);
	} catch (CException exep) {
		Abort(gdata, exep);
	}
#if (PARTICLE_TRANSPORT == CRONOS_ON)
	EulerSolver = new DiffusionConvection1D(gdata, *Problem);
#endif
}


void Environment::pdestep(Data &gdata)
{
	gdata.cfl = 0.;

#ifdef TIME_INTEGRATOR
	for (int n = 0; n < TIME_SUBSTEPS; n++) {
		if(gdata.rank==0 && Problem->checkout(0)){
			cout << "  RKSTEP = " << n+1 << endl;
		}
		try {
			gdata.cfl = (std::max(RKSolver->singlestep(gdata, *gfunc, *Problem, n),
			                      gdata.cfl));
		} catch (CException exep) {

			Abort(gdata, exep);
		}

	}

#endif

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	try
	{
		EulerSolver->singlestep(gdata, *gfunc, *Problem);
	}
	catch (CException exep)
	{
		Abort(gdata, exep);
	}
#endif
}
