#ifndef CRONOS_ALFVEN_WINGS_SPHERICAL_H
#define CRONOS_ALFVEN_WINGS_SPHERICAL_H

#include "problem.H"

// using namespace std;

/*
  Alle Benutzerklassen erben einfach von der Oberklasse Problemtype -
  die kann ich noch in specific.C unterbringen.

  Hier können beliebige benötigte Subroutinen eingebunden werden -
  damit stören sie nicht im Hauptcode
 */

class LinearWaveDecay: public ProblemType {
public:
	LinearWaveDecay (Data &);
	virtual void init_fields(Data &, int [], int []);
	virtual void writePhystestInfo(Data &);
private:
	void get_FluctData(Data &);
	void get_FluctDataDirect(Data &);
	void set_shearAlfven(Data &, int ix, int iy, int iz);
	void set_fastMode(Data &, int ix, int iy, int iz);
	void set_slowMode(Data &, int ix, int iy, int iz);
	void set_soundMode(Data &, int ix, int iy, int iz);
	void set_ShearFlow(Data &, int ix, int iy, int iz);
	void set_entropyMode(Data &, int ix, int iy, int iz);
	void set_soundMode_RHD(Data &, int ix, int iy, int iz);
	void WriteToH5(Hdf5Stream &h5out);

	void (LinearWaveDecay::*set_perturbation)(Data &, int, int, int);
	NumArray<REAL> fields;

	NumArray<REAL> background;
	std::string fieldNames[N_OMINT];

	std::vector<double> timeVelo, rmsVelo;
	std::vector<NumArray<double>> rmsData;

	REAL k_x, k_y;
	REAL v_amp;
	REAL B0, rho0, p0;
	REAL vx0, vy0, vz0;
	REAL lorentz0, eos_h;
	REAL cLambdaR;
	REAL len0;
	REAL cAlfven, cFast, cSlow, cSound;
	REAL cA_k;
	int wave_type;


};

LinearWaveDecay::LinearWaveDecay(Data &gdata)  : ProblemType(gdata)
{
#if (GEOM == CARTESIAN)
	name = " Linear wave test ";
#else
	throw CException("Wrong geometry");
#endif

	this->wave_type = value((char*)"Wave_Type"); // 1 = shear Alvfen, 2 = fast, 3 = slow

	if(wave_type==1) {
		set_perturbation = &LinearWaveDecay::set_shearAlfven;
	} else if (wave_type==2) {
		set_perturbation = &LinearWaveDecay::set_fastMode;
	} else if (wave_type==3) {
		set_perturbation = &LinearWaveDecay::set_slowMode;
	} else if (wave_type==4) {
		set_perturbation = &LinearWaveDecay::set_soundMode;
	} else if (wave_type==5) {
		set_perturbation = &LinearWaveDecay::set_ShearFlow;
	} else if (wave_type==6) {
		set_perturbation = &LinearWaveDecay::set_entropyMode;
	} if (wave_type==11) {
		set_perturbation = &LinearWaveDecay::set_soundMode_RHD;
	} else {
		throw CException(" Unknown Wave Type ");
	}

	this->v_amp = 0.05;
	this->v_amp = value((char*)"Initial_Amplitude");

	// set wave number
	this->k_x = 2.* pi/gdata.Lx[0];
	this->k_y = 2.* pi/gdata.Lx[1];

	// Set correct size for array holding disturbance
	fields.resize(N_OMINT);

	if(value_exists("BackgroundDensity")) {
		rho0 = value((char*)"BackgroundDensity");
	} else {
		rho0 = 1;
	}

	if(value_exists("BackgroundPressure")) {
		p0 = value((char*)"BackgroundPressure");
	} else {
		p0 = 1;
	}

	if(value_exists("BackgroundVelocityX")) {
		vx0 = value("BackgroundVelocityX");
	} else {
		vx0 = 0.;
	}

	if(value_exists("BackgroundVelocityY")) {
		vy0 = value("BackgroundVelocityY");
	} else {
		vy0 = 0.;
	}

	if(value_exists("BackgroundVelocityZ")) {
		vz0 = value("BackgroundVelocityZ");
	} else {
		vz0 = 0.;
	}


#if (FLUID_TYPE == CRONOS_HYDRO || FLUID_TYPE == CRONOS_MHD)
	B0 = 1.;
	cSound = sqrt(gamma*p0/rho0);
	cAlfven = sqrt(B0*B0/(rho0));

	if(value_exists("Alfven_Speed")) {
		cAlfven = value((char*)"Alfven_Speed");
		B0 = sqr(cAlfven)*rho0;
	}
	if(value_exists("Sound_Speed")) {
		cSound  = value((char*)"Sound_Speed");
		p0 = sqr(cSound)*rho0/gamma;
	}

	len0 = gdata.Lx[0];

	cA_k = cAlfven/sqrt(2.);
	REAL theta = M_PI/4;
	cA_k = cAlfven*cos(theta);
	cFast = sqrt(cSound*cSound + cAlfven*cAlfven);
	cFast = 0.5*(sqr(cSound) + sqr(cAlfven) + sqrt(sqr(sqr(cSound) + sqr(cAlfven)) - 4*sqr(cSound)*sqr(cAlfven)*sqr(cos(theta))));
	cFast = sqrt(cFast);
	cSlow = 0.5*(sqr(cSound) + sqr(cAlfven) - sqrt(sqr(sqr(cSound) + sqr(cAlfven)) - 4*sqr(cSound)*sqr(cAlfven)*sqr(cos(theta))));
	cSlow = sqrt(cSlow);

	if(gdata.rank==0) {
		cout << " c_sound " << cSound << endl;
		cout << " c_Alfven " << cAlfven << endl;
		cout << " c_Alfven in k direction " << cA_k << endl;
		cout << " fast mode " << cFast << endl;
		cout << " slow mode " << cSlow << endl;
	}
#elif (FLUID_TYPE == CRONOS_RHD)
	lorentz0 = 1./sqrt(1. - sqr(vx0) - sqr(vy0) - sqr(vz0));
	vx0 *= lorentz0;
	vy0 *= lorentz0;
	vz0 *= lorentz0;

	REAL eos_kappa = gamma/(gamma - 1.) / rho0;
	REAL eos_chi = - gamma/(gamma - 1.) * p0 / sqr(rho0);
	eos_h = 1. + gamma/(gamma - 1.) * p0/ rho0;
	cs2 = rho0 * eos_chi / (eos_h * (1. - eos_kappa * rho0));
	cSound = sqrt(cs2);

	REAL delta = sqr(lorentz0) * (1. - cs2) + cs2;
	cLambdaR = (lorentz0 * vx0 * (1. - cs2) + sqrt(cs2 * (1. + (1. - cs2) * (sqr(vy0) + sqr(vz0))))) / delta;

	REAL period = gdata.Lx[0] / cLambdaR;

	if(gdata.rank==0) {
		cout << " c_sound " << cSound << endl;
		cout << " c_lambda_R " << cLambdaR << endl;
		cout << " oscillation period " << period << endl;
	}



#endif

	// Set correct size for array holding background state and set its values
	// This will be use for the RMS computation
	// Todo include also vector potential
	background.resize(N_OMINT);
	background(q_rho)  = rho0;
	background(q_sx)   = vx0;
	background(q_sy)   = vy0;
	background(q_sz)   = vz0;
	background(q_Eges) = p0 / (gamma - 1.);
}





void LinearWaveDecay::init_fields(Data &gdata, int ibeg[3], int iend[3])
{

	for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {
		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
				(this->*set_perturbation)(gdata, ix, iy, iz);

				gdata.om[q_rho](ix,iy,iz) = background[q_rho] + fields(q_rho);
				gdata.om[q_sx](ix,iy,iz)  = background[q_sx] + fields(q_sx);
				gdata.om[q_sy](ix,iy,iz)  = background[q_sy] + fields(q_sy);
				gdata.om[q_sz](ix,iy,iz ) = background[q_sz] + fields(q_sz);
#if (FLUID_TYPE == CRONOS_MHD)
				gdata.om[q_Bx](ix,iy,iz) = fields(q_Bx);
				gdata.om[q_By](ix,iy,iz) = fields(q_By);
				REAL yPos = gdata.getEdgL_y(iy);
				gdata.om[q_Bz](ix,iy,iz) = B0*yPos + fields(q_Bz);
#endif
				gdata.om[q_Eges](ix,iy,iz) = background[q_Eges] + fields(q_Eges);


			}
		}
	}

#if (FLUID_TYPE == CRONOS_MHD)
	if(gdata.rank==0) {
	cout << " A_x ";
	cout << gdata.om[q_Bx](12,-2,0) << " ";
	cout << gdata.om[q_Bx](12,-1,0) << " ";
	cout << gdata.om[q_Bx](12, 0,0) << " ";
	cout << gdata.om[q_Bx](12, 1,0) << " ";
	cout << endl;
	cout << " A_y ";
	cout << gdata.om[q_By](12,-2,0) << " ";
	cout << gdata.om[q_By](12,-1,0) << " ";
	cout << gdata.om[q_By](12, 0,0) << " ";
	cout << gdata.om[q_By](12, 1,0) << " ";
	cout << endl;
	cout << " A_z ";
	cout << gdata.om[q_Bz](12,-3,0) << " ";
	cout << gdata.om[q_Bz](12,-2,0) << " ";
	cout << gdata.om[q_Bz](12,-1,0) << " ";
	cout << gdata.om[q_Bz](12, 0,0) << " ";
	cout << gdata.om[q_Bz](12, 1,0) << " ";
	cout << gdata.om[q_Bz](12, 2,0) << " ";
	cout << endl;
	}
	if(gdata.rank==1) {
		cout << " A_x1 ";
		cout << gdata.om[q_Bx](12,gdata.mx[1]-3,0) << " ";
		cout << gdata.om[q_Bx](12,gdata.mx[1]-2,0) << " ";
		cout << gdata.om[q_Bx](12,gdata.mx[1]-1,0) << " ";
		cout << gdata.om[q_Bx](12,gdata.mx[1]+0,0) << " ";
		cout << gdata.om[q_Bx](12,gdata.mx[1]+1,0) << " ";
		cout << gdata.om[q_Bx](12,gdata.mx[1]+2,0) << " ";
		cout << endl;
		cout << " A_y1 ";
		cout << gdata.om[q_By](12,gdata.mx[1]-3,0) << " ";
		cout << gdata.om[q_By](12,gdata.mx[1]-2,0) << " ";
		cout << gdata.om[q_By](12,gdata.mx[1]-1,0) << " ";
		cout << gdata.om[q_By](12,gdata.mx[1]+0,0) << " ";
		cout << gdata.om[q_By](12,gdata.mx[1]+1,0) << " ";
		cout << gdata.om[q_By](12,gdata.mx[1]+2,0) << " ";
		cout << endl;
		cout << " A_z1 ";
		cout << gdata.om[q_Bz](12,gdata.mx[1]-3,0) << " ";
		cout << gdata.om[q_Bz](12,gdata.mx[1]-2,0) << " ";
		cout << gdata.om[q_Bz](12,gdata.mx[1]-1,0) << " ";
		cout << gdata.om[q_Bz](12,gdata.mx[1]+0,0) << " ";
		cout << gdata.om[q_Bz](12,gdata.mx[1]+1,0) << " ";
		cout << gdata.om[q_Bz](12,gdata.mx[1]+2,0) << " ";
		cout << gdata.om[q_Bz](12,gdata.mx[1]+3,0) << " ";
		cout << endl;
	}

	gdata.om[q_Bx].rename("A_x");
	gdata.om[q_By].rename("A_y");
	gdata.om[q_Bz].rename("A_z");
#endif

	for (int iq = 0; iq < N_OMINT; ++iq) {
		fieldNames[iq] = gdata.om[iq].getName();
	}
}


void LinearWaveDecay::set_shearAlfven(Data &gdata, int ix, int iy, int iz) {
	/*!
	 * Shear Alfven wave test taken from
	 * RyuEtAl1995
	 * the setup is given in Eq. (3.6) of that paper
	 * */
	REAL xPos = gdata.getCen_x(ix);
	REAL yPos = gdata.getCen_y(iy);


	REAL del_vz = v_amp*cA_k*sin(k_x*xPos + k_y*yPos);
	REAL del_rho = 0.;
	REAL del_p   = 0.;
	REAL del_vx  = 0.;
	REAL del_vy  = 0.;
	REAL del_Ax  = 0.;
	REAL del_Ay  = 0.;
	REAL del_Az  = 0.;

	fields(q_rho) = del_rho;
	fields(q_sx) = del_vx;
	fields(q_sy) = del_vy;
	fields(q_sz) = del_vz;
	fields(q_Bx) = del_Ax;
	fields(q_By) = del_Ay;
	fields(q_Bz) = del_Az;
	fields(q_Eges) = del_p;

}


void LinearWaveDecay::set_fastMode(Data &gdata, int ix, int iy, int iz) {
	/*!
	 * fast-mode wave test taken from
	 * RyuEtAl1995
	 * the setup is given in Eq. (3.10) of that paper
	 * */
	REAL xPos = gdata.getCen_x(ix);
	REAL yPos = gdata.getCen_y(iy);


	REAL cA = 1.;
	REAL del_vx = v_amp*cFast*sin(k_x*xPos + k_y*yPos);
	REAL del_vy = (sqr(cFast)/(sqr(cFast) - sqr(cAlfven)))*v_amp*cFast*sin(k_x*xPos + k_y*yPos);
	REAL del_rho = 0.;
	REAL del_p   = 0.;
	REAL del_vz  = 0.;
	REAL del_Ax  = 0.;
	REAL del_Ay  = 0.;
	REAL del_Az  = 0.;

	fields(q_rho) = del_rho;
	fields(q_sx) = del_vx;
	fields(q_sy) = del_vy;
	fields(q_sz) = del_vz;
	fields(q_Bx) = del_Ax;
	fields(q_By) = del_Ay;
	fields(q_Bz) = del_Az;
	fields(q_Eges) = del_p;

}
void LinearWaveDecay::set_slowMode(Data &gdata, int ix, int iy, int iz) {
	/*!
	 * slow-mode wave test taken from
	 * RyuEtAl1995
	 * the setup is given in Eq. (3.11) of that paper
	 * */
	REAL xPos = gdata.getCen_x(ix);
	REAL yPos = gdata.getCen_y(iy);


	REAL cA = 1.;
	REAL del_vx = v_amp*cSlow*sin(k_x*xPos + k_y*yPos);
	REAL del_vy = (sqr(cSlow)/(sqr(cSlow) - sqr(cAlfven)))*v_amp*cSlow*sin(k_x*xPos + k_y*yPos);
	REAL del_rho = 0.;
	REAL del_p   = 0.;
	REAL del_vz  = 0.;
	REAL del_Ax  = 0.;
	REAL del_Ay  = 0.;
	REAL del_Az  = 0.;

	fields(q_rho) = del_rho;
	fields(q_sx) = del_vx;
	fields(q_sy) = del_vy;
	fields(q_sz) = del_vz;
	fields(q_Bx) = del_Ax;
	fields(q_By) = del_Ay;
	fields(q_Bz) = del_Az;
	fields(q_Eges) = del_p;
}


void LinearWaveDecay::set_soundMode(Data &gdata, int ix, int iy, int iz) {
	/*! Test setup according to:
	 * RyuGoodman1994ApJ422_269
	 * this setup represents a decaying sound wave given in Eq (51) in that paper
	 * */
	REAL xPos = gdata.getCen_x(ix);
	REAL yPos = gdata.getCen_y(iy);

	REAL del_vx = v_amp*cSound*sin(k_x*xPos + k_y*yPos);
	REAL del_vy = del_vx;
	REAL del_rho = 0.;
	REAL del_p   = 0.;
	REAL del_vz  = 0.;
	REAL del_Ax  = 0.;
	REAL del_Ay  = 0.;
	REAL del_Az  = 0.;

	fields(q_rho) = del_rho;
	fields(q_sx) = del_vx;
	fields(q_sy) = del_vy;
	fields(q_sz) = del_vz;
	fields(q_Bx) = del_Ax;
	fields(q_By) = del_Ay;
	fields(q_Bz) = del_Az;
	fields(q_Eges) = del_p;
}

void LinearWaveDecay::set_soundMode_RHD(Data &gdata, int ix, int iy, int iz) {
	/*!
	 * */
	REAL xPos = gdata.getCen_x(ix);
	REAL yPos = gdata.getCen_y(iy);
	REAL amp = v_amp *sin(k_x*xPos);

	REAL del_rho = rho0;
	REAL del_ux  = - cs2 * (vx0 + 1./(vx0 - cLambdaR * lorentz0));
	REAL del_uy  = - cs2 * vy0;
	REAL del_uz  = - cs2 * vz0;
	REAL del_p   = cs2 * rho0 * eos_h;

	fields(q_rho)  = amp * del_rho;
	fields(q_sx)   = amp * del_ux;
	fields(q_sy)   = amp * del_uy;
	fields(q_sz)   = amp * del_uz;
	fields(q_Eges) = amp * del_p / (gamma - 1.);
}


void LinearWaveDecay::set_entropyMode(Data &gdata, int ix, int iy, int iz) {
	/*!
	 * */
	REAL xPos = gdata.getCen_x(ix);
	REAL yPos = gdata.getCen_y(iy);

	REAL del_vx = v_amp*cSound*sin(k_x*xPos + k_y*yPos);
	REAL del_vy = del_vx;
	REAL del_rho = 0.;
	REAL del_p   = 0.;
	REAL del_vz  = 0.;
	REAL del_Ax  = 0.;
	REAL del_Ay  = 0.;
	REAL del_Az  = 0.;

	fields(q_rho) = del_rho;
	fields(q_sx) = del_vx;
	fields(q_sy) = del_vy;
	fields(q_sz) = del_vz;
	fields(q_Bx) = del_Ax;
	fields(q_By) = del_Ay;
	fields(q_Bz) = del_Az;
	fields(q_Eges) = del_p;
}



void LinearWaveDecay::set_ShearFlow(Data &gdata, int ix, int iy, int iz) {
	/*! Test setup according to:
	 * RyuGoodman1994ApJ422_269
	 * this setup represents a planar shear flow, where the gradual decay can be investigated
	 * the specific setup is given in Eq. (54) in that paper
	 * */


	REAL xPos = gdata.getCen_x(ix);
	REAL yPos = gdata.getCen_y(iy);

	REAL del_vx = v_amp*cSound*sin(k_x*(xPos - yPos));
	REAL del_vy = del_vx;
	REAL del_rho = 0.;
	REAL del_p   = 0.;
	REAL del_vz  = 0.;
	REAL del_Ax  = 0.;
	REAL del_Ay  = 0.;
	REAL del_Az  = 0.;

	fields(q_rho) = del_rho;
	fields(q_sx) = del_vx;
	fields(q_sy) = del_vy;
	fields(q_sz) = del_vz;
	fields(q_Bx) = del_Ax;
	fields(q_By) = del_Ay;
	fields(q_Bz) = del_Az;
	fields(q_Eges) = del_p;
}



void LinearWaveDecay::get_FluctData(Data &gdata) {
	int num = 0;
	NumArray<double> rms2(N_OMINT);
	rms2.clear();

	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {
				num += 1;

				for (int iq = 0; iq < N_OMINT; ++iq) {
					rms2(iq) += sqr(gdata.om[iq](ix, iy, iz) - background(iq));
				}
			}
		}
	}

#ifdef parallel
	MPI_Allreduce(MPI_IN_PLACE, &num, 1, MPI_INT, MPI_SUM, gdata.comm3d);
	MPI_Allreduce(MPI_IN_PLACE, rms2, rms2.getLength(), MPI_DOUBLE, MPI_SUM, gdata.comm3d);
#endif

	rms2 /= (double)(num);

	NumArray<double> rms(rms2);
	for (int iq = 0; iq < N_OMINT; ++iq) {
		rms(iq) = sqrt(rms2(iq));
	}

	rmsData.push_back(rms);

	double rms_velo = 0.;
	if(wave_type==1) {
		rms_velo = rms2(q_sz);
	} else if (wave_type==2) {
		rms_velo = sqr(cFast)*rms2(q_sx) + sqr(sqr(cFast)/(sqr(cFast) - sqr(cAlfven)))*rms2(q_sy);
	} else if (wave_type==3) {
		rms_velo = sqr(cSlow)*rms2(q_sx) + sqr(sqr(cSlow)/(sqr(cSlow) - sqr(cAlfven)))*rms2(q_sy);
	} else if (wave_type==4) {
		rms_velo = rms2(q_sx);
	} else if (wave_type==5) {
		rms_velo = rms2(q_sx);
	}

	rmsVelo.push_back(sqrt(rms_velo));
	timeVelo.push_back(gdata.time);
}


void LinearWaveDecay::get_FluctDataDirect(Data &gdata) {
	double amp_vx(0.), amp_vy(0.), amp_vz(0.);
	double amplitude(0.);
	double num(0.);

	for( int ix=0; ix<=gdata.mx[0]; ++ix) {
		for( int iy=0; iy<=gdata.mx[1]; ++iy) {
			if(wave_type==1) {
				set_shearAlfven(gdata, ix,iy,0);
				double val_init = fields(q_sz);
				if(std::abs(val_init)>1.e-12) {
					amp_vz += sqr(gdata.om[q_sz](ix,iy,0)/val_init);
					num += 1.;
				}
			} else {
				if (wave_type==2) {
					set_fastMode(gdata, ix, iy, 0);
				} else if (wave_type==3) {
					set_slowMode(gdata, ix, iy, 0);
				} else if (wave_type==4){
					set_soundMode(gdata, ix, iy, 0);
				} else {
					set_ShearFlow(gdata, ix, iy, 0);
				}
				double val_init = fields(q_sy);
				if(std::abs(val_init)>1.e-12) {
					amp_vy += sqr(gdata.om[q_sy](ix,iy,0)/val_init);
					num += 1.;
				}
			}
//			if(iy==12) {
//				cout << " vals " << amp_vy << " " << fields(q_sy) << " " << gdata.om[q_sy](ix,iy,0) << " " << num << endl;
//			}
		}
	}

#ifdef parallel
	double numGlobal, amp_vxGlobal, amp_vyGlobal, amp_vzGlobal;
	MPI_Allreduce(&num, &numGlobal, 1, MPI_DOUBLE, MPI_SUM, gdata.comm3d);
	MPI_Allreduce(&amp_vx, &amp_vxGlobal, 1, MPI_DOUBLE, MPI_SUM, gdata.comm3d);
	MPI_Allreduce(&amp_vy, &amp_vyGlobal, 1, MPI_DOUBLE, MPI_SUM, gdata.comm3d);
	MPI_Allreduce(&amp_vz, &amp_vzGlobal, 1, MPI_DOUBLE, MPI_SUM, gdata.comm3d);
	MPI_Barrier(gdata.comm3d);
	num = numGlobal;
	amp_vx = amp_vxGlobal;
	amp_vy = amp_vyGlobal;
	amp_vz = amp_vzGlobal;
#endif
	amp_vx /= num;
	amp_vy /= num;
	amp_vz /= num;

	if(wave_type==1) {
		amplitude = amp_vz;
	} else {
		amplitude = amp_vy;
	}

//	cout << " amp: " << amplitude << endl;
//	exit(3);

//	return amplitude;

}

void LinearWaveDecay::writePhystestInfo(Data &gdata) {
	// Here we use this to compute the fluctuation statistics at each relevant step
	get_FluctData(gdata);
//	double rms_velo = get_FluctDataDirect(gdata);
}


void LinearWaveDecay::WriteToH5(Hdf5Stream &h5out) {

	// Indicate that we are using a test case
	h5out.AddGlobalAttr("TestCase", "Linear wave decay");

	// supply normalised values necessary for Parker solution
	// Use own group for that purpose
	hid_t group = h5out.AddGroup("Data/WavePars");

	// supply wave type
	h5out.AddGlobalAttr("Wave_Type", wave_type, group);

	// suppy initial amplitude
	h5out.AddGlobalAttr("Initial_Amplitude", v_amp, group);

	// supply normalised speed of sound
	h5out.AddGlobalAttr("cs", cSound, group);

	// supply normalised Alfven speed
	h5out.AddGlobalAttr("cA", cAlfven, group);

	// supply Alfven wave speed
	h5out.AddGlobalAttr("cWave_Alfven", cA_k, group);

	// supply fast mode speed
	h5out.AddGlobalAttr("cWave_Fast", cFast, group);

	// supply slow mode speed
	h5out.AddGlobalAttr("cWave_Slow", cSlow, group);

	// supply srhd wave speed
	h5out.AddGlobalAttr("cWave_LambdaR", cLambdaR, group);

	// Supply additional extent of domain
	h5out.AddGlobalAttr("Len", len0, group);

	// Background density
	h5out.AddGlobalAttr("BackgroundDensity", rho0, group);


	// Add full array of times and wave amplitudes
	NumArray<float> outData;
	outData.resize(rmsVelo.size());
	for(int i_step=0; i_step<rmsVelo.size(); ++i_step) {
	  outData(i_step) = timeVelo[i_step];
	}
	h5out.WriteNumArray("times",outData,group);
	for(int i_step=0; i_step<rmsVelo.size(); ++i_step) {
	  outData(i_step) = rmsVelo[i_step];
	}
	h5out.WriteNumArray("rmsVelo",outData,group);

	// Now close the new group
	h5out.CloseGroup(group);

	group = h5out.AddGroup("Data/WavePars/rms");
	for (int iq = 0; iq < N_OMINT; ++iq) {
		for(int i_step=0; i_step<rmsData.size(); ++i_step) {
			outData(i_step) = rmsData[i_step](iq);
		}
		h5out.WriteNumArray(fieldNames[iq], outData, group);
	}
	h5out.CloseGroup(group);
}


#endif
