#ifndef CRONOS_MOD_MPITEST_H
#define CRONOS_MOD_MPITEST_H = 1

#include "problem.H"
#include "data.H"
#include "gridgen.H"
#include "boundary.H"


class MpiTest: public ProblemType {
public:
	MpiTest(Data &);
	virtual void init_fields(Data &, int [], int []);
};


MpiTest::MpiTest(Data &gdata) : ProblemType(gdata) {
}

void MpiTest::init_fields(Data &gdata, int ibeg[3], int iend[3])
{
	for (int iq = 0; iq < N_OM; ++iq) {
		for (int k = ibeg[2]; k <= iend[2]; k++){
			for (int j = ibeg[1]; j <= iend[1]; j++){
				for (int i = ibeg[0]; i <= iend[0]; i++){
						gdata.om[iq](i, j, k) = gdata.rank;
				}
			}
		}
	}
	

	Boundary boundary(gdata, N_OM);
	boundary.boundary(gdata, *this, gdata.om, N_OM);

	for (int iq = 0; iq < N_OM; ++iq) {
		for (int k = ibeg[2]; k <= iend[2]; k++){
			for (int j = ibeg[1]; j <= iend[1]; j++){
				for (int i = ibeg[0]; i <= iend[0]; i++){
						gdata.om[iq](i, j, k) = gdata.rank;
				}
			}
		}
	}

	
	// Restrict to magnetic field test:
	boundary.boundary(gdata, *this, &gdata.om[4], 3);
	

	if (gdata.rank == 0) {
		for (int iq = 0; iq < N_OM; ++iq) {
			cout << " Field " << gdata.om[iq].getName() << endl;

			for (int iy = iend[1]; iy >= ibeg[1]; iy--) {
				cout << " iy = " << iy << "  \t";
				for (int ix = ibeg[0]; ix <= iend[0]; ix++) {
					cout << gdata.om[iq](ix, iy, 0) << " ";
				}
				cout << endl;
			}
			cout << endl;
		}
	}

	MPI_Finalize();
	exit(0);
}





#endif
