import sys

from CronosPyLib.ProjHandler import ProjHandler
from CronosPyLib.EmissionHandler import EmissionSky
import os
import numpy as np

from astropy.constants import c
import astropy.units as u
import naima
from naima.models import Synchrotron, PowerLaw, InverseCompton

from matplotlib.colors import LogNorm
from matplotlib import cm

import matplotlib.pyplot as plt

from matplotlib_utils import multipage
import copy


def absorp_loglog(proj, ax):
	sky = proj.emissionData[0].skys[0]
	nEnergies = len(sky.energies)
	nChannels = len(sky.channels)

	ax.loglog(sky.energies,
		sky.getFlux("DummyEmission_Absorption")[:, 0, 0]/
		sky.getFlux("DummyEmission")[:, 0, 0],
		label=f'{proj.name}')

def absorp_semilogx(proj, ax):
	sky = proj.emissionData[0].skys[0]
	nEnergies = len(sky.energies)
	nChannels = len(sky.channels)

	ax.semilogx(sky.energies,
		sky.getFlux("DummyEmission_Absorption")[:, 0, 0]/
		sky.getFlux("DummyEmission")[:, 0, 0],
		label=f'{proj.name}')
	

if __name__ == '__main__':
	rootDir = os.getcwd()


	fig, axs = plt.subplots(2,2, sharex=True)

	projs = [
		ProjHandler(os.path.join(rootDir, 'radial/MonochromaticPointSource')),
		ProjHandler(os.path.join(rootDir, 'radial/MonochromaticExtended')),
		ProjHandler(os.path.join(rootDir, 'radial/BlackbodyPointSource')),
		ProjHandler(os.path.join(rootDir, 'radial/BlackbodyExtended'))
	]

	for proj in projs:
		absorp_semilogx(proj, axs[0,0])

	for proj in projs:
		absorp_loglog(proj, axs[1,0])


	projs = [
		ProjHandler(os.path.join(rootDir, 'perp/MonochromaticPointSource')),
		ProjHandler(os.path.join(rootDir, 'perp/MonochromaticExtended')),
		ProjHandler(os.path.join(rootDir, 'perp/BlackbodyPointSource')),
		ProjHandler(os.path.join(rootDir, 'perp/BlackbodyExtended'))
	]

	for proj in projs:
		absorp_semilogx(proj, axs[0,1])

	for proj in projs:
		absorp_loglog(proj, axs[1,1])


	# cosmetics
	for ax in axs.flatten():
		ax.grid(True)
		ax.set_xlabel("Energy / eV")
		#ax.set_ylabel("Flux / eV^{-1} m^{-2} s^{-1}")
		ax.legend(loc='best', prop={'size': 6})

	multipage(f"absorption_test.pdf")
	plt.show()





