import sys

from CronosPyLib.ProjHandler import ProjHandler
from CronosPyLib.EmissionHandler import EmissionSky
import os
import numpy as np

from astropy.constants import c
import astropy.units as u
import naima
from naima.models import Synchrotron, PowerLaw, InverseCompton

from matplotlib.colors import LogNorm
from matplotlib import cm

import matplotlib.pyplot as plt

from matplotlib_utils import multipage
import copy



if __name__ == '__main__':
	rootDir = os.getcwd()

	
	my_cmap = copy.copy(cm.get_cmap('viridis')) # copy the default cmap
	my_cmap.set_bad(my_cmap.colors[0])


	def work():
		proj = ProjHandler(os.path.join(rootDir, 'BlackbodyExtended'))
		sky = proj.emissionData[0].skys[0]
		nEnergies = len(sky.energies)
		nChannels = len(sky.channels)

		fig, axs = plt.subplots(int(nEnergies/2), 2)

		print(axs, axs.flatten())

		for iEnergy in range(nEnergies):
			ax = axs.flatten(order="F")[iEnergy]
			print(ax, iEnergy)

			absorption = np.transpose(sky.getFlux("DummyEmission_Absorption")[iEnergy]/sky.getFlux("DummyEmission")[iEnergy])

			im = ax.imshow(
				absorption,
				#norm=LogNorm(vmin=1e-10, vmax=1e0),
				#norm=LogNorm(),
				origin='lower',
				extent=sky.getExtent(),
				cmap=my_cmap
			)
			#ax.colorbar(im)
			#plt.colorbar(im, ax=ax)
			ax.text(0.95, 0.95, f'{sky.energies[iEnergy]:.0E}',
					verticalalignment='top', horizontalalignment='right',
					transform=ax.transAxes, fontsize=8, color='red')


	
	work()
	multipage("absorption_point.pdf")
	plt.show()





