#!/usr/bin/env python3

import sys

from CronosPyLib.ProjHandler import ProjHandler
import os
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib import cm

from cxlib.matplotlib_utils import savefigMultipage


def absorp_loglog(proj, ax):
    sky = proj.emissionData[0].skys[0]
    nEnergies = len(sky.energies)
    nChannels = len(sky.channels)

    ax.loglog(sky.energies,
              sky.getFlux("DummyEmission_Absorption")[:, 0, 0] /
              sky.getFlux("DummyEmission")[:, 0, 0],
              label=f'{proj.name}')


def absorp_semilogx(proj, ax):
    sky = proj.emissionData[0].skys[0]
    nEnergies = len(sky.energies)
    nChannels = len(sky.channels)

    ax.semilogx(sky.energies,
                sky.getFlux("DummyEmission_Absorption")[:, 0, 0] /
                sky.getFlux("DummyEmission")[:, 0, 0],
                label=f'{proj.name}')


if __name__ == '__main__':
    rootDir = os.getcwd()

    projNames = ['MonochromaticPointSource', 'MonochromaticExtendedSource',
                 'BlackbodyPointSource', 'BlackbodyExtendedSource']

    projs = []
    for name in projNames:
        try:
            projs.append(ProjHandler(os.path.join(rootDir, name)))
        except:
            pass

    doneGeneratingFigure = False

    for proj in projs:
        if not proj.hasEmissionData:
            continue

        emission = proj.emissionData[0].emissions[0]
        energies = emission.energies
        phis = emission.directions.phi
        thetas = emission.directions.theta
        spectrum = emission.reductions[0]

        def getSpectrum(channelName, iTheta, iPhi): return spectrum[channelName][
            emission.directions.thetaphi2dir[iTheta, iPhi]]

        if not doneGeneratingFigure:
            doneGeneratingFigure = True
            global fig, axs
            nrow = len(thetas)
            ncol = len(phis)
            xScale = 3
            yScale = 2

            fig, axs = plt.subplots(
                len(thetas), len(phis),
                figsize=(xScale*(ncol+1), yScale*(nrow+1)),
                sharey=True, sharex=True,
                gridspec_kw={'hspace': 0, 'wspace': 0}
            )

        for iTheta, theta in enumerate(thetas):
            for iPhi, phi in enumerate(phis):
                if len(thetas) * len(phis) == 1:
                    ax = axs
                elif len(thetas) == 1:
                    ax = axs[iPhi]
                elif len(phis) == 1:
                    ax = axs[iTheta]
                else:
                    ax = axs[iTheta, iPhi]

                flux = getSpectrum("InverseCompton", iTheta, iPhi)
                sed = flux * np.array(energies) ** 2

                ax.loglog(energies, sed)

                ax.text(0.95, 0.95, f'{theta / np.pi:.2f}, {phi / np.pi:.2f}',
                        verticalalignment='top', horizontalalignment='right',
                        transform=ax.transAxes, fontsize=8, color='red')
                # ax.legend(loc='best')
                ax.grid(True)
                ax.set_ylim((5e18,5e21))

    print("saving")
    filename = f"result.pdf"
    #filename = f"{proj.name}.{timestep}.particleSpectrum.pdf"
    # filename = f"{proj.name}.{timestep}.emission_single2.1.pdf"
    savefigMultipage(filename)
    os.system(f'pdfcrop {filename} {filename}')
