import sys

from CronosPyLib.ProjHandler import ProjHandler
from CronosPyLib.EmissionHandler import EmissionSky
import os
import numpy as np

from astropy.constants import c
import astropy.units as u
import naima
from naima.models import Synchrotron, PowerLaw

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib import cm

from matplotlib_utils import multipage
import copy


if __name__ == '__main__':
	rootDir = os.getcwd()
	
	my_cmap = copy.copy(cm.get_cmap('viridis')) # copy the default cmap
	my_cmap.set_bad(my_cmap.colors[0])


	def work():
		proj = ProjHandler(os.path.join(rootDir, 'RHD_Emission_Test_static'))
		#proj = ProjHandler(os.path.join(rootDir, 'RHD_Emission_Test_boosted_100'))


		sky = proj.emissionData[0].skys[0]
		nEnergies = len(sky.energies)
		nChannels = len(sky.channels)

		fig, axs = plt.subplots(nEnergies, nChannels)

		min = 1e-30
		max = 1e-20


		for iChan in range(nChannels):
			try:
				axx = axs[:,iChan]
			except:
				axx = axs
			axx[0].set_title(sky.channels[iChan])

			for iEnergy in range(nEnergies):
				ax = axx[iEnergy]

				mmin = np.min(sky.getFlux(iChan)[iEnergy])
				mmax = np.max(sky.getFlux(iChan)[iEnergy])
				print(mmin, mmax)
				if mmax == 0:
					fig.delaxes(ax)
					plt.draw()
					continue

				im = ax.imshow(
					np.transpose(sky.getFlux(iChan)[iEnergy] * sky.energies[iEnergy]**2),
					#norm=LogNorm(vmin=min, vmax=max),
					norm=LogNorm(),
					origin='lower',
					extent=sky.getExtent(),
					cmap=my_cmap
				)
				#ax.colorbar(im)
				plt.colorbar(im, ax=ax)
				ax.text(0.95, 0.95, f'{sky.energies[iEnergy]:.0E}',
						verticalalignment='top', horizontalalignment='right',
						transform=ax.transAxes, fontsize=8, color='red')

	
	work()
	plt.show()





