#ifndef CRONOS_RHD_PT_TEST_H
#define CRONOS_RHD_PT_TEST_H

#include "problem.H"
#include "randgen.H"
#if (PARTICLE_TRANSPORT != CRONOS_ON)
#error
#endif

#include "ParticleTransport.H"
#include "Adiabatic.H"
#include "Synchrotron.H"
#include "MagneticField_special.H"

// using namespace std;

class RHD_ParticleTransportTest: public ProblemType {
public:
	RHD_ParticleTransportTest (Data &);
	virtual void init_fields(Data &, int [], int []);
	void init_particleTransport(Data &);

private:
	Quantity magneticFieldStrength;
	REAL spectralIndex;
};

RHD_ParticleTransportTest::RHD_ParticleTransportTest(Data &gdata)  : ProblemType(gdata)
{
	spectralIndex = value("spectralIndex");
	magneticFieldStrength = value("magneticFieldStrength") * Gauss;
}


void RHD_ParticleTransportTest::init_fields(Data &gdata, int ibeg[3], int iend[3])
{
	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {
				gdata.om[q_rho](ix, iy, iz) = 1.;
				gdata.om[q_sx](ix, iy, iz) = 0;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = 0.001/(gamma - 1);
			}
		}
	}

}




using namespace CRONOS_CONSTANTS;
class RHD_NORM : public normalisation {
public:
	virtual void setup();
};

void RHD_NORM::setup() {

	set_definition(LEN);
	set_definition(MASS);
	set_definition(VEL, SpeedOfLight);
	set_definition(TEMP);
	set_definition(MAG_IND);
	set_definition(NUM_DENS);

	finish_setup();
}



void RHD_ParticleTransportTest::init_particleTransport(Data & gdata)
{
	// Defining normalisations
	normalisation * normHD =  new RHD_NORM();
	normHD->setup();

	particleTransport.set_normHD(normHD);
	normalisation * normParticle = & particleTransport.get_normParticle();


	// Adding Particles
	Particle * particle = new Particle(gdata, normParticle, 0, 0);
	particleTransport.add_particle(particle);

	for(int iz=-3; iz<=gdata.mx[2]+3; ++iz) {
		for(int iy=-3; iy<=gdata.mx[1]+3; ++iy) {
			for(int ix=-3; ix<=gdata.mx[0]+3; ++ix) {
					particle->set_powerlaw(1, spectralIndex, ix, iy, iz);
			}
		}
	}


	// Adding Loss Processes
	particleTransport.add_lossProcess(new Adiabatic(normHD, normParticle));

	REAL magneticFieldAxis[3] = {0,0,1};

	auto magneticField = std::make_shared<ConstantMagneticField>(normHD, magneticFieldStrength, magneticFieldAxis);
	particleTransport.add_lossProcess(new SynchrotronProcess(normHD, normParticle, magneticField));
}
#endif
