#ifndef CRONOS_RHD_RIEMANN_PROBLEM_2D_H
#define CRONOS_RHD_RIEMANN_PROBLEM_2D_H

#include "problem.H"
#include "randgen.H"

// using namespace std;

/*
  Alle Benutzerklassen erben einfach von der Oberklasse Problemtype -
  die kann ich noch in specific.C unterbringen.

  Hier können beliebige benötigte Subroutinen eingebunden werden -
  damit stören sie nicht im Hauptcode
 */

class RHD_RiemannProblem_2D: public ProblemType {
public:
	RHD_RiemannProblem_2D (Data &);
	virtual void init_fields(Data &, int [], int []);
	//void WriteToH5(Hdf5Stream &);

private:
	REAL rhoUL, rhoUR, rhoLL, rhoLR;
	REAL uxUL, uxUR, uxLL, uxLR;
	REAL uyUL, uyUR, uyLL, uyLR;
	REAL pUL, pUR, pLL, pLR;
	REAL adiabatic_exponent;
};

RHD_RiemannProblem_2D::RHD_RiemannProblem_2D(Data &gdata)  : ProblemType(gdata)
{
	rhoUL = value((char*)"rhoUL");
	rhoUR = value((char*)"rhoUR");
	rhoLL = value((char*)"rhoLL");
	rhoLR = value((char*)"rhoLR");

	uxUL = value((char*)"uxUL");
	uxUR = value((char*)"uxUR");
	uxLL = value((char*)"uxLL");
	uxLR = value((char*)"uxLR");

	uyUL = value((char*)"uyUL");
	uyUR = value((char*)"uyUR");
	uyLL = value((char*)"uyLL");
	uyLR = value((char*)"uyLR");
	
	pUL = value((char*)"pUL");
	pUR = value((char*)"pUR");
	pLL = value((char*)"pLL");
	pLR = value((char*)"pLR");
	
	
	adiabatic_exponent = value((char*)"Adiabatic_exponent");
}




void RHD_RiemannProblem_2D::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	


	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			REAL rho(0.);
			REAL ux(0.);
			REAL uy(0.);
			REAL p(0.);

			if (gdata.getCen_x(ix) > 0) {
				if (gdata.getCen_y(iy) > 0) {
					rho = rhoUR;
					ux = uxUR;
					uy = uyUR;
					p = pUR;
				} else {
					rho = rhoLR;
					ux = uxLR;
					uy = uyLR;
					p = pLR;
				}
			} else {
				if (gdata.getCen_y(iy) > 0) {
					rho = rhoUL;
					ux = uxUL;
					uy = uyUL;
					p = pUL;
				} else {
					rho = rhoLL;
					ux = uxLL;
					uy = uyLL;
					p = pLL;
				}
			}

			REAL W = 1. / sqrt(1. - sqr(ux) - sqr(uy));

			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {
				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux * W;
				gdata.om[q_sy](ix, iy, iz) = uy * W;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(adiabatic_exponent - 1);
			}
		}
	}

}

/*
void RHD_RiemannProblem_2D::WriteToH5(Hdf5Stream &h5out) {
	// Indicate that we are using a test case
	h5out.AddGlobalAttr("TestCase", "rhd 1d riemann problem");

	// supply normalised values necessary for the analytical solution
	// Use own group for that purpose
	hid_t group = h5out.AddGroup("Data/RHDPars");

	h5out.AddGlobalAttr("rhoL", rhoL, group);
	h5out.AddGlobalAttr("uL", uL, group);
	h5out.AddGlobalAttr("pL", pL, group);
	h5out.AddGlobalAttr("rhoR", rhoR, group);
	h5out.AddGlobalAttr("uR", uR, group);
	h5out.AddGlobalAttr("pR", pR, group);
	h5out.AddGlobalAttr("gamma", adiabatic_exponent, group);

	// Now close the new group
	h5out.CloseGroup(group);
}
*/

#endif
