#ifndef CRONOS_SEDOV_3D_H
#define CRONOS_SEDOV_3D_H = 1

#include "problem.H"
#include "randgen.H"
#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include "ParticleTransport.H"
#include "Adiabatic.H"
#endif

// using namespace std;



class RHD_SphericalBlast: public ProblemType {
public:
	RHD_SphericalBlast(Data &);
	void init_fields(Data &, int [], int []);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	virtual void init_particleTransport(Data &);
#endif


private:
	REAL injectionPressure, injectionRadius, injectionDensity;
	REAL backgroundPressure, backgroundDensity;

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	REAL spectralIndex;
#endif
};


RHD_SphericalBlast::RHD_SphericalBlast(Data &gdata) : ProblemType(gdata)
{
#if (GEOM != 1)
	cerr << " Error: Supposed to be a Cartesian problem ";
	cerr << endl;
	exit(2);
#endif

	injectionPressure = value("injectionPressure");
	injectionRadius = value("injectionRadius");
	injectionDensity = value("injectionDensity");
	backgroundPressure = value("backgroundPressure");
	backgroundDensity = value("backgroundDensity");

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	spectralIndex = value("spectralIndex");
#endif
}



void RHD_SphericalBlast::init_fields(Data &gdata, int ibeg[3], int iend[3])
{

	REAL xx, yy, zz, rad;

	for(int k=ibeg[2]; k<=iend[2]; ++k) {
		zz = gdata.getCen_z(k);
		for(int j=ibeg[1]; j<=iend[1]; ++j) {
			yy = gdata.getCen_y(j);
			for(int i=ibeg[0]; i<=iend[0]; ++i) {
				xx = gdata.getCen_x(i);
				rad = sqrt(sqr(xx) + sqr(yy) + sqr(zz));

				gdata.om[q_rho](i,j,k) = backgroundDensity;
				gdata.om[q_sx](i,j,k) = 0.;
				gdata.om[q_sy](i,j,k) = 0.;
				gdata.om[q_sz](i,j,k) = 0.;

				if(rad < injectionRadius) {
					gdata.om[q_Eges](i,j,k) = injectionPressure;
				} else {

					gdata.om[q_Eges](i,j,k) = backgroundPressure;
				}
				
				gdata.om[q_Eges](i,j,k) /= (gamma-1.);
			}
		}
	}


}


#if (PARTICLE_TRANSPORT == CRONOS_ON)
using namespace CRONOS_CONSTANTS;
class RHD_NORM : public normalisation {
public:
	virtual void setup();
};

void RHD_NORM::setup() {

	set_definition(LEN);
	set_definition(MASS);
	set_definition(VEL, SpeedOfLight);
	set_definition(TEMP);
	set_definition(MAG_IND);
	set_definition(NUM_DENS);

	finish_setup();
}



void RHD_SphericalBlast::init_particleTransport(Data & gdata)
{
	// Defining normalisations
	normalisation * normHD =  new RHD_NORM();
	normHD->setup();

	particleTransport.set_normHD(normHD);
	normalisation * normParticle = & particleTransport.get_normParticle();


	// Adding Particles
	Particle * particle = new Particle(gdata, normParticle, 0, 0);
	particleTransport.add_particle(particle);


	for(int iz=-3; iz<=gdata.mx[2]+3; ++iz) {
		for(int iy=-3; iy<=gdata.mx[1]+3; ++iy) {
			for(int ix=-3; ix<=gdata.mx[0]+3; ++ix) {
				particle->set_powerlaw(1, spectralIndex, ix, iy, iz);
			}
		}
	}


	// Adding Loss Processes
	particleTransport.add_lossProcess(new Adiabatic(normHD, normParticle));
}
#endif

#endif
