#ifndef CRONOS_RELATIVISTICJET_H
#define CRONOS_RELATIVISTICJET_H

#include "problem.H"
#include "randgen.H"

// using namespace std;

/*
  Alle Benutzerklassen erben einfach von der Oberklasse Problemtype -
  die kann ich noch in specific.C unterbringen.

  Hier können beliebige benötigte Subroutinen eingebunden werden -
  damit stören sie nicht im Hauptcode
 */

class RelativisticJet: public ProblemType {
public:
	RelativisticJet (Data &);
	virtual void init_fields(Data &, int [], int []);
	virtual void bc_User(Data &, NumMatrix<REAL,3> &, int, int, int, int);
	//void WriteToH5(Hdf5Stream &);

private:
	REAL rho1, vr1, vz1, ur1, uz1, p1, W1;
	REAL rho2, vr2, vz2, ur2, uz2, p2, W2;
	REAL adiabatic_exponent;
};

RelativisticJet::RelativisticJet(Data &gdata)  : ProblemType(gdata)
{
	
	rho1 = value((char*)"rho1");
	vr1 = value((char*)"ur1");
	vz1 = value((char*)"uz1");
	p1 = value((char*)"p1");
	W1 = 1. / sqrt(1. - sqr(vr1) - sqr(vz1));
	ur1 = W1 * vr1;
	uz1 = W1 * vz1;

	rho2 = value((char*)"rho2");
	vr2 = value((char*)"ur2");
	vz2 = value((char*)"uz2");
	p2 = value((char*)"p2");
	W2 = 1. / sqrt(1. - sqr(vr2) - sqr(vz2));
	ur2 = W2 * vr2;
	uz2 = W2 * vz2;
	
	
	adiabatic_exponent = value((char*)"Adiabatic_exponent");
}




void RelativisticJet::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	


	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {
			REAL rho(0.);
			REAL ur(0.);
			REAL uz(0.);
			REAL p(0.);

			if (gdata.getCen_x(ix) < 1 && gdata.getCen_z(iz) < 1) {
				rho = rho1;
				ur = ur1;
				uz = uz1;
				p = p1;
			} else {
				rho = rho2;
				ur = ur2;
				uz = uz2;
				p = p2;
			}

			for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ur;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = uz;
				gdata.om[q_Eges](ix, iy, iz) = p/(adiabatic_exponent - 1);
			}
		}
	}

}





void RelativisticJet::bc_User(Data &gdata, NumMatrix<REAL,3> &omb,
		int dir, int top, int q, int rim)
{
	/*
    Include fixed boundary conditions for the bottom z-Plane:
	Ghost cells will be filled with the initial values
	 */
	if(dir == 2) {
		if (top == 0) {

			REAL qfix(0.);
			if (omb.getName() == "rho") {
				qfix = rho1;
			} else if (omb.getName() == "Etherm") {
				qfix = p1/(adiabatic_exponent - 1);
			} else if (omb.getName() == "v_x") {
				qfix = ur1;
			} else if (omb.getName() == "v_z") {
				qfix = uz1;
			}



			for (int k = -rim; k < 0; ++k) {
				for (int j = -rim; j <= gdata.mx[1]+rim; j++) {
					for (int i = -rim; i <= gdata.mx[0]+rim; ++i) {
						if (gdata.getCen_x(i) < 1) {
							omb(i,j,k) = qfix;
						} else {
							if(gdata.om[q_sz](i,j,0) < 0.) {
									omb(i,j,k) = omb(i,j,k+1);
							} else {
								if(q == q_sz) {
									omb(i,j,k) = -omb(i,j,-k-1);
								} else {
									omb(i,j,k) = omb(i,j,-k-1);
								}
							}
						}
					}
				}
			}

		}

	}
}


#endif
