#ifndef BENCHMARK_H
#define BENCHMARK_H

#include <string>

template <class ret_t>
struct retval_t {
  ret_t retval;
};

template <>
struct retval_t<void> {};

template <class task_t>
auto benchmark(std::string name, task_t&& task, bool showRankFinish = true,
               bool parallelStats = true) -> decltype(task()) {
#ifdef parallel
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif

  CronosTimer timer;
  timer.start();
  ccout << "   Starting: " << name << std::endl;

  using ret_t = decltype(task());
  retval_t<ret_t> retval{task()};

#ifdef parallel
  if (showRankFinish) {
    std::cout << "     rank " << rank << " done after " << timer.time() << std::endl;
  }
#endif

  double timeLocal = timer.dtime();
  double maxTime = timeLocal;
  double meanTime = timeLocal;

#ifdef parallel
  if (parallelStats) {
    MPI_Allreduce(&timeLocal, &maxTime, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
    MPI_Allreduce(MPI_IN_PLACE, &meanTime, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    int size;
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    meanTime /= double(size);
  }
#endif

  timer.stop();
  ccout << "   " << name << " done after " << timer.time() << "\n";
  if (parallelStats) {
    ccout << "    - max = " << maxTime << " mean = " << meanTime << endl;
  }
  return retval.retval;
}

#endif