#ifndef CRONOS_DATA_LINK_H
#define CRONOS_DATA_LINK_H

#include <cassert>

template<class T, int rank>
class DataLink {
public:
	DataLink(
		T * data,
		const int lo[rank],
		const int hi[rank]
	);

	~DataLink();

	const int* getLow() const;
	const int* getHigh() const;
	const int* getDims() const;

	int getLow(int) const;
	int getHigh(int) const;
	int getDims(int) const;

	int getSize() const;

	operator T*();
	operator const T*() const;

	T& operator[](const int* pos);
	T  operator[](const int* pos) const;

	T& operator()(int i, int j, int k);
	T  operator()(int i, int j, int k) const;

protected:
	T * data;
	int size;
	int lo[rank];
	int hi[rank];
	int dims[rank];
};



// -------------------------------------------------------------
// inline functions

// write function
template<class T, int rank>
inline T& DataLink<T, rank>::operator[](const int* pos)
{
	assert((pos[rank - 1] >= lo[rank - 1]) && (pos[rank - 1] <= hi[rank - 1]));
	int p = pos[rank - 1] - lo[rank - 1];

	for (int i = rank - 2; i >= 0; i--) {
		assert((pos[i] >= lo[i]) && (pos[i] <= hi[i]));
		p = p * dims[i] + (pos[i] - lo[i]);
	}

	assert((p >= 0) && (p < size));
	return data[p];
}

// read function
template<class T, int rank>
inline T DataLink<T, rank>::operator[](const int* pos) const
{
	assert((pos[rank - 1] >= lo[rank - 1]) && (pos[rank - 1] <= hi[rank - 1]));
	int p = pos[rank - 1];

	for (int i = rank - 2; i >= 0; i--) {
		assert((pos[i] >= lo[i]) && (pos[i] <= hi[i]));
		p = p * dims[i] + (pos[i] - lo[i]);
	}

	assert((p >= 0) && (p < size));
	return data[p];
}

template<class T, int rank>
inline DataLink<T, rank>::operator T*()
{
	return data;
}

template<class T, int rank>
inline DataLink<T, rank>::operator const T*() const
{
	return data;
}

template<class T, int rank>
inline  T& DataLink<T, rank>::operator ()(int i, int j, int k)
{
	assert(rank == 3);
	assert((i >= lo[0]) && (i <= hi[0]));
	assert((j >= lo[1]) && (j <= hi[1]));
	assert((k >= lo[2]) && (k <= hi[2]));
	return data[(i - lo[0]) + dims[0] * ((j - lo[1]) + dims[1] * (k - lo[2]))];
}

template<class T, int rank>
inline  T DataLink<T, rank>::operator ()(int i, int j, int k) const
{
	assert(rank == 3);
	assert((i >= lo[0]) && (i <= hi[0]));
	assert((j >= lo[1]) && (j <= hi[1]));
	assert((k >= lo[2]) && (k <= hi[2]));
	return data[(i - lo[0]) + dims[0] * ((j - lo[1]) + dims[1] * (k - lo[2]))];
}

#endif