#include <gsl/gsl_roots.h>
#include <gsl/gsl_errno.h>
#include <functional>

class GslRootSolver {
public:
  GslRootSolver();
  virtual ~GslRootSolver();

  double solve_f (double xLower, double xUpper);
  double solve_fdf (double x0);

  std::function<double(double)> fun_f;
  std::function<double(double)> fun_df;
  std::function<void(double, double &, double &)> fun_fdf;

  gsl_root_fsolver * solver_f;
  gsl_root_fdfsolver * solver_fdf;

  int max_iter = 20;
  double tolRel = 1e-12;
  double tolAbs = 1e-12;

private:
  using gsl_params_T = struct {
    std::function<double(double)> * f;
    std::function<double(double)> * df;
    std::function<void(double, double &, double &)> * fdf;
  };
  
  gsl_params_T gsl_params;
  gsl_function function_f;
  gsl_function_fdf function_fdf;

  static double gsl_adapter_f(double, void *);
  static double gsl_adapter_df(double, void *);
  static void gsl_adapter_fdf(double, void *, double *, double *);

  void check_gslStatus(const int status) const;
  gsl_error_handler_t * oldHandler;
};


class GslRootSolverError: public std::runtime_error {
public:
  GslRootSolverError(): runtime_error("GslRootSolverError") {};
  GslRootSolverError(std::string error): runtime_error("GslRootSolverError: " + error) {};
};