#ifndef CRONOS_HALO_BUFFER_H
#define CRONOS_HALO_BUFFER_H

#ifdef MODULE_TEST
	#define REAL double
	#include <cmath>
#else
	#include "constants.H"
#endif

#include <iostream>
#include "DataLink.H"
#include <vector>


class Buffer {
public:
	Buffer();
	~Buffer();

	void resize(
			int maxSize
		);

	void add_field(
			int lo[3],
			int hi[3]
		);

	void clear(
		);

	int get_maxSize(
		) const;

	int get_currentSize(
		) const;

	int get_nFields(
		) const;

	DataLink<REAL, 3> & operator()(
			int iField
		);

	operator REAL*();
	operator const REAL*() const;

protected:
	int maxSize;
	int currentSize;
	REAL * buffer;
	std::vector<DataLink<REAL, 3> > fields;
};



class HaloBuffer {
public:
	HaloBuffer(
			int mxCell[3],
			int rim,
			int numMax
		);

	virtual ~HaloBuffer(
		);

	Buffer & operator()(
			int shiftx,	int shifty,	int shiftz
		);

	void add_field(
		);

	void add_field(
			int loDelta[3],
			int hiDelta[3]
		);

	void clear(
		);

	int get_size(
			int shiftx, int shifty, int shiftz
		) const;

	int get_maxSize(
			int shiftx, int shifty, int shiftz
		) const;

	bool is_face(
			int shiftx,	int shifty,	int shiftz
		) const;

	bool is_edge(
			int shiftx,	int shifty,	int shiftz
		) const;

	bool is_corner(
			int shiftx,	int shifty,	int shiftz
		) const;

	int get_index(
			int shiftx,	int shifty,	int shiftz
		) const;

protected:

	void init_buffers(
		);

	virtual int compute_maxSingleSize(
			int rim,
			int shiftx, int shifty, int shiftz
		) const = 0;

	virtual void get_extent(
			int loDelta[3], int hiDelta[3],
			int shiftx, int shifty, int shiftz,
			int loBuff[3], int hiBuff[3]
		) const = 0;

	int mxCell[3];
	int rim;
	int numMax;

	int nFields;
	Buffer buffer[27];
};





//
//	Derived Buffer Types
//


class SndBuffer: public HaloBuffer {
public:
	SndBuffer(
			int mxCell[3],
			int rim,
			int numMax
		);

protected:
	void get_extent(
			int loDelta[3], int hiDelta[3],
			int shiftx, int shifty, int shiftz,
			int loBuff[3], int hiBuff[3]
		) const;

	int compute_maxSingleSize(
			int rim,
			int shiftx, int shifty, int shiftz
		) const;
};


class RcvBuffer: public HaloBuffer {
public:
	RcvBuffer(
			int mxCell[3],
			int rim,
			int numMax
		);

protected:
	void get_extent(
			int loDelta[3], int hiDelta[3],
			int shiftx, int shifty, int shiftz,
			int loBuff[3], int hiBuff[3]
		) const;

	int compute_maxSingleSize(
			int rim,
			int shiftx, int shifty, int shiftz
		) const;
};




#endif
