#include "boundary.H"

using namespace std;

/*-----------------------------------------------------------------
  Data connected functions
  -----------------------------------------------------------------*/

Boundary::Boundary(Data &gdata, int numMax)
{
	this->numMax = numMax;

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	n_om = gdata.fluids->get_N_OM();
	n_omInt = gdata.fluids->get_N_OMINT();
	n_omUser = gdata.fluids->get_N_OM_USER();
	n_omIntUser = gdata.fluids->get_N_OMINT_USER();
	n_omIntAll = gdata.fluids->get_N_OMINT_ALL();

	q_Bx = gdata.fluids->get_q_Bx();
	q_By = gdata.fluids->get_q_By();
	q_Bz = gdata.fluids->get_q_Bz();
#else
#if (FLUID_TYPE == CRONOS_MHD)
	q_Bx = gdata.fluid.get_q_Bx();
	q_By = gdata.fluid.get_q_By();
	q_Bz = gdata.fluid.get_q_Bz();
#else
	// Use a value that is not used (-1 used by user fields -> so, I am using -2)
	q_Bx = -2;
	q_By = -2;
	q_Bz = -2;
#endif

	n_om = N_OM;
	n_omInt = N_OMINT;
#if (OMS_USER == TRUE)
	n_omUser = N_OM_USER;
	n_omIntUser = N_OMINT_USER;
#else
	n_omUser = 0;
	n_omIntUser = 0;
#endif
	n_omIntAll = N_OMINT_ALL;
#endif

	// Load boundary types
	bc_Type[0] = int(value((char*)"bc_x_bot"));
	bc_Type[1] = int(value((char*)"bc_x_top"));
	bc_Type[2] = int(value((char*)"bc_y_bot"));
	bc_Type[3] = int(value((char*)"bc_y_top"));
	bc_Type[4] = int(value((char*)"bc_z_bot"));
	bc_Type[5] = int(value((char*)"bc_z_top"));

	rimDefault = B;

	is_halo.resize(Index::set(-1,-1,-1), Index::set( 1, 1, 1));
	i_min.resize(DIM);
	i_max.resize(DIM);



#ifndef parallel
	// No MPI boundaries in sequential simulations
	is_halo.set_constVal(0);
	// if necessary: compute positions for partners on grid axis (so
	// far only for cylindrical grids
#if (GEOM == CYLINDRICAL)
	if(gdata.get_singularity_treatment(0) > 0) {
		get_AxisPartners(gdata);
		bc_Type[0] = 6;
	}
#elif (GEOM == SPHERICAL)
	if(gdata.get_singularity_treatment(2) > 0) {
		get_AxisPartners(gdata, 0); // Partners at bottom
		bc_Type[2] = 6;
	}
	if(gdata.get_singularity_treatment(3) > 0) {
		get_AxisPartners(gdata, 1); // Partners at top
		bc_Type[3] = 6;
	}



#endif

#else
	// Free the send requests (set them to the ok-state)
	// This is needed to ensure that the initial call
	// of checkSendBuff does not block
	for (int ireq = 0; ireq < 27; ++ireq) {
		requests_send[ireq] = MPI_REQUEST_NULL;
//		MPI_Request_free(&requests_send[ireq]); // Not necessary?
	}

	for(int bound=0; bound<6; ++bound) {
		if(bc_Type[bound] > 5) {
			cerr << " boundary type 6 not to be assigned by user " << endl;
			cerr << " EXITING " << endl;
#ifdef parallel
			MPI_Finalize();
#endif
			exit(-29);
		}
	}



// 	bool atAxis = false;
// #if (GEOM == CYLINDRICAL)
// 	REAL eps(1.e-5*(gdata.global_xe[0]-gdata.global_xb[0]));
// 	bc_Type[0] = -1;
// 	if(gdata.xb[0] >= gdata.global_xb[0]-eps &&
// 	   gdata.xb[0] <= gdata.global_xb[0]+eps) {
// 		atAxis = true;
// 	}
// #elif (GEOM == SPHERICAL)
// 	REAL eps(1.e-5*(gdata.global_xe[1]-gdata.global_xb[1]));
// 	if((gdata.xb[1] >= gdata.global_xb[1]-eps &&
// 	    gdata.xb[1] <= gdata.global_xb[1]+eps) ||
// 	   (gdata.xe[1] >= gdata.global_xe[1]-eps &&
// 	    gdata.xe[1] <= gdata.global_xe[1]+eps)) {
// 		atAxis = true;
// 	}
// #endif


	// if(atAxis) {
#if (GEOM == CYLINDRICAL)
	REAL eps(1.e-5*(gdata.global_xe[0]-gdata.global_xb[0]));
		bc_Type[0] = int(value((char*)"bc_x_bot"));
		if(gdata.get_singularity_treatment(0) > 0) {
			get_AxisPartners(gdata);
			//			if(gdata.axis_treatment(gdata.phi_dir) == 1) {
			// if(gdata.axis_treatment(0) == 1) {
			if(gdata.get_singularity_treatment(0)==1) {
				bc_Type[0] = 6;
			}
		}
#elif (GEOM == SPHERICAL)
		REAL eps(1.e-5*(gdata.global_xe[1]-gdata.global_xb[1]));
		// if(gdata.get_singularity_treatment(2) > 0 ||
		// 		gdata.get_singularity_treatment(3) > 0) {
		// 	get_AxisPartners(gdata,0);
		// 	get_AxisPartners(gdata,1);
		// }
		if(gdata.get_singularity_treatment(2) == 1) {
			get_AxisPartners(gdata,0);
			cout << " involved " << gdata.rank << endl;
		}

		if(gdata.get_singularity_treatment(3) == 1) {
			cout << " doing this " << gdata.rank << endl;
			get_AxisPartners(gdata,1);
		}


		// if(gdata.get_singularity_treatment(2) > 0) {
		if(gdata.get_singularity_treatment(2) == 1) {
			bc_Type[2] = 6;
		}

		// if(gdata.get_singularity_treatment(3) > 0) {
		if(gdata.get_singularity_treatment(3) == 1) {
			bc_Type[3] = 6;
		}
#else
		REAL eps(0.);
#endif
	// }
	// Ranks that do not participate in axis-stuff have to wait...
	MPI_Barrier(gdata.comm3d);


	// Identify outer boundaries (only necessary if NOT fully parallel)
	is_halo.resize(Index::set(-1,-1,-1), Index::set( 1, 1, 1));
	// check x dimension
	if(gdata.check_MpiLowest(0)) {
		is_halo(-1, 0, 0) = 0;
	} else {
		is_halo(-1, 0, 0) = 1;
	}
	if(gdata.check_MpiHighest(0)) {
		is_halo( 1, 0, 0) = 0;
	} else {
		is_halo( 1, 0, 0) = 1;
	}

	// check y dimension
	if(gdata.check_MpiLowest(1)) {
		is_halo( 0,-1, 0) = 0;
	} else {
		is_halo( 0,-1, 0) = 1;
	}
	if(gdata.check_MpiHighest(1)) {
		is_halo( 0, 1, 0) = 0;
	} else {
		is_halo( 0, 1, 0) = 1;
	}

	// check z dimension
	if(gdata.check_MpiLowest(2)) {
		is_halo( 0, 0,-1) = 0;
	} else {
		is_halo( 0, 0,-1) = 1;
	}
	if(gdata.check_MpiHighest(2)) {
		is_halo( 0, 0, 1) = 0;
	} else {
		is_halo( 0, 0, 1) = 1;
	}

	// Treat special case of axis bcs
	if(bc_Type[0]==6) {
		is_halo(-1, 0, 0) = false;
	}
	if(bc_Type[2]==6) {
			is_halo( 0,-1, 0) = false;
	}
	if(bc_Type[3]==6) {
		is_halo( 0, 1, 0) = false;
	}

	for(int dir0=-1; dir0<=1; dir0+=2) {
		for(int dir1=-1; dir1<=1; dir1+=2) {

			// Now determine line boundaries
			if(is_halo(dir0, 0, 0) && is_halo( 0,dir1, 0)) {
				is_halo(dir0,dir1, 0) = true;
			} else {
				is_halo(dir0,dir1, 0) = false;
			}

			if(is_halo(dir0, 0, 0) && is_halo( 0, 0,dir1)) {
				is_halo(dir0, 0,dir1) = true;
			} else {
				is_halo(dir0, 0,dir1) = false;
			}

			if(is_halo( 0,dir0, 0) && is_halo( 0, 0,dir1)) {
				is_halo(0, dir0,dir1) = true;
			} else {
				is_halo( 0,dir0,dir1) = false;
			}

			// ... and corner boundaries
			for(int dir2=-1; dir2<=1; dir2+=2) {
				if(is_halo(dir0, 0, 0) && is_halo( 0,dir1, 0) && is_halo( 0, 0,dir2)) {
					is_halo(dir0,dir1,dir2) = true;
				} else {
					is_halo(dir0,dir1,dir2) = false;
				}
			}

		}
	}
#endif

	// Set all bc types to fields:
	int bc_TypeLow[DIM], bc_TypeHigh[DIM];
	for(int dir=0; dir<DIM; ++dir) {
		bc_TypeLow[dir] = bc_Type[2*dir];
		bc_TypeHigh[dir] = bc_Type[2*dir+1];
	}

	for(int q=0; q<n_om+N_P; ++q) {
		gdata.om[q].set_bcType(bc_TypeLow, bc_TypeHigh);
	}
#if (OMS_USER == TRUE)
	for(int q=0; q<n_omUser; ++q) {
		gdata.om_user[q].set_bcType(bc_TypeLow, bc_TypeHigh);
	}
#endif

	// For constant boundaries set all fields that are not directly
	// accessed by the user to extrapolation
	for(int dir=0; dir<DIM; ++dir) {
		for(int q=n_omInt; q<n_om+N_P; ++q) {
			if(gdata.om[q].get_bcTypeLow(dir) == 7) {
				gdata.om[q].set_bcTypeLow(dir, 2);
			}
			if(gdata.om[q].get_bcTypeHigh(dir) == 7) {
				gdata.om[q].set_bcTypeHigh(dir, 2);
			}
		}
#if (OMS_USER == TRUE)
		for(int q=n_omIntUser; q<n_omUser; ++q) {
			if(gdata.om_user[q].get_bcTypeLow(dir) == 7) {
				gdata.om_user[q].set_bcTypeLow(dir, 2);
			}
			if(gdata.om_user[q].get_bcTypeHigh(dir) == 7) {
				gdata.om_user[q].set_bcTypeHigh(dir, 2);
			}
		}
#endif
	}


	// Prepare size of Send and Receive arrays
	// (currently only for case of edge gridding)

	if(gdata.get_EdgeGridding()) {

		int rim = 3;

		int iminSend = 1;
		int imaxSend = rim;

//		Send_x.resize(Index::set(iminSend, -rim, -rim),
//				Index::set(imaxSend, gdata.mx[1]+rim, gdata.mx[2]+rim));
//		SendArr_x.resize((imaxSend-iminSend+1)*(gdata.mx[1]+2*rim+1)*(gdata.mx[2]+2*rim+1));
//		SendArrSmall_x.resize((imaxSend-iminSend)*(gdata.mx[1]+2*rim+1)*(gdata.mx[2]+2*rim+1));
//
//		Send_y.resize(Index::set(-rim, iminSend, -rim),
//				Index::set(gdata.mx[0]+rim, imaxSend, gdata.mx[2]+rim));
//		SendArr_y.resize((gdata.mx[0]+2*rim+1)*(imaxSend-iminSend+1)*(gdata.mx[2]+2*rim+1));
//		SendArrSmall_y.resize((gdata.mx[0]+2*rim+1)*(imaxSend-iminSend)*(gdata.mx[2]+2*rim+1));
//
//		Send_z.resize(Index::set(-rim, -rim, iminSend),
//				Index::set(gdata.mx[0]+rim, gdata.mx[1]+rim, imaxSend));
//		SendArr_z.resize((gdata.mx[0]+2*rim+1)*(gdata.mx[1]+2*rim+1)*(imaxSend-iminSend+1));
//		SendArrSmall_z.resize((gdata.mx[0]+2*rim+1)*(gdata.mx[1]+2*rim+1)*(imaxSend-iminSend));
//
//		Send_empty.resize(Index::set(0,0,0), Index::set(0,0,0));
//		SendArr_empty.resize(0);

		int size_x = (imaxSend-iminSend+1)*(gdata.mx[1]+2*rim+1)*(gdata.mx[2]+2*rim+1);
		int size_y = (gdata.mx[0]+2*rim+1)*(imaxSend-iminSend+1)*(gdata.mx[2]+2*rim+1);
		int size_z = (gdata.mx[0]+2*rim+1)*(gdata.mx[1]+2*rim+1)*(imaxSend-iminSend+1);
		int size = std::max(size_x, std::max(size_y, size_z));

//		SendArr_buff.resize(size);
//		RecvArr_buff.resize(size);
//
//		SendArr_xRL.resize(size_x);
//		SendArr_xLR.resize(size_x);
//		SendArr_yRL.resize(size_y);
//		SendArr_yLR.resize(size_y);
//		SendArr_zRL.resize(size_z);
//		SendArr_zLR.resize(size_z);
//
//		RecvArr_xRL.resize(size_x);
//		RecvArr_xLR.resize(size_x);
//		RecvArr_yRL.resize(size_y);
//		RecvArr_yLR.resize(size_y);
//		RecvArr_zRL.resize(size_z);
//		RecvArr_zLR.resize(size_z);
//
//		SendArrRL.resize(size);
//		SendArrLR.resize(size);
//
//		RecvArrRL.resize(size);
//		RecvArrLR.resize(size);

		// imaxRecv = rim-1;
		int iminRecv = 1; // Corresponding to mx[0]+1
		int imaxRecv = rim;
//		Recv_x.resize(Index::set(iminRecv,-rim,-rim),
//				Index::set(imaxRecv,gdata.mx[1]+rim,gdata.mx[2]+rim));
//		RecvArr_x.resize((imaxSend-iminSend+1)*(gdata.mx[1]+2*rim+1)*(gdata.mx[2]+2*rim+1));
//		RecvArrSmall_x.resize((imaxSend-iminSend)*(gdata.mx[1]+2*rim+1)*(gdata.mx[2]+2*rim+1));
//
//		Recv_y.resize(Index::set(-rim,iminRecv,-rim),
//				Index::set(gdata.mx[0]+rim,imaxRecv,gdata.mx[2]+rim));
//		RecvArr_y.resize((gdata.mx[0]+2*rim+1)*(imaxSend-iminSend+1)*(gdata.mx[2]+2*rim+1));
//		RecvArrSmall_y.resize((gdata.mx[0]+2*rim+1)*(imaxSend-iminSend)*(gdata.mx[2]+2*rim+1));
//
//		Recv_z.resize(Index::set(-rim,-rim,iminRecv),
//				Index::set(gdata.mx[0]+rim,gdata.mx[1]+rim,imaxRecv));
//		RecvArr_z.resize((gdata.mx[0]+2*rim+1)*(gdata.mx[1]+2*rim+1)*(imaxSend-iminSend+1));
//		RecvArrSmall_z.resize((gdata.mx[0]+2*rim+1)*(gdata.mx[1]+2*rim+1)*(imaxSend-iminSend));
//
//		Recv_empty.resize(Index::set(0,0,0), Index::set(0,0,0));
//		RecvArr_empty.resize(0);

	}




//	cout << " Done prep " << gdata.rank << endl;
	// for(int dir=0; dir<DIM; ++dir) {
	// 	for(int q=0; q<n_om+N_P; ++q) {
	// 		if(bc_Type[2*dir] == 7) {
	// 		} else {
	// 			om[q].set
	// 		}
	// 	}
	// }

	// for(int ibound=0; ibound<2*DIM; ++ibound) {
	// 	if(bc_Type[ibound] == 7) {
	// 		for(int q=0; q<n_omInt; ++q) {
	// 			om.q
	// 		}
	// 	}
	// }
//	cout << " My upper type: " << gdata.rank << " " << bc_Type[2] << " " << bc_Type[3] << endl;
}


int Boundary::get_bc_Type(int num)
{
	int bc_type_ret(0);
	if(num < 8) {
		bc_type_ret = bc_Type[num];
	} else {
		cerr << " no such boundary: " << num << endl;
		exit(-32);
		bc_type_ret = -99;
	}
	return bc_type_ret;
}



void Boundary::prep_boundaries(Data &gdata, ProblemType &pt) {

	if(bc_Type[0] == 7) { // x, bottom

		int lbound[3] = {-B,-B,-B};
		int ubound[3] = {-1,gdata.mx[1]+B,gdata.mx[2]+B};


		for(int q=0; q<n_omInt; ++q) {
			bcVals_xb[q].resize(lbound, ubound);

			// Setting field at lower x bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=-1; ++ix) {
						bcVals_xb[q](ix, iy, iz) = gdata.om[q](ix, iy, iz);
					}
				}
			}
		}
#if (OMS_USER == TRUE)

		for(int q=0; q<n_omIntUser; ++q) {
			bcVals_User_xb[q].resize(lbound, ubound);

			// Setting user field at lower x bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=-1; ++ix) {
						bcVals_User_xb[q](ix, iy, iz) =
							gdata.om_user[q](ix, iy, iz);
					}
				}
			}
		}
#endif
	}

	if(bc_Type[1] == 7) { // x, top

		int lbound[3] = {gdata.mx[0]+1,-B,-B};
		int ubound[3] = {gdata.mx[0]+B,gdata.mx[1]+B,gdata.mx[2]+B};

		for(int q=0; q<n_omInt; ++q) {
			bcVals_xe[q].resize(lbound, ubound);

			// Setting field at upper x bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=gdata.mx[0]+1; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_xe[q](ix, iy, iz) = gdata.om[q](ix, iy, iz);
					}
				}
			}
		}
#if (OMS_USER == TRUE)
		for(int q=0; q<n_omIntUser; ++q) {
			bcVals_User_xe[q].resize(lbound, ubound);

			// Setting user field at upper x bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=gdata.mx[0]+1; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_User_xe[q](ix, iy, iz) = 
							gdata.om_user[q](ix, iy, iz);
					}
				}
			}
		}
#endif
	}

	
	if(bc_Type[2] == 7) { // y, bottom

		int lbound[3] = {-B,-B,-B};
		int ubound[3] = {gdata.mx[1]+B,-1,gdata.mx[2]+B};

		for(int q=0; q<n_omInt; ++q) {
			bcVals_yb[q].resize(lbound, ubound);

			// Setting field at lower y bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=-1; ++iy) {
					for(int ix=-B; ix<=gdata.mx[1]+B; ++ix) {
						bcVals_yb[q](ix, iy, iz) = gdata.om[q](ix, iy, iz);
					}
				}
			}
		}
#if (OMS_USER == TRUE)
		for(int q=0; q<n_omIntUser; ++q) {
			bcVals_User_yb[q].resize(lbound, ubound);

			// Setting user field at lower y bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=-1; ++iy) {
					for(int ix=-B; ix<=gdata.mx[1]+B; ++ix) {
						bcVals_User_yb[q](ix, iy, iz) = 
							gdata.om_user[q](ix, iy, iz);
					}
				}
			}
		}
#endif
	}
	
	if(bc_Type[3] == 7) { // y, top

		int lbound[3] = {-B,gdata.mx[1]+1,-B};
		int ubound[3] = {gdata.mx[0]+B,gdata.mx[1]+B,gdata.mx[2]+B};

		for(int q=0; q<n_omInt; ++q) {
			bcVals_ye[q].resize(lbound, ubound);

			// Setting field at uppter y bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=gdata.mx[1]+1; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_ye[q](ix, iy, iz) = gdata.om[q](ix, iy, iz);
					}
				}
			}
		}
#if (OMS_USER == TRUE)
		for(int q=0; q<n_omIntUser; ++q) {
			bcVals_User_ye[q].resize(lbound, ubound);

			// Setting user field at uppter y bound:
			for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=gdata.mx[1]+1; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_User_ye[q](ix, iy, iz) = 
							gdata.om_user[q](ix, iy, iz);
					}
				}
			}
		}
#endif
	}


	if(bc_Type[4] == 7) { // z, bottom

		int lbound[3] = {-B,-B,-B};
		int ubound[3] = {gdata.mx[0]+B,gdata.mx[1]+B,-1};

		for(int q=0; q<n_omInt; ++q) {
			bcVals_zb[q].resize(lbound, ubound);

			// Setting field at lower z bound:
			for(int iz=-B; iz<=-1; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_zb[q](ix, iy, iz) = gdata.om[q](ix, iy, iz);
					}
				}
			}
		}
#if (OMS_USER == TRUE)
		for(int q=0; q<n_omIntUser; ++q) {
			bcVals_User_zb[q].resize(lbound, ubound);
			
			// Setting user field at lower z bound:
			for(int iz=-B; iz<=-1; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_User_zb[q](ix, iy, iz) = 
							gdata.om_user[q](ix, iy, iz);
					}
				}
			}
		}
#endif
	}

	if(bc_Type[5] == 7) { // z, top

		int lbound[3] = {-B,-B,gdata.mx[2]+1};
		int ubound[3] = {gdata.mx[0]+B,gdata.mx[1]+B,gdata.mx[2]+B};

		for(int q=0; q<n_omInt; ++q) {
			bcVals_ze[q].resize(lbound, ubound);

			// Setting field at upper z bound:
			for(int iz=gdata.mx[2]+1; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_ze[q](ix, iy, iz) = gdata.om[q](ix, iy, iz);
					}
				}
			}
		}
#if (OMS_USER == TRUE)
		for(int q=0; q<n_omIntUser; ++q) {
			bcVals_User_ze[q].resize(lbound, ubound);

			// Setting user field at upper z bound:
			for(int iz=gdata.mx[2]+1; iz<=gdata.mx[2]+B; ++iz) {
				for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
					for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
						bcVals_User_ze[q](ix, iy, iz) = 
							gdata.om_user[q](ix, iy, iz);
					}
				}
			}
		}
#endif
	}
	

}





void Boundary::boundary(Data &gdata, ProblemType &Problem, NumMatrix<double,3> omb[], int num, int iFluid, bool non_generic)
{

	/*
	  Determine type of boundary condition for the individual directions:
	  (-1) -> MPI-periodic (only for internal use)
	  (0) -> Periodic
	  (1) -> Periodic (with data at mx to be overwritten)
	  (2) -> Extrapolation
	  (3) -> Outflow
	  (4) -> User defined local bcs (bc_User)
	  (5) -> Reflecting boundaries
	  (6) -> singularity axis boundary conditions
	  (7) -> constant boundaries (kept from initial conditions) - still under construction
	  */

#ifdef parallel
	// Send and Receive for MPI boundaries
	bc_MPI_Init(gdata, omb, num);
#endif

#ifdef parallel
	// Wait for MPI boundaries
	bc_MPI_Finalize(omb, num);
#endif

	// Treatment of outer boundaries

	for(int qi=0; qi<num; ++qi) {
		int qIndex = -1;
		if(!non_generic) {
#if(FLUID_TYPE==CRONOS_MULTIFLUID)
			int iFluid = gdata.fluids->get_FluidIndex(qi);
			string fluidName = gdata.fluids->fluids[iFluid].get_Name();
			qIndex = Problem.get_q(omb[qi], fluidName);
#else
			qIndex = Problem.get_q(omb[qi]);
#endif
		}

		//-----------------------------------------------------------
		// x-direction
		//-----------------------------------------------------------

		// Non-MPI periodic boundary conditions
#ifndef parallel
//		bc_Periodic_serial(gdata, Problem, omb[qIndex], 0, qIndex, rimDefault);
		bc_Periodic_serial(gdata, Problem, omb[qi], 0, qIndex, rimDefault);
#endif

		if(!is_halo(-1, 0 ,0)) {
//			apply_bc_outer(gdata, Problem, omb[qIndex],0,0,qIndex,rimDefault);
			apply_bc_outer(gdata, Problem, omb[qi],0,0,qIndex,rimDefault);
		}

		if(!is_halo( 1, 0, 0)) {
//			apply_bc_outer(gdata, Problem, omb[qIndex],0,1,qIndex,rimDefault);
			apply_bc_outer(gdata, Problem, omb[qi],0,1,qIndex,rimDefault);
		}


		//-----------------------------------------------------------
		// y-direction
		//-----------------------------------------------------------

		// Non-MPI periodic boundary conditions
	#ifndef parallel
//		bc_Periodic_serial(gdata, Problem, omb[qIndex], 1, qIndex, rimDefault);
		bc_Periodic_serial(gdata, Problem, omb[qi], 1, qIndex, rimDefault);
	#endif

		if(!is_halo(0, -1 ,0)) {
//			apply_bc_outer(gdata, Problem, omb[qIndex],1,0,qIndex,rimDefault);
			apply_bc_outer(gdata, Problem, omb[qi],1,0,qIndex,rimDefault);
		}

		if(!is_halo(0,  1, 0)) {
//			apply_bc_outer(gdata, Problem, omb[qIndex],1,1,qIndex,rimDefault);
			apply_bc_outer(gdata, Problem, omb[qi],1,1,qIndex,rimDefault);
		}


		//-----------------------------------------------------------
		// z-direction
		//-----------------------------------------------------------

		// Non-MPI periodic boundary conditions
	#ifndef parallel
//		bc_Periodic_serial(gdata, Problem, omb[qIndex], 2, qIndex, rimDefault);
		bc_Periodic_serial(gdata, Problem, omb[qi], 2, qIndex, rimDefault);
	#endif

		if(!is_halo(0, 0, -1)) {
//			apply_bc_outer(gdata, Problem, omb[qIndex],2,0,qIndex,rimDefault);
			apply_bc_outer(gdata, Problem, omb[qi],2,0,qIndex,rimDefault);
		}

		if(!is_halo(0, 0,  1)) {
//			apply_bc_outer(gdata, Problem, omb[qIndex],2,1,qIndex,rimDefault);
			apply_bc_outer(gdata, Problem, omb[qi],2,1,qIndex,rimDefault);
		}


		// Do correction for magnetic field on axis after all
		// magnetic boundaries have been applied
//		if(omb[qIndex].getName()=="B_z") {
		if(omb[qi].getName()=="B_z") {

	#if (GEOM == CYLINDRICAL)
			if(gdata.get_singularity_treatment(0) > 0) {
				do_AxisValCorrectionCyl(gdata, Problem);
			}
	#else
			// Correction at lower theta boundary
			if(gdata.get_singularity_treatment(2) > 0) {
				do_AxisValCorrectionSph(gdata, Problem, 0);
			}
			// Correction at lower theta boundary
			// if(gdata.get_singularity_treatment(2) > 0) {
			// 	do_AxisValCorrectionSph(gdata, Problem, 2);
			// }
			if(gdata.get_singularity_treatment(3) > 0) {
				do_AxisValCorrectionSph(gdata, Problem, 1);
			}
	#endif
		}


	} // loop over q index


//#ifdef parallel
//	// Wait for MPI boundaries
//	bc_MPI_Finalize(omb, num);
//#endif

}

#ifdef parallel
void Boundary::fill_MPI_SendBuff(NumMatrix<double,3> omb[], Buffer & sBuff, int num, int shift0, int shift1, int shift2) {

	// Loop over all fields
	for(int qi=0; qi<num; ++qi) {
		DataLink<REAL, 3> & field = sBuff(qi);
		// Loop over related region of grid:
		for (int kk = field.getLow(2); kk <= field.getHigh(2); ++kk){
			for (int jj = field.getLow(1); jj <= field.getHigh(1); ++jj){
				for (int ii = field.getLow(0); ii <= field.getHigh(0); ++ii){
					field(ii, jj, kk) = omb[qi](ii, jj, kk);
				} // ii
			} // jj
		} // kk

	} // qIndex
}


void Boundary::bc_MPI_checkSendBuff(Buffer & sBuff, int shift0, int shift1, int shift2) {
	// Test if MPI is needed
	if(is_halo(shift0, shift1, shift2)) {
		// Request for send direction
		int i_req_send = SendBuff->get_index(shift0, shift1, shift2);

		// Wait for communication to complete
		MPI_Wait(& requests_send[i_req_send], MPI_STATUS_IGNORE);
	}
}


void Boundary::bc_MPI_sendBuff(Data &gdata, Buffer & sBuff, int num, int shift0, int shift1, int shift2) {
	// get rank where to send data:
	int into = gdata.Neighbour(shift0, shift1, shift2);

	int i_req = SendBuff->get_index(shift0, shift1, shift2);

	// Initialize specific request:
	requests_send[i_req] =  MPI_REQUEST_NULL;
	int tag = 42; // Only one communication with specific other rank -> so tag is not important, here
	tag = SendBuff->get_index(shift0, shift1, shift2); // Ignore previous line -> that is bullshit

	MPI_Isend((double *)sBuff, sBuff.get_currentSize(), MPI_DOUBLE, into, tag,
				          gdata.comm3d, &requests_send[i_req]);
}


void Boundary::process_MPI_Send(Data &gdata, NumMatrix<double,3> omb[], int num, int shift0, int shift1, int shift2) {

	// Check if send is necessary:
	if(is_halo(shift0, shift1, shift2)) {

		Buffer & sBuff = (*SendBuff)(shift0, shift1, shift2);

		// Ensure that the send of the last iteration is done before writing
		bc_MPI_checkSendBuff(sBuff, shift0, shift1, shift2);

		// Fill send buffer
		fill_MPI_SendBuff(omb, sBuff, num, shift0, shift1, shift2);

		// Send data
		bc_MPI_sendBuff(gdata, sBuff, num, shift0, shift1, shift2);

	} // if is_halo
}

void Boundary::bc_MPI_RecvAll(Data &gdata, int num) {
	// get rank where to send data:

	// Loop over all communication directions
	for(int shift2 = -1; shift2 <= 1; ++shift2) {
		for(int shift1 = -1; shift1 <= 1; ++shift1) {
			for(int shift0 = -1; shift0 <= 1; ++shift0) {

				// Jump this since 0,0,0 is not a halo block (it referes to the actual physical domain)
				if (shift0==0 && shift1==0 && shift2==0) {
					continue;
				}

				// Check if communication in given direction is necessary:
				if(is_halo(shift0, shift1, shift2)) {
					// Get reference to receive buffer
					Buffer & rBuff = (*RecvBuff)(shift0, shift1, shift2);

					// get rank from where to receive data
					int from = gdata.Neighbour(shift0, shift1, shift2);
					int i_req = RecvBuff->get_index(shift0, shift1, shift2);

					// Initialize specific request:
					requests_recv[i_req] =  MPI_REQUEST_NULL;

					int tag = 42; // Only one communication with specific other rank -> so tag is not important, here
					tag = SendBuff->get_index(-shift0, -shift1, -shift2); // Ignore previous line -> that is bullshit

					MPI_Irecv((double *)rBuff, rBuff.get_currentSize(), MPI_DOUBLE, from, tag,
									          gdata.comm3d, &requests_recv[i_req]);

				} // if is_halo

			} // shift0
		} // shift1
	} // shift2
}

void Boundary::bc_MPI_InitBufferFields(Data &gdata, NumMatrix<double, 3> omb[], int num) {
	if(SendBuff == nullptr || RecvBuff == nullptr) {
		SendBuff = std::unique_ptr<SndBuffer>(new SndBuffer(gdata.mx, B, numMax));
		RecvBuff = std::unique_ptr<RcvBuffer>(new RcvBuffer(gdata.mx, B, numMax));
	}

	SendBuff->clear();
	RecvBuff->clear();

	// Loop over all fields
	for (int qi = 0; qi < num; ++qi) {
		int loDelta[3] = { 0,0,0 };
		int hiDelta[3] = { 0,0,0 };

		if (omb[qi].getName() == "B_x") {
			loDelta[0] = -1;
		}
		else if (omb[qi].getName() == "B_y") {
			loDelta[1] = -1;
		}
		else if (omb[qi].getName() == "B_z") {
			loDelta[2] = -1;
		}
		else if (omb[qi].getName() == "A_x") {
			loDelta[1] = -1;
			loDelta[2] = -1;
		}
		else if (omb[qi].getName() == "A_y") {
			loDelta[0] = -1;
			loDelta[2] = -1;
		}
		else if (omb[qi].getName() == "A_z") {
			loDelta[0] = -1;
			loDelta[1] = -1;
		}

		SendBuff->add_field(loDelta, hiDelta);
		RecvBuff->add_field(loDelta, hiDelta);

	}

	// Assert that sizes match
	for(int shift2 = -1; shift2 <= 1; ++shift2) {
		for(int shift1 = -1; shift1 <= 1; ++shift1) {
			for(int shift0 = -1; shift0 <= 1; ++shift0) {
				assert((*SendBuff)(shift0, shift1, shift2).get_currentSize() ==
						(*RecvBuff)(-shift0, -shift1, -shift2).get_currentSize());
			}
		}
	}

}

void Boundary::bc_MPI_Init(Data &gdata, NumMatrix<double,3> omb[], int num) {

	// Initialize Buffer Fields
	bc_MPI_InitBufferFields(gdata, omb, num);

	// Start all necessary non-blocking receive operations
	bc_MPI_RecvAll(gdata, num);

	// Fill all send buffers together with ensuing send operations

	// Start with smallest fields -> Corners
	for(int shift2 = -1; shift2 <= 1; ++shift2) {
		for(int shift1 = -1; shift1 <= 1; ++shift1) {
			for(int shift0 = -1; shift0 <= 1; ++shift0) {

				if(SendBuff->is_corner(shift0, shift1, shift2)) {
					process_MPI_Send(gdata, omb, num, shift0, shift1, shift2);
				}
			} // shift0
		} // shift1
	} // shift2

	// Now edges
	for(int shift2 = -1; shift2 <= 1; ++shift2) {
		for(int shift1 = -1; shift1 <= 1; ++shift1) {
			for(int shift0 = -1; shift0 <= 1; ++shift0) {

				if(SendBuff->is_edge(shift0, shift1, shift2)) {
					process_MPI_Send(gdata, omb, num, shift0, shift1, shift2);
				}
			} // shift0
		} // shift1
	} // shift2


	// Finally, address cell faces
	for(int shift2 = -1; shift2 <= 1; ++shift2) {
		for(int shift1 = -1; shift1 <= 1; ++shift1) {
			for(int shift0 = -1; shift0 <= 1; ++shift0) {

				if(SendBuff->is_face(shift0, shift1, shift2)) {
					process_MPI_Send(gdata, omb, num, shift0, shift1, shift2);
				}
			} // shift0
		} // shift1
	} // shift2
}


void Boundary::write_MPI_RecvBuff(NumMatrix<double,3> omb[], Buffer & rBuff, int num, int shift0, int shift1, int shift2) {

	// Loop over all fields
	for (int qi = 0; qi<num; ++qi) {
		DataLink<REAL, 3> & field = rBuff(qi);

		// Loop over related region of grid:
		for (int kk = field.getLow(2); kk <= field.getHigh(2); ++kk) {
			for (int jj = field.getLow(1); jj <= field.getHigh(1); ++jj) {
				for (int ii = field.getLow(0); ii <= field.getHigh(0); ++ii) {
					omb[qi](ii, jj, kk) = field(ii, jj, kk);
				} // ii
			} // jj
		} // kk
	} // qIndex
}


void Boundary::bc_MPI_checkRecvBuff(Buffer &rBuff, int shift0, int shift1, int shift2) {

	// Test if MPI is needed
	if(is_halo(shift0, shift1, shift2)) {
		// Request for receive direction
		int i_req_recv = RecvBuff->get_index(shift0, shift1, shift2);

		// Wait for communication to complete
		MPI_Wait(& requests_recv[i_req_recv], MPI_STATUS_IGNORE);
	}
}

void Boundary::process_MPI_Recv(NumMatrix<double,3> omb[], int num, int shift0, int shift1, int shift2) {

	// Check if send is necessary:
	if(is_halo(shift0, shift1, shift2)) {

		Buffer & rBuff = (*RecvBuff)(shift0, shift1, shift2);

		// Wait for receive buffer to be filled
		bc_MPI_checkRecvBuff(rBuff, shift0, shift1, shift2);

		// Fill send buffer
		write_MPI_RecvBuff(omb, rBuff, num, shift0, shift1, shift2);


	} // if is_halo
}


void Boundary::bc_MPI_Finalize(NumMatrix<double,3> omb[], int num) {
	// Check all receive buffers, ensuing receive operations and write data

	// Start with smallest fields -> Corners
	for(int shift2 = 1; shift2 >= -1; --shift2) {
		for(int shift1 = 1; shift1 >= -1; --shift1) {
			for(int shift0 = 1; shift0 >= -1; --shift0) {

				if(SendBuff->is_corner(shift0, shift1, shift2)) {
					process_MPI_Recv(omb, num, shift0, shift1, shift2);
				}
			} // shift0
		} // shift1
	} // shift2

	// Now edges
	for(int shift2 = 1; shift2 >= -1; --shift2) {
		for(int shift1 = 1; shift1 >= -1; --shift1) {
			for(int shift0 = 1; shift0 >= -1; --shift0) {

				if(SendBuff->is_edge(shift0, shift1, shift2)) {
					process_MPI_Recv(omb, num, shift0, shift1, shift2);
				}
			} // shift0
		} // shift1
	} // shift2


	// Finally, address cell faces
	for(int shift2 = 1; shift2 >= -1; --shift2) {
		for(int shift1 = 1; shift1 >= -1; --shift1) {
			for(int shift0 = 1; shift0 >= -1; --shift0) {

				if(SendBuff->is_face(shift0, shift1, shift2)) {
					process_MPI_Recv(omb, num, shift0, shift1, shift2);
				}
			} // shift0
		} // shift1
	} // shift2

}



#endif

void Boundary::get_bcBuffRange(Data &gdata, int &range_min, int & range_max,
		int dir, int leftToRight, int q, int rim ) {
	if(leftToRight) {

		range_min = 1;
		range_max = rim;

		if(!gdata.get_EdgeGridding()) {
			range_min = 1;
			if(bc_Type[2*dir] == 1 || gdata.time < 0.01*gdata.dt) {
				range_min = 0;
			}
		}

	} else {

		range_min = -rim;
		range_max = -1;

		if((bc_Type[2*dir+1] > -1 || gdata.time < 0.01*gdata.dt) &&
				((dir==0 && q==q_Bx) || (dir==1 && q==q_By) || (dir==2 && q==q_Bz))) {
			range_max = -2;
		}

	}
}


void Boundary::bc_Periodic_serial(Data &gdata, ProblemType &Problem,
		NumMatrix<REAL,3> &omb, int dir, int q, int rim) {

	int shift(0);
	if(gdata.get_EdgeGridding()) {
		shift = 1;
	}

	int i_min, i_max;

	// Check if direction is periodic
	if((bc_Type[2*dir] < 2) && (bc_Type[2*dir+1] < 2)) {

		// x-direction
		if(dir == 0) {
			// From left to right:
			get_bcBuffRange(gdata, i_min, i_max, 0, 1, q, rim);
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; j++){
					for (int i = i_min; i <= i_max; i++){
						omb(gdata.mx[0]+i,j,k) = omb(i-shift,j,k);
					}
				}
			}
			// From right to left:
			get_bcBuffRange(gdata, i_min, i_max, 0, 0, q, rim);
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; j++){
					for (int i = i_max; i >= i_min; --i){
						omb(i,j,k) = omb(gdata.mx[0]+i+shift,j,k);
					}
				}
			}

			// y-direction
		} else if (dir == 1) {
			// From left to right:
			get_bcBuffRange(gdata, i_min, i_max, 1, 1, q, rim);
			for (int j = i_min; j <= i_max; j++){
				for (int k = -rim; k <= gdata.mx[2]+rim; k++){
					for (int i = -rim; i <= gdata.mx[0]+rim; i++){
						// y_min -> y_max
						omb(i,gdata.mx[1]+j,k) = omb(i,j-shift,k);
					}
				}
			}
			// From right to left:
			get_bcBuffRange(gdata, i_min, i_max, 1, 0, q, rim);
			for (int j = i_max; j >= -rim; --j){
				for (int k = -rim; k <= gdata.mx[2]+rim; k++){
					for (int i = -rim; i <= gdata.mx[0]+rim; i++){
						omb(i,j,k) = omb(i,gdata.mx[1]+j+shift,k);
					}
				}
			}

			// z-direction
		} else {
			// From left to right:
			get_bcBuffRange(gdata, i_min, i_max, 2, 1, q, rim);
			for (int k = i_min; k <= i_max; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; j++){
					for (int i = -rim; i <= gdata.mx[0]+rim; i++){
						// z_min -> z_max
						omb(i,j,gdata.mx[2]+k) = omb(i,j,k-shift);
					}
				}
			}
			// From right to left:
			get_bcBuffRange(gdata, i_min, i_max, 2, 0, q, rim);
			for (int k = i_max; k >= -rim; --k){
				for (int j = -rim; j <= gdata.mx[1]+rim; j++){
					for (int i = -rim; i <= gdata.mx[0]+rim; i++){
						omb(i,j,k) = omb(i,j,gdata.mx[2]+k+shift);
					}
				}
			}
		}

	}
}






void Boundary::ExchangePositions(Data &gdata, int top) {
	// First we need to find if we have a corresponding opposite rank
	// (should moslty be the case - only when deviding by odd number
	// would we expect problems)

	// cerr << " Exchanging " << gdata.rank << " " << gdata.coords[0] << endl;

#if (GEOM != CARTESIAN)
	PhiPos[top].resize(Index::set(-B),
	                   Index::set(gdata.global_mx[gdata.phi_dir]+B));

#ifdef parallel
	if(gdata.nproc[gdata.phi_dir] > 1) {
#if(GEOM == CYLINDRICAL)
		if(gdata.coords[0] == 0) {
#elif (GEOM == SPHERICAL)
		   if(top==0 &&gdata.coords[1]==0 || top==1 && gdata.coords[1]==gdata.nproc[1]-1) {
#endif

			// Prepare array holding nproc entries in phi-direction
			NumMatrix<REAL, 1> PhiPos_loc[gdata.nproc[gdata.phi_dir]];
			
#if(GEOM == CYLINDRICAL)
			int rankOwn = gdata.rank_constz;
#elif (GEOM == SPHERICAL)
			int rankOwn = gdata.rank_constxy;
#endif

			// Resize and clear local array
			PhiPos_loc[rankOwn].resize(Index::set(-B), 
			                           Index::set(gdata.mx[gdata.phi_dir]+B));
			PhiPos_loc[rankOwn].clear();

			// Fill array with phi-data
			for(int ip = -B; ip <= gdata.mx[gdata.phi_dir]+B; ++ip) { 
				double phi = gdata.getCen(gdata.phi_dir, ip);
			
				// PhiPos_loc[gdata.rank_constz](ip) = phi;
				// MyData(ip) = phi;
				PhiPos_loc[rankOwn](ip) = phi;
				
			}

 
			// Prepare mpi-communication:
			int size = (PhiPos_loc[rankOwn].getHigh(0) - 
			            PhiPos_loc[rankOwn].getLow(0) + 1);
		
			int req_num = 2*(gdata.nproc[gdata.phi_dir]-1);
			MPI_Request requests[req_num];
			MPI_Status statusrl[req_num];
			
			for (int i=0; i<req_num; i++){
				requests[i] = MPI_REQUEST_NULL;
			}
			
			// Loop over all processes with same z at origin:
			int number = 0;
			
			//		NumMatrix<REAL, 1> PhiPos_loc[gdata.nproc[gdata.phi_dir]];
// #if(GEOM == CYLINDRICAL)
// 			int rankOwn = gdata.rank_constz;
// #elif (GEOM == SPHERICAL)
// 			int rankOwn = gdata.rank_constx;
// #endif
			for(int irank=0; irank<gdata.nproc[gdata.phi_dir]; ++irank) {
			
				int rankother;
#if(GEOM == CYLINDRICAL)
				rankother = gdata.rankpos_z(gdata.coords[0],irank);
				MPI_Comm my_comm = gdata.comm_constz;
#elif (GEOM == SPHERICAL)
				rankother = gdata.rankpos_xy(irank);
				MPI_Comm my_comm = gdata.comm_constxy;
#endif

				// Treat data from other ranks
				if(rankOwn != rankother) {

					PhiPos_loc[irank].resize(Index::set(-B), 
					                         Index::set(gdata.mx[gdata.phi_dir]+B));
					PhiPos_loc[irank].clear();


					int tag = rankother; // message tag
					
					MPI_Irecv((double *)PhiPos_loc[irank], size, MPI_DOUBLE,
					          rankother, tag, my_comm, &requests[number]);
					number++;
					
					tag = gdata.rank_constxy;
					
					// Send data:
					MPI_Isend((double *)PhiPos_loc[rankOwn], size, 
					          MPI_DOUBLE, rankother, tag, my_comm, 
					          &requests[number]);
					number++;

				// } else {
				// 	// If this ever becomes a problem I have to store the local rank
				// 	// separately.
				// 	if(irank != rankOwn) {
				// 		cerr << " Mixup with local ranks along phi " << endl;
				// 		cerr << " Exiting " << endl;
				// 		exit(3);
				// 	}
				}
				// } else {
				// 	// Is this correct??

				// 	PhiPos_loc[irank].resize(Index::set(-B), 
				// 	                         Index::set(gdata.mx[gdata.phi_dir]+B));
				// 	PhiPos_loc[irank].clear();
				// 	for(int ip = -B; ip <= gdata.mx[gdata.phi_dir]+B; ++ip) { 
				// 		double phi = gdata.getCen(gdata.phi_dir, ip);
						
				// 		// PhiPos_loc[gdata.rank_constz](ip) = phi;
				// 		// MyData(ip) = phi;
				// 		PhiPos_loc[irank](ip) = phi;
				// 	}

				// 	//					PhiPos_loc[irank] = irank;
				// }
			
			}

			if(req_num > 0) {
				MPI_Waitall(req_num, requests, statusrl);
			}

			for(int irank=0; irank<gdata.nproc[gdata.phi_dir]; ++irank) {
				int shift = irank*gdata.get_RankWidth(gdata.phi_dir);
				for(int ip = -B; ip <= gdata.mx[gdata.phi_dir]+B; ++ip) { 
					PhiPos[top](ip+shift) = PhiPos_loc[irank](ip);
				}
			}
			
		   }			
			

		} else {
#endif

	// Compute Phi-positions also for serial run
	for(int ip = -B; ip <= gdata.mx[gdata.phi_dir]+B; ++ip) { 
		double phi = gdata.getCen(gdata.phi_dir, ip);
		
		PhiPos[top](ip) = phi;
		
	}
	

#ifdef parallel

	}

#endif

#endif // not cartesian

		}


#ifdef parallel

void Boundary::get_AxisData(Data &gdata,
                            NumMatrix<REAL,3> &myBound,
                            NumMatrix<REAL,3> &global_bcData,
                            int top) {
	// For parallel case: exchange data an build global array of
	// boundary data. For serial data - just pipe through?? No! That
	// would double the array data - argh So maybe avoid that
	// altogether. Maybe for the serial case something like & operator
	// in the external routine?

	// Array holding the global version of the data (full phi-extent)
	// NumMatrix<REAL,3> global_bcData;

#if (GEOM != CARTESIAN)

#if (GEOM == CYLINDRICAL) // phi in y-direction
	int lbound[DIM] = { 0, -B, -B};
	int ubound_global[DIM] = { 2, gdata.global_mx[1]+B, gdata.mx[2]+B };
	int ubound_local[DIM]  = { 2, gdata.mx[1]+B, gdata.mx[2]+B };

	if(myBound.getName() == "B_x") {
		lbound[0] = -1;
		ubound_global[0] = 1;
		ubound_local[0] = 1;
	}
#elif (GEOM == SPHERICAL) // phi in z-direction

	int lbound[DIM] = { -B, 0, -B};
	int ubound_global[DIM] = {gdata.mx[0]+B, 2, gdata.global_mx[2]+B };
	int ubound_local[DIM]  = {gdata.mx[0]+B, 2, gdata.mx[2]+B };

	if(myBound.getName() == "B_y") {
		lbound[1] = -1;
		ubound_global[1] = 1;
		ubound_local[1] = 1;
	}

	// At upper boundary:
	if(top==1) {
		int mxtet = gdata.mx[1];
		lbound[1] = mxtet-B+1;
		ubound_global[1] = mxtet;
		ubound_local[1] = mxtet;

		if(myBound.getName() == "B_y") {
			lbound[1] -= 1;
			ubound_global[1] -= 1;
			ubound_local[1] -= 1;
		} else if (myBound.getName() == "A_x" || myBound.getName() == "A_z") {
		  ubound_global[1] += 1;
		  ubound_local[1] += 1;
		}
	}
#endif

#if (GEOM != CARTESIAN)
	global_bcData.resize(lbound, ubound_global);
#endif

	// So far only for cylindrical coords
#if (GEOM == CYLINDRICAL)

	// cout << " My coords: " << gdata.coords[0] << " " << gdata.rank << " " << gdata.rank_constz << " " << gdata.rankpos(gdata.coords[0],1, gdata.coords[2]) << " " << gdata.coords[2] << endl;

	// MPI_Finalize();
	// exit(3);

	// Only applicable for axis-adjacent grid points
	if(gdata.coords[0] == 0) {

		int ownRank(gdata.rank_constz);

		// Assiging local boundary data array:
		NumMatrix<REAL,3> local_bcData[gdata.nproc[gdata.phi_dir]];
		local_bcData[ownRank].resize(lbound, ubound_local);

		for(int iz=lbound[2]; iz<=ubound_local[2]; ++iz) {
			for(int iy=lbound[1]; iy<=ubound_local[1]; ++iy) {
				for(int ix=lbound[0]; ix<=ubound_local[0]; ++ix) {
					local_bcData[ownRank](ix, iy, iz) = myBound(ix, iy, iz);
				}
			}
		}

		// Preparing MPI-communication
		int size = ((local_bcData[ownRank].getHigh(0) -
		             local_bcData[ownRank].getLow(0) + 1)*
		            (local_bcData[ownRank].getHigh(1) -
		             local_bcData[ownRank].getLow(1) + 1)*
		            (local_bcData[ownRank].getHigh(2) -
		             local_bcData[ownRank].getLow(2) + 1));
		
		// number of MPI communication events (one less, because no
		// self-communication)
		int req_num = 2*(gdata.nproc[gdata.phi_dir]-1);
		MPI_Request requests[req_num];
		MPI_Status statusrl[req_num];

		// Initialise requests
		for (int i=0; i<req_num; i++){
			requests[i] = MPI_REQUEST_NULL;
		}

		// Loop over all processes with same z at origin:
		int number = 0;

#if(GEOM == CYLINDRICAL)
		int rankOwn = gdata.rank_constz;
		MPI_Comm my_comm = gdata.comm_constz;
#elif (GEOM == SPHERICAL)
		int rankOwn = gdata.rank_constxy;
		MPI_Comm my_comm = gdata.comm_constxy;
#endif

		for(int irank=0; irank<gdata.nproc[gdata.phi_dir]; ++irank) {

			int rankother;
			// obtain position of rank irank
			if(gdata.phi_dir == 1) {
				rankother = gdata.rankpos_z(gdata.coords[0],irank);
			}

			// Get other data
			// Avoid self-communication

			// if(gdata.rank == 2) {
			// 	cerr << " The others: " << rankother << endl;
			// }

			if(rankOwn != rankother) {

				// Prepare array to hold other boundary data
				local_bcData[irank].resize(lbound, ubound_local);

				int tag = rankother; // message tag

				MPI_Irecv((double *)local_bcData[irank], size, MPI_DOUBLE,
				          rankother, tag, my_comm, &requests[number]);
				number++;

				tag = gdata.rank;

				// Send data:
				MPI_Isend((double *)local_bcData[ownRank], size, MPI_DOUBLE, 
				          rankother, tag, my_comm, &requests[number]);
				number++;

			}

		}

		MPI_Waitall(req_num, requests, statusrl);

		for(int irank=0; irank<gdata.nproc[gdata.phi_dir]; ++irank) {
			store_AxisData(gdata, lbound, ubound_local, irank,
			               local_bcData[irank], global_bcData);
		}
	
		// if(gdata.rank == 1 && myBound.getName() == "v_z") {
		// 	cout << " other rank " << 3 << endl;
		// 	for(int j = -B; j <= gdata.mx[1]+B; ++j) {
		// 		cout << j << " fu " << (local_bcData[0](1,j,2)-30)/pi << " " << (local_bcData[1](1,j,2)-30)/pi << " " << gdata.get_z(2) << " " << ownRank << endl;
		// 	}
			
		// }




	} // if coords[0] == 0
	// MPI_Finalize();
	// exit(3);
#else // Now spherical(?!)
	if((top==0 && gdata.coords[1] == 0 )||
	   (top==1 && gdata.coords[1]==gdata.nproc[1]-1)) {

		// Communicate with ranks with same x-values
		int ownRank(gdata.rank_constxy);

		// Assiging local boundary data array:
		NumMatrix<REAL,3> local_bcData[gdata.nproc[gdata.phi_dir]];
		local_bcData[ownRank].resize(lbound, ubound_local);

		for(int iz=lbound[2]; iz<=ubound_local[2]; ++iz) {
			for(int iy=lbound[1]; iy<=ubound_local[1]; ++iy) {
				for(int ix=lbound[0]; ix<=ubound_local[0]; ++ix) {
					local_bcData[ownRank](ix, iy, iz) = myBound(ix, iy, iz);
				}
			}
		}

		// Preparing MPI-communication
		int size = ((local_bcData[ownRank].getHigh(0) - local_bcData[ownRank].getLow(0) + 1)*
				(local_bcData[ownRank].getHigh(1) - local_bcData[ownRank].getLow(1) + 1)*
				(local_bcData[ownRank].getHigh(2) - local_bcData[ownRank].getLow(2) + 1));

		// number of MPI communication events (one less, because no
		// self-communication)
		int req_num = 2*(gdata.nproc[gdata.phi_dir]-1);
		MPI_Request requests[req_num];
		MPI_Status statusrl[req_num];

		// Initialise requests
		for (int i=0; i<req_num; i++){
			requests[i] = MPI_REQUEST_NULL;
		}

		// Loop over all processes with same z at origin:
		int number = 0;

#if(GEOM == CYLINDRICAL)
		int rankOwn = gdata.rank_constz;
		MPI_Comm my_comm = gdata.comm_constz;
#elif (GEOM == SPHERICAL)
		int rankOwn = gdata.rank_constxy;
		MPI_Comm my_comm = gdata.comm_constxy;
#endif

		for(int irank=0; irank<gdata.nproc[gdata.phi_dir]; ++irank) {

			int rankother;
			// obtain position of rank irank
			// AT ME: does this have to be 2?
			if(gdata.phi_dir == 2) {
				rankother = gdata.rankpos_xy(irank);
			}

			// Get other data
			// Avoid self-communication

			if(rankOwn != rankother) {

				// Prepare array to hold other boundary data
				local_bcData[irank].resize(lbound, ubound_local);

				int tag = rankother; // message tag

				MPI_Irecv((double *)local_bcData[irank], size, MPI_DOUBLE,
						rankother, tag, my_comm, &requests[number]);
				number++;

				tag = gdata.rank_constxy;

				// Send data:
				MPI_Isend((double *)local_bcData[ownRank], size, MPI_DOUBLE,
						rankother, tag, my_comm, &requests[number]);
				number++;

			}

		}

		MPI_Waitall(req_num, requests, statusrl);

		for(int irank=0; irank<gdata.nproc[gdata.phi_dir]; ++irank) {
			store_AxisData(gdata, lbound, ubound_local, irank,
					local_bcData[irank], global_bcData);
		}
	}

	// exit(3);
#endif
#endif // not Cartesian

}


void Boundary::store_AxisData(Data &gdata, int lbound[DIM],
                              int ubound_local[DIM], int rank_local,
                              NumMatrix<REAL,3> &local_bcData,
                              NumMatrix<REAL,3> &global_bcData) {

	// Use data by other rank (or own) to store in global data array

#if (GEOM == CYLINDRICAL) // phi in y-direction
	
	//	int shift = rank_local*gdata.mx[1];
	int shift = rank_local*gdata.get_RankWidth(1);

	// set data into global array
	for(int iz=lbound[2]; iz<=ubound_local[2]; ++iz) {
		for(int iy=lbound[1]; iy<=ubound_local[1]; ++iy) {
			for(int ix=lbound[0]; ix<=ubound_local[0]; ++ix) {
				global_bcData(ix, iy+shift, iz) = local_bcData(ix, iy, iz);
			}
		}
	}

#elif (GEOM == SPHERICAL) // phi in z-direction

	int shift = rank_local*gdata.get_RankWidth(2);

	// set data into global array
	for(int iz=lbound[2]; iz<=ubound_local[2]; ++iz) {
		for(int iy=lbound[1]; iy<=ubound_local[1]; ++iy) {
			for(int ix=lbound[0]; ix<=ubound_local[0]; ++ix) {
				global_bcData(ix, iy, iz+shift) = local_bcData(ix, iy, iz);
			}
		}
	}

#endif

#if (FALSE)
	// General Version:
	// Axis for cyl coords
	if(gdata.phi_dir == 1) { // for cyl coords
		// Compute shift in phi-direction (depends on local rank)
		// in this case rank_local == rank_constz

		int shift[3] = {0,0,0};

		shift[gdata.phi_dir] = rank_local*gdata.mx[gdata.phi_dir];

		// set data into global array
		for(int iz=-B; iz<=gdata.mx[2]+B; ++iz) {
			for(int iy=-B; iy<=gdata.mx[1]+B; ++iy) {
				for(int ix=-B; ix<=gdata.mx[0]+B; ++ix) {
					global_bcData(ix+shift[0], iy+shift[1], iz+shift[2]) = 
						local_bcData(ix, iy, iz);
				}
			}
		}

	}
#endif

}


void Boundary::ExchangeData(Data &gdata) {
	// First we need to find if we have a corresponding opposite rank
	// (should moslty be the case - only when deviding by odd number
	// would we expect problems)

}
#endif


void Boundary::get_AxisPartners(Data &gdata, int top) {
	// Computing the partners for axis boundary conditions - so far
	// only for cylindrical coordinates

#if (GEOM != CARTESIAN)
	ExchangePositions(gdata, top);

	// Array over full phi-domain
#ifdef parallel
	AxisPartnersPhi[top].resize(Index::set(-B),
			Index::set(gdata.global_mx[gdata.phi_dir]+B));
#else
	AxisPartnersPhi[top].resize(Index::set(-B),
			Index::set(gdata.mx[gdata.phi_dir]+B));
#endif

	double Pi2 = 2.*M_PI;
	double eps = (1.e-5)*(gdata.xe[gdata.phi_dir]-
	                      gdata.xb[gdata.phi_dir])/gdata.global_mx[gdata.phi_dir];

	// loop over azimuthal domain:
	// look for mapping p.s
	for (int ip = -B; ip <= gdata.global_mx[gdata.phi_dir]+B; ++ip) { 

		AxisPartnersPhi[top](ip) = -100;   // default value to flag a failure
		// 'phi' = required partner position

		double phi = PhiPos[top](ip) + M_PI;
		phi -= Pi2*static_cast<int>(phi/Pi2);    // keep phi in [0,2pi]

		// if (phi >= gdata.xb[gdata.phi_dir] and   // desired mp inside grid
		//     phi <= gdata.xe[gdata.phi_dir]) {   //  => loop candidates
		if (phi >= gdata.global_xb[gdata.phi_dir] and  // desired mp inside grid
		    phi <= gdata.global_xe[gdata.phi_dir]) {   //  => loop candidates

			// loop candidates
			for (int imp = 0; imp <= gdata.global_mx[gdata.phi_dir]; ++imp) { 
				//if (abs(gdata.getCen(gdata.phi_dir, imp) - phi) < eps) {
				if (abs(PhiPos[top](imp) - phi) < eps) {
					AxisPartnersPhi[top](ip) = imp; // found one
					break;
				}
			}
			if (AxisPartnersPhi[top](ip) == -100) {  // mp needed, but none found
				cerr << "error: singularity present, "
				     << "but cells cannot be mapped." << endl;
				exit(1);
			}
		} else {
			AxisPartnersPhi[top](ip) = -10;  // this means "no mp needed"
		}


		// Workaround for single grid point in phi-direction:
		if(gdata.global_mx[gdata.phi_dir]==0) {
			AxisPartnersPhi[top](ip) = 0;
		}
//		 if(gdata.rank==0) {
//		 	cout << " My partner: " << ip << " " << AxisPartnersPhi[top](ip) << " ";
//		 	cout << phi << " " << (phi-M_PI)/M_PI;
//		 	cout << endl;
//		 }
	}
#endif	

	// Additional convenience functions holding sin and cos of positions
	// along the grid
	for(unsigned int dir=0; dir<3; ++dir) {
		cosPos[dir].resize(Index::set(-B), Index::set(gdata.mx[dir]+B));
		sinPos[dir].resize(Index::set(-B), Index::set(gdata.mx[dir]+B));
		// Fill values of arrays
		for(int iPos=-B; iPos<=gdata.mx[dir]+B; ++iPos) {
			sinPos[dir](iPos) = sin(gdata.get_pos(dir, iPos, 0));
			cosPos[dir](iPos) = cos(gdata.get_pos(dir, iPos, 0));
		}
	}

}

void Boundary::bc_Axis(Data &gdata, ProblemType &Problem,
		NumMatrix<REAL,3> &omb, int dir, int top, int q, int rim) {
#if (GEOM == CYLINDRICAL)
	if(dir==0) {
		bc_AxisCyl(gdata, Problem, omb, q, rim);
	}
#elif (GEOM == 3)
	if(dir==1) {
		bc_AxisSph(gdata, Problem, omb, top, q, rim);
	}
#endif
}


void Boundary::bc_AxisCyl(Data &gdata, ProblemType &Problem,
		NumMatrix<REAL,3> &omb, int q, int rim) {

	// Special treatment of coordinate axis - so far only for
	// cylindrical coordinates.

#if (GEOM == CYLINDRICAL)
	// Check for necessity of sign inversion at the origin:
	string qname = omb.getName(); 
	string subsp = qname.substr(qname.size()-2,2); // last two letters
	short sign = (subsp == "_x" || subsp == "_y") ? -1 : +1; // sign inv.

	// if(q == 7) {
	// 	cout << " Partner: " << AxisPartnersPhi(5) << " ";
	// 	cout << (gdata.mx[1]+1)/2 + 5 << endl;
	// }

#ifdef parallel
	// Parallel version -- need to work with extra global array
	NumMatrix<REAL,3> omb_global; // additional global array
	get_AxisData(gdata, omb, omb_global);
#else
	NumMatrix<REAL,3> & omb_global = omb;
#endif

	for (int k = -rim; k <= gdata.mx[2]+rim; ++k) {
		for (int j = -rim; j <= gdata.mx[1]+rim; ++j) {
			if (AxisPartnersPhi[0](j) != -10) {
				for (int i = -rim; i <= -1; i++) {
					if (qname == "B_x") {
						omb(i,j,k) = -1*omb_global(-2-i,AxisPartnersPhi[0](j),k);
					} else {
						omb(i,j,k) = sign*omb_global(-1-i,AxisPartnersPhi[0](j),k);
					}

				}
			}
		}
	}

#endif

}


void Boundary::bc_AxisSph(Data &gdata, ProblemType &Problem,
		NumMatrix<REAL,3> &omb, int top, int q, int rim) {
#if (GEOM == 3)

#ifdef parallel
	// Parallel version -- need to work with extra global array
	NumMatrix<REAL,3> omb_global; // additional global array
	get_AxisData(gdata, omb, omb_global, top);
#else
	NumMatrix<REAL,3> & omb_global = omb;
#endif

	string qname = omb.getName();
	string subsp = qname.substr(qname.size()-2,2); // last two letters


	const int mxtet = gdata.mx[1];
	short si = (subsp == "_y" or subsp == "_z") ? -1 : +1;
	for (int iz = -rim; iz <= gdata.mx[2]+rim; ++iz) {
		if (AxisPartnersPhi[top](iz) != -10) {
			for (int ix = -rim; ix <= gdata.mx[0]+rim; ++ix) {
				if (top == 0 and gdata.get_singularity_treatment(2)==1) {
					for (int iy = -1; iy >= -rim; --iy) {
						if (qname == "B_y") {
							omb(ix,iy,iz) = -1*omb_global(ix,-iy-2,AxisPartnersPhi[top](iz));
						} else if (qname == "A_x") { // means A_r
						  if(iy>-3) {
							omb(ix,iy,iz) = omb_global(ix,-iy,AxisPartnersPhi[top](iz));
						  }
						} else if (qname == "A_z") { // means A_phi
						  if(iy>-3) {
							omb(ix,iy,iz) = -omb_global(ix,-iy,AxisPartnersPhi[top](iz));
						  }
						} else {
							omb(ix,iy,iz) = si*omb_global(ix,-iy-1,AxisPartnersPhi[top](iz));
						}
					}
				} else if (top == 1 and gdata.get_singularity_treatment(3)==1) {
					for (int iy = 1; iy <= rim; ++iy) {
						if (qname == "B_y") {
							omb(ix,mxtet+iy,iz) = -1*omb_global(ix,mxtet-iy,AxisPartnersPhi[top](iz));
						} else if (qname == "A_x") { // means A_r
						  omb(ix,mxtet+iy,iz) = -omb_global(ix,mxtet+2-iy,AxisPartnersPhi[top](iz));
						} else if (qname == "A_z") { // means A_phi
						  omb(ix,mxtet+iy,iz) = -1*omb_global(ix,mxtet+2-iy,AxisPartnersPhi[top](iz));
						} else {
							omb(ix,mxtet+iy,iz) = si*omb_global(ix,mxtet+1-iy,AxisPartnersPhi[top](iz));
						}
					}
				}
			} // ix loop
		}
	}
#endif

}


void Boundary::do_AxisValCorrectionCyl(Data &gdata, ProblemType &Problem) {
#if (GEOM == CYLINDRICAL)
	int jnum = gdata.get_RankWidth(1);
	for (int iz = -B; iz <= gdata.mx[2]+B; ++iz) {
		REAL Bcx_sum = 0.;  // new values for Cartesian Bx, By
		REAL Bcy_sum = 0.;  //  on the singular axis
		int N_terms(0);
			//		for (int j = 0; j <= gdata.mx[1]; j++) {
			// Do not use phi=0 twice!
		if(gdata.get_singularity_treatment(0) == 1) { // at lower r-bound
			for (int iy = 0; iy < jnum; iy++) {
				// br0_j, bp0_j: projected to (r,phi)=(0,phi_j) via avg.
				REAL br0_j = (+gdata.om[q_Bx]( 0,iy,  iz)
				              +gdata.om[q_Bx](-2,iy,  iz))/2.;
				REAL bp0_j = (+gdata.om[q_By]( 0,iy-1,iz)
				              +gdata.om[q_By]( 0,iy  ,iz)
				              +gdata.om[q_By](-1,iy-1,iz)
				              +gdata.om[q_By](-1,iy  ,iz))/4.;
				REAL phi = gdata.getCen(1, iy);
				Bcx_sum += br0_j * cos(phi) - bp0_j * sin(phi);
				Bcy_sum += br0_j * sin(phi) + bp0_j * cos(phi);
				// if(gdata.rank==0 && k==1) {
				// 	cout << " axis: " << j << " " << br0_j << " " << bp0_j << " " << Bcx_sum << " " << Bcy_sum << endl;
				// 	cout << "    " << gdata.om[q_By]( 0,j-1,k) << " ";
				// 	cout << gdata.om[q_By]( 0,j  ,k) << " ";
				// 	cout << gdata.om[q_By](-1,j  ,k) << " ";
				// 	cout << gdata.om[q_By](-1,j-1,k) << " ";
				// 	cout << endl;
				// }
				N_terms++;
			}

		} else {
			Bcx_sum = 0.;
			Bcy_sum = 0.;
			N_terms = 0.;
		}
#ifdef parallel
		gdata.MpiSum(2, Bcx_sum); // Sum only for ranks with same z
		gdata.MpiSum(2, Bcy_sum); // Sum only for ranks with same z
		gdata.MpiSum(2, N_terms); // Sum only for ranks with same z
#endif
		REAL Bcx_ave(Bcx_sum/N_terms);
		REAL Bcy_ave(Bcy_sum/N_terms);

		// if((gdata.rank==0 && k==gdata.mx[2]) ||
		//    (gdata.rank==1 && k==-1)) {
		// 	cout << setiosflags(ios::scientific) << setprecision(12);
		// 	cout << " corr val: " << gdata.rank << " " << Bcx_ave << " " <<	Bcy_ave << endl;
		// 	cout << resetiosflags(ios::scientific) << setprecision(6);
		// }

		if(gdata.get_singularity_treatment(0) == 1) { // at lower r-bound
			for (int iy = -B; iy <= gdata.mx[1]+B; iy++) { // set Br(0)s
				REAL phi = gdata.getCen(1, iy);
				REAL Br0_ave = ( Bcx_ave * cos(phi) +
				                 Bcy_ave * sin(phi) );
				gdata.om[q_Bx](-1,iy,iz) = Br0_ave;
			} // j
		} // check if r==0
	} // k
	// exit(3);


#endif
}
void Boundary::do_AxisValCorrectionSph(Data &gdata, ProblemType &Problem, bool top) {
	//! Correction for mag field values on polar axis

//#if (GEOM == SPHERICAL)
	// Check if singularity is on current rank
	if(gdata.get_singularity_treatment(2+top) == 1) {
		REAL cost0 = 1.-2.*top;      // cos({0|Pi}) = {+1|-1}
		int jS = top*(gdata.mx[1]+1)-1; // {-1|mxtet} (sing. index)
		for (int ix = -B; ix <= gdata.mx[0]+B; ++ix) {
			REAL Bcx_sum = 0.;  // new values for Cartesian Bx, By
			REAL Bcy_sum = 0.;  //  on the singular axis
			for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
				// bt0_k, bp0_k: projected to (tet,phi)=(0,phi_k) via avg.
				REAL bt0_k = (+gdata.om[q_By](ix,jS-1, iz  )
						+gdata.om[q_By](ix,jS+1, iz  ))*0.5;
				REAL bp0_k = (+gdata.om[q_Bz](ix,jS  , iz-1)
						+gdata.om[q_Bz](ix,jS  , iz  )
						+gdata.om[q_Bz](ix,jS+1, iz-1)
						+gdata.om[q_Bz](ix,jS+1, iz  ))*0.25;

				Bcx_sum += bt0_k * cost0 * cosPos[2](iz) - bp0_k * sinPos[2](iz);
				Bcy_sum += bt0_k * cost0 * sinPos[2](iz) + bp0_k * cosPos[2](iz);
			}
			int N_terms = gdata.mx[2]+1;

#ifdef parallel
			// ToDo: Communicate [Bc_sum, N_terms] as above
			gdata.MpiSum(0, Bcx_sum); // Sum only for ranks with same x
			gdata.MpiSum(0, Bcy_sum); // Sum only for ranks with same x
			gdata.MpiSum(0, N_terms); // Sum only for ranks with same x

#endif // [parallel]
			REAL Bcx_ave(Bcx_sum/N_terms);
			REAL Bcy_ave(Bcy_sum/N_terms);

			for (int iz = -B; iz <= gdata.mx[2]+B; iz++) { // set Bt(0)s
				REAL Bt0_ave = ( Bcx_ave * cost0 * cosPos[2](iz) +
						Bcy_ave * cost0 * sinPos[2](iz) );
				//         NB:  -Bcz_sum * sint0 ~ sin({0|Pi}) = 0
				gdata.om[q_By](ix,jS,iz) = Bt0_ave;
				// gdata.om[q_By](i,jS,k) = (gdata.om[q_By](i,jS-1,k)+
				//                           gdata.om[q_By](i,jS+1,k))/2.;
			} // k
		} // i
	} else { // Other ranks have to communicate, too
#ifdef parallel
		double dummy(0.);
		int dummy_int(0);
		for (int ix = -B; ix <= gdata.mx[0]+B; ++ix) {
			dummy = 0.;
			gdata.MpiSum(0, dummy); // Sum only for ranks with same x
			dummy = 0.;
			gdata.MpiSum(0, dummy); // Sum only for ranks with same x
			dummy_int = 0;
			gdata.MpiSum(0, dummy_int); // Sum only for ranks with same x
		}
#endif
	}

//#endif
}


void Boundary::apply_bc_outer(
			Data &gdata, ProblemType &problem,
			NumMatrix<REAL,3> &omb,
			int dir, int above, int q, int rim
			)
{
	int bcNum = dir * 2 + above;
	int bcType = bc_Type[bcNum];

	if(bcType > 1) {
		switch(bcType) {
			case 2: bc_Extrapolate(gdata, problem, omb, dir, above, q, rim); 	break;
			case 3: bc_Outflow(gdata, problem, omb, dir, above, q, rim); 		break;
			case 4: problem.bc_User(gdata, omb, dir, above, q, rim); 			break;
			case 5: bc_Reflecting(gdata, problem, omb, dir, above, q, rim);		break;
			case 6: bc_Axis(gdata, problem, omb, dir, above, q, rim); 			break;
			case 7: if(omb.get_bcTypeLow(dir) == 7) {
					} else {
						bc_Fixed(gdata, problem, omb, dir, above, q,rim);
					}; 															break;
			default: cerr << " No such boundary condition " <<  bcNum << endl; 	break;
		}
	}
}



void Boundary::bc_Extrapolate(Data &gdata, ProblemType &Problem,
                              NumMatrix<REAL,3> &omb,
                              int dir, int above, int q, int rim)
{

	string qname = omb.getName();

	if(dir == 0) {
		if(above == 0) {
			int imin = -1;
//			if(q == Problem.q_Bx) {
			if(qname=="B_x") {
				imin = -2;
			}
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; ++j){
					for (int i = imin; i >= -rim; --i){
						omb(i,j,k) = omb(i+1,j,k);    // x_min
						// omb(i,j,k) = 2.*omb(i+1,j,k) - omb(i+2,j,k);  // x_min
					}
				}
			}

		} else if (above == 1) {
			int imax = 1;
			if(omb.getName() == "A_y" || omb.getName() == "A_z") {
				imax = 2;
			}
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; j++){
					for (int i = imax; i <= +rim; i++){
						omb(gdata.mx[0]+i,j,k) = omb(gdata.mx[0]+i-1,j,k); //  x_max
						// omb(gdata.mx[0]+i,j,k) = 2.*omb(gdata.mx[0]+i-1,j,k) - omb(gdata.mx[0]+i-2,j,k); //  x_max
					}
				}
			}
		}
	}
	
	// if(q == 7) {
	// 	cout << " Dir: " << dir << " " << above << " " << omb(0,0,0) << endl;
	// }

	if(dir == 1) {
		if(above == 0) {
			int jmin = -1;
//			if(q == Problem.q_By) {
			if(qname=="B_y") {
				jmin = -2;
			}
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int j = jmin; j >= -rim; --j){
						omb(i,j,k) = omb(i,j+1,k);    // y_min
						// omb(i,j,k) = 2.*omb(i,j+1,k) - omb(i,j+2,k);    // y_min
						// if(i==396 && k==0) {
						// 	cout << endl << " ext: ";
						// 	cout << j << " ";
						// 	cout << omb(i,j,k) << " ";
						// 	cout << omb(i,j+1,k) << " ";
						// 	cout << omb(i,j+2,k) << endl;
						// }
					}
				}
			}
		} else if (above == 1) {
			int jmax = 1;
			if(omb.getName() == "A_x" || omb.getName() == "A_z") {
				jmax = 2;
			}
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int j = jmax; j <= rim; j++){
						omb(i,gdata.mx[1]+j,k) = omb(i,gdata.mx[1]+j-1,k); //  y_max
						// omb(i,gdata.mx[1]+j,k) = 2.*omb(i,gdata.mx[1]+j-1,k) - omb(i,gdata.mx[1]+j-2,k); //  y_max
					}
				}
			}
		}
	}


	if(dir == 2) {
		if(above == 0) {
			int kmin = -1;
//			if(q == Problem.q_Bz) {
			if(qname=="B_z") {
				kmin = -2;
			}
			for (int j = -rim; j <= gdata.mx[1]+rim; j++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int k = kmin; k >= -rim; --k){
						omb(i,j,k) = omb(i,j,k+1);     // z_min
						// omb(i,j,k) = 2.*omb(i,j,k+1) - omb(i,j,k+2);     // z_min
					}
				}
			}
		} else if (above == 1) {
			int kmax = 1;
			if(omb.getName() == "A_x" || omb.getName() == "A_y") {
				kmax = 2;
			}
			for (int j = -rim; j <= gdata.mx[1]+rim; j++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int k = kmax; k <= rim; k++){
						omb(i,j,gdata.mx[2]+k) = omb(i,j,gdata.mx[2]+k-1);    // z_max
						// omb(i,j,gdata.mx[2]+k) = 2.*omb(i,j,gdata.mx[2]+k-1) - omb(i,j,gdata.mx[2]+k-2);    // z_max
					}
				}
			}
		}
	}
}




void Boundary::bc_Outflow(Data &gdata, ProblemType &pr,
                          NumMatrix<REAL,3> &omb,
                          int dir, int above, int q, int rim)
{

	string qname = omb.getName();
	bool is_mag = false;
	if(qname=="B_x" || qname=="B_y" || qname=="B_z") {
		is_mag = true;
	}
	bool is_sx = false;
	if(qname=="s_x") {
		is_sx = true;
	}
	bool is_sy = false;
	if(qname=="s_y") {
		is_sy = true;
	}
	bool is_sz = false;
	if(qname=="s_z") {
		is_sz = true;
	}

	if(dir == 0) {
		if(above == 0) {
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; ++j){
					for (int i = -1; i >= -rim; --i){
						if(gdata.om[pr.q_sx](0,j,k) < 0.) {
//							if(q >= pr.q_Bx && q <= pr.q_Bz) {
							if(is_mag) {
								omb(i,j,k) = (3.*omb(i+1,j,k) -
								              3.*omb(i+2,j,k) +
								              omb(i+3,j,k));
							} else {
								omb(i,j,k) = omb(i+1,j,k);  // x_min
							}
						} else {
//							if(q == pr.q_sx) {
							if(is_sx) {
								omb(i,j,k) = -omb(-i-1,j,k);  // x_min
							} else {
								omb(i,j,k) =  omb(-i-1,j,k);  // x_min
							}
						}
					}
				}
			}
		} else if (above == 1) {
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; j++){
					for (int i = 1; i <= +rim; i++){
						if(gdata.om[pr.q_sx](gdata.mx[0],j,k) > 0.) {
//							if(q >= pr.q_Bx && q <= pr.q_Bz) {
							if(is_mag) {
								omb(gdata.mx[0]+i,j,k) = (3.*omb(gdata.mx[0]+i-1,j,k) -
								                          3.*omb(gdata.mx[0]+i-2,j,k) +
								                          omb(gdata.mx[0]+i-3,j,k));
							} else {
								omb(gdata.mx[0]+i,j,k) = omb(gdata.mx[0]+i-1,j,k); //  x_max
							}
						} else {
//							if(q == pr.q_sx) {
							if(is_sx) {
								omb(gdata.mx[0]+i,j,k) = -omb(gdata.mx[0]-i+1,j,k);
							} else {
								omb(gdata.mx[0]+i,j,k) =  omb(gdata.mx[0]-i+1,j,k);
							}
						}
					}
				}
			}
		}
	}


	if(dir == 1) {
		if(above == 0) {
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int j = -1; j >= -rim; --j){
						if(gdata.om[pr.q_sy](i,0,k) < 0.) {
//							if(q >= pr.q_Bx && q <= pr.q_Bz) {
							if(is_mag) {
								omb(i,j,k) = (3.*omb(i,j+1,k) -
								              3.*omb(i,j+2,k) +
								              omb(i,j+3,k));
							} else {
								omb(i,j,k) = omb(i,j+1,k);    // y_min
							}
						} else {
//							if(q == pr.q_sy) {
							if(is_sy) {
								omb(i,j,k) = -omb(i,-j-1,k);
							} else {
								omb(i,j,k) =  omb(i,-j-1,k);
							}
						}
					}
				}
			}
		} else if (above == 1) {
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int j = 1; j <= rim; j++){
						if(gdata.om[pr.q_sy](i,gdata.mx[1],k) > 0.) {
//							if(q >= pr.q_Bx && q <= pr.q_Bz) {
							if(is_mag) {
								omb(i,gdata.mx[1]+j,k) = (3.*omb(i,gdata.mx[1]+j-1,k) -
								                          3.*omb(i,gdata.mx[1]+j-2,k) +
								                          omb(i,gdata.mx[1]+j-3,k));
							} else {
								omb(i,gdata.mx[1]+j,k) = omb(i,gdata.mx[1]+j-1,k); //  y_max
							}
						} else {
//							if(q == pr.q_sy) {
							if(is_sy) {
								omb(i,gdata.mx[1]+j,k) = -omb(i,gdata.mx[1]-j+1,k);
							} else {
								omb(i,gdata.mx[1]+j,k) =  omb(i,gdata.mx[1]-j+1,k);
							}
						}
					}
				}
			}
		}
	}


	if(dir == 2) {
		int kmin = -1;
//		if(q == pr.q_Bz) {
		if(qname=="B_x") {
			kmin = -2;
		}
		if(above == 0) {
			for (int j = -rim; j <= gdata.mx[1]+rim; j++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int k = kmin; k >= -rim; --k){
						if(gdata.om[pr.q_sz](i,j,0) < 0.) {
//							if(q >= pr.q_Bx && q <= pr.q_Bz) {
							if(is_mag) {
								omb(i,j,k) = (3.*omb(i,j,k+1) -
								              3.*omb(i,j,k+2) +
								              omb(i,j,k+3));
							} else {
								omb(i,j,k) = omb(i,j,k+1);     // z_min
							}
						} else {
//							if(q == pr.q_sz) {
							if(is_sz) {
								omb(i,j,k) = -omb(i,j,-k-1);
							} else {
								omb(i,j,k) = omb(i,j,-k-1);
							}
						}
					}
				}
			}
			//       }
		} else if (above == 1) {
			for (int j = -rim; j <= gdata.mx[1]+rim; j++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int k = 1; k <= rim; k++){
						if(gdata.om[pr.q_sz](i,j,gdata.mx[2]) > 0.) {
//							if(q >= pr.q_Bx && q <= pr.q_Bz) {
							if(is_mag) {
								omb(i,j,gdata.mx[2]+k) = (3.*omb(i,j,gdata.mx[2]+k-1) -
								                          3.*omb(i,j,gdata.mx[2]+k-2) +
								                          omb(i,j,gdata.mx[2]+k-3));
							} else {
								omb(i,j,gdata.mx[2]+k) = omb(i,j,gdata.mx[2]+k-1);    // z_max
							}
						} else {
//							if(q == pr.q_sz) {
							if(is_sz) {
								omb(i,j,gdata.mx[2]+k) = -omb(i,j,gdata.mx[2]-k+1);
							} else {
								omb(i,j,gdata.mx[2]+k) = omb(i,j,gdata.mx[2]-k+1);
							}
						}
					}
				}
			}
		}
	}
}




void Boundary::bc_Reflecting(Data &gdata, ProblemType &pr,
                             NumMatrix<REAL,3> &omb,
                             int dir, int above, int q, int rim)
{
	/*
	  Reflecting boundaries: inversion of velocity at the
	  boundaries. All other quantities are extrapolated via zero
	  gradient.
	*/

	string qname = omb.getName();
	bool is_sx = false;
	if(qname=="s_x") {
		is_sx = true;
	}
	bool is_sy = false;
	if(qname=="s_y") {
		is_sy = true;
	}
	bool is_sz = false;
	if(qname=="s_z") {
		is_sz = true;
	}

	if(dir == 0) {
		if(above == 0) {
			int imin = -1;
//			if(q == pr.q_Bx) {
			if(qname=="B_x") {
				imin = -2;
			}
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; ++j){
					for (int i = imin; i >= -rim; --i){
//						if(q == pr.q_sx) {
						if(is_sx) {
							omb(i,j,k) = -omb(-i-1,j,k);  // x_min
						} else {
							omb(i,j,k) = omb(i+1,j,k);  // x_min
						}
					}
				}
			}
		} else if (above == 1) {
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int j = -rim; j <= gdata.mx[1]+rim; j++){
					for (int i = 1; i <= +rim; i++){
//						if(q == pr.q_sx) {
						if(is_sx) {
							omb(gdata.mx[0]+i,j,k) = -omb(gdata.mx[0]-i+1,j,k);
						} else {
							omb(gdata.mx[0]+i,j,k) =  omb(gdata.mx[0]+i-1,j,k);
						}
					}
				}
			}
		}
	}


	if(dir == 1) {
		if(above == 0) {
			int jmin = -1;
//			if(q == pr.q_By) {
			if(qname=="B_y") {
				jmin = -2;
			}
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int j = jmin; j >= -rim; --j){
//						if(q == pr.q_sy) {
						if(is_sy) {
							omb(i,j,k) = -omb(i,-j-1,k);
						} else {
							omb(i,j,k) =  omb(i,j+1,k);
						}
					}
				}
			}
		} else if (above == 1) {
			for (int k = -rim; k <= gdata.mx[2]+rim; k++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int j = 1; j <= rim; j++){
//						if(q == pr.q_sy) {
						if(is_sy) {
							omb(i,gdata.mx[1]+j,k) = -omb(i,gdata.mx[1]-j+1,k);
						} else {
							omb(i,gdata.mx[1]+j,k) =  omb(i,gdata.mx[1]+j-1,k);
						}
					}
				}
			}
		}
	}


	if(dir == 2) {
		int kmin = -1;
//		if(q == pr.q_Bz) {
		if(qname=="B_z") {
			kmin = -2;
		}
		if(above == 0) {
			for (int j = -rim; j <= gdata.mx[1]+rim; j++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int k = kmin; k >= -rim; --k){
//						if(q == pr.q_sz) {
						if(is_sz) {
							omb(i,j,k) = -omb(i,j,-k-1);
						} else {
							omb(i,j,k) = omb(i,j,k-1);
						}
					}
				}
			}
			//       }
		} else if (above == 1) {
			for (int j = -rim; j <= gdata.mx[1]+rim; j++){
				for (int i = -rim; i <= gdata.mx[0]+rim; i++){
					for (int k = 1; k <= rim; k++){
//						if(q == pr.q_sz) {
						if(is_sz) {
							omb(i,j,gdata.mx[2]+k) = -omb(i,j,gdata.mx[2]-k+1);
						} else {
							omb(i,j,gdata.mx[2]+k) = omb(i,j,gdata.mx[2]+k-1);
						}
					}
				}
			}
		}
	}
}



void Boundary::bc_Fixed(Data &gdata, ProblemType &pr,
                        NumMatrix<REAL,3> &omb,
                        int dir, int above, int q, int rim)
{
	/*
	  Fixed type boundary conditions. The initial values are stored
	  into an array and they are just copied into the boundaries for
	  every call of this bc.
	 */

	// Still missing some details: destinction between generic and
	// non-generic vars

	bool generic_vars = true;
#if (OMS_USER == TRUE)
	if(gdata.is_userField(omb)) {
		generic_vars = false;
	}
#endif

	if(dir == 0) {
		if(above == 0) {

			int imin[3] = {-rim, -rim, -rim};
			int imax[3] = {-1, gdata.mx[1]+rim, gdata.mx[2]+rim};

			// Do not overwrite B on face between computing domain and
			// ghost cells.
			if(q==pr.q_Bx) {
				imax[0] = -2;
			}

			if(generic_vars) {
				NumMatrix<REAL,3> & bcVals = bcVals_xb[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#if (OMS_USER == TRUE)
			} else {
				NumMatrix<REAL,3> & bcVals = bcVals_User_xb[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#endif
			}

		} else {

			int imin[3] = {gdata.mx[0]+1, -rim, -rim};
			int imax[3] = {gdata.mx[0]+rim, gdata.mx[1]+rim, gdata.mx[2]+rim};

			if(generic_vars) {
				NumMatrix<REAL,3> & bcVals = bcVals_xe[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#if (OMS_USER == TRUE)
			} else {
				NumMatrix<REAL,3> & bcVals = bcVals_User_xe[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#endif
			}

		}
	} else if(dir == 1) { // y-direction
		if(above == 0) {

			int imin[3] = {-rim, -rim, -rim};
			int imax[3] = {gdata.mx[0]+rim, -1, gdata.mx[2]+rim};

			// Do not overwrite B on face between computing domain and
			// ghost cells.
			if(q==pr.q_Bx) {
				imax[1] = -2;
			}

			if(generic_vars) {
				NumMatrix<REAL,3> & bcVals = bcVals_yb[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#if (OMS_USER == TRUE)
			} else {
				NumMatrix<REAL,3> & bcVals = bcVals_User_yb[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#endif
			}

		} else {

			int imin[3] = {-rim, gdata.mx[1]+1, -rim};
			int imax[3] = {gdata.mx[0]+rim, gdata.mx[1]+rim, gdata.mx[2]+rim};

			if(generic_vars) {
				NumMatrix<REAL,3> & bcVals = bcVals_ye[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#if (OMS_USER == TRUE)
			} else {
				NumMatrix<REAL,3> & bcVals = bcVals_User_ye[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#endif
			}

		}
	} else { // z-direction
		if(above == 0) {

			int imin[3] = {-rim, -rim, -rim};
			int imax[3] = {gdata.mx[0]+rim, gdata.mx[1]+rim, -1};

			// Do not overwrite B on face between computing domain and
			// ghost cells.
			if(q==pr.q_Bz) {
				imax[2] = -2;
			}

			if(generic_vars) {
				NumMatrix<REAL,3> & bcVals = bcVals_zb[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#if (OMS_USER == TRUE)
			} else {
				NumMatrix<REAL,3> & bcVals = bcVals_User_zb[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#endif
			}

		} else {

			int imin[3] = {-rim, -rim, gdata.mx[2]+1};
			int imax[3] = {gdata.mx[0]+rim, gdata.mx[1]+rim, gdata.mx[2]+rim};

			if(generic_vars) {
				NumMatrix<REAL,3> & bcVals = bcVals_ze[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#if (OMS_USER == TRUE)
			} else {
				NumMatrix<REAL,3> & bcVals = bcVals_User_ze[q];
				// Doing bcs:
				bc_Fixed_general(omb, bcVals, imin, imax);
#endif
			}

		}
	}


}

void Boundary::bc_Fixed_general(NumMatrix<REAL,3> &omb,
                                NumMatrix<REAL,3> &bcVals,
                                int imin[3], int imax[3]) {

	for (int iz = imin[2]; iz <= imax[2]; iz++){
		for (int iy = imin[1]; iy <= imax[1]; ++iy){
			for (int ix = imin[0]; ix <= imax[0]; ++ix){

				omb(ix,iy,iz) = bcVals(ix,iy,iz);

			}
		}
	}
}


