#include <iomanip>
#include <sstream>
#include "gridgen.H"
#include "data.H"
#include "problem.H"
// #include "Hdf5File.H"

/*-----------------------------------------------------------------
  Simplified matrix class
  -----------------------------------------------------------------*/

Pot::Pot() {
	//FieldType = "undetermined";
}

Pot::Pot(const int *mx)
{
	// FieldType = "undetermined";
	resize(mx);
}

Pot::Pot(const NumMatrix<DATA_TYPE, DIM>& _matr)
	: NumMatrix<DATA_TYPE, DIM>(_matr) {
	// FieldType = "undetermined";
}

void Pot::resize(const int *mx)
{
	int l[DIM], h[DIM];
	for (int d = 0; d < DIM; d++) {
		l[d] = -B;
		h[d] = mx[d]+B;
	}
	NumMatrix<DATA_TYPE,DIM>::resize(l,h);
}


void Pot::resize(const int *mx, int rim)
{
	int l[DIM], h[DIM];
	for (int d = 0; d < DIM; d++) {
		l[d] = -rim;
		h[d] = mx[d]+rim;
	}
	NumMatrix<DATA_TYPE,DIM>::resize(l,h);
}


void Pot::resize_ll(const int *ib, const int *ie)
{
	const int* ib_old = NumMatrix<DATA_TYPE,DIM>::getLow();
	const int* ie_old = NumMatrix<DATA_TYPE,DIM>::getHigh();

	int ib_loc[DIM], ie_loc[DIM];

	for(int i=0; i<DIM; ++i) {
		ib_loc[i]=std::max(ib_old[i], ib[i]);
		ie_loc[i]=std::min(ie_old[i], ie[i]);
	}

	NumMatrix<DATA_TYPE, DIM> matr_back(*this);

	for (int i = 0; i < size; i++)
		matr_back[i] = matr[i];

	NumMatrix<DATA_TYPE,DIM>::resize(ib,ie);

	if(DIM == 3) {
		for (int k=ib_loc[2]; k<=ie_loc[2]; ++k) {
			for (int j=ib_loc[1]; j<=ie_loc[1]; ++j) {
				for (int i=ib_loc[0]; i<=ie_loc[0]; ++i) {
					NumMatrix<DATA_TYPE,DIM>::operator()(i,j,k) = matr_back(i,j,k);
				}
			}
		}
	} else if (DIM == 2) {
		for (int j=ib_loc[1]; j<=ie_loc[1]; ++j) {
			for (int i=ib_loc[0]; i<=ie_loc[0]; ++i) {
				NumMatrix<DATA_TYPE,DIM>::operator()(i,j) = matr_back(i,j);
			}
		}
	} else if (DIM == 1) {
		for (int i=ib_loc[0]; i<=ie_loc[0]; ++i) {
			NumMatrix<DATA_TYPE,DIM>::operator()(i) = matr_back(i);
		}
	} else {
		cerr << " resize_ll not implemented for DIM > " << DIM << endl;
		exit(-3);
	}


}


void Pot::set_max(const double &val)
{
	for(int i=0; i < this->size; i++) {
		if(this->matr[i] > val) {
			this->matr[i] = val;
		}
	}
	return;
}


REAL Pot::get_max()
{
	REAL maximum = static_cast<REAL>(this->matr[0]);
	for(int i=0; i<this->size; i++) {
		maximum = std::max(maximum, static_cast<REAL>(this->matr[i]));
	}
	return maximum;
}


Pot& Pot::operator=(const double& val)
{
	for (int i = 0; i < size; i++)
		matr[i] = val;
	return *this;
}

void Pot::set_min(const double &val)
{
	for(int i=0; i < this->size; i++) {
		if(this->matr[i] < val) {
			this->matr[i] = val;
		}
	}
	return;
}


REAL Pot::get_min()
{
	REAL minimum = static_cast<REAL>(this->matr[0]);
	for(int i=0; i<this->size; i++) {
		minimum = std::min(minimum, static_cast<REAL>(this->matr[i]));
	}
	return minimum;
} 


// void Pot::set_FieldType(string name) {
// 	this->FieldType = name;
// }

// string Pot::get_FieldType() {
// 	// Get name of variable used in program
// 	return FieldType;
// }

/*-----------------------------------------------------------------
  Data container -- in current version just holding the actual data
  -----------------------------------------------------------------*/


Data::Data()
{

	// Start runtime timer
	gettimeofday(&tick_start, 0);

//	cout << " Making new class " << endl << endl << endl;
#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	int nFluids = NUM_FLUIDS;
	fluids = new CronosMultifluid(nFluids, ENERGETICS, OMS_USER);
#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	fluids->set_dualEnergy(AUX_ENERGY);
#else
	fluids->unset_dualEnergy();
#endif
	fluids->setup();
	fluids->compute_Variables(N_ADD, N_SUBS);
	int N_OM = fluids->get_N_OM();
	int n_omInt = fluids->get_N_OMINT();
	int n_omUser = fluids->get_N_OM_USER();
#else
#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	fluid.set_dualEnergy(AUX_ENERGY);
#else
	fluid.unset_dualEnergy();
#endif
	fluid.setup(FLUID_TYPE, ENERGETICS, N_ADD, N_SUBS, N_OMINT_USER);
	int N_OM = fluid.get_N_OM();
	int n_omInt = fluid.get_N_OMINT();
	int n_omUser = fluid.get_N_OM_USER();
//	N_OM = 8;
#endif

	om = new Pot[N_OM+N_P];
	om_old = new Pot[N_OM+N_P];

	nom = new NumMatrix<REAL,3> [n_omInt];

	for (int q = 0; q < n_omInt; ++q) {
		nom[q].resize(Index::set(0,0,0), Index::set(mx[0],mx[1],mx[2]));
	}

//	exit(3);
//	om.clear();
//	om.reserve(N_OM+N_P);
	for(int q=0; q<N_OM+N_P; ++q) {
//		cout << " kuh " << q << endl;
		// Make om fields
//		om.push_back(Pot(mx));
//		cout << " next " << endl;
		om[q].clear();
		om_old[q].clear();
		// om[q].set_VariableName("om");
	}

#if (OMS_USER == TRUE)
#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	int N_OM_USER = fluids->get_N_OM_USER();
	int n_omIntUser = gdata.fluids->get_N_OMINT_USER();
#else
	int n_omIntUser = fluid.get_N_OMINT_USER();
#endif

	om_user = new Pot[n_omUser];
//	nom_user = new NumMatrix<REAL,3> [n_omUser];
	nom_user = new NumMatrix<REAL,3> [n_omIntUser];

//	for (int q = 0; q < n_omUser; ++q) {
	for (int q = 0; q < n_omIntUser; ++q) {
		nom_user[q].resize(Index::set(0,0,0), Index::set(mx[0],mx[1],mx[2]));
	}
	for(int q=0; q<n_omUser; ++q) {
		om_user[q].clear();
		// om_user[q].set_VariableName("om_user");
	}
#else
	nom_user = NULL;
#endif
  
	
	t_end = value((char*)"t_end");
	dt = value((char*)"dt");
	cfl = 0.;
	tstep = 0; // Timestep

// #ifdef parallel
// 	GetMpi();
// #else
// #ifndef parallel
// 	rank = 0;
// 	for(int i=0; i<DIM; ++i) {
// 		coords[i] = 0;
// 	}
// #endif


	for (int ll=0; ll<N_OM+N_P; ++ll) {
		string name = "om";
		char cll[255];
		sprintf(cll,"%2.2d",ll);
		name += cll;
    
//		om.push_back(new Pot(mx));
		om[ll].resize(mx);
		om[ll].rename(name);
		om[ll].clear();

		om_old[ll].resize(mx);
		om_old[ll].rename(name);
		om_old[ll].clear();
	}

#if (OMS_USER == TRUE)
    for (int q=0; q<n_omUser; ++q) {
		string name = "om_user";
		char cll[255];
		sprintf(cll,"%2.2d",q);
		name += cll;
    
		om_user[q].resize(mx);
		om_user[q].rename(name);
		om_user[q].clear();
    }
#endif

#if (GEOM == SPHERICAL)
    massFlux.resize(Index::set(0), Index::set(mx[0]+1));

    // Fill with corresponding area
    radialArea = 0.;
    for(int iPhi=0; iPhi<=mx[2]; ++iPhi) {
    	for(int iTheta=0; iTheta<=mx[1]; ++iTheta) {
#if (NON_LINEAR_GRID == CRONOS_OFF)
    		REAL f_geom_x = h2(0.5,iTheta,iPhi)/h1(0.5,iTheta,iPhi);
#else
    		REAL f_geom_x = h2(0,iTheta,iPhi,1,0,0)/h1(0,iTheta,iPhi,1,0,0);
#endif
    		radialArea += f_geom_x*getCen_dx(1,iTheta)*getCen_dx(2,iPhi);
    	}
    }

#ifdef parallel
    double radialArea_global;
	MPI_Reduce(&radialArea, &radialArea_global, 1, MPI_DOUBLE,
			MPI_SUM, 0, comm_constx);
	MPI_Bcast(&radialArea_global, 1, MPI_DOUBLE, 0, comm_constx);
//	MPI_Barrier(comm_constx);
	radialArea = radialArea_global;
#endif

#endif

	// Check whether thermal pressure is needed
    if(value_exists("use_carbuncleFlag")) {
    	use_carbuncleFlag = true;
    } else {
    	use_carbuncleFlag = false;
    }
	if(use_carbuncleFlag) {
		storePressure = true;
		// Resize pressure matrix
		pTherm.resize(Index::set(-B,-B,-B), Index::set(mx[0]+B,mx[1]+B,mx[2]+B));
		carbuncleFlag.resize(Index::set(-B,-B,-B), Index::set(mx[0]+B,mx[1]+B,mx[2]+B));

	} else {
		storePressure = false;
	}

	// Check whether a flag is needed to store the admissibility of states
	if(value_exists("write_admissibleIndicator")) {
		write_admissibleIndicator = true;
	} else {
		write_admissibleIndicator = false;
	}
	if(write_admissibleIndicator) {
		admissibleIndicator.resize(Index::set(-B,-B,-B), Index::set(mx[0]+B,mx[1]+B,mx[2]+B));
	}

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	entropyFlag.resize(Index::set(-B,-B,-B), Index::set(mx[0]+B,mx[1]+B,mx[2]+B));
	entropySwitch2.resize(Index::set(-B,-B,-B), Index::set(mx[0]+B,mx[1]+B,mx[2]+B));
	entropySwitch3.resize(Index::set(-B,-B,-B), Index::set(mx[0]+B,mx[1]+B,mx[2]+B));
	
	entropyFlag.clear();
	entropySwitch2.clear();
	entropySwitch3.clear();
#endif

    mag = value((char*)"mag");
    time = 0.;

#if (FLUID_TYPE==CRONOS_RHD && PARTICLE_TRANSPORT == TRUE)
	lorentz.resize(mx);
	lorentz.rename("lorentz");
	lorentz.clear();
	h5FileAuxFields.push_back(&lorentz);

	injection.resize(mx);
	injection.rename("injection");
	injection.clear();
	h5FileAuxFields.push_back(&injection);

	du_cov.resize(mx);
	du_cov.rename("du_cov");
	du_cov.clear();
	h5FileAuxFields.push_back(&du_cov);

	du_cov2.resize(mx);
	du_cov2.rename("du_cov2");
	du_cov2.clear();
	h5FileAuxFields.push_back(&du_cov2);

	sync_fac.resize(mx);
	sync_fac.rename("sync_fac");
	sync_fac.clear();
	h5FileAuxFields.push_back(&sync_fac);

	ic_fac.resize(mx);
	ic_fac.rename("ic_fac");
	ic_fac.clear();
	h5FileAuxFields.push_back(&ic_fac);

	lorentz_old.resize(mx);
	lorentz_old.rename("lorentz_old");
	lorentz_old.clear();

	gammaMin.resize(mx);
	gammaMin.rename("gammaMin");
	gammaMin.clear();
	h5FileAuxFields.push_back(&gammaMin);

	gammaMax.resize(mx);
	gammaMax.rename("gammaMax");
	gammaMax.clear();
	h5FileAuxFields.push_back(&gammaMax);

	injectionK.resize(mx);
	injectionK.rename("injectionK");
	injectionK.clear();
	h5FileAuxFields.push_back(&injectionK);

  fluidSpecEnergy.resize(mx);
  fluidSpecEnergy.rename("fluidSpecEnergy");
  fluidSpecEnergy.clear();
  h5FileAuxFields.push_back(&fluidSpecEnergy);

  powerlawSpecEnergy.resize(mx);
  powerlawSpecEnergy.rename("powerlawSpecEnergy");
  powerlawSpecEnergy.clear();
  h5FileAuxFields.push_back(&powerlawSpecEnergy);

  injectionGammaMaxwellian.resize(mx);
  injectionGammaMaxwellian.rename("injectionGammaMaxwellian");
  injectionGammaMaxwellian.clear();
  h5FileAuxFields.push_back(&injectionGammaMaxwellian);

	magneticField_LAB.resize(mx);
	magneticField_LAB.rename("magneticField_LAB");
	magneticField_LAB.clear();
	h5FileAuxFields.push_back(&magneticField_LAB);

	magneticField_COM.resize(mx);
	magneticField_COM.rename("magneticField_COM");
	magneticField_COM.clear();
	h5FileAuxFields.push_back(&magneticField_COM);

	negativeParticles.resize(mx);
	negativeParticles.rename("negativeParticles");
	negativeParticles.clear();
	h5FileAuxFields.push_back(&negativeParticles);

//	momentumLoss_c1_int.resize(mx);
//	momentumLoss_c1_int.rename("momentumLoss_c1_int");
//	momentumLoss_c1_int.clear();
//
//	momentumLoss_c2.resize(mx);
//	momentumLoss_c2.rename("momentumLoss_c2");
//	momentumLoss_c2.clear();
//
//	momentumLoss_c2_old.resize(mx);
//	momentumLoss_c2_old.rename("momentumLoss_c2_old");
//	momentumLoss_c2_old.clear();
#endif
}

Data::~Data() {
	delete [] om;
	delete [] om_old;
	delete [] nom;
#if (OMS_USER == TRUE)
	delete [] om_user;
	delete [] nom_user;
#endif

	// Timer for code finalisation
	gettimeofday(&tock_end, 0);

	double full_time_sec = ((tock_end.tv_sec + tock_end.tv_usec/1.e6) -
			(tick_start.tv_sec + tick_start.tv_usec/1.e6));
	double time_per_step = full_time_sec/(1.*tstep);
	if(rank==0) {
		cout << "======================================================" << endl;
		cout << " Total run time               " << git_humanReadable(full_time_sec) << endl;
		cout << " Average time step duration   " << git_humanReadable(time_per_step) << endl;
		cout << "======================================================" << endl;
	}
}



std::string Data::git_humanReadable(double t_sec) {
	int t_int_sec = static_cast<int>(t_sec);
	int full_time_min = t_int_sec/60;
	int full_time_hours = full_time_min/60;
	int full_time_days = full_time_hours/24;
	double full_time_sec = t_sec - 60*full_time_min;
	full_time_min -= 60*full_time_hours;
	full_time_hours -= 24*full_time_days;

	std::ostringstream t_string_d;
	t_string_d << full_time_days << "d:" << full_time_hours << "h:" << full_time_min << "m:" << full_time_sec << "s";
	std::ostringstream t_string_h;
	t_string_h << full_time_hours << "h:" << full_time_min << "m:" << full_time_sec << "s";
	std::ostringstream t_string_m;
	t_string_m << full_time_min << "m:" << full_time_sec << "s";
	std::ostringstream t_string_s;
	t_string_s << full_time_sec << "s";
	std::string t_string = t_string_s.str();
	if(full_time_days > 0) {
		t_string = t_string_d.str();
	} else if (full_time_hours > 0) {
		t_string = t_string_h.str();
	} else if (full_time_min > 0) {
		t_string = t_string_m.str();
	}
	return t_string;
}



#ifdef parallel



void Data::MpiSum(double &variable) const {
	double variableGlobal = variable;
	MPI_Reduce(&variable, &variableGlobal, 1, MPI_DOUBLE,
	           MPI_SUM, 0, comm3d);
	MPI_Bcast(&variableGlobal, 1, MPI_DOUBLE, 0, comm3d);
//	MPI_Barrier(comm3d);
	variable = variableGlobal;
}


void Data::MpiSum(int &variable) const {
	int variableGlobal = variable;
	MPI_Reduce(&variable, &variableGlobal, 1, MPI_INT,
	           MPI_SUM, 0, comm3d);
	MPI_Bcast(&variableGlobal, 1, MPI_INT, 0, comm3d);
//	MPI_Barrier(comm3d);
	variable = variableGlobal;
}


void Data::MpiSum(REAL variable[3]) const {
	REAL variableGlobal[3] = {variable[0], variable[1], variable[2]};
	MPI_Reduce(&variable[0], &variableGlobal[0], 3, MPI_DOUBLE,
	           MPI_SUM, 0, comm3d);
	MPI_Bcast(&variableGlobal[0], 3, MPI_DOUBLE, 0, comm3d);
//	MPI_Barrier(comm3d);
	variable[0] = variableGlobal[0];
	variable[1] = variableGlobal[1];
	variable[2] = variableGlobal[2];
}


void Data::MpiSum(REAL variable[], int num) const {
	REAL variableGlobal[num];
	for(int i=0; i<num; ++i) {
		variableGlobal[i] = variable[i];
	}
//	MPI_Barrier(comm3d);
	MPI_Reduce(&variable[0], &variableGlobal[0], num, MPI_DOUBLE,
	           MPI_SUM, 0, comm3d);
	MPI_Bcast(&variableGlobal[0], num, MPI_DOUBLE, 0, comm3d);
//	MPI_Barrier(comm3d);
	for(int i=0; i<num; ++i) {
		variable[i] = variableGlobal[i];
	}
}

void Data::MpiSum(int dir, int &variable) const {
	int variableGlobal = variable;
	if(dir == 0) { // Use plane in x-dir:
		MPI_Reduce(&variable, &variableGlobal, 1, MPI_INT,
		           MPI_SUM, 0, comm_constx);
 		MPI_Bcast(&variableGlobal, 1, MPI_INT, 0, comm_constxy);
//		MPI_Barrier(comm_constx);
	} else if(dir == 2) { // Use plane in z-dir:
		MPI_Reduce(&variable, &variableGlobal, 1, MPI_INT,
		           MPI_SUM, 0, comm_constz);
		MPI_Bcast(&variableGlobal, 1, MPI_INT, 0, comm_constz);
//		MPI_Barrier(comm_constz);
	}
	variable = variableGlobal;
}

void Data::MpiSum(int dir, double &variable) const {
	double variableGlobal = variable;
	if(dir == 0) { // Use plane in x-dir:
		MPI_Reduce(&variable, &variableGlobal, 1, MPI_DOUBLE,
		           MPI_SUM, 0, comm_constx);
 		MPI_Bcast(&variableGlobal, 1, MPI_DOUBLE, 0, comm_constxy);
//		MPI_Barrier(comm_constx);
	} else if(dir == 2) { // Use plane in z-dir:
		MPI_Reduce(&variable, &variableGlobal, 1, MPI_DOUBLE,
		           MPI_SUM, 0, comm_constz);
		MPI_Bcast(&variableGlobal, 1, MPI_DOUBLE, 0, comm_constz);
//		MPI_Barrier(comm_constz);
	}
	variable = variableGlobal;
}


void Data::MpiAvergage(double & value) const {
	MPI_Allreduce(MPI_IN_PLACE, &value, 1, MPI_DOUBLE, MPI_SUM, comm3d);
	int size;
	MPI_Comm_size(comm3d, &size);
	value /= double(size);
}

void Data::MpiMax(double &variable) const {
	MPI_Allreduce(MPI_IN_PLACE, &variable, 1, MPI_DOUBLE, MPI_MAX, comm3d);
}

#endif


double Data::computeInt(int q)
{
	double sum = 0.;
	REAL dV(0.);
#if (NON_LINEAR_GRID == CRONOS_OFF)
	dV = get_CellVolume(0,0,0);
#endif
	for(int k=0; k<numCellsEff[2]; ++k) {
		for(int j=0; j<numCellsEff[1]; ++j) {
			for(int i=0; i<numCellsEff[0]; ++i) {
#if (GEOM != CARTESIAN || NON_LINEAR_GRID == CRONOS_ON)
				dV = get_CellVolume(i,j,k);
#endif
				sum += om[q](i,j,k)*dV;
			}
		}
	}
#ifdef parallel
	MpiSum(sum);
#endif
	return sum;
}


double Data::computeRMS(int q)
{
	double sum(0.);
	double dV(dx[0]*dx[1]*dx[2]);
	double Vol(0.);
	for(int k=0; k<numCellsEff[2]; ++k) {
		for(int j=0; j<numCellsEff[1]; ++j) {
			for(int i=0; i<numCellsEff[0]; ++i) {
				sum += sqr(om[q](i,j,k))*dV;
				Vol += dV;
			}
		}
	}
#ifdef parallel
	MpiSum(sum);
	MpiSum(Vol);
#endif
	sum /= Vol;
	return sqrt(sum);
}


NumMatrix<double,3> Data::computeAbs(int qmin, int qmax) const
{
	int ibeg[3] = {om[qmin].getLow(0), om[qmin].getLow(1), om[qmin].getLow(2)};
	int iend[3] = {om[qmin].getHigh(0), om[qmin].getHigh(1), om[qmin].getHigh(2)};
  
	NumMatrix<double,3> Abs;
	Abs.resize(Index::set(ibeg[0],ibeg[1],ibeg[2]),
	           Index::set(iend[0],iend[1],iend[2]));
	Abs.clear();

	for(int q=qmin; q<=qmax; ++q) {
		for(int k=ibeg[2]; k<=iend[2]; ++k) {
			for(int j=ibeg[1]; j<=iend[1]; ++j) {
				for(int i=ibeg[0]; i<=iend[0]; ++i) {
					Abs(i,j,k) += sqr(om[q](i,j,k));
				}
			}
		}
	}

	for(int k=ibeg[2]; k<=iend[2]; ++k) {
		for(int j=ibeg[1]; j<=iend[1]; ++j) {
			for(int i=ibeg[0]; i<=iend[0]; ++i) {
				Abs(i,j,k) = sqrt(Abs(i,j,k));
			}
		}
	}

	return Abs;
}


double Data::getMin(int q)
{
	int ibeg[3] = {om[q].getLow(0), om[q].getLow(1), om[q].getLow(2)};
	int iend[3] = {om[q].getHigh(0), om[q].getHigh(1), om[q].getHigh(2)};
	double minimum(abs(om[q](ibeg[0],ibeg[1],ibeg[2])));

	for(int k=ibeg[2]; k<=iend[2]; ++k) {
		for(int j=ibeg[1]; j<=iend[1]; ++j) {
			for(int i=ibeg[0]; i<=iend[0]; ++i) {

				minimum = std::min(minimum, om[q](i,j,k));

			}
		}
	}

	return minimum;
}


double Data::getMin(int pos[3], int q)
{
	int ibeg[3] = {om[q].getLow(0), om[q].getLow(1), om[q].getLow(2)};
	int iend[3] = {om[q].getHigh(0), om[q].getHigh(1), om[q].getHigh(2)};
	pos[0] = ibeg[0];
	pos[1] = ibeg[1];
	pos[2] = ibeg[2];
	double minimum(abs(om[q](ibeg[0],ibeg[1],ibeg[2])));

	for(int k=ibeg[2]; k<=iend[2]; ++k) {
		for(int j=ibeg[1]; j<=iend[1]; ++j) {
			for(int i=ibeg[0]; i<=iend[0]; ++i) {

				if(om[q](i,j,k) < minimum) {
					minimum = om[q](i,j,k);
					pos[0] = i;
					pos[1] = j;
					pos[2] = k;
				}
			}
		}
	}

	return minimum;
}


double Data::getMax(int q)
{
	int ibeg[3] = {om[q].getLow(0), om[q].getLow(1), om[q].getLow(2)};
	int iend[3] = {om[q].getHigh(0), om[q].getHigh(1), om[q].getHigh(2)};
	double maximum(abs(om[q](ibeg[0],ibeg[1],ibeg[2])));

	for(int k=ibeg[2]; k<=iend[2]; ++k) {
		for(int j=ibeg[1]; j<=iend[1]; ++j) {
			for(int i=ibeg[0]; i<=iend[0]; ++i) {

				maximum = std::max(maximum, om[q](i,j,k));

			}
		}
	}

	return maximum;
}



double Data::getMax(int pos[3], int q)
{
	int ibeg[3] = {om[q].getLow(0), om[q].getLow(1), om[q].getLow(2)};
	int iend[3] = {om[q].getHigh(0), om[q].getHigh(1), om[q].getHigh(2)};
	pos[0] = ibeg[0];
	pos[1] = ibeg[1];
	pos[2] = ibeg[2];
	double maximum(abs(om[q](ibeg[0],ibeg[1],ibeg[2])));

	for(int k=ibeg[2]; k<=iend[2]; ++k) {
		for(int j=ibeg[1]; j<=iend[1]; ++j) {
			for(int i=ibeg[0]; i<=iend[0]; ++i) {

				if(om[q](i,j,k) > maximum) {
					maximum = om[q](i,j,k);
					pos[0] = i;
					pos[1] = j;
					pos[2] = k;
				}
			}
		}
	}

	return maximum;
}






NumMatrix<float,3> Data::float_data(int q, int R, bool generic)
{

	NumMatrix<float,3> datafloat(Index::set(-R,-R,-R),
	                             Index::set(mx[0]+R,mx[1]+R,mx[2]+R));
  
#if (OMS_USER == TRUE)
	if(generic) {
#endif
		for (int k = -R; k <= mx[2]+R; k++) {
			for (int j = -R; j <= mx[1]+R; j++) {
				for (int i = -R; i <= mx[0]+R; i++) {
					datafloat(i,j,k) = (q < 0) ? 0. : static_cast<float>(om[q](i,j,k));
				}
			}
		}
#if (OMS_USER == TRUE)
	} else {
		for (int k = -R; k <= mx[2]+R; k++) {
			for (int j = -R; j <= mx[1]+R; j++) {
				for (int i = -R; i <= mx[0]+R; i++) {
					datafloat(i,j,k) = (q < 0) ? 0. : static_cast<float>(om_user[q](i,j,k));
				}
			}
		}
	}
#endif
	return datafloat;

}



void Data::floatom_out(int R, Hdf5Stream & h5file, int om_lo, int om_hi)
{
	double xbvar[3];
	xbvar[0] = xb[0]-dx[0]*R;
	xbvar[1] = xb[1]-dx[1]*R;
	xbvar[2] = xb[2]-dx[2]*R;
	//   string omName;
	NumMatrix<float,3> data(Index::set(-R,-R,-R),
	                        Index::set(mx[0]+R,mx[1]+R,mx[2]+R));

	for (int q = om_lo; q <= om_hi; q++) {
		for (int k = -R; k <= mx[2]+R; k++) {
			for (int j = -R; j <= mx[1]+R; j++) {
				for (int i = -R; i <= mx[0]+R; i++) {
					data(i,j,k) = (q < 0) ? 0. : (float)om[q](i,j,k);
				}
			}
		}
		if(!h5file.Write3DMatrix(om[q].getName(), data, xbvar, dx)){
			cerr << " Writing not successful for om: " << om[q].getName() << endl;
			exit(-15);
		}
	}
}



void Data::CheckNan(int pos)
{
	char message[255];
	sprintf(message,"Check: %2.2d om is NAN",pos);
#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	int N_OMINT = fluids->get_N_OMINT();
#endif

	int found_NAN = 0;
	int qNAN(-10), iNAN(-10), jNAN(-10), kNAN(-10);

	for (int q = 0; q < N_OMINT; ++q){

		int lo[3], up[3];
		lo[0] = om[q].getLow(0);
		lo[1] = om[q].getLow(1);
		lo[2] = om[q].getLow(2);
    
		up[0] = om[q].getHigh(0);
		up[1] = om[q].getHigh(1);
		up[2] = om[q].getHigh(2);
      
		for (int k = lo[2]; k <= up[2]; k++){
			for (int j = lo[1]; j <= up[1]; j++){
				for (int i = lo[0]; i <= up[0]; i++){
					if(std::isnan(om[q](i,j,k)) && !found_NAN) {
						found_NAN = 1;
						qNAN = q;
						iNAN = i;
						jNAN = j;
						kNAN = k;
					}
				}
			}
		}
	}
#ifdef parallel
	int found_NAN_global;
	MPI_Allreduce(&found_NAN, &found_NAN_global, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
	found_NAN = found_NAN_global;
//	if(found_NAN) {
//		int qNAN_global, iNAN_global, jNAN_global, kNAN_global;
//
//		MPI_Allreduce(&qNAN, &qNAN_global, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
//		MPI_Allreduce(&iNAN, &iNAN_global, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
//		MPI_Allreduce(&jNAN, &jNAN_global, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
//		MPI_Allreduce(&kNAN, &kNAN_global, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
//
//		qNAN = qNAN_global;
//		iNAN = iNAN_global;
//		jNAN = jNAN_global;
//		kNAN = kNAN_global;
//	}
#endif
	if(found_NAN) {
		throw CException(message,pos,qNAN,iNAN,jNAN,kNAN);
	}
}



void Data::set_fieldIds() {
	// std::map<NumMatrix<REAL,3> *,int> id_numbers;



#if(FLUID_TYPE==CRONOS_MULTIFLUID)
	int n_omInt = fluids->get_N_OMINT();
	int n_Omega = fluids->get_N_OMEGA();

	// Loop over normal fields
	for(int q=0; q<n_omInt; ++q) {
		string name = om[q].getName();
		int iFluid = fluids->get_FluidIndex(q);
		string nameFluid = fluids->fluids[iFluid].get_Name();
		name = nameFluid + "/" + name;
		cout << " name: " << name << " " << iFluid << endl;
		om_ids.enter_fieldId(name);
	}

	// Loop over extra-fields:
	for(int q=n_omInt; q<n_omInt+N_ADD+N_SUBS; ++q) {
		string name = om[q].getName();
		cout << " name var " << name << endl;
		om_ids.enter_fieldId(name);
	}

	// Loop over storage fields
	for(int q=0; q<n_omInt; ++q) {
		string name = om[q+n_Omega].getName();
		int iFluid = fluids->get_FluidIndex(q);
		string nameFluid = fluids->fluids[iFluid].get_Name();
		name = nameFluid + "/" + name;
		cout << " name var var " << name << endl;
		om_ids.enter_fieldId(name);
	}

#else
	for(int q=0; q<N_OM+N_P; ++q) {
		string name = om[q].getName();
		om_ids.enter_fieldId(name);
	}
#endif


#if (OMS_USER == TRUE)
#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	int N_OM_USER = fluids->get_N_OM_USER();
#endif
	for(int q=0; q<N_OM_USER; ++q) {

		om_ids.enter_fieldId(om_user[q].getName());

	}
#endif

	// Print map to screen
	if(rank == 0) {
		cout << "------- Field ids-------------------------------------";
		cout << endl;
		om_ids.print_map();
		// cout << "------------------------------------------------------" << endl;
		// cout << endl;
	}

}

bool Data::is_userField(NumMatrix<REAL,3> &omField) {
	
	if(om_ids.get_fieldId(omField) >= om_ids.max_generic && 
	   om_ids.get_fieldId(omField) < om_ids.fieldIds.size()) {
		return true;
	} else {
		return false;
	}

}



id_handler::id_handler() {
	max_generic = N_OM+N_P;
}


void id_handler::enter_fieldId(string FieldName) {
	// Add entry to fieldIds map - only if not already used:
	unsigned int used = fieldIds.count(FieldName);

	if(used) {
		
		typedef map<std::string, unsigned int> stringType;
		stringType::iterator stringIter;
		stringIter = fieldIds.find(FieldName);

		cerr << FieldName << " already in uses as ";
		cerr << stringIter->second;
		cerr << endl;
		exit(-3);

	} else {

		unsigned int id_number = fieldIds.size();
		fieldIds[FieldName] = id_number;

	}
	

}



bool id_handler::is_generic(NumMatrix<REAL,3> &omField) {
	
	if(get_fieldId(omField) < max_generic) {
		return true;
	} else {
		return false;
	}

}



unsigned int id_handler::get_userFieldId(NumMatrix<REAL,3> &omField) {
	return get_fieldId(omField)-(max_generic);
}

unsigned int id_handler::get_fieldId(NumMatrix<REAL,3> &omField) {

	// get unique id number for field
	typedef map<std::string, unsigned int> id_map;
	id_map::iterator idIter;
	idIter = fieldIds.find(omField.getName());

	return idIter->second;

}


void id_handler::print_map() {

	typedef map<std::string, unsigned int>::iterator idIter;

	for(idIter id=fieldIds.begin(); id!=fieldIds.end(); ++id) {
		// cout << id->second << " -> " << id->first << endl;
		cout << setiosflags(ios::left);
		cout << setw(9) << id->first << " -> ";
		cout << resetiosflags(ios::left);
		cout << setiosflags(ios::right )<< setw(2) << id->second << endl;
		cout << resetiosflags(ios::right);
	}
	// cout << resetiosflags(ios::left);


}
