#ifndef CRONOS_DATA_H
#define CRONOS_DATA_H = 1

#ifdef parallel
#include "mpi.h"
#endif

#include <map>
#include <time.h>

#include "matrix.H"
#include "Hdf5File_cbase.H"
#include "grid.H"
#include "multifluid.H"
#include "constants.H"


#ifndef DATA_TYPE
#define DATA_TYPE REAL
#define REAL_FLAG
#endif

class Pot : public NumMatrix<DATA_TYPE,DIM>
{
public:
	Pot();
	Pot(const int *);
	Pot(const NumMatrix<DATA_TYPE,DIM>&);
	void resize(const int *);
	void resize(const int *, int);
	void resize_ll(const int *, const int *);
	void set_max(const double &);
	void set_min(const double &);
	Pot& operator=(const double&);  
	REAL get_max();
	REAL get_min();
// 	std::string get_FieldType();
// 	void set_FieldType(std::string);
// private:
// 	std::string FieldType; // Indicates the name used in the program
};



class id_handler {
public:
	id_handler();
	unsigned int get_fieldId(NumMatrix<REAL,3> &omField);
	unsigned int get_userFieldId(NumMatrix<REAL,3> &omField);
	bool is_generic(NumMatrix<REAL,3> &omField);
	void enter_fieldId(std::string);
	void print_map();
	std::map<std::string, unsigned int> fieldIds;
	unsigned int max_generic;
	unsigned int max_generic_int;
	unsigned int max_user;
};

// class FieldHandler {
// public:
// 	FieldHandler
// }


class Data: public Grid {
public:
	Data();
	~Data();
	void floatom_out(int R, Hdf5Stream & file, int om_lo, int om_hi);
	void float_out(int, bool);
	void CheckNan(int);
	double computeInt(int);
	double computeRMS(int);
	NumMatrix<double,3>  computeAbs(int, int) const;
	double getMin(int);
	double getMin(int [], int);
	double getMax(int);
	double getMax(int [], int);

	void set_fieldIds();
	bool is_userField(NumMatrix<REAL,3> &omField);

	NumMatrix<float,3> float_data(int, int, bool);

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	CronosMultifluid *fluids;
#else
	CronosFluid fluid;
#endif

#ifdef parallel
	void MpiSum(int &) const;
	void MpiSum(REAL &) const;
	void MpiSum(REAL [3]) const;
	void MpiSum(REAL [], int) const;
	void MpiSum(int dir, int &) const;
	void MpiSum(int dir, REAL &) const;

	void MpiAvergage(REAL &) const;
	void MpiMax(REAL &) const;
#endif
	Pot *om;
	NumMatrix<REAL,3> *nom;
//	Pot om[N_OM+N_P];
//	Pot om[31];
//	std::vector<Pot> om;
//	std::vector<Pot*> om;
#if (OMS_USER == TRUE)
	Pot *om_user;
#endif

	// This is set to NULL for OMS_USER == false (has to exist due to design decisions)
	NumMatrix<REAL,3> *nom_user;

	// Storage for the previous timestep
	Pot *om_old;

#if (GEOM == SPHERICAL)
	NumMatrix<double,1> massFlux;
	double radialArea;
#endif

	NumMatrix<double,3> pTherm; // can store thermal pressure
	NumMatrix<int,3> carbuncleFlag; // flag, where carbuncle problem might occur

	NumMatrix<double,3> admissibleIndicator; // flag, which yields the admissibility of the state

#if (CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	NumMatrix<int,3> entropyFlag;
	NumMatrix<bool,3> entropySwitch2;
	NumMatrix<bool,3> entropySwitch3;
#endif

#if (FLUID_TYPE==CRONOS_RHD && PARTICLE_TRANSPORT == TRUE)
	Pot lorentz;
	Pot lorentz_old;
	Pot du_cov, du_cov2;
	Pot injection;
	Pot powerlawSpecEnergy;
	Pot fluidSpecEnergy;
	Pot sync_fac, ic_fac;
	Pot gammaMin, gammaMax;
	Pot injectionK;
	Pot injectionGammaMaxwellian;

	/* For debugging */
	Pot magneticField_LAB;
	Pot magneticField_COM;

	Pot negativeParticles;
#endif

	std::vector<const Pot *> h5FileAuxFields;


	REAL time, dt, t_end;
	REAL cfl;
	int tstep;

	bool mag, storePressure, use_carbuncleFlag, write_admissibleIndicator;

private:
	/**
	 * Display time in user friendly fashion
	 * @PARAM t_sec full time in seconds
	 * */
	std::string git_humanReadable(double t_sec);

	/** Start and stop time of program for performance analysis
	 * */
	timeval tick_start,tock_end;
	id_handler om_ids;
};



#endif
