#include <stdlib.h>
#include <fstream>
#include <matrix.H>

#define ISODD(i) ((i)&1)     

#include <sys/types.h>
#include <sys/stat.h>   
#include <utime.h>

#include <fenv.h>
#include <signal.h>

#include "gridgen.H"
#ifdef parallel
#include "mpi.h"
#include "MpiErrorHandler.H"
#endif

cronos_ostream ccout(std::cout, 0);
cronos_ostream ccerr(std::cerr, 0);

int main(int argc, char *argv[])
{
#ifdef parallel
	int ntasks;
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
	MPI_Errhandler errhandler;
	MPI_Errhandler_create(& MPI_Errorhandler_cronos, & errhandler);
	MPI_Comm_set_errhandler(MPI_COMM_WORLD, errhandler);
	cout << " Number of tasks: " << ntasks << endl;

	ccout.checkRank();
	ccerr.checkRank();
#endif
	/*
	  Additional arguments:
	  (a) Programm name
	  (b) data directory, Program name
	*/
	if(argc == 2) {
		setenv("pname",argv[1],1);
	} else if (argc == 3) {
		setenv("poub",argv[1],1);
		setenv("pname",argv[2],1);
	}

	feclearexcept(FE_ALL_EXCEPT);  		// Clear all exceptions
	signal(SIGFPE, SIG_IGN);       		// Ignore Floating point errors (zero division)
	signal(SIGSEGV, stacktraceHandler); // install the stacktracer for SEGFAULTS
	signal(SIGABRT, stacktraceHandler); // install the stacktracer for aborts (e.g. by failed asserts)

#ifdef RK_STEPS

	// N_P is given in constants.H - number of poisson-steps
	if (N_P) {
		cerr << "N_P should be 0 for Runge Kutta" << endl;
		exit(2);
	}
#endif

	Data gdata;

	// Environment *solver;
	// solver = new Environment(gdata);
	Environment solver(gdata);
	//  solver->setType(gdata);


	bool RESTART = bool(value((char*)"restart"));

	if (RESTART) {
		solver.LoadData(gdata);
		try {
			solver.RKSolver->restart(gdata, *solver.gfunc, *solver.Problem);
			for(auto& eulersolver: solver.EulerSolvers) {
				eulersolver->restart(gdata, *solver.gfunc, *solver.Problem);
			}
		} catch (CException exep) {
			solver.Abort(gdata, exep);
		}
	} else {
		gdata.time = 0.;
		try {
			solver.RKSolver->init(gdata, *solver.gfunc, *solver.Problem);
			for(auto& eulersolver: solver.EulerSolvers) {
				eulersolver->init(gdata, *solver.gfunc, *solver.Problem);
			}
		} catch (CException exep) {
			solver.Abort(gdata, exep);
		}
		solver.InitOutput(gdata);
	}

	int EndProgram(0);
	// for (;;) {
	while(EndProgram == 0) {
		EndProgram = solver.integrate(gdata);
		// EndProgram = solver.Finalize(gdata, static_cast<string>("Stopping for gdm"));

	}
#ifdef parallel
//	MPI_Barrier(gdata.comm3d);
	MPI_Finalize();
#endif
// #ifdef parallel
// 	MPI_Finalize();
// #endif

	if (EndProgram == 1) {
		// Everything is fine -> exit normally
		return 0;
	} else {
		return EndProgram;
	}
}










